unit GL3DPrimatives;

{3D Primatives for XPL0 intrinsics}

interface

uses   Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, paintpan, StdCtrls, SpinButtons, intspin,Contnrs,DSupport,Vector,
  GL3DFigures,GLViewer, GLRenderer,GLFrame,GLMesh,GLLines,GLTexture,GLCubeMap,
  GLLibrary,GLFonts,GLColorArray,GLLights,OpenGL15;

type T3DObjectType = (t3dFrame, t3dMesh, t3dLines, t3dText);
type T3DObjHandle = integer;
type TTextureParams = (tpScale2D,tpOffset2D,tpScale3D,tpOffset3D,tRSPlane,tpSPlane,tpTPlane,tpRadius,tpYaw,tpPitch,tpRoll);
type TTextureFlags = (tfEnable,tfHighlight);


procedure GLIntializeGraphics(Viewer: TGLViewer);

procedure GLViewerEnable(Enable: boolean);
procedure GLViewerWindow(Left,Top,Width,Height: integer);
procedure GLSetRenderMode(Handle: T3DObjHandle; Item, Value: integer);
procedure GLSetViewerMode(Item, Value: integer);
function GLAddObject(Parent: T3DObjHandle; Name: string; ObjType: T3DObjectType): T3DObjHandle;
function GLGetObjectName(Handle: T3DObjHandle): string;
procedure GLDeleteObject(Handle: T3DObjHandle);
procedure GLPositionObject(Handle: T3DObjHandle; X,Y,Z: double);
procedure GLRotateObject(Handle: T3DObjHandle; Pitch,Yaw,Roll: double);
procedure GLOrientObject(Handle: T3DObjHandle; Mode: integer; Dir,Up: T3DVector);

function GLAddTriangle(Handle: T3DObjHandle; T: TTriangle; Color: TColor): integer; overload;
function GLAddTriangle(Handle: T3DObjHandle; P1,P2,P3: T3DVector; Color: TColor): integer; overload;
function GLAddTriangle(Handle: T3DObjHandle; X1,Y1,Z1, X2,Y2,Z2, X3,Y3,Z3: double; Color: TColor): integer; overload;
procedure GLAddRectangle(Handle: T3DObjHandle; T: T3DRectangle; Color: TColor); overload;
procedure GLAddRectangle(Handle: T3DObjHandle; P1,P2,P3,P4: T3DVector; Color: TColor); overload;
procedure GLAddRectangle(Handle: T3DObjHandle; X1,Y1,Z1, X2,Y2,Z2, X3,Y3,Z3, X4,Y4,Z4: double; Color: TColor); overload;


procedure GLAddLine(Handle: T3DObjHandle; X1,Y1,Z1, X2,Y2,Z2: double; Width: integer; Color: TColor);
procedure GLUpdateViewer;
procedure GLSetNormals(Handle: T3DObjHandle; CreaseAngle: double);
procedure GLClearData(Handle: T3DObjHandle);
function GLGetViewerInfo(InfoIndex: integer): integer;
procedure GLSetBackgroundColor(Color: TColor);

procedure GLSetAmbient(Color: TColor);
procedure GLResetTransforms(Handle: integer);
function GLGetCount(Handle,Index: integer): integer;
function GLGetListHandle(ParentFrame: T3DObjHandle; ListType,Index: integer): T3DObjHandle;
procedure GLSetItemColor(Handle,Index: integer; C1,C2,C3: TColor);

procedure GLSetLights(Index,Enable,Mode: integer; X,Y,Z: double; Color: TColor);
procedure GLSetObjectColor(Handle: integer; Color: TColor);
procedure GLSetObjectAlpha(Handle, Enable,DepthCheck,BackBlend,Flags: integer; Trans: double);
procedure GLSetBackgroundBM(BM: TBitmap; Enable: boolean; Pos: T3DVector; Scale: T2DVector);
procedure GLLoadFile(Filename: string);
procedure GLSaveFile(Filename: string);
procedure GLFitToWindow(Handle,Index: integer);
procedure GLCopyObjects(ToObj,FromObj: integer);
function GLHitTest(X,Y: integer; var Handle, TriIndex: integer): boolean;
procedure GLSetText(Handle: integer; Text: string; Font: TFont; Depth: double);
procedure GLGetObjectInfo(Handle: T3DObjHandle; var Parent: T3DObjHandle; var ObjType: T3DObjectType);
procedure GLSetObjectScale(Handle: integer; Scale: T3DVector);

procedure GLSetTexture(Handle: T3DObjHandle; Mode: TTextureMode; SubMode: integer; Bitmaps: array of TBitmap);
procedure GLSetTextureParams(Handle: T3DObjHandle; Variable: TTextureParams; X,Y,Z,W: double);
procedure GLSetTextureFlags(Handle: T3DObjHandle; Variable: TTextureFlags; Enable: boolean);
procedure GLGetObjectPos(Handle: integer; var X, Y, Z, Pitch, Yaw, Roll: double);
procedure GLSetObjectData(Handle: T3DObjHandle; DoubleArray: PDoubleArray; Count,Color,Width: integer);
procedure GLSetMaterials(Handle: T3DObjHandle; Mode: integer; Color: TColor);
procedure GLShowNormals(Handle: T3DObjHandle; Enable: boolean);
procedure GLSetStrokeText(Handle: integer; UsePos: boolean; Text: string; Pos: T3DVector; Scale: double; Color: TColor);


implementation

uses IDEProgressDlg;


var GLObjects: TObjectList;
var WrkViewer: TGLViewer;



function HandleToGLObject(Handle: T3DObjHandle): TGLObject;
begin
if Handle=-1 then Result:=WrkViewer.BaseFrame
else if (Handle<0) or (Handle>=GLObjects.Count) then  raise Exception.Create('Invalid Handle.')
else Result:=TGLObject(GLObjects[Handle]);
end;



function GLObjectToHandle(O: TGLObject): T3DObjHandle;
begin
Result:=GLObjects.IndexOf(O);
end;




procedure GLViewerEnable(Enable: boolean);
begin
WrkViewer.Visible:=Enable;
WrkViewer.Update;
WrkViewer.Enabled:=Enable;
end;



procedure GLViewerWindow(Left,Top,Width,Height: integer);
begin
WrkViewer.SetBounds(Left,Top,Width,Height);
end;


procedure GLSetViewerMode(Item, Value: integer);
begin
case Item of
 0: WrkViewer.BackCulling:=Value<>0;
 end;
end;



procedure GLSetRenderMode(Handle: T3DObjHandle; Item, Value: integer);
var P: TGLObject;
begin
P:=HandleToGLObject(Handle);
if not (P is TGLRenderer) then raise Exception.Create('Object doesn''t support this action.');
case Item of
 0:
	begin
	TGLRenderer(P).RenderMode:=TRenderMode(Value);
	WrkViewer.WireFrameLight:=TGLRenderer(P).RenderMode = rmWireFrame;
	end;
 1: TGLRenderer(P).DoTwoSidedLighting:=Value<>0;
 end;
end;




function GLAddObject(Parent: T3DObjHandle; Name: string; ObjType: T3DObjectType): T3DObjHandle;
var P: TGLObject;
var Mesh: TGLMesh;
var Frame: TGLFrame;
var Lines: TGLLines;
var GLString: TGL3DString;
begin
P:=HandleToGLObject(Parent);
if not (P is TGLFrame) then raise Exception.Create('Parent Must Be Frame');

case ObjType of
 t3dFrame:
	begin
	TGLFrame(P).FrameList.Add(Frame);
	Frame.Name:=Name;
	Result:=GLObjects.Add(Frame);
	end;
 t3dMesh:
	begin
	TGLFrame(P).MeshList.Add(Mesh);
	Mesh.Name:=Name;
	Result:=GLObjects.Add(Mesh);
	end;
 t3dLines:
	begin
	TGLFrame(P).LineList.Add(Lines);
	Lines.Name:=Name;
	Result:=GLObjects.Add(Lines);
	end;
 t3dText:
	begin
	TGLFrame(P).StringList3D.Add(GLString);
	GLString.Name:=Name;
	Result:=GLObjects.Add(GLString);
	end;
 end;
end;



function GLGetObjectName(Handle: T3DObjHandle): string;
var P: TGLObject;
begin
P:=HandleToGLObject(Handle);
if not (P is TGLObject) then raise Exception.Create('Invalid Handle.');
Result:=P.Name;
end;




procedure GLDeleteObject(Handle: T3DObjHandle);
var G: TGLObject;
var Inx: integer;
begin
G:=TGLObject(GLObjects[Handle]);
if G.ParentObject is TGLFrame then
	begin
	if G is TGLFrame then
		begin
		Inx:=TGLFrame(G.ParentObject).FrameList.IndexOf(G);
                TGLFrame(G.ParentObject).FrameList.Delete(Inx);
                end
	else if G is TGLMesh then
		begin
		Inx:=TGLFrame(G.ParentObject).MeshList.IndexOf(G);
                TGLFrame(G.ParentObject).MeshList.Delete(Inx);
                end
	else if G is TGLLines then
		begin
		Inx:=TGLFrame(G.ParentObject).LineList.IndexOf(G);
                TGLFrame(G.ParentObject).LineList.Delete(Inx);
                end


	end;
GLObjects.Delete(Handle);
end;


procedure GLPositionObject(Handle: T3DObjHandle; X,Y,Z: double);
var P: TGLObject;
begin
P:=HandleToGLObject(Handle);
if not (P is TGLObject) then raise Exception.Create('Object doesn''t support this action.');
TGLObject(P).Position:=MakeVector3D(X,Y,Z);
end;



procedure GLRotateObject(Handle: T3DObjHandle; Pitch,Yaw,Roll: double);
var P: TGLObject;
begin
P:=HandleToGLObject(Handle);
if not (P is TGLObject) then raise Exception.Create('Object doesn''t support this action.');
TGLObject(P).SetRotation(Pitch,Yaw,Roll);
TGLObject(P).RollInc:=Roll;
TGLObject(P).YawInc:=Yaw;
TGLObject(P).PitchInc:=Pitch;
end;


procedure GLOrientObject(Handle: T3DObjHandle; Mode: integer; Dir,Up: T3DVector);
var P: TGLObject;
begin
P:=HandleToGLObject(Handle);
if not (P is TGLObject) then raise Exception.Create('Object doesn''t support this action.');
case Mode of
 0: TGLObject(P).SetDirection(Dir);
 1: TGLObject(P).SetDirection(Dir,Up);
 end;
end;


{-------------------------------------------------------------------------------}

function GLAddTriangle(Handle: T3DObjHandle; T: TTriangle; Color: TColor): integer; overload;
var P: TGLObject;
begin
P:=HandleToGLObject(Handle);
if not (P is TGLMesh) then raise Exception.Create('Object doesn''t support this action.');
Result:=TGLMesh(P).AddTriangle(T, Color);
end;


function GLAddTriangle(Handle: T3DObjHandle; P1,P2,P3: T3DVector; Color: TColor): integer; overload;
begin
Result:=GLAddTriangle(Handle,MakeTriangle(P1,P2,P3), Color);
end;

function GLAddTriangle(Handle: T3DObjHandle; X1,Y1,Z1, X2,Y2,Z2, X3,Y3,Z3: double; Color: TColor): integer; overload;
begin
Result:=GLAddTriangle(Handle,MakeVector3D(X1,Y1,Z1), MakeVector3D(X2,Y2,Z2), MakeVector3D(X3,Y3,Z3), Color);
end;


{-------------------------------------------------------------------------------}

procedure GLAddRectangle(Handle: T3DObjHandle; T: T3DRectangle; Color: TColor); overload;
var P: TGLObject;
begin
P:=HandleToGLObject(Handle);
if not (P is TGLMesh) then raise Exception.Create('Object doesn''t support this action.');
TGLMesh(P).AddRectangle(T, Color);
end;


procedure GLAddRectangle(Handle: T3DObjHandle; P1,P2,P3,P4: T3DVector; Color: TColor); overload;
begin
GLAddRectangle(Handle,MakeRectangle(P1,P2,P3,P4), Color);
end;

procedure GLAddRectangle(Handle: T3DObjHandle; X1,Y1,Z1, X2,Y2,Z2, X3,Y3,Z3, X4,Y4,Z4: double; Color: TColor); overload;
begin
GLAddRectangle(Handle,MakeVector3D(X1,Y1,Z1), MakeVector3D(X2,Y2,Z2), MakeVector3D(X3,Y3,Z3),  MakeVector3D(X4,Y4,Z4), Color);
end;


{-------------------------------------------------------------------------------}

procedure GLAddLine(Handle: T3DObjHandle; X1,Y1,Z1, X2,Y2,Z2: double; Width: integer; Color: TColor);
var P: TGLObject;
var L: T3DLine;
begin
P:=HandleToGLObject(Handle);
if not (P is TGLLines) then raise Exception.Create('Object doesn''t support this action.');
L:=MakeLine3D(MakeVector3D(X1,Y1,Z1), MakeVector3D(X2,Y2,Z2));
TGLLines(P).AddLine(L, Color, Width);
end;




procedure GLIntializeGraphics(Viewer: TGLViewer);
begin
WrkViewer:=Viewer;
WrkViewer.Reset;
WrkViewer.BackTexture.Clear;
GLViewerEnable(False);
GLObjects.Clear;
end;


procedure GLUpdateViewer;
begin
WrkViewer.Repaint;
end;



procedure GLSetNormals(Handle: T3DObjHandle; CreaseAngle: double);
var P: TGLObject;
begin
P:=HandleToGLObject(Handle);
if not (P is TGLMesh) then raise Exception.Create('Object doesn''t support this action.');
TGLMesh(P).SetAllNormals(True,CreaseAngle);
end;



procedure GLClearData(Handle: T3DObjHandle);
var P: TGLObject;
begin
P:=HandleToGLObject(Handle);
if P is TGLFrame then TGLFrame(P).Clear
else if P is TGLMesh then TGLMesh(P).ClearData
else if P is TGLLines then TGLLines(P).ClearData
else if P is TGL3DString then TGL3Dstring(P).ClearData
else raise Exception.Create('Object doesn''t support this action.');
end;



function GLGetViewerInfo(InfoIndex: integer): integer;
begin
case InfoIndex of
 0: Result:=Trunc(WrkViewer.FramesPerSec * 10);
 end;
end;



procedure GLSetBackgroundColor(Color: TColor);
begin
WrkViewer.ScreenColor:=Color;
end;



procedure GLSetAmbient(Color: TColor);
var R,G,B: double;
begin
{OpenGL amnient light values go from -1.0 to +1.0}
R:=(GetRValue(Color)/127)-1;
G:=(GetGValue(Color)/127)-1;
B:=(GetBValue(Color)/127)-1;
WrkViewer.GlobalAmbientLight:=MakeGLArrayf4(R,G,B,1);
end;





procedure GLResetTransforms(Handle: integer);
var P: TGLObject;
begin
P:=HandleToGLObject(Handle);
if P is TGLFrame then TGLFrame(P).ResetTransforms
else if P is TGLMesh then TGLMesh(P).ResetTransforms
else if P is TGLLines then TGLLines(P).ResetTransforms
else if P is TGL3DString then TGL3DString(P).ResetTransforms
else raise Exception.Create('Object doesn''t support this action.');
end;



function GLGetCount(Handle,Index: integer): integer;
var P: TGLObject;
begin
Result:=0;
P:=HandleToGLObject(Handle);
if P is TGLFrame then
	begin
	case Index of
	 0: Result:=TGLFrame(P).FrameList.Count;
	 1: Result:=TGLFrame(P).MeshList.Count;
	 2: Result:=TGLFrame(P).LineList.Count;
	 3: Result:=TGLFrame(P).StringList3D.Count;
	 end;
	end
else if P is TGLMesh then Result:=TGLMesh(P).FaceColor.Count
else if P is TGLLines then Result:=TGLLines(P).LineColor.Count
else if P is TGL3DString then Result:=Length(TGL3DString(P).Text)
else raise Exception.Create('Object doesn''t support this action.');
end;




function GLGetListHandle(ParentFrame: T3DObjHandle; ListType,Index: integer): T3DObjHandle;
var P: TGLObject;

	function OutOfRange(Index: integer; OL: TObjectList): boolean;
	begin
	Result:=(Index<0) or (Index>=OL.Count);
	end;


begin
P:=HandleToGLObject(ParentFrame);
if not (P is TGLFrame) then raise Exception.Create('Parent must be Frame.');
case ListType of
 0:
	begin
	if OutOfRange(Index,TGLFrame(P).FrameList) then raise Exception.Create('Index out of range.');
	Result:=GLObjectToHandle(TGLFrame(P).FrameList[Index]);
	end;
 1:	begin
 	if OutOfRange(Index,TGLFrame(P).MeshList) then raise Exception.Create('Index out of range.');
 	Result:=GLObjectToHandle(TGLFrame(P).MeshList[Index]);
	end;
 2:	begin
 	if OutOfRange(Index,TGLFrame(P).LineList) then raise Exception.Create('Index out of range.');
 	Result:=GLObjectToHandle(TGLFrame(P).LineList[Index]);
	end;
 3:	begin
 	if OutOfRange(Index,TGLFrame(P).StringList3D) then raise Exception.Create('Index out of range.');
 	Result:=GLObjectToHandle(TGLFrame(P).StringList3D[Index]);
	end;
 else raise Exception.Create('Unknown list type.');
 end;

end;



procedure GLSetItemColor(Handle,Index: integer; C1,C2,C3: TColor);
var P: TGLObject;
var GLC: TGLFaceColor;
begin
P:=HandleToGLObject(Handle);
if P is TGLMesh then
	begin
	GLC.C1:=C1; GLC.C2:=C2; GLC.C3:=C3;
	TGLMesh(P).FaceColor[Index]:=GLC;
	end
else if P is TGLLines then TGLLines(P).LineColor[Index]:=C1
else if P is TGL3DString then TGL3DString(P).Color:=C1
else raise Exception.Create('Object doesn''t support this action.');
end;


procedure GLSetLights(Index,Enable,Mode: integer; X,Y,Z: double; Color: TColor);
var V: T3DVector;
begin
V:=MakeVector3D(X,Y,Z);
WrkViewer.Lights[Index].Enabled:=Enable<>0;
WrkViewer.Lights[Index].LightType:=TLightType(Mode);
WrkViewer.Lights[Index].SpotDirection:=GLLibrary.Vector3DToGLArray(V);
WrkViewer.Lights[Index].DirectionalDir:=GLLibrary.MakeGLArrayf4(X,Y,Z,0);
WrkViewer.Lights[Index].Position:=MakeGLArrayf4(X,Y,Z,1);
WrkViewer.Invalidate;
end;



procedure GLSetObjectColor(Handle: integer; Color: TColor);
var P: TGLObject;
var GLC: TGLFaceColor;
begin
P:=HandleToGLObject(Handle);
if P is TGLMesh then TGLMesh(P).Color:=Color
else if P is TGLLines then TGLLines(P).Color:=Color
else if P is TGL3DString then TGL3DString(P).Color:=Color
else raise Exception.Create('Object doesn''t support this action.');
end;




procedure GLSetObjectAlpha(Handle, Enable,DepthCheck,BackBlend,Flags: integer; Trans: double);
var P: TGLObject;
var SourceBlend,DestBlend,Equation: integer;
begin
P:=HandleToGLObject(Handle);
if P is TGLRenderer then
	begin
	TGLRenderer(P).DoAlphaBlend:=Enable<>0;
	TGLRenderer(P).TransparentDepth:=DepthCheck<>0;
	TGLRenderer(P).BlendBackground:=BackBlend<>0;
	TGLRenderer(P).Transparency:=Trans;
	TGLRenderer(P).UserBlend:=(Flags and $FF000000)<>0;
	Equation:=(Flags and $00FF0000) shr 16;
	SourceBlend:=Flags and $FF;
	DestBlend:=(Flags shr 8) and $FF00;

	case Equation of
	 0: TGLRenderer(P).BlendEquation:=GL_FUNC_ADD;
	 1: TGLRenderer(P).BlendEquation:=GL_FUNC_SUBTRACT;
	 2: TGLRenderer(P).BlendEquation:=GL_FUNC_REVERSE_SUBTRACT;
	 3: TGLRenderer(P).BlendEquation:=GL_MIN;
	 4: TGLRenderer(P).BlendEquation:=GL_MAX;
	 end;

	case SourceBlend of
	 0: TGLRenderer(P).SrcUserBlend:=GL_ZERO;
	 1: TGLRenderer(P).SrcUserBlend:=GL_ONE;
	 2: TGLRenderer(P).SrcUserBlend:=GL_DST_COLOR;
	 3: TGLRenderer(P).SrcUserBlend:=GL_ONE_MINUS_DST_COLOR;
	 4: TGLRenderer(P).SrcUserBlend:=GL_SRC_ALPHA;
	 5: TGLRenderer(P).SrcUserBlend:=GL_ONE_MINUS_SRC_ALPHA;
	 6: TGLRenderer(P).SrcUserBlend:=GL_DST_ALPHA;
	 7: TGLRenderer(P).SrcUserBlend:=GL_ONE_MINUS_DST_ALPHA;
	 8: TGLRenderer(P).SrcUserBlend:=GL_SRC_ALPHA_SATURATE;
	 end;

	case DestBlend of
	 0: TGLRenderer(P).DestUserBlend:=GL_ZERO;
	 1: TGLRenderer(P).DestUserBlend:=GL_ONE;
	 2: TGLRenderer(P).DestUserBlend:=GL_SRC_COLOR;
	 3: TGLRenderer(P).DestUserBlend:=GL_ONE_MINUS_SRC_COLOR;
	 4: TGLRenderer(P).DestUserBlend:=GL_SRC_ALPHA;
	 5: TGLRenderer(P).DestUserBlend:=GL_ONE_MINUS_SRC_ALPHA;
	 6: TGLRenderer(P).DestUserBlend:=GL_DST_ALPHA;
	 7: TGLRenderer(P).DestUserBlend:=GL_ONE_MINUS_DST_ALPHA;
	 end;
	end
else raise Exception.Create('Object doesn''t support this action.');
end;


procedure GLSetBackgroundBM(BM: TBitmap; Enable: boolean; Pos: T3DVector; Scale: T2DVector);
begin
if BM<>nil then WrkViewer.BackTexture.LoadBitmap(BM);
WrkViewer.BackTexture.Enabled:=Enable;
WrkViewer.BackTexture.Position:=Pos;
WrkViewer.BackTexture.Zoom:=Scale;
WrkViewer.Invalidate;
end;



procedure HandleProgress(Percent: integer);
begin
SetProgress(Percent);
end;



procedure GLSaveFile(Filename: string);
begin
ShowProgressDlg(True);
try
SetProgress('Initializing',0);
if LowerCase(ExtractFileExt(Filename))='.x' then WrkViewer.SaveXFile(Filename,HandleProgress)
else  WrkViewer.SaveXMLFile(Filename,HandleProgress);
SetProgress('Done',100);
finally ShowProgressDlg(False); end;
end;



procedure GLLoadFile(Filename: string);
var OldInx,I: integer;
begin
ShowProgressDlg(True);
try
SetProgress('Loading: '+ExtractFilename(Filename),0);
OldInx:=WrkViewer.BaseFrame.MasterList.Count;
if LowerCase(ExtractFileExt(Filename))='.x' then WrkViewer.LoadXFile(Filename,HandleProgress,False)
else WrkViewer.LoadXMLFile(Filename,HandleProgress,False);
SetProgress('Done',100);
{Add all newly loaded meshes to our handle list}
for I:=OldInx to WrkViewer.BaseFrame.MasterList.Count-1 do
	GLObjects.Add(WrkViewer.BaseFrame.MasterList[I]);
finally ShowProgressDlg(False); end;
end;



procedure GLFitToWindow(Handle,Index: integer);
var P: TGLObject;
begin
if (WrkViewer.GLWindow=nil) or (WrkViewer.GLWindow.RealWidth<=0) or
   (WrkViewer.GLWindow.RealHeight<=0) then raise Exception.Create('Cannot fit/scale. Window Size = 0.');
if Handle=-2 then
	begin
	case Index of
	 0: WrkViewer.FitSceneByCamera;
	 1: WrkViewer.FitSceneByDistance;
	 2: WrkViewer.FitSceneByScaling;
	 end
	end
else
	begin
	P:=HandleToGLObject(Handle);
	if P is TGLObject then
		begin
		case Index of
		 0: WrkViewer.FitObjectByDistance(TGLObject(P));
		 1: WrkViewer.FitObjectByScaling(TGLObject(P));
		 end;
		end
	else raise Exception.Create('Object doesn''t support this action.');
	end;
end;




procedure GLCopyObjects(ToObj,FromObj: integer);
var PTo,PFrom: TGLObject;
begin
PTo:=HandleToGLObject(ToObj);
PFrom:=HandleToGLObject(FromObj);
if not (PTo is TGLObject) then raise Exception.Create('To Object doesn''t support this action.');
if not (PFrom is TGLObject) then raise Exception.Create('From Object doesn''t support this action.');
PTo.Assign(PFrom);
end;


function GLHitTest(X,Y: integer; var Handle, TriIndex: integer): boolean;
var Pos: TPoint;
var TriInfo: TTriangleIntersect;
begin
TriInfo.Valid:=False;
WrkViewer.BaseFrame.FindNearestTriangle(X,Y,TriInfo);
Result:=TriInfo.Valid;
Handle:=GLObjects.IndexOf(TriInfo.Mesh);
TriIndex:=TriInfo.Index;
end;


procedure GLSetText(Handle: integer; Text: string; Font: TFont; Depth: double);
var P: TGLObject;
begin
P:=HandleToGLObject(Handle);
if P is TGL3DString then
	begin
	TGL3DString(P).Text:=Text;
	TGL3DString(P).GL3DFont.Font:=Font;
	TGL3DString(P).GL3DFont.FontDepth:=Depth;
	end
else raise Exception.Create('Object doesn''t support this action.');
end;



procedure GLGetObjectInfo(Handle: T3DObjHandle; var Parent: T3DObjHandle; var ObjType: T3DObjectType);
var P: TGLObject;
begin
P:=HandleToGLObject(Handle);
if P is TGLFrame then ObjType:=t3dFrame
else if P is TGLMesh then ObjType:=t3dMesh
else if P is TGLLines then ObjType:=t3dLines
else if P is TGL3DString then ObjType:=t3dText;
Parent:=GLObjects.IndexOf(TGLObject(P).ParentObject);
end;




procedure GLSetObjectScale(Handle: integer; Scale: T3DVector);
var P: TGLObject;
begin
P:=HandleToGLObject(Handle);
if P is TGLRenderer then
	begin
	TGLRenderer(P).Scale3D:=Scale;
        if P is TGLMesh then TGLMesh(P).NormalizeScale;
	end
else raise Exception.Create('Object doesn''t support this action.');
end;



procedure GLSetTexture(Handle: T3DObjHandle; Mode: TTextureMode; SubMode: integer; Bitmaps: array of TBitmap);
var P: TGLObject;
var T: TGLObjectTexture;
var Cnt: integer;
begin
P:=HandleToGLObject(Handle);
if not (P is TGLRenderer) then raise Exception.Create('Object doesn''t support this action.');
T:=TGLRenderer(P).Texture;
T.TextureMode:=Mode;
case Mode of
 tmAuto: T.AutoTextureMap:=TAutoTextureMap(SubMode);
 tmManual: T.ManualTextureMap:=TManualTextureMap(SubMode);
 tmCubeMap: T.CubeMap.MapMode:=TCubeTextureMap(SubMode);
 end;
Cnt:=Length(Bitmaps);

if Mode=tmCubeMap then
	begin
	T.CubeMap.MapMode:=TCubeTextureMap(SubMode);
	if Cnt>0 then T.CubeMap.LeftMap.LoadBitmap(Bitmaps[0]);
	if Cnt>1 then T.CubeMap.RightMap.LoadBitmap(Bitmaps[1]);
	if Cnt>2 then T.CubeMap.TopMap.LoadBitmap(Bitmaps[2]);
	if Cnt>3 then T.CubeMap.BottomMap.LoadBitmap(Bitmaps[3]);
	if Cnt>4 then T.CubeMap.FrontMap.LoadBitmap(Bitmaps[4]);
	if Cnt>5 then T.CubeMap.BackMap.LoadBitmap(Bitmaps[5]);
	end
else if (Mode=tmAuto) or (Mode=tmManual) and (Cnt>0) then T.LoadBitmap(Bitmaps[0]);
end;



procedure GLSetTextureParams(Handle: T3DObjHandle; Variable: TTextureParams; X,Y,Z,W: double);
var P: TGLObject;
var T: TGLObjectTexture;
var Cnt: integer;
begin
P:=HandleToGLObject(Handle);
if not (P is TGLRenderer) then raise Exception.Create('Object doesn''t support this action.');
T:=TGLRenderer(P).Texture;
case Variable of
 tpScale2D: T.Scale:=MakeVector2D(X,Y);
 tpOffset2D: T.Offset:=MakeVector2D(X,Y);
 tpScale3D: T.CubeMap.Scale:=MakeVector3D(X,Y,Z);
 tpOffset3D: T.CubeMap.Offset:=MakeVector3D(X,Y,Z);
 tRSPlane: T.CubeMap.RPlane:=MakeGLArrayf4(X,Y,Z,W);
 tpSPlane: begin T.CubeMap.SPlane:=MakeGLArrayf4(X,Y,Z,W); T.SPlane:=MakeGLArrayf4(X,Y,Z,W); end;
 tpTPlane: begin T.CubeMap.TPlane:=MakeGLArrayf4(X,Y,Z,W); T.TPlane:=MakeGLArrayf4(X,Y,Z,W); end;
 tpRadius:  T.Radius:=X;
 tpYaw: T.Yaw:=X;
 tpPitch: T.Pitch:=X;
 tpRoll: T.Roll:=X;
 end;
end;




procedure GLSetTextureFlags(Handle: T3DObjHandle; Variable: TTextureFlags; Enable: boolean);
var P: TGLObject;
var T: TGLObjectTexture;
begin
P:=HandleToGLObject(Handle);
if not (P is TGLRenderer) then raise Exception.Create('Object doesn''t support this action.');
T:=TGLRenderer(P).Texture;
case Variable of
 tfEnable: T.Enabled:=Enable;
 tfHighlight: T.Highlight:=Enable;
 end;
end;



procedure GLGetObjectPos(Handle: integer; var X, Y, Z, Pitch, Yaw, Roll: double);
var P: TGLObject;
begin
P:=HandleToGLObject(Handle);
if not (P is TGLObject) then raise Exception.Create('Object doesn''t support this action.');
X:=P.Position.X; Y:=P.Position.Y; Z:=P.Position.Z;
Pitch:=P.Pitch;
Yaw:=P.Yaw;
Roll:=P.Roll;
end;




procedure GLSetObjectData(Handle: T3DObjHandle; DoubleArray: PDoubleArray; Count,Color,Width: integer);
var I: integer;
var P: TGLObject;
var T: TTriangle;
var L: T3DLine;

	function GetTriangle(var I: integer; var T: TTriangle): boolean;
	begin
	Result:=False;
	if I>=Count then exit;
	T:=PTriangle(@DoubleArray[I])^;
	Inc(I,9);
	Result:=True;
	end;


	function GetLine(var I: integer; var T: T3DLine): boolean;
	begin
	Result:=False;
	if I>=Count then exit;
	T:=P3DLine(@DoubleArray[I])^;
	Inc(I,3);
	Result:=True;
	end;

begin
I:=0;
P:=HandleToGLObject(Handle);
if P is TGLMesh then
	begin
	while GetTriangle(I,T) do TGLMesh(P).AddTriangle(T, Color);
	end
else if P is TGLLines then
	begin
	while GetLine(I,L) do TGLLines(P).AddLine(L, Color, Width);
	end
else raise Exception.Create('Object must Mesh or Lines.');
end;





procedure GLSetMaterials(Handle: T3DObjHandle; Mode: integer; Color: TColor);
var P: TGLObject;
var GLA: TGLArrayf4;
begin
P:=HandleToGLObject(Handle);
if not (P is TGLRenderer) then raise Exception.Create('Object doesn''t support this action.');
GLA:=ColorToGLArrayf4(Color);
case Mode of
 0: TGLRenderer(P).LightEmission:=GLA;
 1: TGLRenderer(P).LightSpecular:=GLA;
 2: TGLRenderer(P).LightShininess:=GLA[0];
 end;
end;



procedure GLShowNormals(Handle: T3DObjHandle; Enable: boolean);
var P: TGLObject;
var GLA: TGLArrayf4;
begin
P:=HandleToGLObject(Handle);
if not (P is TGLMesh) then raise Exception.Create('Operation Only Applies To Meshes.');
TGLMesh(P).ShowNormals:=Enable;
end;



procedure GLSetStrokeText(Handle: integer; UsePos: boolean; Text: string; Pos: T3DVector; Scale: double; Color: TColor);
var P: TGLObject;
begin
P:=HandleToGLObject(Handle);
if P is TGLLines then
	begin
	TGLLines(P).Color:=Color;
	TGLLines(P).StrokeFont.Scale:=Scale;
	if UsePos then TGLLines(P).StrokeFont.DrawStr(Pos,Text)
	else TGLLines(P).StrokeFont.DrawStr(Text);
	end
else raise Exception.Create('Object doesn''t support this action.');
end;


initialization
GLObjects:=TObjectList.Create;
GLObjects.OwnsObjects:=False;
finalization
GLObjects.Free;
end.
