unit I2LConsoleIO;

interface

uses Forms,Windows,SysUtils,Printers,Dialogs,DSupport,I2LGlobal,I2LDataTypes,XPLDataTypes,
     ConsoleWindow,ComSetup,CommPort;

{*** DO APEX LIKE DEVICE HANDLER DISPATCH ****}
{Device number in NOWDEV, Function in NOWFUN, byte passed in AL}

var Dev6Mode: integer;	{Mode that contolls device 6 window options}
var WinCorners: TRect = (Left: 0; Top: 0; Right: 80; Bottom: 25);	{Corners for the device 6 window}

var NOWDEV: TI2LWord;	{CURRENT DEVICE CHANNEL}
var NOWFUN: TI2LWord;	{CURRENT FUNCTION}

{DEFINE KHAND  FUNCTIONS, PUT FUNCTION IN NOWFUN BEFORE CALLING KHAND}

const INIFUN=0;		{INITIALIZE INPUT FUNCTION}
const INOFUN=1;		{INITIALIZE OUTPUT FUNCTION}
const INPFUN=2;		{BYTE INPUT FUNCTION}
const OUTFUN=3;		{BYTE OUTPUT FUNCTION}
const CLOFUN=4;		{CLOSE OUTPUT FUNCTION}
const CLIFUN=5;		{CLOSE INPUT FUNCTION}
const OIFFUN=6;		{OPEN INPUT FILE FUNCTION}
const OOFFUN=7;		{OPEN OUTPUT FILE  FUNCTION}


var OutputFlag: boolean;	{Flags }

var PseudoBuffer: string;	{Storage for device-8 pseudo chanel}

var ComPort: TCommPort;		{Serial port}
var EOFFlag: boolean;		{Flags that an EOF has been added to Disk Input}

function KHAND(C: char): char;
procedure KHandString(S: string);
procedure InitializeConsole;
procedure AbortConsoleFunctions;

procedure SetSerialParams(SubType, Value: integer);
function GetSerialParams(SubType: integer): integer;
procedure ClearConsole(UseAttribute: boolean);


implementation

uses IDEDebugInfo, I2LRawPrint, I2lIntrinsics;




{Prototype Device handler - all I/O functions must use this format}

type TDeviceHandler = function(var C: char): boolean;

{-------------- Low Level Routines --------------------------------------------}


procedure ClearConsole(UseAttribute: boolean);
begin
GlobalConsole.VGABox.UseAttribute:=UseAttribute;
GlobalConsole.VGABox.Clear;
GlobalConsole.VGABox.Update;
end;



procedure CharOut(C: char; ExpandTabs,UseAttribute: boolean);
begin
GlobalConsole.VGABox.ExpandTabs:=ExpandTabs;
GlobalConsole.VGABox.UseAttribute:=UseAttribute;
GlobalConsole.VGABox.CharOut(C);
end;







{------------- Support Routines -----------------------------------------------}


function GetIOFunctionStr(I: integer): string;
begin
case I of
 INIFUN: Result:='Init Input';
 INOFUN: Result:='Init Output';
 INPFUN: Result:='Byte In';   
 OUTFUN: Result:='Byte Out';
 CLOFUN: Result:='Close Output';
 CLIFUN: Result:='Close Input';
 OIFFUN: Result:='Open Input';
 OOFFUN: Result:='Open Output';
 else Result:='';
 end;
end;


procedure ResetDevice6;
begin
GlobalConsole.VGABox.ResetWindowMode;
WinCorners:=GlobalConsole.VGABox.ClipWindow;
end;



procedure EnableDev6Modes(Enable: boolean);
begin
if Enable then
	begin
	if (Dev6Mode and $FF00)<>0 then GlobalConsole.VGABox.ShowCursor:=False;
	GlobalConsole.VGABox.ClipWindow:=WinCorners;
	case Dev6Mode and $F of
	 0: GlobalConsole.VGABox.WinMode:=wmScroll;
	 1: GlobalConsole.VGABox.WinMode:=wmWrap;
	 2: GlobalConsole.VGABox.WinMode:=wmClip;
	 else GlobalConsole.VGABox.WinMode:=wmScroll;
	 end;
	end
else
	begin
        GlobalConsole.VGABox.ResetClipWindow;
	GlobalConsole.VGABox.WinMode:=wmScroll;
	end;
end;


{----------------- Access Serial Port Parameter -------------------------------}



procedure SetSerialParams(SubType, Value: integer);
begin
case SubType of
 0: ComPort.BaudRate:=TBaudRate(Value);
 1: ComPort.DataBits:=TDataBits(Value);
 2: ComPort.StopBits:=TStopBits(Value);
 3: ComPort.Parity:=TParity(Value);
 4: ComPort.Port:=TSerialPort(Value);
 5: begin ComPort.ReadBufferSize:=Value; ComPort.WriteBufferSize:=Value; end;
 6: ;
 end;
end;



function GetSerialParams(SubType: integer): integer;
begin
case SubType of
 0: Result:=Integer(ComPort.BaudRate);
 1: Result:=Integer(ComPort.DataBits);
 2: Result:=Integer(ComPort.StopBits);
 3: Result:=Integer(ComPort.Parity);
 4: Result:=Integer(ComPort.Port);
 5: Result:=ComPort.ReadBufferSize;
 6: Result:=ComPort.InputCount;
 end;
end;





{-----------------------------------------------------------------------------}

{------------- CHAN 0 FUNCTIONS, BUFFERED CONSOLE -----------------------------}

function RESBUF(var C: char): boolean;
{Func - 0 INTIALIZE INPUT}
begin
GlobalConsole.ResetKeyInBuffer;
ResetRedirection;
C:=#0;
Result:=True;
end;



function DUMMIO(var C: char): boolean;
{Func - 1 INTIALIZE OUTPUT}
begin
C:=#0;
Result:=True;
end;



function BUFKEY(var C: char): boolean;
{Func - 2 INPUT BYTE}
var C2: Char;
begin
EnableDev6Modes(False);
{if the input buffer is empty, check for redirection}
if Length(GlobalConsole.KeyInBuffer)<1 then
	begin
	GlobalConsole.EchoFlag:=True;
	{Device-0 redirects data up to end of line}
	while GetRedirectInput(C2) and
	      not Application.Terminated and
	      not GlobalConsole.CtrlCPressed do
		begin
		{No line-feeds in the input stream for Device-0}
		if C2=LF_Char then
			begin
			{Echo LF without putting it in the input buffer}
			GlobalConsole.VGABox.CharOut(C2);
			break;
			end;
		GlobalConsole.InjectKeystroke(C2);
		end;
	end;
C:=GlobalConsole.BuffererKeyIn;
Result:=True;
end;



function CONOUT(var C: char): boolean;
{Func - 3 OUTPUT BYTE}
begin
EnableDev6Modes(False);
if not RedirectOutput(C) then
	begin
	case C of
	 FF_Char: ClearConsole(False);
	 CtrlH: GlobalConsole.VGABox.MoveCursor(-1,0);
	 Bell_Char: Beep;
	 else CharOut(C,True,True);
	 end;
	end;
Result:=True;
end;

{Func - 4 CLOSE OUTPUT - Dummy}
{Func - 5 CLOSE INPUT - Dummy}


{-------CHAN 1 FUNCTIONS, UNBUFFERED CONSOLE-------}


function RESKEY(var C: char): boolean;
{Func - 0 INTIALIZE INPUT}
begin
GlobalConsole.ResetKeyInBuffer;
ResetRedirection;
C:=#0;
Result:=True;
end;

{Func - 1 INTIALIZE OUTPUT - Dummy}


function KEYIN(var C: char): boolean;
{Func - 2 INPUT BYTE}
begin
EnableDev6Modes(False);
if not GetRedirectInput(C) then C:=GlobalConsole.KeyBoardIn;
Result:=True;
end;



function TVOUT(var C: char): boolean;
{Func - 3 OUTPUT BYTE}
begin
EnableDev6Modes(False);
if not RedirectOutput(C) then
	begin
	case C of
	 CtrlH: GlobalConsole.VGABox.MoveCursor(-1,0);
	 else CharOut(C,False,True);
	 end;
	end;
C:=#0;
Result:=True;
end;


{Func - 4 CLOSE OUTPUT - Dummy}
{Func - 5 CLOSE INPUT - Dummy}


{-------CHAN 2 FUNCTIONS, PRINTER CHANNEL-------}

var PrintChan: TextFile;

var PrintStr: string;

var PrinterSetupDlg: TPrinterSetupDialog;


function PrinterInitializeInput(var C: char): boolean;
{Func - 0 INTIALIZE INPUT - Execute Printer Setup dialog}
begin
Result:=PrinterSetupDlg.Execute;
C:=#0;
end;



function PrinterInitializeOutput(var C: char): boolean;
{Func - 1 INTIALIZE OUTPUT }
begin
PrintStr:='';
AssignPrn(PrintChan);
Rewrite(PrintChan);
C:=#0;
Result:=True;
end;

{Func - 2 INPUT BYTE - Dummy}


function PrinterOutput(var C: char): boolean;
{Func - 3 OUTPUT BYTE}
begin
PrintStr:=PrintStr + C;
C:=#0;
Result:=True;
end;



function PrinterCloseOutput(var C: char): boolean;
{Func - 4 CLOSE OUTPUT }
begin
Write(PrintChan,PrintStr);
System.Close(PrintChan);
C:=#0;
Result:=True;
end;


{Func - 5 CLOSE INPUT - Dummy}


{---------- CHAN 3 FUNCTIONS, DISK I/O--------------------}

var DiskChar: char;


function DOPENI(var C: char): boolean;
{Func - 0 INTIALIZE INPUT}
begin
DiskChar:=#0;
EOFFlag:=False;
Result:=FileRewind(INHAND);
end;



function DOPENO(var C: char): boolean;
{Func - 1 INTIALIZE OUTPUT}
begin
Result:=FileRewind(OTHAND);
end;





function DSKIN(var C: char): boolean;
{Func - 2 INPUT BYTE}
begin
Result:=INHAND>=0;
if Result then
	begin
	Result:=FileRead(INHAND,C,1)>0;
	if not Result then
		begin
		C:=EOF_Char;
		if not EOFFlag then Result:=True;
		EOFFlag:=True;
		end;
	DiskChar:=C;
	end;
end;



function DSKOUT(var C: char): boolean;
{Func - 3 OUTPUT BYTE}
var Cnt: integer;
begin
Result:=OTHAND>=0;
Cnt:=1;
if Result then
	begin
	Cnt:=FileWrite(OTHAND,C,Cnt);
	Result:=Cnt>0;
	end;
end;




function DCLOSO(var C: char): boolean;
{Func - 4 CLOSE OUTPUT}
begin
C:=#0;
DiskChar:=#0;
Result:=True;
end;

{Func - 5 CLOSE INPUT - Dummy}


{-------CHAN 4 FUNCTIONS, SERIAL I/O-------}



function SerialInitIn(var C: char): boolean;
{Func -0  Handles input initialization}
begin
C:=#0;
DoCommSettingsDlg(ComPort);
Result:=True;
end;


function SerialInitOut(var C: char): boolean;
{Func - Handles either 0 and 1, input/output initialization}
begin
C:=#0;
ComPort.PortActive:=True;
Result:=True;
end;



function SERIN(var C: char): boolean;
{Func - 2 INPUT BYTE}
begin
C:=ComPort.ReadChar;
Result:=True;
end;



function SEROUT(var C: char): boolean;
{Func - 3 OUTPUT BYTE}
begin
ComPort.WriteChar(C);
C:=#0;
Result:=True;
end;




function SERCLOSE(var C: char): boolean;
{Func - 4 CLOSE OUTPUT - Dummy}
begin
ComPort.PortActive:=False;
C:=#0;
Result:=True;
end;


{Func - 5 CLOSE INPUT - Dummy}


{-------CHAN 5 Raw, Low-Level printer output -------}


var RawPrintStr: string;

function RawPrinterInitIn(var C: char): boolean;
{Func - 0 Intialize Input - Execute Printer Setup Dialog}
begin
Result:=PrinterSetupDlg.Execute;
C:=#0;
end;


{Func - 1 INTIALIZE OUTPUT - Clear output buffer}

function RawPrinterInitOut(var C: char): boolean;
{Func - 1 Intialize Output }
begin
RawPrintStr:='';
C:=#0;
Result:=True;
end;



{Func - 2 INPUT BYTE - Dummy}

function RawPrinterOutput(var C: char): boolean;
{Func - 3 OUTPUT BYTE}
begin
RawPrintStr:=RawPrintStr + C;
C:=#0;
Result:=True;
end;


function RawPrinterCloseOut(var C: char): boolean;
{Func - 4 CLOSE OUTPUT }
begin
C:=#0;
RawPrintText(RawPrintStr);
RawPrintStr:='';
Result:=True;
end;


{Func - 5 CLOSE INPUT - Dummy}


{------- CHAN 6 TEXT WINDOWING FUNCTIONS-------}


{Func - 0 INTIALIZE INPUT - Dummy}


function RESWND(var C: char): boolean;
{Func - 1 INTIALIZE OUTPUT}
begin
C:=#0;
ResetDevice6;
Result:=True;
end;



function WNDCHN(var C: char): boolean;
{Func - 2 INPUT BYTE}
begin
GlobalConsole.SuppressCtrlC:=True;
try
C:=GlobalConsole.KeyBoardIn;
EnableDev6Modes(True);
CharOut(C,False,True);
Result:=True;
finally GlobalConsole.SuppressCtrlC:=True; end;
end;



function WNDOUT(var C: char): boolean;
{Func - 3 OUTPUT BYTE}
var P: TPoint;
begin
EnableDev6Modes(True);
CharOut(C,False,True);
C:=#0;
Result:=True;
end;



{Func - 4 CLOSE OUTPUT - Dummy}
{Func - 5 CLOSE INPUT - Dummy}


{-------CHAN 7 FUNCTIONS, NULL DEVICE-------}

{Func - 0 INTIALIZE INPUT - Dummy}
{Func - 1 INTIALIZE OUTPUT - Dummy}
{Func - 2 INPUT BYTE - Dummy}
{Func - 3 OUTPUT BYTE - Dummy}
{Func - 4 CLOSE OUTPUT - Dummy}
{Func - 5 CLOSE INPUT - Dummy}


{-------CHAN 8 FUNCTIONS, PSEUDO DEVICE-------}


var PseudoInx: integer;

function PSOPNI(var C: char): boolean;
{Func - 0 INTIALIZE INPUT}
begin
PseudoInx:=1;
Result:=True;
end;



function PSOPNO(var C: char): boolean;
{Func - 1 INTIALIZE OUTPUT}
begin
PseudoBuffer:='';
PseudoInx:=1;
Result:=True;
end;



function PSIN(var C: char): boolean;
{Func - 2 INPUT BYTE}
begin
Result:=True;
if PseudoInx<=Length(PseudoBuffer) then
	begin
	C:=PseudoBuffer[PseudoInx];
	Inc(PseudoInx);
	end
else C:=EOF_Char;
end;



function PSOUT(var C: char): boolean;
{Func - 3 OUTPUT BYTE}
begin
PseudoBuffer:=PseudoBuffer+C;
Result:=True;
end;



{Func - 4 CLOSE OUTPUT - Dummy}
{Func - 5 CLOSE INPUT - Dummy}

{-------CHAN 9 FUNCTIONS, Graphics I/O Channel -------}



function GraphicOut(var C: char): boolean;
{Func - 3 OUTPUT BYTE}
begin
GlobalConsole.VGABox.GraphicsTextOut(C);
C:=#0;
Result:=True;
end;


{---------------------------------------------}

const DeviceChannels: array [0..10,0..5] of TDeviceHandler =
	(
	{CHAN 0 FUNCTIONS, BUFFERED CONSOLE}
	(RESBUF,	{INTIALIZE INPUT}
	DUMMIO,		{INTIALIZE OUTPUT}
	BUFKEY,		{INPUT BYTE}
	CONOUT,		{OUTPUT BYTE}
	DUMMIO,		{CLOSE OUTPUT}
	DUMMIO),	{CLOSE INPUT}


{CHAN 1 FUNCTIONS, UNBUFFERED CONSOLE}

	(RESKEY,	{INTIALIZE INPUT}
	DUMMIO,		{INTIALIZE OUTPUT}
	KEYIN,		{INPUT BYTE}
	TVOUT,		{OUTPUT BYTE}
	DUMMIO,		{CLOSE OUTPUT}
	DUMMIO),	{CLOSE INPUT}


{CHAN 2 FUNCTIONS, PRINTER CHANNEL}

	(PrinterInitializeInput,	{INTIALIZE INPUT}
	PrinterInitializeOutput,	{INTIALIZE OUTPUT}
	DUMMIO,				{INPUT BYTE}
	PrinterOutput,			{OUTPUT BYTE}
	PrinterCloseOutput,		{CLOSE OUTPUT}
	DUMMIO),			{CLOSE INPUT}


{CHAN 3 FUNCTIONS, DISK I/O}

	(DOPENI,	{INTIALIZE INPUT}
	DOPENO,		{INTIALIZE OUTPUT}
	DSKIN,		{INPUT BYTE}
	DSKOUT,		{OUTPUT BYTE}
	DCLOSO,		{CLOSE OUTPUT}
	DUMMIO),	{CLOSE INPUT}


{CHAN 4 FUNCTIONS, SERIAL I/O}

	(SerialInitIn,	{INTIALIZE INPUT}
	SerialInitOut,	{INTIALIZE OUTPUT}
	SERIN,		{INPUT BYTE}
	SEROUT,		{OUTPUT BYTE}
	SERCLOSE,	{CLOSE OUTPUT}
	DUMMIO),	{CLOSE INPUT}


{CHAN 5 BIOS LEVEL PRINTER OUTPUT}

	(RawPrinterInitIn,	{INTIALIZE INPUT}
	RawPrinterInitOut,	{INTIALIZE OUTPUT}
	DUMMIO,			{INPUT BYTE}
	RawPrinterOutput,	{OUTPUT BYTE}
	RawPrinterCloseOut,	{CLOSE OUTPUT}
	DUMMIO),	   	{CLOSE INPUT}

{CHAN 6 TEXT WINDOWING FUNCTIONS}

	(DUMMIO,	{INTIALIZE INPUT}
	RESWND,		{INTIALIZE OUTPUT}
	WNDCHN,		{INPUT BYTE}
	WNDOUT,		{OUTPUT BYTE}
	DUMMIO,		{CLOSE OUTPUT}
	DUMMIO),	{CLOSE INPUT}


{CHAN 7 FUNCTIONS, NULL DEVICE}

	(DUMMIO,	{INTIALIZE INPUT}
	DUMMIO,		{INTIALIZE OUTPUT}
	DUMMIO,		{INPUT BYTE}
	DUMMIO,		{OUTPUT BYTE}
	DUMMIO,		{CLOSE OUTPUT}
	DUMMIO),	{CLOSE INPUT}


{CHAN 8 FUNCTIONS, PSEUDO DEVICE}

	(PSOPNI,	{INTIALIZE INPUT}
	PSOPNO,		{INTIALIZE OUTPUT}
	PSIN,		{INPUT BYTE}
	PSOUT,		{OUTPUT BYTE}
	DUMMIO,		{CLOSE OUTPUT}
	DUMMIO),	{CLOSE INPUT}

{CHAN 9 FUNCTIONS, PEN TEXT}

	(DUMMIO,	{INTIALIZE INPUT}
	RESWND,		{INTIALIZE OUTPUT}
	WNDCHN,		{INPUT BYTE}
	GraphicOut,	{OUTPUT BYTE}
	DUMMIO,		{CLOSE OUTPUT}
	DUMMIO),	{CLOSE INPUT}

{CHAN 10 FUNCTIONS, DEBUG DEVICE}

	(OpenDebugIn,	{INTIALIZE INPUT}
	OpenDebugOut,		{INTIALIZE OUTPUT}
	DebugIn,		{INPUT BYTE}
	DebugOut,		{OUTPUT BYTE}
	CloseDebugOut,		{CLOSE OUTPUT}
	CloseDebugIn)		{CLOSE INPUT}
	);


function KHAND(C: char): char;
var S: string;
begin
if NowFun=OUTFUN then OutputFlag:=True;
if not DeviceChannels[NowDev,NowFun](C) then
	begin
	{I/O ERROR}
	S:='I/O Error. Dev: '+IntToStr(NowDev)+' Func: '+GetIOFunctionStr(NowFun);
	HandleI2LError(S,ieIOError);
	Result:=C;
	end
else Result:=C;
end;


{----------------- Optimized KHand for strings ---------------------------------}




procedure ConsoleStringOut(var S: string);
{Output string to console Device}
var I: integer;
var S2: string;
begin
if RedirectString(S) then exit;
GlobalConsole.VGABox.UseAttribute:=True;
GlobalConsole.VGABox.ExpandTabs:=True;
EnableDev6Modes(False);
S2:='';
for I:=1 to Length(S) do
	begin
	case S[I] of
	 FF_Char:
		begin
		GlobalConsole.VGABox.StringOut(S2);
		S2:='';
		ClearConsole(False);
		end;
	 Bell_Char:
		begin
		GlobalConsole.VGABox.StringOut(S2);
		S2:='';
		Beep;
		end;
	 else S2:=S2+S[I];
	 end;
	end;
GlobalConsole.VGABox.StringOut(S2);
end;





procedure TVStringOut(var S: string);
{Output String to "TV" device}
begin
if RedirectString(S) then exit;
GlobalConsole.VGABox.ExpandTabs:=False;
GlobalConsole.VGABox.UseAttribute:=True;
EnableDev6Modes(False);
GlobalConsole.VGABox.StringOut(S);
end;




procedure WindowStringOut(var S: string);
{Output string in Windowed Mode}
begin
GlobalConsole.VGABox.ExpandTabs:=False;
GlobalConsole.VGABox.UseAttribute:=True;
EnableDev6Modes(True);
GlobalConsole.VGABox.StringOut(S);
end;



procedure KHandString(S: string);
{Specialized version of KHand exclusively for strings}
var I: integer;
begin
case NowDev of
 0: ConsoleStringOut(S);	{Chan 0 Buffered COnsole}
 1: TVStringOut(S);		{CHAN 1 Unbuffered Console}
 6: WindowStringOut(S);		{CHAN 6 Text Windowing Output}
 else for I:=1 to Length(S) do KHAND(S[I]);
 end;
end;




procedure CommandLineToDev8;
var I: integer;
begin
PseudoBuffer:='';
{Device 8 doesn't use the first argument, which is the program name}
for I:=1 to ComArgs.Count-1 do
	begin
	if I<>1 then PseudoBuffer:=PseudoBuffer+' ';
	PseudoBuffer:=PseudoBuffer+ComArgs[I];
	end;
PseudoBuffer:=PseudoBuffer+CRLF_Char;
PseudoInx:=1;
end;



procedure InitializeConsole;
begin
ResetDevice6;
GlobalConsole.ResetKeyinBuffer;
CommandLineToDev8;
end;


procedure AbortConsoleFunctions;
begin
Comport.AbortWait:=True;
Comport.PortActive:=False;
if GlobalConsole<>nil then GlobalConsole.AbortInput:=True;
end;



initialization
PrinterSetupDlg:=TPrinterSetupDialog.Create(nil);
CommSetupDlg:=TCommSetupDlg.Create(nil);
ComPort:=TCommPort.Create(nil);
finalization
ComPort.Free;
CommSetupDlg.Free;
PrinterSetupDlg.Free;
end.

