unit I2lIntrinsics;

interface

uses Forms,Windows,Messages,Graphics,Classes,Contnrs,SysUtils,Math,DSupport,WaveLib,Vector,
 MMSystem,BitmapLib,ConsoleWindow,LowDiskIO,WinSupport;

{Array to hold information about all open files. We use them to }
{make sure the all handles are closed when the program terminates}

type TFileInfo = record
 Name: string;
 Handle: integer;
 Mode: integer;
 end;

type TFileArray = array of TFileInfo;

var FileArray: TFileArray;

var ComArgs: TStringList;


procedure CallIntrinsic(Num: integer);
procedure InitializeIntrinsics(Console: TConsoleWindow);
procedure FinalizeIntrinsics;

procedure GetWinCommandLine;
procedure SetCommandLine(Strs: TStrings);

procedure GetCodeDeclaration(var S: string; Proto: boolean);
procedure GetPrototypeList(var SL: TStrings);
function NameToIntrinNumber(Name: string): integer;
function NameToProtoType(Name: string): string;
function GetProtoType(Index: integer; DoResult: boolean): string;
function TranslateColor(R,G,B: integer): TColor;

implementation

uses I2LGlobal,I2LDataTypes, I2LInterp, I2LConsoleIO, XPLDataTypes,
  I2L3DIntrinsics;


var BackupFlag: boolean;		{Flags that we re-read the last character}
var LastChar: Char;			{Last character read the the CHIN routine}

var HexDigits: cardinal = IntNibbles;	{Number of hex digits output by HexOut}





{------------------------------------------------------------------------------}



procedure AddFile(Name: string; Handle,Mode: integer);
{Insert File Info in dynamic array}
begin
SetLength(FileArray,Length(FileArray)+1);
FileArray[High(FileArray)].Name:=Name;
FileArray[High(FileArray)].Handle:=Handle;
FileArray[High(FileArray)].Mode:=Mode;
end;



procedure DeleteFile(Inx: integer);
{Delete file in dynamic array}
var I: integer;
begin
for I:=Inx to High(FileArray)-1 do FileArray[I]:=FileArray[I+1];
SetLength(FileArray,Length(FileArray)-1);
end;



procedure ClearFileArray;
{Clear dynamic array}
var I: integer;
begin
for I:=0 to High(FileArray) do FileClose(FileArray[I].Handle);
SetLength(FileArray,0);
end;



function FindFileHandle(Handle: integer): integer;
begin
for Result:=0 to High(FileArray) do
 if FileArray[Result].Handle=Handle then exit;
Result:=-1;
end;



{------------------------------------------------------------------------------}




procedure GetWinCommandLine;
var I: integer;
begin
ComArgs.Clear;
for I:=0 to ParamCount do ComArgs.Add(ParamStr(I));
end;



procedure SetCommandLine(Strs: TStrings);
begin
ComArgs.Assign(Strs);
end;


{---------------- Support Routines -------------------------------------------}


function GetFilteredChar: char;
begin
repeat Result:=KHAND(Result) until Result<>'_';
end;




function StripSignChars(S: string): string;
var MinusCnt,I: integer;
var Leading: boolean;
begin
Result:='';
Leading:=True;
for I:=1 to Length(S) do
 if Leading then
	begin
	Leading:=S[I] in ['+','-'];
	case S[I] of
	 '-': Inc(MinusCnt);
	 '+': ;
	 else Result:=Result+S[I];
	 end;
	end
 else Result:=Result+S[I];
if (MinusCnt and 1)<>0 then Result:='-'+Result;
end;




function GetInteger: integer;
var S: string;
var C: char;
var Num: boolean;
const Numerics = ['0'..'9','+','-'];
begin
S:='';
Num:=False;
while true do
	begin
	C:=GetFilteredChar;
	if C=CtrlZ then break;
	if Num and not (C in Numerics) then break
	else Num:=C in Numerics;
	if Num then S:=S+C;
	end;
S:=StripSignChars(S);
Result:=StrToIntDef(S,0);
end;



function GetReal: double;
{ Read a real from the input chanel }
var IP,FP,EP: String[128];
var C: char;
begin
IP:=''; FP:=''; EP:='';
{Get integer portion}
C:=GetFilteredChar;
while ((C<'0') or (C>'9')) and (C<>'.') do
	begin
	if C=EOF_Char then begin Result:=0; exit; end;
	if (C = '-') or (C = '+') then IP:=IP+C
	else IP:='';
	C:=GetFilteredChar;
	end;
{Get fractional portion}
IP:=StripSignChars(IP);
while (C>='0') and (C<='9') or (C='.') do
	begin
	FP:=FP+C;
	C:=GetFilteredChar;
	if C=EOF_Char then break;
	end;
{Get exponent portion}
if C in ['E','e'] then
	begin
        FP:=FP+C;
	C:=GetFilteredChar;
	while (C>='0') and (C<='9') or
	      (C='+') or (C='-') do
		begin
		EP:=EP+C;
		C:=GetFilteredChar;
		if C=EOF_Char then break;
		end;
	end;
EP:=StripSignChars(EP);
Result:=StrToFloat(IP+FP+EP);
end;



function GetHexNumber: integer;
var S: string[8];
var C: char;
var Num: boolean;
begin
S:='';
Num:=False;
while true do
	begin
	C:=GetFilteredChar;
	if C=CtrlZ then break;
	if Num and not IsHexDigit(C) then break
	else Num:=IsHexDigit(C);
	if Num then
		begin
		S:=S+C;
		{Don't read more than 4/8 characters}
		if Length(S)>=IntNibbles then Break;
		end;
	end;
Result:=StrToIntDef('$'+S,0);
end;



procedure StringOut(S: string);
var I: integer;
begin
KHandString(S);
end;



procedure IntegerOut(I: TI2LWord);
begin
StringOut(IntToStr(I));
end;



procedure RealOut(D: double);
begin
StringOut(GetXPLRealStr(D));
end;



procedure HexadecimalOut(I: TI2LWord);
begin
StringOut(IntToHex(I,HexDigits));
end;



function GetRawString(Inx: TI2lWord): string;
var I: integer;
var B: byte;
begin
Result:='';
for I:=Inx to ProgramStore.TotalSize-1 do
	begin
	B:=ProgramStore.DataByte[I];
	if (B=$A0) or (B=$00) then break;
	Result:=Result + Char(B);
	end;
end;


function WinExecAndWait(Path: String; var ExitCode: DWord): boolean;
{Execute the specified program and wait}
var Start: TStartupInfo;
var ProcessInfo: TProcessInformation;
begin
FillChar(Start,Sizeof(Start),0);
start.cb:=Sizeof(start);
Result:=CreateProcess(Nil,PChar(Path),Nil,Nil,True,NORMAL_PRIORITY_CLASS,Nil,Nil, Start, ProcessInfo);
if Result then
	begin
	try
	repeat Application.ProcessMessages
	until WaitForSingleObject(ProcessInfo.hProcess, 100) <> WAIT_TIMEOUT;
	finally
	 GetExitCodeProcess(ProcessInfo.hProcess, ExitCode);
	 CloseHandle(ProcessInfo.hProcess);
	 CloseHandle(ProcessInfo.hThread);
	 end;
	end;
end;


var EquipAddress: TI2LWord;	{Points to address of Equipment list}
var FontInfo: TI2LWord;		{Points to the address of font information}
var CursorInfo: TI2LWord;	{Point to the address of Cursor Information}

procedure SetupFontInfo;
begin
{Setup dummy info to build array -
real data will be provided by the GetFontInfo instrinic}
FontInfo:=ProgramStore.AddLocal(12);	{Width}
ProgramStore.AddLocal(34);		{Height}
end;


procedure SetupCursorInfo;
begin
{Setup dummy info to build array -
real data will be provided by the GetCursorInfo instrinic}
CursorInfo:=ProgramStore.AddLocal(XPLTrue);	{Cursor Visible}
ProgramStore.AddLocal($0D0E);			{Cursor size info - DOS style}
end;



procedure SetupEquipment;
var Freq: Int64;
var Mhz: integer;
begin
QueryPerformanceFrequency(Freq);
Mhz:=trunc(Freq/1E6);
{If Mhz=0 then it is an old pentium with a performance counter}
{In that case we'll assume it is a 1-Ghz machine}
if Mhz=0 then Mhz:=1000;
EquipAddress:=ProgramStore.AddLocal(586);
ProgramStore.AddLocal(387);
ProgramStore.AddLocal($1818);
ProgramStore.AddLocal(Mhz);

ProgramStore.AddLocal(30);
ProgramStore.AddLocal(byte('I'));
end;


procedure SetupLocalData;
{Setup local data area to be passed back intrinsics}
begin
ProgramStore.LocalPtr:=0;
SetupEquipment;
SetupFontInfo;
SetupCursorInfo;
end;


{------------- Intrinsics -------------------------------------------------}


procedure ABSFUN;
{0 ABSOLUTE VALUE}
begin
PushStack(Abs(PopStack));
end;



procedure RANFUN;
{1 RANDOM NUMBER}
var V: TI2LWord;
begin
V:=PopStack;
if V=0 then
	begin
	RandSeed:=2537;
	PushStack(0);
	end
else if V<0 then Randomize;
PushStack(Random(Abs(V)));
end;



procedure REMFUN;
{2 REMAINDER OF LAST DIVIDE}
begin
PopStack;
PushStack(Remainder);
end;



procedure RESERV;
{3 RESERVE ARRAY}
var R: integer;
begin
R:=PopStack;
PushStack(ProgramStore.HeapPtr);
ReserveMemory(R);
end;



procedure SWAPFUN;
{4 SWAP BYTES}
begin
PushStack(SwapByteArgs(PopStack));
end;



procedure EXTEN;
{5 EXTEND SIGN FROM LOW BYTE}
begin
PushStack(ShortInt(Lo(PopStack)));
end;



procedure RESTRT;
{6 RESTART CURRENT PROGRAM}
begin
InitializeInterpreter(GlobalConsole);
ReRunFlag:=True;
end;



procedure CHIN;
{7 INPUT A BYTE}
var C: char;
begin
NOWDEV:=PopStack;
NOWFUN:=INPFUN;
if BackupFlag then C:=LastChar
else C:=KHAND(C);
PushStack(TI2LWord(C));
LastChar:=C;
BackupFlag:=False;
end;



procedure CHOUT;
{8 OUTPUT A BYTE}
var C: char;
begin
C:=char(PopStack);
NOWDEV:=PopStack;
NOWFUN:=OUTFUN;
KHAND(C);
end;



procedure SKIP;
{9 NEW LINE (CRLF)}
begin
NOWDEV:=PopStack;
NOWFUN:=OUTFUN;
KHAND(CR_Char); KHAND(LF_Char);
end;





procedure INTIN;
{10 INPUT AN INTEGER}
begin
NOWDEV:=PopStack;
NOWFUN:=INPFUN;
PushStack(GetInteger);
end;



procedure INTOUT;
{11 OUTPUT AN INTEGER}
var I: TI2LWord;
begin
I:=PopStack;
NOWDEV:=PopStack;
NOWFUN:=OUTFUN;
IntegerOut(I);
end;



procedure TEXT;
{12 OUTPUT A STRING}
var Inx: TI2LWord;
var S: string;
begin
Inx:=PopStack;
NOWDEV:=PopStack;
NOWFUN:=OUTFUN;
S:=ProgramStore.GetDataString(Inx,UseMSBStrings);
StringOut(S);
end;



procedure IDEVIN;
{13 INITIALIZE INPUT DEVICE}
var C: char;
begin
NOWDEV:=PopStack;
NOWFUN:=INIFUN;
KHAND(C);
end;



procedure ODEVIN;
{14 INITIALIZE OUTPUT DEVICE}
var C: char;
begin
NOWDEV:=PopStack;
NOWFUN:=INOFUN;
KHAND(C);
end;



procedure ODEVCL;
{15 CLOSE AN OUTPUT DEVICE}
var C: char;
begin
NOWDEV:=PopStack;
NOWFUN:=CLOFUN;
KHAND(C);
end;




procedure AbortProg;
{16 ABORT PROGRAM (LIKE A CTRL-P)}
begin
abort;
end;



procedure TRAP;
{17 SET TRAP FLAG}
begin
Traps:=PopStack;
end;



procedure SPACE;
{18 DETERMINE REMAINING SPACE}
begin
PushStack(ProgramStore.HeapSize-ProgramStore.HeapPtr);
end;



procedure RERUN;
{19 TEST RERUN FLAG}
begin
PushXPLBoolean(ReRunFlag);
end;



procedure TSTHP;
{20 GET CURRENT HEAP POINTER}
begin
PushStack(ProgramStore.HeapPtr);
end;



procedure SETHP;
{21 SET HEAP POINTER}
begin
ProgramStore.HeapPtr:=PopStack;
end;



procedure GETERR;
{22 GET I2L ERROR NUMBER}
begin
PushStack(TI2LWord(ErrorNum));
ErrorNum:=TI2LError(0);
end;



procedure CURSOR;
{23 SET TV CURSOR POSITION}
var X,Y: TI2LWord;
begin
Y:=PopStack;
X:=PopStack;
GlobalConsole.VGABox.Cursor(X,Y);
end;



procedure FSET;
{24 SET A FILE HANDLE}
var Mode,Handle: TI2LWord;
begin
Mode:=PopStack;
Handle:=PopStack;
if char(Lo(Mode)) in ['o','O'] then OTHAND:=Handle
else INHAND:=Handle;
EOFFlag:=False;
end;



procedure SETRUN;
{25 SET THE RERUN FLAG}
begin
ReRunFlag:=PopStack<>0;
end;



procedure HEXI;
{26 INPUT HEX INTEGER}
begin
NOWDEV:=PopStack;
NOWFUN:=INPFUN;
PushStack(GetHexNumber);
end;




procedure HEXO;
{27 OUTPUT HEX INTEGER}
var I: TI2LWord;
begin
I:=PopStack;
NOWDEV:=PopStack;
NOWFUN:=OUTFUN;
HexadecimalOut(I);
end;




procedure CHAIN;
{28 CHAIN TO A EXTERNAL PROGRAM}
var Filename: string;
var ExitCode: DWord;
var Inx: TI2LWord;
begin
Inx:=PopStack;
Filename:=ProgramStore.GetDataString(Inx,UseMSBStrings);
WinExecAndWait(Filename,ExitCode);
end;




procedure FOPEN;
{29 OPEN NEW FILE}
var Mode: TI2LWord;
var Inx: TI2LWord;
var Filename: string;
var Handle: integer;
begin
Mode:=PopStack;
Inx:=PopStack;
Filename:=ProgramStore.GetDataString(Inx,UseMSBStrings);
case Mode of
 0: Handle:=FileOpen(Filename,fmOpenRead);
 1: Handle:=FileCreate(Filename);
 else HandleI2LError('Illegal File Attribute',ieIOError);
 end;
PushStack(Handle);
if Handle=-1 then HandleI2LError('Unable To Open File: '+ExtractFilename(Filename),ieIOError)
else AddFile(Filename,Handle,Mode);
end;



procedure FWRITE;
{30 WRITE DISK BLOCKS}
{**** Currently Unimplemented}
var SectorCnt,Buffer,Start,Drive: TI2LWord;
begin
{Dummy function - get arguments and do nothing}
SectorCnt:=PopStack;
Buffer:=PopStack;
Start:=PopStack;
Drive:=PopStack;
end;



procedure FREAD;
{31 READ DISK BLOCKS}
var SectorCnt,Buffer,Start,Drive: TI2LWord;
var Data: TSector;
var I,J: integer;
begin
{Dummy function - get arguments and do nothing}
SectorCnt:=PopStack;
Buffer:=PopStack;
Start:=PopStack;
Drive:=PopStack;
for I:=0 to SectorCnt-1 do
	begin
	LowReadSector(char(Drive + $41), Start+I,Data);
	for J:=0 to 512 do ProgramStore.DataByte[Buffer+J]:=Data[J];
	end;
end;




procedure FCLOSE;
{32 CLOSE DISK FILE}
var Handle: integer;
var Inx: integer;
begin
Handle:=PopStack;
FileClose(Handle);
Inx:=FindFileHandle(Handle);
if Inx>=0 then DeleteFile(Inx);
end;



procedure CHKKEY;
{33 CHECK FOR KEY STRIKE}
begin
Application.ProcessMessages;
if GlobalConsole.KeyPressed then PushStack(XPLTrue)
else PushStack(XPLFalse);
end;



procedure SFTINT;
{34 DO SOFTWARE INTERRUPT}
{**** Currently Unimplemented}
var Interrupt: TI2LWord;
begin
Interrupt:=PopStack;
end;



procedure GETREG;
{35 GET REGISTER ADDRESS}
{**** Currently Unimplemented}
begin
PushStack(0);
end;



procedure BLKTRN;
{36 BLOCK TRANSFER}
{**** Currently Unimplemented}
var Size,ToAdd,ToSeg,FromAdd,FromSeg: TI2LWord;
begin
Size:=PopStack;
ToAdd:=PopStack;
ToSeg:=PopStack;
FromAdd:=PopStack;
FromSeg:=PopStack;
end;



procedure PEEK;
{37 PEEK MEMORY}
{**** Currently Unimplemented}
var Add,Seg: TI2LWord;
begin
Add:=PopStack;
Seg:=PopStack;
PushStack(0);
end;



procedure POKE;
{38 POKE INTO MEMORY}
{**** Currently Unimplemented}
var Add,Seg,Value: TI2LWord;
begin
Add:=PopStack;
Seg:=PopStack;
Value:=PopStack;
end;



procedure SOUND;
{39 SQUEAK THE SPEAKER}
var Freq,Len,SoundOn: TI2LWord;
var RF,Sec: double;
begin
Freq:=PopStack;
Len:=PopStack;
SoundOn:=PopStack;
RF:=1119000/Freq;
Sec:=Len/18.2;
if SoundOn<>0 then SynthesizeSound8(RF,Sec,True)
else DelayProcessMessages(round(Sec*1000));
end;



procedure VCLEAR;
{40 CLEAR VIDEO DISPLAY}
begin
ClearConsole(True);
end;



procedure DOT;
{41 PLOT HI-RES POINT}
var Color,Y,X: TI2LWord;
begin
Color:=PopStack;
Y:=PopStack;
X:=PopStack;
GlobalConsole.VGABox.SetLineStyle(Color);
GlobalConsole.VGABox.DrawPoint(X,Y,GlobalConsole.VGABox.VGAIndexToColor(Color));
end;



procedure LINE;
{42 DRAW A HI-RES LINE}
var Color,Y,X: TI2LWord;
begin
Color:=PopStack;
Y:=PopStack;
X:=PopStack;
GlobalConsole.VGABox.SetLineStyle(Color);
GlobalConsole.VGABox.LineTo(X,Y,GlobalConsole.VGABox.VGAIndexToColor(Color));
GlobalConsole.VGABox.DrawPoint(X,Y,GlobalConsole.VGABox.VGAIndexToColor(Color));
end;



procedure MOVE;
{43 HI-RES MOVE}
var Y,X: TI2LWord;
begin
Y:=PopStack;
X:=PopStack;
GlobalConsole.VGABox.MoveTo(X,Y);
end;



procedure REDDOT;
{44 READ BACK PIXEL}
var Color,Y,X: TI2LWord;
begin
Y:=PopStack;
X:=PopStack;
Color:=GlobalConsole.VGABox.Canvas.Pixels[X,Y];
PushStack(GlobalConsole.VGABox.ColorToVGAIndex(Color));
end;



procedure SETVID;
{45 SET VIDEO MODE}
begin
GlobalConsole.VGABox.ResetWindowMode;
GlobalConsole.VGABox.VGAMode:=PopStack;
WinCorners:=GlobalConsole.VGABox.ClipWindow;
end;



procedure FLRES;
{46 RESERVE MEMORY FOR REALS}
var I,Size: integer;
begin
Size:=PopStack;
PushStack(ProgramStore.HeapPtr);
{Storing integer into a real - so pad rest of real with zero}
for I:=0 to IntsPerReal-2 do PushStack(0);
ReserveMemory(Size * RealSize);
end;




procedure FLIN;
{47 INPUT A REAL NUMBER}
begin
NOWDEV:=PopStack;
NOWFUN:=INPFUN;
PushDouble(GetReal);
end;



procedure FLOUT;
{48 OUTPUT A REAL NUMBER}
var D: double;
begin
D:=PopDouble;
NOWDEV:=PopStack;
NOWFUN:=OUTFUN;
RealOut(D);
end;



procedure FLTFUN;
{49 CONVERT INTEGER TO REAL}
begin
PushDouble(PopStack);
end;



procedure FIXFUN;
{50 CONVERT REAL TO INTEGER}
var D: double;
begin
D:=PopDouble;
PushStack(round(D));
if (D<Low(TI2LWord)) or (D>High(TI2LWord)) then  HandleI2LError('Fix Overflow',ieFixError);
end;



procedure FLABS;
{51 ABSOLUTE VALUE OF REAL}
begin
PushDouble(abs(PopDouble));
end;



procedure FMTFUN;
{52 PLACES BEFORE & AFTER DEC POINT}
var I1,I2: integer;
begin
I2:=PopStack;
I1:=PopStack;
RealFormat(I1,I2);
end;




procedure FLSQRT;
{53 TOS:=SQRT(TOS)}
var D: double;
begin
D:=PopDouble;
PushDouble(sqrt(D));
if D<0 then HandleI2LError('Sqrt < 0',ieSquareroot);
end;



procedure FLLN;
{54 TOS:=LN(TOS)}
var D: double;
begin
D:=PopDouble;
PushDouble(Ln(D));
if D<=0 then HandleI2LError('LOG <= 0',ieLogError);
end;



procedure FLEXP;
{55 TOS:=EXP(TOS)}
var D: double;
begin
D:=PopDouble;
PushDouble(Exp(D));
if (Get8087Status(True) and $1C)<>0 then HandleI2LError('EXP Overflow',ieExpError);
end;



procedure FLSIN;
{56 TOS:=SIN(TOS)}
begin
PushDouble(sin(PopDouble));
end;



procedure FLAT2;
{57 TOS:=ATAN(NOS/TOS)}
var Tos,Nos: double;
begin
Tos:=PopDouble;
Nos:=PopDouble;
PushDouble(Math.ArcTan2(Nos,Tos));
if (Tos=0) and (Nos=0) then HandleI2LError('ATAN2(0,0)',ieArcTan);
end;



procedure FLMOD;
{58 TOS:= NOS MOD TOS}
var Tos,Nos: double;
begin
Tos:=PopDouble;
Nos:=PopDouble;
PushDouble(FloatMod(Nos,Tos));
end;



procedure FLLOG;
{59 TOS:=LOG(TOS)}
var D: double;
begin
D:=PopDouble;
PushDouble(log(D));
if D<=0 then HandleI2LError(' LOG <= 0',ieLogError);
end;



procedure FLCOS;
{60 TOS:=COS(TOS)}
begin
PushDouble(cos(PopDouble));
end;



procedure FLTAN;
{61 TOS:=TAN(TOS)}
begin
PushDouble(tan(PopDouble));
end;



procedure FLASIN;
{62 TOS:=ASIN(TOS)}
begin
PushDouble(ArcSin(PopDouble));
end;



procedure FLACOS;
{63 TOS:=ACOS(TOS)}
begin
PushDouble(ArcCos(PopDouble));
end;



procedure FloorFunction;
{64 TOS:=Floor(TOS)}
begin
PushDouble(Floor(PopDouble));
end;



procedure CeilFunction;
{65 TOS:=Ceil(TOS)}
begin
PushDouble(Ceil(PopDouble));
end;



procedure PowerFunction;
{66 EXIT TOS:=Power(NOS,TOS)}
var Tos,Nos: double;
begin
Tos:=PopDouble;
Nos:=PopDouble;
PushDouble(Power(Nos,Tos));
end;


procedure EXTJMP;
{67 JUMP TO FAR ROUTINE}
{**** Currently Unimplemented}
var Add,Seg: TI2LWord;
begin
Add:=PopStack;
Seg:=PopStack;
end;




procedure EXTCAL;
{68 CALL FAR ROUTINE}
{**** Currently Unimplemented}
var Add,Seg: TI2LWord;
begin
Add:=PopStack;
Seg:=PopStack;
end;



procedure  SETATT;
{69 SET WINDOW ATTRIBUTE}
begin
GlobalConsole.VGABox.Attribute:=PopStack;
end;




procedure SETWND;
{70 SET DIMENSIONS AND MODE}
var Fill: TI2LWord;
begin
Fill:=PopStack;
Dev6Mode:=PopStack;
WinCorners.BottomRight.Y:=PopStack;
WinCorners.BottomRight.X:=PopStack;
WinCorners.TopLeft.Y:=PopStack;
WinCorners.TopLeft.X:=PopStack;
GlobalConsole.VGABox.Cursor(WinCorners.TopLeft.X,WinCorners.TopLeft.Y);
if Fill<>0 then GlobalConsole.VGABox.ClearBlock(WinCorners);
if (Dev6Mode and $FF00)<>0 then GlobalConsole.VGABox.ShowCursor:=False;
end;



procedure RAWTXT;
{71 OUTPUT RAW TEXT STRING}
var Inx: TI2LWord;
var S: string;
begin
Inx:=PopStack;
NOWDEV:=PopStack;
NOWFUN:=OUTFUN;
S:=GetRawString(Inx);
StringOut(S);
end;



procedure HLIGHT;
{72 HIGHLIGHT PART OF SCREEN}
var Fill: TI2LWord;
var X1,Y1,X2,Y2: integer;
begin
Fill:=PopStack;
Y2:=PopStack; X2:=PopStack;
Y1:=PopStack; X1:=PopStack;
GlobalConsole.VGABox.HighlightText(X1,Y1,X2,Y2,Fill);
//GlobalConsole.Invalidate;
end;




procedure MALLOC;
{73 Allocate Memory from OS}
var Size,Seg: cardinal;
begin
Size:=PopStack shl 4;
if Size>MaxInteger then raise Exception.Create('Malloc Exceeds Addressable Space: '+FloatToStrF(Size,ffNumber,18,0));
Seg:=MemManager.Allocate(Size);
PushStack(Seg);
end; 




procedure RELEAS;
{74 RELEASE MEMORY TO DOS}
var Seg: TI2LWord;
begin
Seg:=PopStack;
MemManager.Deallocate(Seg);
end;



procedure TRAPC;
{75 TURN ON OR OFF CTR-C TRAP}
begin
TrapControlC:=PopStack<>0;
end;




procedure TESTC;
{76 TEST FOR RECENT CTR-C}
begin
PushXPLBoolean(ControlCHit);
ControlCHit:=False;
end;



procedure GETEQP;
{77 GET EQUIPMENT LIST}
begin
PushStack(ProgramStore.LocalLow+EquipAddress);
end;



procedure SHRINK;
{78 SHRINK MEMORY ALLOCATION}
{**** Currently Unimplemented}
var Keep: TI2LWord;
begin
Keep:=PopStack;
end;



procedure RANSEED;
{79 SET RANDOM NUMBER SEED}
begin
RandSeed:=PopStack;
end;



procedure IRQ;
{80 TURN ON OR OFF INTERRUPTS}
{**** Currently Unimplemented}
var Flag: TI2LWord;
begin
Flag:=PopStack;
end;


procedure SetCloseOnTerminate;
{81 Set the CloseOnTerminate Flag}
begin
CloseOnTerminate:=PopStack<>0;
end;




procedure SetHexDigits;
{84 Set the SetHexDigits Flag}
begin
HexDigits:=PopStack;
end;



{------------------ Additional Non-traditional Intrinsics ---------------------}


function TranslateColor(R,G,B: integer): TColor;
{Translate color if R<0 then G is taken as VGA color Index}
{Otherwise the color is taken to be normal R,G,B}
begin
if R<0 then Result:=GlobalConsole.VGABox.VGAIndexToColor(G)
else Result:=RGB(R,G,B);
end;


procedure HandleConsoleError(E: TObject);
{Handle exceptions generated by the Console window and pass them through}
var EMsg: string;
begin
if E is Exception then EMsg:=Exception(E).Message else EMsg:='';
HandleI2LError(EMsg,ieIOError);
end;


procedure GetTime;
{82 - Get milliseconds Tick count}
{Prototype: TickCount:=GetTime; }
begin
PushStack(GetHiResTick and MaxInteger);
end;


procedure Backup;
{83 - Set backup flag, so we re-read last character }
{Prototype: Backup; }
begin
BackupFlag:=True;
end;




procedure ReallocMem;
{85 Reallocate Memory from OS}
var Size,Inx: cardinal;
begin
Size:=PopStack shl 4;
Inx:=PopStack;
if Size>MaxInteger then raise Exception.Create('Malloc Exceeds Addressable Space: '+FloatToStrF(Size,ffNumber,18,0));
Inx:=MemManager.Reallocate(Inx,Size);
PushStack(Inx);
end;




procedure ShowCursor;
{88 - Set cursor visibility}
{Prototype: ShowCursor(Visible)}
begin
GlobalConsole.VGABox.ShowCursor:=PopStack<>0;
end;



procedure GetCursorInfo;
{89 - Show set cursor visibility}
{Prototype: GetCursorAttributes(addr X, addr Y; addr Attrib)}
var X,Y,Attrib: TI2LWord;
begin
Attrib:=PopStack;
Y:=PopStack;
X:=PopStack;

ProgramStore.InsertLocal(CursorInfo,GetXPLBoolean(GlobalConsole.VGABox.ShowCursor));
ProgramStore.InsertLocal(CursorInfo+IntBytes,$0D0E);
ProgramStore.SetDataWord(X,GlobalConsole.VGABox.CursorX);
ProgramStore.SetDataWord(Y,GlobalConsole.VGABox.CursorY);
ProgramStore.SetDataWord(Attrib,ProgramStore.LocalLow+CursorInfo);
end;




procedure GetCommandLine;
{90: Returns command line information}
{Prototype: Valid:=CommandLine(Index,String,MaxLength); }
var Index,StrInx,MaxLen: integer;
var S: string;
var Valid: boolean;
begin
MaxLen:=PopStack;
StrInx:=PopStack;
Index:=PopStack;
Valid:=(Index>=0) and (Index<ComArgs.Count);
if Valid then
	begin
	S:=ComArgs[Index];
	if Length(S)>MaxLen then SetLength(S,MaxLen);
	ProgramStore.SetDataString(S,StrInx);
	end;
if Valid then PushStack(XPLTrue)
else  PushStack(XPLFalse);
end;



procedure GetFileInfo;
{91: Returns file attributes}
{Prototype: GetFileInfo(Handle, addr Size); }
var Handle,SizeInx: integer;
var Size,High: cardinal;
begin
SizeInx:=PopStack;
Handle:=PopStack;
Size:=GetFileSize(Handle,@High);
ProgramStore.SetDataWord(SizeInx,Size);
ProgramStore.SetDataWord(SizeInx+2,Size shr 16);
ProgramStore.SetDataWord(SizeInx+4,High);
ProgramStore.SetDataWord(SizeInx+6,High shr 16);
end;



procedure GetPathInfo;
{92: Returns default path}
{Prototype: GetPathInfo(Drive,String,MaxLength); }
var Drive,StrInx,MaxLen: integer;
var S: string;
begin
MaxLen:=PopStack;
StrInx:=PopStack;
Drive:=PopStack;
GetDir(Drive,S);
if Length(S)>MaxLen then SetLength(S,MaxLen);
ProgramStore.SetDataString(S,StrInx);
end;




procedure SetPathInfo;
{93: Specifies the default path}
{Prototype: SetPathInfo(String); }
var StrInx: integer;
var S: string;
begin
StrInx:=PopStack;
S:=ProgramStore.GetDataString(StrInx,UseMSBStrings);
ChDir(S);
end;




procedure DelayUS;
{94 - Delay a specified number of Microseconds}
{Prototype: DelayUS(US); }
var US: TI2LWord;
begin
US:=PopStack;
HiResDelayProcessMessages(US);
end;



procedure GetParams;
{95: Returns the specified System Parameter}
{Prototype: GetParams(MainType,SubType); }
var SubType,MainType: integer;
begin
SubType:=PopStack;
MainType:=PopStack;
case MainType of
 0: PushStack(GetSerialParams(SubType));
 end;
end;




procedure SetParams;
{96: Sets the Specified System Parameter}
{Prototype: SetParams(MainType,SubType,Value); }
var Value,SubType,MainType: integer;
begin
Value:=PopStack;
SubType:=PopStack;
MainType:=PopStack;
case MainType of
 0: SetSerialParams(SubType, Value);
 end;
end;





procedure GetKeyState;
{98 - Get the state of the specified key}
{Prototype: }
var Key,State: TI2LWord;
var R: smallint;
begin
Key:=PopStack;
R:=GetAsyncKeyState(Key);
State:=SetClearBits(IsBitSet(R,$8000),0,1);
State:=SetClearBits(IsBitSet(R,$0001),State,2);
PushStack(State);
end;




procedure SetCharAttrib;
{99 - Set character and attribute under cursor}
{Prototype: SetScreen(X,Y,C); }
var X,Y,C: TI2LWord;
begin
C:=PopStack;
Y:=PopStack;
X:=PopStack;
GlobalConsole.VGABox.SetCharAttrib(X,Y,C);
OutputFlag:=True;
end;



procedure CursorChar;
{100 - Get character and attribute under cursor}
{Prototype: Addribute:=CursorChar(X,Y); }
var X,Y: TI2LWord;
begin
Y:=PopStack;
X:=PopStack;
with GlobalConsole.VGABox do
 begin
 if Y>High(TextArray[0]) then Y:=0;
 if X>High(TextArray) then X:=0;
 PushStack(TextArray[X,Y]);
 end;
end;



procedure DelayMS;
{101 - Delay a specified number of Milliseconds}
{Prototype: Delay(MS); }
var MS: TI2LWord;
begin
MS:=PopStack;
DelayProcessMessages(MS);
end;



procedure MultiplyDivide;
{102 - Perform 32-bit X*Y/Z}
{Prototype:  MulDiv(X,Y,Z) }
var X,Y,Z: TI2LWord;
begin
Z:=PopStack; Y:=PopStack; X:=PopStack;
PushStack(MulDiv(X,Y,Z));
end;



procedure GetDateTime;
{103 Read system time}
{Prototype: GetSysTime(GMT,addr Year,addr Month,addr Day,addr WeekDay,addr Hour,addr Min,addr Sec,addr MS); }
var SystemTime: TSystemTime;
var Year,Month,WeekDay,Day,Hour,Minute,Second,MSec,GMT: TI2LWord;
begin
MSec:=PopStack;
Second:=PopStack;
Minute:=PopStack;
Hour:=PopStack;
WeekDay:=PopStack;
Day:=PopStack;
Month:=PopStack;
Year:=PopStack;
GMT:=PopStack;
if GMT<>0 then GetSystemTime(SystemTime)
else GetLocalTime(SystemTime);

ProgramStore.SetDataWord(MSec,SystemTime.wMilliseconds);
ProgramStore.SetDataWord(Second,SystemTime.wSecond);
ProgramStore.SetDataWord(Minute,SystemTime.wMinute);
ProgramStore.SetDataWord(Hour,SystemTime.wHour);
ProgramStore.SetDataWord(Day,SystemTime.wDay);
ProgramStore.SetDataWord(WeekDay,SystemTime.wDayOfWeek);
ProgramStore.SetDataWord(Month,SystemTime.wMonth);
ProgramStore.SetDataWord(Year,SystemTime.wYear);
end;



procedure GetScancode;
{104: Read scancode from keyboard}
{Prototype: KeyPressed:=GetScanCode(addr RawCode,addr VirtualCode,addr BIOSCode,Wait);}
var RawInx,VirtInx,DOSInx,Wait: TI2LWord;
var RawCode,VirtualCode,BIOSCode: integer;
begin
Wait:=PopStack;

DOSInx:=PopStack;
VirtInx:=PopStack;
RawInx:=PopStack;
PushXPLBoolean(GlobalConsole.GetScanCode(RawCode,VirtualCode,BIOSCode,Wait<>0));
ProgramStore.SetDataWord(RawInx,RawCode);
ProgramStore.SetDataWord(VirtInx,VirtualCode);
ProgramStore.SetDataWord(DOSInx,BIOSCode);
end;




procedure GetMouseState;
{105 Read current mouse state}
{Prototype: GetMouseState(addr State, addr X, addr Y);}
var AddState,AddX,AddY: TI2LWord;
var State: TI2LWord;
begin
Application.ProcessMessages;
AddY:=PopStack;
AddX:=PopStack;
AddState:=PopStack;
State:=0;
with GlobalConsole do
 begin
 if MouseState.Left then State:=SetBits(State,1) else State:=ClearBits(State,1);
 if MouseState.Right then State:=SetBits(State,2) else State:=ClearBits(State,2);
 if MouseState.Middle then State:=SetBits(State,4) else State:=ClearBits(State,4);
 if MouseState.Click then State:=SetBits(State,8) else State:=ClearBits(State,8);
 if MouseState.DoubleClick then State:=SetBits(State,16) else State:=ClearBits(State,16);
 ProgramStore.SetDataWord(AddState,State);
 ProgramStore.SetDataWord(AddX,MouseState.Pos.X);
 ProgramStore.SetDataWord(AddY,MouseState.Pos.Y);
 MouseState.Click:=False;
 MouseState.DoubleClick:=False;
 end;
end;




procedure GetLastKey;
{106 Read last key press}
{Prototype: Char:=GetLastKey;}
var Clear: boolean;
begin
Clear:=PopStack<>0;
PushStack(byte(GlobalConsole.GetLastKey(Clear)));
end;



procedure GetVideoMode;
{107 Read current video mode value}
{Prototype: Mode:=GetVideoMode;}
begin
PushStack(GlobalConsole.VGABox.VGAMode);
end;


procedure SetPalette;
{108 Set palette }
{Prototype: SetPalette(Inx,R,G,B); }
var Inx,R,G,B: TI2LWord;
begin
B:=PopStack; G:=PopStack; R:=PopStack;
Inx:=PopStack;
GlobalConsole.VGABox.SetPaletteItem(Inx,TranslateColor(R,G,B));
GlobalConsole.Invalidate;
end;



procedure TimeSlice;
{109 Set Relinquish current time slice operating system }
{Prototype: TimeSlice(Delay); }
begin
Sleep(PopStack);
Application.ProcessMessages;
end;



{------------------------------------------------------------------------------}




procedure AllocateBitmap;
{110 Allocate bitmap and return handle}
{Prototype: Handle:=AllocateBitmap; }
begin
try
PushStack(GlobalConsole.VGABox.BitmapManager.Allocate);
except HandleConsoleError(ExceptObject); end;
end;



procedure DeAllocateBitmap;
{111 deallocate bitmap handle}
{Prototype: DeAllocateBitmap(Handle); }
var Handle: TI2LWord;
begin
try
Handle:=PopStack;
GlobalConsole.VGABox.BitmapManager.DeAllocate(Handle);
except HandleConsoleError(ExceptObject); end;
end;



procedure LoadBitmap;
{112 Load bitmap from file}
{Prototype: SaveBitmap(Handle,Filename); }
var Filename: string;
var Handle: TI2LWord;
var Inx: TI2LWord;
begin
try
Inx:=PopStack;
Handle:=PopStack;
Filename:=ProgramStore.GetDataString(Inx,UseMSBStrings);
GlobalConsole.VGABox.BitmapManager.LoadBitmap(Handle,Filename);
except HandleConsoleError(ExceptObject); end;
end;


procedure SaveBitmap;
{113 Save bitmap to a file}
{Prototype: SaveBitmap(Handle,Filename); }
var Filename: string;
var Handle: TI2LWord;
var Inx: TI2LWord;
begin
try
Inx:=PopStack;
Handle:=PopStack;
Filename:=ProgramStore.GetDataString(Inx,UseMSBStrings);
GlobalConsole.VGABox.BitmapManager.SaveBitmap(Handle,Filename);
except HandleConsoleError(ExceptObject); end;
end;



procedure CopyImage;
{114 Copy Image}
{Prototype:  CopyImage(SourceHandle,Left,Top,Right,Bottom,DestHandle,Left,Top,Right,Bottom); }
var SourceHandle, DestHandle: TI2LWord;
var SourceRect, DestRect: TRect;
begin
try
DestRect.Bottom:=PopStack;
DestRect.Right:=PopStack;
DestRect.Top:=PopStack;
DestRect.Left:=PopStack;
DestHandle:=PopStack;
SourceRect.Bottom:=PopStack;
SourceRect.Right:=PopStack;
SourceRect.Top:=PopStack;
SourceRect.Left:=PopStack;
SourceHandle:=PopStack;
GlobalConsole.VGABox.BitmapManager.CopyImage(DestHandle,SourceHandle,DestRect,SourceRect);
if DestHandle=-1 then OutputFlag:=True;
except HandleConsoleError(ExceptObject); end;
end;


procedure PaintImage;
{115 Paint Image}
{Prototype:  PaintImage(SourceHandle,DestHandle,Left,Top); }
var SourceHandle, DestHandle: TI2LWord;
var Left,Top: integer;
begin
try
Top:=PopStack;
Left:=PopStack;
DestHandle:=PopStack;
SourceHandle:=PopStack;
GlobalConsole.VGABox.BitmapManager.DrawImage(DestHandle,SourceHandle,Left,Top);
if DestHandle=-1 then OutputFlag:=True;
except HandleConsoleError(ExceptObject); end;
end;



procedure SetCopyMode;
{116 Set bitmap copy mode}
{Prototype:  SetCopyMode(Mode,Alpha,R,G,B); }
var Mode,Alpha,R,G,B: TI2LWord;
begin
try
B:=PopStack; G:=PopStack; R:=PopStack;
Alpha:=PopStack;
Mode:=PopStack;
GlobalConsole.VGABox.BitmapManager.CopyMode.TransColor:=TranslateColor(R,G,B);
GlobalConsole.VGABox.BitmapManager.CopyMode.Mode:=Mode;
GlobalConsole.VGABox.BitmapManager.CopyMode.Alpha:=Alpha;
GlobalConsole.VGABox.BitmapManager.CopyMode.PerPixel:=(Alpha and $100)<>0;
except HandleConsoleError(ExceptObject); end;
end;



procedure GetBitmapSize;
{117 Get bitmap size}
{Prototype:  GetBitmapSize(Handle,addr X,addr Y); }
var Handle,X,Y: TI2LWord;
var Width,Height: integer;
begin
try
Y:=PopStack;
X:=PopStack;
Handle:=PopStack;
GlobalConsole.VGABox.BitmapManager.GetBitmapSize(Handle,Width,Height);
ProgramStore.SetDataWord(X,Width);
ProgramStore.SetDataWord(Y,Height);
except HandleConsoleError(ExceptObject); end;
end;



procedure SetBitmapSize;
{118 Set Bitmap Size}
{Prototype:  SetBitmapSize(Handle,X,Y); }
var Handle,X,Y: TI2LWord;
begin
try
Y:=PopStack;
X:=PopStack;
Handle:=PopStack;
GlobalConsole.VGABox.BitmapManager.SetBitmapSize(Handle,X,Y);
except HandleConsoleError(ExceptObject); end;
end;




procedure SetPixel;
{119 Set pixel color}
{Prototype: SetPixel(Handle,X,Y,R,G,B); }
var Handle,X,Y,R,G,B: TI2LWord;
begin
try
B:=PopStack; G:=PopStack; R:=PopStack;
Y:=PopStack; X:=PopStack;
Handle:=PopStack;
GlobalConsole.VGABox.BitmapManager.SetPixel(Handle,X,Y,TranslateColor(R,G,B));
OutputFlag:=True;
except HandleConsoleError(ExceptObject); end;
end;



procedure GetPixel;
{120 Set pixel color}
{Prototype: GetPixel(Handle,X,Y,addr R,addr G,addr B); }
var C: TColor;
var Handle,X,Y,R,G,B: TI2LWord;
begin
try
B:=PopStack; G:=PopStack; R:=PopStack;
Y:=PopStack; X:=PopStack;
Handle:=PopStack;
C:=GlobalConsole.VGABox.BitmapManager.GetPixel(Handle,X,Y);
ProgramStore.SetDataWord(R,GetRValue(C));
ProgramStore.SetDataWord(G,GetGValue(C));
ProgramStore.SetDataWord(B,GetBValue(C));
except HandleConsoleError(ExceptObject); end;
end;



procedure ClearImage;
{121 Clear image}
{Prototype: ClearImage(Handle); }
var Handle,R,G,B: TI2LWord;
begin
try
B:=PopStack; G:=PopStack; R:=PopStack;
Handle:=PopStack;
GlobalConsole.VGABox.BitmapManager.ClearImage(Handle,TranslateColor(R,G,B));
except HandleConsoleError(ExceptObject); end;
end;



procedure DrawCircle;
{122 Draw Circle}
{Prototype: DrawCircle(Hanle,X,Y,Radius); }
var Handle,X,Y,Radius: TI2LWord;
begin
try
Radius:=PopStack;
Y:=PopStack;
X:=PopStack;
Handle:=PopStack;
GlobalConsole.VGABox.BitmapManager.DrawCircle(Handle,X,Y,Radius);
except HandleConsoleError(ExceptObject); end;
end;



procedure DrawEllipse;
{123 Draw Elipse}
{Prototype: DrawElipse(Handle,X1,Y1,X2,Y2); }
var Handle,X1,Y1, X2, Y2: TI2LWord;
begin
try
Y2:=PopStack;
X2:=PopStack;
Y1:=PopStack;
X1:=PopStack;
Handle:=PopStack;
GlobalConsole.VGABox.BitmapManager.DrawEllipse(Handle,X1,Y1,X2,Y2);
except HandleConsoleError(ExceptObject); end;
end;


procedure DrawRectangle;
{124 Draw Rectangle}
{Prototype: DrawRectangle(Handle,X1,Y1,X2,Y2); }
var Handle,X1,Y1, X2, Y2: TI2LWord;
begin
try
Y2:=PopStack;
X2:=PopStack;
Y1:=PopStack;
X1:=PopStack;
Handle:=PopStack;
GlobalConsole.VGABox.BitmapManager.DrawRectangle(Handle,X1,Y1,X2,Y2);
except HandleConsoleError(ExceptObject); end;
end;


procedure MoveTo;
{125 Move pen position}
{Prototype: MoveTo(Handle,X,Y); }
var Handle,X,Y: TI2LWord;
begin
try
Y:=PopStack;
X:=PopStack;
Handle:=PopStack;
GlobalConsole.VGABox.BitmapManager.MoveTo(Handle,X,Y);
except HandleConsoleError(ExceptObject); end;
end;



procedure Lineto;
{126 Draw line to new position}
{Prototype: Lineto(Handle,X,Y); }
var Handle,X,Y: TI2LWord;
begin
try
Y:=PopStack;
X:=PopStack;
Handle:=PopStack;
GlobalConsole.VGABox.BitmapManager.LineTo(Handle,X,Y);
except HandleConsoleError(ExceptObject); end;
end;



procedure Polyline;
{127 Draw polylines }
{Prototype: Polyline(Handle,Mode,addr Points, Count); }
var Handle,Mode,Points,Count: TI2LWord;
var PointArray: array of TPoint;
var I,Inx: integer;
begin
try
Count:=PopStack;
Points:=PopStack;
Mode:=PopStack;
Handle:=PopStack;
SetLength(PointArray,Count);
for I:=0 to High(PointArray) do
	begin
	Inx:=Points+I*IntBytes * 2;
	PointArray[I].X:=I2LProgram.GetDataWord(Inx);
	PointArray[I].Y:=I2LProgram.GetDataWord(Inx+IntBytes);
	end;
GlobalConsole.VGABox.BitmapManager.Polyline(Handle,Mode,PointArray);
except HandleConsoleError(ExceptObject); end;
end;



procedure GetPenPosition;
{128 Get pen's current position}
{Prototype: GetPenPosition(Handle,addr X,addr Y); }
var Handle,X,Y: TI2LWord;
var P: TPoint;
begin
try
Y:=PopStack; X:=PopStack;
Handle:=PopStack;
P:=GlobalConsole.VGABox.BitmapManager.GetPenPosition(Handle);
ProgramStore.SetDataWord(X,P.X);
ProgramStore.SetDataWord(Y,P.Y);
except HandleConsoleError(ExceptObject); end;
end;



procedure SetFill;
{129 Set fill mode and color}
{Prototype: SetFill(Handle,Mode,R,G,B) }
var Handle,Mode,R,G,B: TI2LWord;
begin
try
B:=PopStack; G:=PopStack; R:=PopStack;
Mode:=PopStack;
Handle:=PopStack;
GlobalConsole.VGABox.BitmapManager.SetFill(Handle,Mode,TranslateColor(R,G,B));
except HandleConsoleError(ExceptObject); end;
end;



procedure SetPen;
{130 Set pen style, color and mode}
{Prototype: SetPen(Hande,Mode,Width,Style,R,G,B) }
var Handle,Mode,Width,Style,R,G,B: TI2LWord;
begin
try
B:=PopStack; G:=PopStack; R:=PopStack;
Style:=PopStack;
Width:=PopStack;
Mode:=PopStack;
Handle:=PopStack;
GlobalConsole.VGABox.BitmapManager.SetPen(Handle,Mode,Width,Style,TranslateColor(R,G,B));
except HandleConsoleError(ExceptObject); end;
end;



procedure TextOut;
{131 Output text to image}
{Prototype: TextOut(Hande,X,Y,Text) }
var Handle,Inx,X,Y: TI2LWord;
var S: string;
begin
try
Inx:=PopStack;
Y:=PopStack; X:=PopStack;
S:=ProgramStore.GetDataString(Inx,UseMSBStrings);
Handle:=PopStack;
GlobalConsole.VGABox.BitmapManager.DoTextOut(-1,X,Y,S);
except HandleConsoleError(ExceptObject); end;
end;



procedure SetFontFace;
{132 Set image font face}
{Prototype: SetFontFace(Hande,FaceName) }
var Handle,Inx: TI2LWord;
var S: string;
begin
try
Inx:=PopStack;
S:=ProgramStore.GetDataString(Inx,UseMSBStrings);
Handle:=PopStack;
GlobalConsole.VGABox.BitmapManager.SetFont(Handle,S);
except HandleConsoleError(ExceptObject); end;
end;



procedure SetFontColor;
{133 Set image font color}
{Prototype: SetFontColor(Hande,R,G,B) }
var Handle: TI2LWord;
var R,G,B: TI2LWord;
begin
try
B:=PopStack;
G:=PopStack;
R:=PopStack;
Handle:=PopStack;
GlobalConsole.VGABox.BitmapManager.SetFontColor(Handle,TranslateColor(R,G,B));
except HandleConsoleError(ExceptObject); end;
end;



procedure SetFontStyle;
{134 Set image font style}
{Prototype: SetFontStyle(Hande,Size,Bold,Italic,Underline,Strike) }
var Handle: TI2LWord;
var Size, Bold,Italic,Underline,Strike: TI2LWord;
begin
try
Strike:=PopStack;
Underline:=PopStack;
Italic:=PopStack;
Bold:=PopStack;
Size:=PopStack;
Handle:=PopStack;
GlobalConsole.VGABox.BitmapManager.SetFontStyle(Handle,Size,Bold<>0,Italic<>0,Underline<>0,Strike<>0);
except HandleConsoleError(ExceptObject); end;
end;


procedure GetFontInfo;
{135 Get image font properties}
{Prototype: Array:=GetFontInfo(Hande,String) }
var Handle,Inx: TI2LWord;
var Width,Height: integer;
var S: string;
begin
Inx:=PopStack;
Handle:=PopStack;
S:=ProgramStore.GetDataString(Inx,UseMSBStrings);
GlobalConsole.VGABox.BitmapManager.GetFontProperties(Handle,S,Width,Height);
ProgramStore.InsertLocal(FontInfo,Width);
ProgramStore.InsertLocal(FontInfo+IntBytes,Height);
PushStack(ProgramStore.LocalLow+FontInfo);
end;



procedure CopyFont;
{136 Copy font properties from one Image to another}
{Prototype: CopyFont(ToHandle,FromHandle) }
var FromHandle,ToHandle: TI2LWord;
begin
FromHandle:=PopStack;
ToHandle:=PopStack;
GlobalConsole.VGABox.BitmapManager.CopyFont(ToHandle,FromHandle);
end;



procedure SetBorderColor;
{138 Set the color of the VGA border }
{Prototype: SetBorderColor(R,G,B); }
var R,G,B: TI2LWord;
var C: TColor;
begin
B:=PopStack; G:=PopStack; R:=PopStack;
try
C:=TranslateColor(R,G,B);
GlobalConsole.BorderColor:=C;
except HandleConsoleError(ExceptObject); end;
end;




procedure EnableRefresh;
{139 Enable or Disable Refreshing}
{Prototype: EnableRefresh(Enable); }
var Enable: TI2LWord;
begin
try
Enable:=PopStack;
GlobalConsole.EnableRefresh(Enable<>0);
except HandleConsoleError(ExceptObject); end;
end;



{------------------------------------------------------------------------------}



procedure PlaySoundFile;
{140 Play sound file}
{Prototype: PlaySoundFile(Filename); }
var SoundFile: string;
var Inx: TI2LWord;
begin
Inx:=PopStack;
SoundFile:=ProgramStore.GetDataString(Inx,UseMSBStrings);
SoundFile:=GetExpandedEnviroString(SoundFile);
if FileExists(SoundFile) then
	begin
	PlaybackSoundFile(SoundFile);
	end
else StopSoundFile;
end;



procedure PlaySound(Data: pointer; Size,Channels,Bits,SampPerSec,Align: integer; Wait: boolean);
var OutFormat: TWaveFormatEx;
begin
OutFormat.nChannels:=Channels;
OutFormat.nSamplesPerSec:=SampPerSec;
OutFormat.nAvgBytesPerSec:=SampPerSec;
OutFormat.nBlockAlign:=Align;
OutFormat.wBitsPerSample:=Bits;
OutFormat.cbSize:=0;
OutFormat.wFormatTag:=WAVE_FORMAT_PCM;
SetupOutputFormat(OutFormat);
WaveBufferOut(Data,Size,Wait);
end;



procedure PlaySoundData;
{141 Play Sound data}
{Prototype: PlaySoundData(Data,Size,Chans,Bits,DataPerSec,Align,Wait); }
var Data,Size,Channels,Bits,DataPerSec,Align: TI2LWord;
var Wait: boolean;
begin
Wait:=PopStack<>0;
Align:=PopStack;
DataPerSec:=PopStack;
Bits:=PopStack;
Channels:=PopStack;
Size:=PopStack;
Data:=PopStack;
PlaySound(@MemManager.SegArray[Data,0],Size,Channels,Bits,DataPerSec,Align,Wait);
end;



procedure AbortSound;
{142 Abort any all sounds being played}
{Prototype: AbortSound; }
begin
WaveOutShutdown;
end;



procedure GetSoundStatus;
{143 Get status data being played a}
{Prototype: Status:=GetSoundStatus; }
var Data: TI2LWord;
begin
Data:=PendingOutBuffers and $FF;
PushStack(Data)
end;




procedure SoundVolume;
{143 Get status data being played a}
{Prototype: Volume:=SoundVolume(Set,Volume); }
var V: TI2LWord;
begin
V:=PendingOutBuffers and $FF;
PushStack(V)
end;


{------------------------------------------------------------------------------}

procedure PrintImage;
{145 Print Image}
{Prototype:  PrintImage(SourceHandle,Left,Top,Right,Bottom); }
var SourceHandle: TI2LWord;
var DestRect: TRect;
begin
try
DestRect.Bottom:=PopStack;
DestRect.Right:=PopStack;
DestRect.Top:=PopStack;
DestRect.Left:=PopStack;
SourceHandle:=PopStack;
GlobalConsole.VGABox.BitmapManager.PrintImage(SourceHandle,DestRect);
except HandleConsoleError(ExceptObject); end;
end;



{------------------------------------------------------------------------------}


procedure SetScreenScale;
{146 SetScreenScale}
{Prototype: SetScreenScale(ScaleFactor); }
var Scale: double;
begin
try
Scale:=PopDouble;
GlobalConsole.ScaleFactor:=Scale;
except HandleConsoleError(ExceptObject); end;
end;


procedure GetScreenScale;
{147 GetScreenScale}
{Prototype: ScaleFactor:=GetScreenScale; }
var Scale: double;
begin
Scale:=GlobalConsole.ScaleFactor;
PushDouble(Scale);
end;




procedure GetSysMetrics;
{148 - Get specified system value. }
{Prototype: Metric:=GetSysMetrics(MetricType}
var MetricInx,Value: TI2LWord;
begin
MetricInx:=PopStack;
case MetricInx of
 0: Value:=GetSystemMetrics(SM_CXSCREEN);
 1: Value:=GetSystemMetrics(SM_CYSCREEN);
 2: Value:=GetSystemMetrics(SM_CXFRAME);
 3: Value:=GetSystemMetrics(SM_CYCAPTION);
 4: Value:=GetSystemMetrics(SM_CYMENUSIZE);
 else Value:=0;
 end;
PushStack(Value);
end;




procedure GetKeyboardParams;
{200 - Get current keyborad repeat-rate and response delay}
{Prototype: GetKeyboardParams(addr Delay, addr Speed}
var DelayInx,SpeedInx: TI2LWord;
var Speed,Delay: integer;
begin
SpeedInx:=PopStack;
DelayInx:=PopStack;
WinSupport.GetKeyboardDynamics(Delay,Speed);
ProgramStore.SetDataWord(DelayInx,Delay);
ProgramStore.SetDataWord(SpeedInx,Speed);
end;



procedure SetKeyboardParams;
{201 - Set keyborad repeat-rate and response delay}
{Prototype: SetKeyboardParams(Delay, Speed}
var Speed,Delay: integer;
begin
Speed:=PopStack;
Delay:=PopStack;
WinSupport.setKeyboardDynamics(Delay,Speed)
end;

{------------------------------------------------------------------------------}


var LastIntrinsic: integer;

procedure UnusedIntrinsic;
begin
HandleI2LError('Invalid Intrinsic: '+IntToStr(LastIntrinsic),ieIntrinsic);
end;


type TIntrinInfo = record
 Inx: integer;
 Real: boolean;
 Proc: procedure;
 Name: string;
 RetType: string;
 ArgStr: string;
 ArgTypes: string;
 end;

type TIntrinAlias = record
 Inx: integer;
 Name: string;
 end;

const IntCnt = 256;

{
ArgTypes contains type specification for the parameters and return value of the intrinsic.
Each character in the string specifies an type. The first characters is the return type.
This is followed by a "Colon." All the remaining characters are the arguments. An empty
string specifies an unused intrinsic. The values are as follows:

I = int
A = address, although treated as int
i = int pointer
R = real
r = real pointer
V = Void or no argument - only used for the return value

}

const IntrinArray: array [0..IntCnt-1] of TIntrinInfo =(
(Inx:0;  Real: False; Proc: ABSFUN;              Name: 'Abs';                RetType: 'int';     ArgStr: 'value';                        ArgTypes: 'I:I'),
(Inx:1;  Real: False; Proc: RANFUN;              Name: 'Ran';                RetType: 'int';     ArgStr: 'value';                        ArgTypes: 'I:I'),
(Inx:2;  Real: False; Proc: REMFUN;              Name: 'Rem';                RetType: 'int';     ArgStr: 'expression';                   ArgTypes: 'I:I'),
(Inx:3;  Real: False; Proc: RESERV;              Name: 'Reserve';            RetType: 'address'; ArgStr: 'value';                        ArgTypes: 'A:I'),
(Inx:4;  Real: False; Proc: SWAPFUN;             Name: 'Swap';               RetType: 'int';     ArgStr: 'value';                        ArgTypes: 'I:I'),
(Inx:5;  Real: False; Proc: EXTEN;               Name: 'Extend';             RetType: 'int';     ArgStr: 'value';                        ArgTypes: 'I:I'),
(Inx:6;  Real: False; Proc: RESTRT;              Name: 'Restart';            RetType: '';        ArgStr: '';                             ArgTypes: 'V:'),
(Inx:7;  Real: False; Proc: CHIN;                Name: 'ChIn';               RetType: 'int';     ArgStr: 'device';                       ArgTypes: 'I:I'),
(Inx:8;  Real: False; Proc: CHOUT;               Name: 'ChOut';              RetType: '';        ArgStr: 'device, byte';                 ArgTypes: 'V:IA'),
(Inx:9;  Real: False; Proc: SKIP;                Name: 'CrLf';               RetType: '';        ArgStr: 'device';                       ArgTypes: 'V:I'),
(Inx:10; Real: False; Proc: INTIN;               Name: 'IntIn';              RetType: 'int';     ArgStr: 'device';                       ArgTypes: 'I:I'),
(Inx:11; Real: False; Proc: INTOUT;              Name: 'IntOut';             RetType: '';        ArgStr: 'device, value';                ArgTypes: 'V:II'),
(Inx:12; Real: False; Proc: TEXT;                Name: 'Text';               RetType: '';        ArgStr: 'device, address';              ArgTypes: 'V:IA'),
(Inx:13; Real: False; Proc: IDEVIN;              Name: 'OpenI';              RetType: '';        ArgStr: 'device';                       ArgTypes: 'V:I'),
(Inx:14; Real: False; Proc: ODEVIN;              Name: 'OpenO';              RetType: '';        ArgStr: 'device';                       ArgTypes: 'V:I'),
(Inx:15; Real: False; Proc: ODEVCL;              Name: 'Close';              RetType: '';        ArgStr: 'device';                       ArgTypes: 'V:I'),
(Inx:16; Real: False; Proc: AbortProg;           Name: 'Abort';              RetType: '';        ArgStr: '';                             ArgTypes: 'V:'),
(Inx:17; Real: False; Proc: TRAP;                Name: 'Trap';               RetType: '';        ArgStr: 'integer';                      ArgTypes: 'V:I'),
(Inx:18; Real: False; Proc: SPACE;               Name: 'Free';               RetType: 'int';     ArgStr: '';                             ArgTypes: 'I:'),
(Inx:19; Real: False; Proc: RERUN;               Name: 'Rerun';              RetType: 'boolean'; ArgStr: '';                             ArgTypes: 'I:'),
(Inx:20; Real: False; Proc: TSTHP;               Name: 'GetHp';              RetType: 'address'; ArgStr: '';                             ArgTypes: 'A:'),
(Inx:21; Real: False; Proc: SETHP;               Name: 'SetHp';              RetType: '';        ArgStr: 'address';                      ArgTypes: 'V:I'),
(Inx:22; Real: False; Proc: GETERR;              Name: 'GetErr';             RetType: 'int';     ArgStr: '';                             ArgTypes: 'I:'),
(Inx:23; Real: False; Proc: CURSOR;              Name: 'Cursor';             RetType: '';        ArgStr: 'X, Y';                         ArgTypes: 'V:II'),
(Inx:24; Real: False; Proc: FSET;                Name: 'FSet';               RetType: '';        ArgStr: 'handle, mode';                 ArgTypes: 'V:II'),
(Inx:25; Real: False; Proc: SETRUN;              Name: 'SetRun';             RetType: '';        ArgStr: 'boolean';                      ArgTypes: 'V:I'),
(Inx:26; Real: False; Proc: HEXI;                Name: 'HexIn';              RetType: 'int';     ArgStr: 'device';                       ArgTypes: 'I:I'),
(Inx:27; Real: False; Proc: HEXO;                Name: 'HexOut';             RetType: '';        ArgStr: 'device, value';                ArgTypes: 'V:II'),
(Inx:28; Real: False; Proc: CHAIN;               Name: 'Chain';              RetType: '';        ArgStr: '"path\filename.ext"';          ArgTypes: 'V:A'),
(Inx:29; Real: False; Proc: FOPEN;               Name: 'FOpen';              RetType: 'handle';  ArgStr: 'FilePath, mode';               ArgTypes: 'I:AI'),
(Inx:30; Real: False; Proc: FWRITE;              Name: 'Write';              RetType: '';        ArgStr: 'drive, sector, buffer, size';  ArgTypes: 'V:IIII'),
(Inx:31; Real: False; Proc: FREAD;               Name: 'Read';               RetType: '';        ArgStr: 'drive, sector, buffer, size';  ArgTypes: 'V:IIII'),
(Inx:32; Real: False; Proc: FCLOSE;              Name: 'FClose';             RetType: '';        ArgStr: 'handle';                       ArgTypes: 'V:I'),
(Inx:33; Real: False; Proc: CHKKEY;              Name: 'ChkKey';             RetType: 'boolean'; ArgStr: '';                             ArgTypes: 'I:'),
(Inx:34; Real: False; Proc: SFTINT;              Name: 'SoftInt';            RetType: '';        ArgStr: 'interrupt';                    ArgTypes: 'V:I'),
(Inx:35; Real: False; Proc: GETREG;              Name: 'GetReg';             RetType: 'address'; ArgStr: '';                             ArgTypes: 'A:'),
(Inx:36; Real: False; Proc: BLKTRN;              Name: 'Blit';               RetType: '';        ArgStr: 'srcseg, srcoff, desseg, desoff, size'; ArgTypes: 'V:IIIII'),
(Inx:37; Real: False; Proc: PEEK;                Name: 'Peek';               RetType: 'int';     ArgStr: 'segment, offset';              ArgTypes: 'I:II'),
(Inx:38; Real: False; Proc: POKE;                Name: 'Poke';               RetType: '';        ArgStr: 'segment, offset, value';       ArgTypes: 'V:III'),
(Inx:39; Real: False; Proc: SOUND;               Name: 'Sound';              RetType: '';        ArgStr: 'volume, duration, frequency';  ArgTypes: 'V:III'),
(Inx:40; Real: False; Proc: VCLEAR;              Name: 'Clear';              RetType: '';        ArgStr: '';                             ArgTypes: 'V:'),
(Inx:41; Real: False; Proc: DOT;                 Name: 'Point';              RetType: '';        ArgStr: 'X, Y, color';                  ArgTypes: 'V:III'),
(Inx:42; Real: False; Proc: LINE;                Name: 'Line';               RetType: '';        ArgStr: 'X, Y, color';                  ArgTypes: 'V:III'),
(Inx:43; Real: False; Proc: MOVE;                Name: 'Move';               RetType: '';        ArgStr: 'X, Y';                         ArgTypes: 'V:II'),
(Inx:44; Real: False; Proc: REDDOT;              Name: 'ReadPix';            RetType: 'color';   ArgStr: 'X, Y';                         ArgTypes: 'I:II'),
(Inx:45; Real: False; Proc: SETVID;              Name: 'SetVid';             RetType: '';        ArgStr: 'mode';                         ArgTypes: 'V:I'),
(Inx:46; Real: True;  Proc: FLRES;               Name: 'RlRes';              RetType: 'real';    ArgStr: 'size';                         ArgTypes: 'R:I'),
(Inx:47; Real: True;  Proc: FLIN;                Name: 'RlIn';               RetType: 'real';    ArgStr: 'device';                       ArgTypes: 'R:I'),
(Inx:48; Real: True;  Proc: FLOUT;               Name: 'RlOut';              RetType: '';        ArgStr: 'device, real';                 ArgTypes: 'V:IR'),
(Inx:49; Real: True;  Proc: FLTFUN;              Name: 'Float';              RetType: 'real';    ArgStr: 'integer';                      ArgTypes: 'R:I'),
(Inx:50; Real: False; Proc: FIXFUN;              Name: 'Fix';                RetType: 'int';     ArgStr: 'real';                         ArgTypes: 'I:R'),
(Inx:51; Real: True;  Proc: FLABS;               Name: 'RlAbs';              RetType: 'real';    ArgStr: 'real';                         ArgTypes: 'R:R'),
(Inx:52; Real: True;  Proc: FMTFUN;              Name: 'Format';             RetType: '';        ArgStr: 'integer, integer';             ArgTypes: 'V:II'),
(Inx:53; Real: True;  Proc: FLSQRT;              Name: 'Sqrt';               RetType: 'real';    ArgStr: 'real';                         ArgTypes: 'R:R'),
(Inx:54; Real: True;  Proc: FLLN;                Name: 'Ln';                 RetType: 'real';    ArgStr: 'real';                         ArgTypes: 'R:R'),
(Inx:55; Real: True;  Proc: FLEXP;               Name: 'Exp';                RetType: 'real';    ArgStr: 'real';                         ArgTypes: 'R:R'),
(Inx:56; Real: True;  Proc: FLSIN;               Name: 'Sin';                RetType: 'real';    ArgStr: 'real';                         ArgTypes: 'R:R'),
(Inx:57; Real: True;  Proc: FLAT2;               Name: 'ATan2';              RetType: 'real';    ArgStr: 'Y, X';                         ArgTypes: 'R:RR'),
(Inx:58; Real: True;  Proc: FLMOD;               Name: 'Mod';                RetType: 'real';    ArgStr: 'real, real';                   ArgTypes: 'R:RR'),
(Inx:59; Real: True;  Proc: FLLOG;               Name: 'Log';                RetType: 'real';    ArgStr: 'real';                         ArgTypes: 'R:R'),
(Inx:60; Real: True;  Proc: FLCOS;               Name: 'Cos';                RetType: 'real';    ArgStr: 'real';                         ArgTypes: 'R:R'),
(Inx:61; Real: True;  Proc: FLTAN;               Name: 'Tan';                RetType: 'real';    ArgStr: 'real';                         ArgTypes: 'R:R'),
(Inx:62; Real: True;  Proc: FLASIN;              Name: 'ASin';               RetType: 'real';    ArgStr: 'real';                         ArgTypes: 'R:R'),
(Inx:63; Real: True;  Proc: FLACOS;              Name: 'ACos';               RetType: 'real';    ArgStr: 'real';                         ArgTypes: 'R:R'),
(Inx:64; Real: True;  Proc: FloorFunction;       Name: 'Floor';              RetType: 'real';    ArgStr: 'real';                         ArgTypes: 'R:R'),
(Inx:65; Real: True;  Proc: CeilFunction;        Name: 'Ceil';               RetType: 'real';    ArgStr: 'real';                         ArgTypes: 'R:R'),
(Inx:66; Real: True;  Proc: PowerFunction;       Name: 'Pow';                RetType: 'real';    ArgStr: 'base,exp';                     ArgTypes: 'R:RR'),
(Inx:67; Real: False; Proc: EXTJMP;              Name: 'ExtJmp';             RetType: '';        ArgStr: 'segment, offset';              ArgTypes: 'V:II'),
(Inx:68; Real: False; Proc: EXTCAL;              Name: 'ExtCal';             RetType: '';        ArgStr: 'segment, offset';              ArgTypes: 'V:II'),
(Inx:69; Real: False; Proc: SETATT;              Name: 'Attrib';             RetType: '';        ArgStr: 'colors';                       ArgTypes: 'V:I'),
(Inx:70; Real: False; Proc: SETWND;              Name: 'SetWind';            RetType: '';        ArgStr: 'X0, Y0, X1, Y1, mode, fill';   ArgTypes: 'V:IIIIII'),
(Inx:71; Real: False; Proc: RAWTXT;              Name: 'RawText';            RetType: '';        ArgStr: 'device, address';              ArgTypes: 'V:II'),
(Inx:72; Real: False; Proc: HLIGHT;              Name: 'Hilight';            RetType: '';        ArgStr: 'X0, Y0, X1, Y1, attribute';    ArgTypes: 'V:IIIII'),
(Inx:73; Real: False; Proc: MALLOC;              Name: 'Malloc';             RetType: 'address'; ArgStr: 'paragraphs';                   ArgTypes: 'A:I'),
(Inx:74; Real: False; Proc: RELEAS;              Name: 'Release';            RetType: '';        ArgStr: 'address';                      ArgTypes: 'V:I'),
(Inx:75; Real: False; Proc: TRAPC;               Name: 'TrapC';              RetType: '';        ArgStr: 'boolean';                      ArgTypes: 'V:I'),
(Inx:76; Real: False; Proc: TESTC;               Name: 'TestC';              RetType: 'boolean'; ArgStr: '';                             ArgTypes: 'I:'),
(Inx:77; Real: False; Proc: GETEQP;              Name: 'Equip';              RetType: 'address'; ArgStr: '';                             ArgTypes: 'A:'),
(Inx:78; Real: False; Proc: SHRINK;              Name: 'Shrink';             RetType: '';        ArgStr: 'value';                        ArgTypes: 'V:I'),
(Inx:79; Real: False; Proc: RANSEED;             Name: 'RanSeed';            RetType: '';        ArgStr: 'seed';                         ArgTypes: 'V:I'),
(Inx:80; Real: False; Proc: IRQ;                 Name: 'Irq';                RetType: '';        ArgStr: 'enable';                       ArgTypes: 'V:I'),
(Inx:81; Real: False; Proc: SetCloseOnTerminate; Name: 'CloseOnTerminate';   RetType: '';        ArgStr: 'boolean';                      ArgTypes: 'V:I'),
(Inx:82; Real: False; Proc: GetTime;             Name: 'GetTime';            RetType: 'int';     ArgStr: '';                             ArgTypes: 'I:'),
(Inx:83; Real: False; Proc: Backup;              Name: 'Backup';             RetType: '';        ArgStr: '';                             ArgTypes: 'V:'),
(Inx:84; Real: False; Proc: SetHexDigits;        Name: 'SetHexDigits';       RetType: '';        ArgStr: 'digits';                       ArgTypes: 'V:I'),
(Inx:85; Real: False; Proc: ReallocMem;          Name: 'ReallocMem';         RetType: 'address'; ArgStr: '?';                            ArgTypes: 'A:II'),
(Inx:86; Real: False; Proc: UnusedIntrinsic;     Name: '?';                  RetType: '';        ArgStr: '?';                            ArgTypes: ''),
(Inx:87; Real: False; Proc: UnusedIntrinsic;     Name: '?';                  RetType: '';        ArgStr: '?';                            ArgTypes: ''),
(Inx:88; Real: False; Proc: ShowCursor;          Name: 'ShowCursor';         RetType: '';        ArgStr: 'Visible';                      ArgTypes: 'V:I'),
(Inx:89; Real: False; Proc: GetCursorInfo;       Name: 'GetCursorInfo';      RetType: '';        ArgStr: 'addr X, addr Y; addr Info';    ArgTypes: 'V:iii'),
(Inx:90; Real: False; Proc: GetCommandLine;      Name: 'CommandLine';        RetType: 'boolean'; ArgStr: 'Inx,String,MaxLen';            ArgTypes: 'I:IAI'),
(Inx:91; Real: False; Proc: GetFileInfo;         Name: 'GetFileInfo';        RetType: '';        ArgStr: 'Handle, addr Size';            ArgTypes: 'V:Ii'),
(Inx:92; Real: False; Proc: GetPathInfo;         Name: 'GetPathInfo';        RetType: '';        ArgStr: 'Drive,String,MaxLen';          ArgTypes: 'V:IAI'),
(Inx:93; Real: False; Proc: SetPathInfo;         Name: 'SetPathInfo';        RetType: '';        ArgStr: 'Path';                         ArgTypes: 'V:A'),
(Inx:94; Real: False; Proc: DelayUS;             Name: 'DelayUS';            RetType: '';        ArgStr: 'MS';                           ArgTypes: 'V:I'),
(Inx:95; Real: False; Proc: GetParams;           Name: 'GetParams';          RetType: 'int';     ArgStr: 'MainType,SubType';             ArgTypes: 'I:II'),
(Inx:96; Real: False; Proc: SetParams;           Name: 'SetParams';          RetType: '';        ArgStr: 'MainType,SubType,Value';       ArgTypes: 'V:III'),
(Inx:97; Real: False; Proc: UnusedIntrinsic;     Name: '?';                  RetType: '';        ArgStr: '?';                            ArgTypes: ''),
(Inx:98; Real: False; Proc: GetKeyState;         Name: 'GetKeyState';        RetType: 'Int';     ArgStr: 'Key';                          ArgTypes: 'I:I'),
(Inx:99; Real: False; Proc: SetCharAttrib;       Name: 'SetCharAttrib';      RetType: '';        ArgStr: 'X,Y,C';                        ArgTypes: 'V:III'),
(Inx:100; Real: False; Proc: CursorChar;         Name: 'CursorChar';         RetType: 'int';     ArgStr: 'X,Y';                          ArgTypes: 'I:II'),
(Inx:101; Real: False; Proc: DelayMS;            Name: 'DelayMS';            RetType: '';        ArgStr: 'Milliseconds';                 ArgTypes: 'V:I'),
(Inx:102; Real: False; Proc: MultiplyDivide;     Name: 'MulDiv';             RetType: 'int';     ArgStr: 'X,Y,Z';                        ArgTypes: 'I:III'),
(Inx:103; Real: False; Proc: GetDateTime;        Name: 'GetSysTime';         RetType: '';        ArgStr: 'GMT,addr Year,addr Month,addr Day,addr WeekDay, addr Hour,addr Min,addr Sec,addr MS';  ArgTypes: 'V:Iiiiiiiii'),
(Inx:104; Real: False; Proc: GetScanCode;        Name: 'GetScanCode';        RetType: 'int';     ArgStr: 'addr RawCode,addr VirtualCode,addr BIOSCode, Wait';        ArgTypes: 'I:iiiI'),
(Inx:105; Real: False; Proc: GetMouseState;      Name: 'GetMouseState';      RetType: '';        ArgStr: 'addr State, addr X, addr Y';   ArgTypes: 'V:iii'),
(Inx:106; Real: False; Proc: GetLastKey;         Name: 'GetLastKey';         RetType: 'int';     ArgStr: 'Clear';                        ArgTypes: 'I:I'),
(Inx:107; Real: False; Proc: GetVideoMode;       Name: 'GetVideoMode';       RetType: 'int';     ArgStr: '';                             ArgTypes: 'I:'),
(Inx:108; Real: False; Proc: SetPalette;         Name: 'SetPalette';         RetType: '';        ArgStr: 'Inx,R,G,B';                    ArgTypes: 'V:IIII'),
(Inx:109; Real: False; Proc: TimeSlice;          Name: 'TimeSlice';          RetType: '';        ArgStr: 'Delay';                        ArgTypes: 'V:I'),
(Inx:110; Real: False; Proc: AllocateBitmap;     Name: 'AllocateBitmap';     RetType: 'int';     ArgStr: '';                             ArgTypes: 'I:'),
(Inx:111; Real: False; Proc: DeAllocateBitmap;   Name: 'DeAllocBitmap';      RetType: '';        ArgStr: 'Handle';                       ArgTypes: 'V:I'),
(Inx:112; Real: False; Proc: LoadBitmap;         Name: 'LoadBitmap';         RetType: '';        ArgStr: 'Handle,Filename';              ArgTypes: 'V:IA'),
(Inx:113; Real: False; Proc: SaveBitmap;         Name: 'SaveBitmap';         RetType: '';        ArgStr: 'Handle,Filename';              ArgTypes: 'V:IA'),
(Inx:114; Real: False; Proc: CopyImage;          Name: 'CopyImage';          RetType: '';        ArgStr: 'SourceHandle,SX1,SY1,SX2,SY2, DestHandle,DX1,DY1,DX2,DY2'; ArgTypes: 'V:IIIIIIIIII'),
(Inx:115; Real: False; Proc: PaintImage;         Name: 'PaintImage';         RetType: '';        ArgStr: 'SourceHandle,DestHandle,X,Y';              ArgTypes: 'V:IIII'),
(Inx:116; Real: False; Proc: SetCopyMode;        Name: 'SetCopyMode';        RetType: '';        ArgStr: 'Mode,Alpha,R,G,B';                         ArgTypes: 'V:IIIII'),
(Inx:117; Real: False; Proc: GetBitmapSize;      Name: 'GetBitmapSize';      RetType: '';        ArgStr: 'Handle,addr Width,addr Height';            ArgTypes: 'V:Iii'),
(Inx:118; Real: False; Proc: SetBitmapSize;      Name: 'SetBitmapSize';      RetType: '';        ArgStr: 'Handle,Width,Height';                      ArgTypes: 'V:III'),
(Inx:119; Real: False; Proc: SetPixel;           Name: 'SetPixel';           RetType: '';        ArgStr: 'Handle,X,Y,Red,Green,Blue';                ArgTypes: 'V:IIIIII'),
(Inx:120; Real: False; Proc: GetPixel;           Name: 'GetPixel';           RetType: '';        ArgStr: 'Handle,X,Y,addr Red,addr Green,addr Blue'; ArgTypes: 'V:IIIiii'),
(Inx:121; Real: False; Proc: ClearImage;         Name: 'ClearImage';         RetType: '';        ArgStr: 'Handle,R,G,B';                                ArgTypes: 'V:IIII'),
(Inx:122; Real: False; Proc: DrawCircle;         Name: 'DrawCircle';         RetType: '';        ArgStr: 'Handle,X,Y,Radius';                           ArgTypes: 'V:IIII'),
(Inx:123; Real: False; Proc: DrawEllipse;        Name: 'DrawEllipse';        RetType: '';        ArgStr: 'Handle,X1,Y1,X2,Y2';                          ArgTypes: 'V:IIIII'),
(Inx:124; Real: False; Proc: DrawRectangle;      Name: 'DrawRectangle';      RetType: '';        ArgStr: 'Handle,X1,Y1,X2,Y2';                          ArgTypes: 'V:IIIII'),
(Inx:125; Real: False; Proc: MoveTo;             Name: 'MoveTo';             RetType: '';        ArgStr: 'Handle,X,Y';                                  ArgTypes: 'V:III'),
(Inx:126; Real: False; Proc: Lineto;             Name: 'Lineto';             RetType: '';        ArgStr: 'Handle,X,Y';                                  ArgTypes: 'V:III'),
(Inx:127; Real: False; Proc: PolyLine;           Name: 'Polyline';           RetType: '';        ArgStr: 'Handle,Mode,addr Points, Count';              ArgTypes: 'V:IIiI'),
(Inx:128; Real: False; Proc: GetPenPosition;     Name: 'GetPenPos';          RetType: '';        ArgStr: 'Handle,addr X,addr Y';                        ArgTypes: 'V:Iii'),
(Inx:129; Real: False; Proc: SetFill;            Name: 'SetFill';            RetType: '';        ArgStr: 'Handle,Mode,R,G,B';                           ArgTypes: 'V:IIIII'),
(Inx:130; Real: False; Proc: SetPen;             Name: 'SetPen';             RetType: '';        ArgStr: 'Handle,Mode,Width,Style,R,G,B';               ArgTypes: 'V:IIIIIII'),
(Inx:131; Real: False; Proc: TextOut;            Name: 'TextOut';            RetType: '';        ArgStr: 'Handle,X,Y,Text';                             ArgTypes: 'V:IIIA'),
(Inx:132; Real: False; Proc: SetFontFace;        Name: 'SetFontFace';        RetType: '';        ArgStr: 'Handle,FaceName';                             ArgTypes: 'V:II'),
(Inx:133; Real: False; Proc: SetFontColor;       Name: 'SetFontColor';       RetType: '';        ArgStr: 'Handle,R,G,B';                                ArgTypes: 'V:IIII'),
(Inx:134; Real: False; Proc: SetFontStyle;       Name: 'SetFontStyle';       RetType: '';        ArgStr: 'Handle,Size,Bold,Italic,Underline,Strike';    ArgTypes: 'V:IIIIII'),
(Inx:135; Real: False; Proc: GetFontInfo;        Name: 'GetFontInfo';        RetType: 'address'; ArgStr: 'Handle,String';                               ArgTypes: 'A:IA'),
(Inx:136; Real: False; Proc: CopyFont;           Name: 'CopyFont';           RetType: '';        ArgStr: 'ToHandle,FromHandle';                         ArgTypes: 'V:II'),
(Inx:137; Real: False; Proc: UnusedIntrinsic;    Name: '?';                  RetType: '';        ArgStr: '?';                                           ArgTypes: ''),
(Inx:138; Real: False; Proc: SetBorderColor;     Name: 'SetBorderColor';     RetType: '';        ArgStr: 'R,G,B';                                       ArgTypes: 'V:III'),
(Inx:139; Real: False; Proc: EnableRefresh;      Name: 'EnableRefresh';      RetType: '';        ArgStr: 'Enabled';                                     ArgTypes: 'V:I'),
(Inx:140; Real: False; Proc: PlaySoundFile;      Name: 'PlaySoundFile';      RetType: '';        ArgStr: 'Filename';                                    ArgTypes: 'V:A'),
(Inx:141; Real: False; Proc: PlaySoundData;      Name: 'PlaySoundData';      RetType: '';        ArgStr: 'Data,Size,Chans,Bits,DataPerSec,Align,Wait';  ArgTypes: 'V:IIIIIII'),
(Inx:142; Real: False; Proc: AbortSound;         Name: 'AbortSound';         RetType: '';        ArgStr: '';                                            ArgTypes: 'V:'),
(Inx:143; Real: False; Proc: GetSoundStatus;     Name: 'GetSoundStatus';     RetType: 'int';     ArgStr: '';                                            ArgTypes: 'I:'),
(Inx:144; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';                                           ArgTypes: ''),
(Inx:145; Real: False; Proc: PrintImage;         Name: 'PrintImage';         RetType: '';        ArgStr: 'SourceHandle,Left,Top,Right,Bottom';          ArgTypes: 'V:IIIII'),
(Inx:146; Real: False; Proc: SetScreenScale;     Name: 'SetScreenScale';     RetType: '';        ArgStr: 'ScaleFactor';                                 ArgTypes: 'V:R'),
(Inx:147; Real: False; Proc: GetScreenScale;     Name: 'GetScreenScale';     RetType: 'real';    ArgStr: '?';                                           ArgTypes: 'R:'),
(Inx:148; Real: False; Proc: GetSysMetrics;      Name: 'GetSysMetrics';      RetType: 'int';     ArgStr: 'MetricType';                                  ArgTypes: 'I:I'),
(Inx:149; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';                                           ArgTypes: ''),
(Inx:150; Real: False; Proc: Enable3DViewer;     Name: 'Enable3DViewer';     RetType: '';        ArgStr: 'Enable';                                      ArgTypes: 'V:I'),
(Inx:151; Real: False; Proc: Set3DWindow;        Name: 'Set3DWindow';        RetType: '';        ArgStr: 'Left,Top,Width,Height';                       ArgTypes: 'V:IIII'),
(Inx:152; Real: False; Proc: Set3DRenderMode;    Name: 'Set3DRenderMode';    RetType: '';        ArgStr: 'Handle,Item,Mode';                            ArgTypes: 'V:III'),
(Inx:153; Real: False; Proc: Add3DObject;        Name: 'Add3DObject';        RetType: 'int';     ArgStr: 'Parent,ObjType,Name';                         ArgTypes: 'I:IIA'),
(Inx:154; Real: False; Proc: Delete3DObject;     Name: 'Delete3DObject';     RetType: '';        ArgStr: 'Handle';                                      ArgTypes: 'V:I'),
(Inx:155; Real: False; Proc: Position3DObject;   Name: 'Position3DObject';   RetType: '';        ArgStr: 'Handle,X,Y,Z';                                ArgTypes: 'V:IRRR'),
(Inx:156; Real: False; Proc: Rotate3DObject;     Name: 'Rotate3DObject';     RetType: '';        ArgStr: 'Handle,Pitch,Yaw,Roll';                       ArgTypes: 'V:IRRR'),
(Inx:157; Real: False; Proc: Add3DTriangle;      Name: 'Add3DTriangle';      RetType: '';        ArgStr: 'Handle,X1,Y1,Z1, X2,Y2,Z2, X3,Y3,Z3,Color';              ArgTypes: 'V:IRRRRRRRRRI'),
(Inx:158; Real: False; Proc: Add3DRectangle;     Name: 'Add3DRectangle';     RetType: '';        ArgStr: 'Handle,X1,Y1,Z1, X2,Y2,Z2, X3,Y3,Z3, X4,Y4,Z4, Color';   ArgTypes: 'V:IRRRRRRRRRRRRI'),
(Inx:159; Real: False; Proc: Add3DLine;          Name: 'Add3DLine';          RetType: '';        ArgStr: 'Handle,X1,Y1,Z1, X2,Y2,Z2, Width, Color';                ArgTypes: 'V:IRRRRRRII'),
(Inx:160; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';                                               ArgTypes: ''),
(Inx:161; Real: False; Proc: Update3DViewer;     Name: 'Update3DViewer';     RetType: '';        ArgStr: '';                                                ArgTypes: 'V:'),
(Inx:162; Real: False; Proc: Set3DNormals;       Name: 'Set3DNormals';       RetType: '';        ArgStr: 'Handle,CreaseAngle';                              ArgTypes: 'V:IR'),
(Inx:163; Real: False; Proc: Clear3DData;        Name: 'Clear3DData';        RetType: '';        ArgStr: 'Handle';                                          ArgTypes: 'V:I'),
(Inx:164; Real: False; Proc: Get3DViewerInfo;    Name: 'Get3DViewerInfo';    RetType: 'int';     ArgStr: 'InfoIndex';                                       ArgTypes: 'I:I'),
(Inx:165; Real: False; Proc: Set3DBackColor;     Name: 'Set3DBackColor';     RetType: '';        ArgStr: 'Color';                                           ArgTypes: 'V:I'),
(Inx:166; Real: False; Proc: Reset3DTransforms;  Name: 'Reset3DTransforms';  RetType: '';        ArgStr: 'Handle';                                          ArgTypes: 'V:I'),
(Inx:167; Real: False; Proc: Set3DAmbient;       Name: 'Set3DAmbient';       RetType: '';        ArgStr: 'Color';                                           ArgTypes: 'V:I'),
(Inx:168; Real: False; Proc: Get3DCount;         Name: 'Get3DCount';         RetType: 'int';     ArgStr: 'Handle,Index';                                    ArgTypes: 'I:II'),
(Inx:169; Real: False; Proc: Set3DItemColor;     Name: 'Set3DItemColor';     RetType: '';        ArgStr: 'Handle,Index,C1,C2,C3';                           ArgTypes: 'V:IIIII'),
(Inx:170; Real: False; Proc: RGBToColor;         Name: 'RGBToColor';         RetType: 'int';     ArgStr: 'R,G,B,A';                                         ArgTypes: 'I:IIII'),
(Inx:171; Real: False; Proc: Set3DLights;        Name: 'Set3DLights';        RetType: '';        ArgStr: 'Index,Enable,Mode,X,Y,Z,Color';                   ArgTypes: 'V:IIIRRRI'),
(Inx:172; Real: False; Proc: Set3DObjectColor;   Name: 'Set3DObjectColor';   RetType: '';        ArgStr: 'Handle,Color';                                    ArgTypes: 'V:II'),
(Inx:173; Real: False; Proc: Set3DObjectAlpha;   Name: 'Set3DObjectAlpha';   RetType: '';        ArgStr: 'Handle,Enable,DepthCheck,BackBlend,Flags,Trans';  ArgTypes: 'V:IIIIIR'),
(Inx:174; Real: False; Proc: Set3DBackground;    Name: 'Set3DBackground';    RetType: '';        ArgStr: 'BMHandle,Enable,PosX,PosY,PosZ,ScaleX,ScaleY';    ArgTypes: 'V:IIRRRRR'),
(Inx:175; Real: False; Proc: Load3DFile;         Name: 'Load3DFile';         RetType: '';        ArgStr: 'Filename';                                        ArgTypes: 'V:A'),
(Inx:176; Real: False; Proc: Save3DFile;         Name: 'Save3DFile';         RetType: '';        ArgStr: 'Filename';                                        ArgTypes: 'V:A'),
(Inx:177; Real: False; Proc: FitTo3DWindow;      Name: 'FitTo3DWindow';      RetType: '';        ArgStr: 'Handle,Index;';                                   ArgTypes: 'V:II'),
(Inx:178; Real: False; Proc: Copy3DObjects;      Name: 'Copy3DObjects';      RetType: '';        ArgStr: 'ToObj,FromObj';                                   ArgTypes: 'V:II'),
(Inx:179; Real: False; Proc: Test3DHit;          Name: 'Test3DHit';          RetType: 'boolean'; ArgStr: 'X,Y, addr Handle, addr TriIndex';                 ArgTypes: 'I:IIii'),
(Inx:180; Real: False; Proc: Mouse3DControl;     Name: 'Mouse3DControl';     RetType: '';        ArgStr: 'Enable,AutoUpdate,AutoAnimate';                   ArgTypes: 'V:III'),
(Inx:181; Real: False; Proc: Set3DText;          Name: 'Set3DText';          RetType: '';        ArgStr: 'Handle,String,Depth';                             ArgTypes: 'V:IAR'),
(Inx:182; Real: False; Proc: Get3DObjectInfo;    Name: 'Get3DObjectInfo';    RetType: '';        ArgStr: 'Handle,addr Parent,addr ObjType';                 ArgTypes: 'V:Iii'),
(Inx:183; Real: False; Proc: Scale3DObject;      Name: 'Scale3DObject';      RetType: '';        ArgStr: 'Handle,X,Y,Z';                                    ArgTypes: 'V:IRRR'),
(Inx:184; Real: False; Proc: Set3DTexture;       Name: 'Set3DTexture';       RetType: '';        ArgStr: 'Handle, Mode, SubMode, addr BMHandles, BMPCount'; ArgTypes: 'V:IIIiI'),
(Inx:185; Real: False; Proc: Set3DTextureParams; Name: 'Set3DTextureParams'; RetType: '';        ArgStr: 'Handle, VarIndex, X,Y,Z,W';                       ArgTypes: 'V:IIRRRR'),
(Inx:186; Real: False; Proc: Set3DTextureFlags;  Name: 'Set3DTextureFlags';  RetType: '';        ArgStr: 'Handle, VarIndex, Enable';                        ArgTypes: 'V:III'),
(Inx:187; Real: False; Proc: Set3DViewerMode;    Name: 'Set3DViewerMode';    RetType: '';        ArgStr: 'Item,Mode';                                       ArgTypes: 'V:II'),
(Inx:188; Real: False; Proc: Get3DObjPosRot;     Name: 'Get3DObjPosRot';     RetType: '';        ArgStr: 'Handle, addr X, addr Y, addr Z, addr Pitch, addr Yaw, addr Roll'; ArgTypes: 'V:Irrrrrr'),
(Inx:189; Real: False; Proc: Set3DObjectData;    Name: 'Set3DObjectData';    RetType: '';        ArgStr: 'Handle,addr Array,Count,Color,Width';             ArgTypes: 'V:IrIII'),
(Inx:190; Real: False; Proc: Orient3DObject;     Name: 'Orient3DObject';     RetType: '';        ArgStr: 'Handle,Mode,DirX,DirY,DirZ,UpX,UpY,UpZ';          ArgTypes: 'V:IIRRRRRR'),
(Inx:191; Real: False; Proc: Set3DMaterials;     Name: 'Set3DMaterials';     RetType: '';        ArgStr: 'Handle,Mode,Color';                               ArgTypes: 'V:III'),
(Inx:192; Real: False; Proc: Show3DNormals;      Name: 'Show3DNormals';      RetType: '';        ArgStr: 'Handle,Enable';                                   ArgTypes: 'V:II'),
(Inx:193; Real: False; Proc: Get3DListHandle;    Name: 'Get3DListHandle';    RetType: 'int';     ArgStr: 'ParentFrame,ListType,Index';                      ArgTypes: 'I:III'),
(Inx:194; Real: False; Proc: Set3DStrokeText;    Name: 'Set3DStrokeText';    RetType: '';        ArgStr: 'Handle, UsePos, Text, PosX, PosY, PosZ, Scale, Color'; ArgTypes: 'V:IIARRRRI'),
(Inx:195; Real: False; Proc: Get3DObjectName;    Name: 'Get3DObjectName';    RetType: '';        ArgStr: 'Handle, Name, MaxLength'; ArgTypes: 'V:IAI'),
(Inx:196; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:197; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:198; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:199; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),

(Inx:200; Real: False; Proc: GetKeyboardParams;  Name: 'GetKeyboardParams';  RetType: '';        ArgStr: 'addr Delay, addr Speed';  ArgTypes: 'V:ii'),
(Inx:201; Real: False; Proc: SetKeyboardParams;  Name: 'SetKeyboardParams';  RetType: '';        ArgStr: 'Delay, Speed';  ArgTypes: 'V:II'),
(Inx:202; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:203; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:204; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:205; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:206; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:207; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:208; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:209; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:210; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:211; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:212; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:213; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:214; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:215; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:216; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:217; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:218; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:219; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:220; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:221; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:222; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:223; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:224; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:225; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:226; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:227; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:228; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:229; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:230; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:231; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:232; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:233; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:234; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:235; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:236; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:237; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:238; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:239; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:240; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:241; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:242; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:243; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:244; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:245; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:246; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:247; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:248; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:249; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:250; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:251; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:252; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:253; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:254; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: ''),
(Inx:255; Real: False; Proc: UnusedIntrinsic;    Name: '';                   RetType: '';        ArgStr: '?';  ArgTypes: '')
);

const IntrinAlias: array [0..1-1] of TIntrinAlias =(
(Inx:33; Name: 'KeyHit')
);



{------------------------------------------------------------------------------}


procedure CallIntrinsic(Num: integer);
begin
OutputFlag:=False;
LastIntrinsic:=Num;
IntrinArray[Num].Proc;
{Give time to the main thread if we are outputing to the screen}
if not GlobalConsole.VGABox.BlockPaint and OutputFlag then Application.ProcessMessages;
end;



{------------------------------------------------------------------------------}



function GetProtoType(Index: integer; DoResult: boolean): string;
begin
with IntrinArray[Index] do
 begin
 if @Proc=@UnusedIntrinsic then Result:='?'
 else
	begin
	if DoResult and (RetType<>'') then Result:=RetType+':=' else Result:='';
	Result:=Result+Name;
	if ArgStr<>'' then Result:=Result+ '(' +ArgStr+ ')';
	Result:=Result + ';';
	end;
 end;
end;


procedure GetPrototypeList(var SL: TStrings);
var I: integer;
var S: string;
begin
SL.Clear;
for I:=0 to High(IntrinArray) do
	begin
	S:=GetProtoType(I,False);
	if S<>'?' then SL.AddObject(S,TObject(I));
	end;
end;



function NameToProtoType(Name: string): string;
var I: integer;
begin
Result:='';
for I:=0 to High(IntrinArray) do
  if AnsiCompareText(IntrinArray[I].Name,Name)=0 then
	begin
	Result:=GetProtoType(I,True);
	break;
	end;
end;


function NameToIntrinNumber(Name: string): integer;
var I: integer;
begin
Result:=-1;
for I:=0 to High(IntrinArray) do
  if AnsiCompareText(IntrinArray[I].Name,Name)=0 then
	begin
	Result:=IntrinArray[I].Inx;
	break;
	end;
end;



procedure GetCodeDeclaration(var S: string; Proto: boolean);
var I: integer;
var S1,IntCode,RealCode: string;
var Line: string;


	procedure AddTabs(var S: string; TargetCols: integer);
	var Col: integer;
	begin
	Col:=CountColums(S,8);
	while Col<TargetCols do
		begin
		S:=S+Tab_Char;
		Col:=CountColums(S,8);
		end
	end;



	procedure StoreDeclaration(S: string; var CodeStore: string);
	var Col: integer;
	begin
	if Length(Line)<1 then Line:=Tab_Char+S
	else
		begin
		Col:=CountColums(Line,8);
		if Col>=56 then
			begin
			CodeStore:=CodeStore+Line+CRLF_Char;
			Line:=Tab_Char+S;
			end
		else if Col>=32 then
			begin
			AddTabs(Line,56);
			Line:=Line+S;
			end
		else
			begin
			AddTabs(Line,32);
			Line:=Line+S;
			end;
		end;
	end;



	function GetArgStr(C: char): string;
	begin
	case C of
	 'I','A': Result:='int';
	 'i','a': Result:='@int';
	 'R': Result:='real';
	 'r': Result:='@real';
	 'V': Result:='void';
	 end;
	end;


	function GetPrototype(Inx: integer): string;
	var S: string;
	var I: integer;

		procedure AddStr(S: string);
		begin
		if (Result<>'') and (Result[Length(Result)]<>':') then Result:=Result+',';
		Result:=Result+S;
		end;

	begin
	S:=IntrinArray[Inx].ArgTypes;
	if S='' then exit;
	AddStr(GetArgStr(S[1]));
	Result:=Result+':';
	for I:=3 to Length(S) do
		begin
		AddStr(GetArgStr(S[I]));
		end;
	Result:='('+Result+')';
	end;


	function GetAlias(Inx: integer): integer;
	var I: integer;
	begin
	for Result:=0 to High(IntrinAlias) do
	 if IntrinAlias[Result].Inx=Inx then exit;
	Result:=-1;
	end;

	function ExtractDeclaration(I: integer; var S: string; Last: boolean): boolean;
	var AI: integer;
	begin
	Result:=@IntrinArray[I].Proc<>@UnusedIntrinsic;
	if not Result then exit;
	S:=IntrinArray[I].Name;
	if Proto then S:=S+GetPrototype(I);
	S:=S+'='+IntToStr(IntrinArray[I].Inx);
	AI:=GetAlias(I);
	if AI>=0 then
		begin
		S:=S+',	'+IntrinAlias[AI].Name;
		if Proto then S:=S+GetPrototype(I);
		S:=S+'='+IntToStr(IntrinArray[I].Inx);
		end;
	if Last then S:=S+';' else S:=S+',';
	end;


	function GetTabStr(S: string): string;
	var J,TabCnt: integer;
	begin
	Result:='';
	TabCnt:=3 - (Length(S1) div 8);
	for J:=0 to TabCnt-1 do Result:=Result+Tab_Char;
	end;


begin
Line:='';
IntCode:='code'+CRLF_Char;
for I:=0 to High(IntrinArray) do
	begin
	if not ExtractDeclaration(I,S1,I>=High(IntrinArray)) then continue;
	if IntrinArray[I].Real then continue;
	StoreDeclaration(S1,IntCode);
	end;
if Line<>'' then IntCode:=IntCode+Line;
IntCode:=Trim(IntCode);
IntCode[Length(IntCode)]:=';';


Line:='';
RealCode:='code real'+CRLF_Char;
for I:=0 to High(IntrinArray) do
	begin
	if not ExtractDeclaration(I,S1,I>=High(IntrinArray)) then continue;
	if not IntrinArray[I].Real then continue;
	StoreDeclaration(S1,RealCode);
	end;
if Line<>'' then RealCode:=RealCode+Line;
RealCode:=Trim(RealCode);
RealCode[Length(RealCode)]:=';';

S:=IntCode+CRLF_Char+RealCode;
end;


var KeyDelay,KeySpeed: integer;


procedure InitializeIntrinsics(Console: TConsoleWindow);
begin
SetupLocalData;
GlobalInitialize(Console);
HexDigits:=IntNibbles;
Randomize;
RealFormat(5,5);
GlobalConsole.Reset;
Intialize3DGraphics(GlobalConsole.GLViewer);
WinSupport.GetKeyboardDynamics(KeyDelay,KeySpeed);
end;



procedure FinalizeIntrinsics;
begin
WinSupport.SetKeyboardDynamics(KeyDelay,KeySpeed);
ClearFileArray;
MemManager.Clear;
AbortConsoleFunctions;
GlobalConsole.VGABox.ShowCursor:=False;
GlobalConsole.EnableRefresh(True);
GlobalConsole.VGABox.Repaint;
end;


initialization
ClearFileArray;
ComArgs:=TStringList.Create;
GetWinCommandLine;
finalization
ComArgs.Free;
ClearFileArray;
end.
