unit IDEBitmapDlg;

interface

uses Windows, SysUtils, Classes, Graphics, Forms, Controls, StdCtrls, 
  Buttons, Dialogs, ExtCtrls, DSupport, BitmapLib, IDEVariables, IDEVariableLib, I2LInterp;

type
  TBitmapDlg = class(TForm)
    Panel1: TPanel;
    CancelBtn: TButton;
    Panel2: TPanel;
    ScrollBox1: TScrollBox;
    Image1: TImage;
    Panel10: TPanel;
    Panel13: TPanel;
    MouseDis: TLabel;
    Panel3: TPanel;
    SizeDis: TLabel;
    BitBtn1: TBitBtn;
    procedure CancelBtnClick(Sender: TObject);
    procedure Image1MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormCreate(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var BitmapDlg: TBitmapDlg;


procedure ShowBitmap(Info: TVariableInfo);
procedure UpdateBitmapDisplay; overload;
procedure UpdateBitmapDisplay(Info: TVariableInfo); overload;
procedure ResetBitmapDisplay;


implementation

uses EBXMain, XPLDataTypes;

{$R *.dfm}


var WatchInfo: TVariableInfo;


function GetArrayValue(Inx: integer; Add: integer; Info: TVariableInfo): integer;
begin
case Info.VarType of
 vtInteger: Result:=I2LProgram.GetDataWord(Add+Inx*IntBytes);
 vtIntConst: Result:=Info.Offset;
 vtRealConst: Result:=Trunc(Info.RealConst);
 vtChar: Result:=I2LProgram.DataByte[Add+Inx];
 vtReal: Result:=I2LProgram.GetDataWord(Add+Inx*8);
 end;
end;


type TBitmapStatus = (bsValid,bsInvalid,bsNonInteger,bsAccessError,bsScopeError);


function GetBitmapHandle(VI: TVariableInfo; var Handle: integer): TBitmapStatus;
var Status: TVarStatus;
var Add: integer;
begin
Result:=bsNonInteger;
if WatchInfo.VarType=vtRealConst then exit;
Result:=bsValid;
Handle:=VI.Offset;
if WatchInfo.VarType=vtIntConst then exit;
Status:=GetVariableAddress(VI,Add);
case Status of
 vsScopeError: Result:=bsScopeError;
 vsAccessError: Result:=bsAccessError;
 end;
if Status<>vsValidAddress then exit;
Result:=bsAccessError;
if not I2LProgram.RangeCheck(Add) then exit;
Result:=bsNonInteger;
if VI.DefFormat.DisplayType<>dsInteger then exit;
Handle:=I2LProgram.GetDataWord(Add);
Result:=bsInvalid;
if (Handle<0) or (Handle>=Form1.ConsoleWnd.VGABox.BitmapManager.BitmapList.Count) then exit;
Result:=bsValid
end;



procedure UpdateBitmapDisplay;
var BMHandle: integer;
var Status: TBitmapStatus;
var Msg: string;
var BM: TBitmap;
begin
if not BitmapDlg.Visible then exit;
Status:=GetBitmapHandle(WatchInfo,BMHandle);
Msg:=WatchInfo.BaseName+GetIndicesString(WatchInfo)+': '+IntToStr(BMHandle);
BitmapDlg.SizeDis.Caption:='?';
case Status of
 bsValid:
	begin
	BM:=TBitmap(Form1.ConsoleWnd.VGABox.BitmapManager.BitmapList[BMHandle]);
	if BM=nil then Msg:=Msg+' - Deallocated Handle'
	else
		begin
		BitmapDlg.Image1.Picture.Bitmap.Assign(BM);
		BitmapDlg.SizeDis.Caption:=IntToStr(BM.Width)+' X '+IntToStr(BM.Height);
		end;
	end;
 bsInvalid: Msg:=Msg+' - Invalid Handle';
 bsNonInteger: Msg:=Msg+' - Non-Integer';
 bsAccessError: Msg:=Msg+' - Access Error';
 bsScopeError: Msg:=Msg+' - Out of Scope';
 end;
if (Status<>bsValid) or (BM=nil) then
	begin
	BitmapDlg.Image1.Picture.Bitmap.ReleaseHandle;
	BitmapDlg.Image1.Invalidate;
	end;
BitmapDlg.Caption:=Msg;
end;



procedure UpdateBitmapDisplay(Info: TVariableInfo);
begin
WatchInfo:=Info;
UpdateBitmapDisplay;
end;


procedure ShowBitmap(Info: TVariableInfo);
begin
WatchInfo:=Info;
BitmapDlg.Show;
UpdateBitmapDisplay;
end;



procedure ResetBitmapDisplay;
begin
end;



procedure TBitmapDlg.CancelBtnClick(Sender: TObject);
begin
Close;
end;



procedure TBitmapDlg.Image1MouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
begin
MouseDis.Caption:=IntToStr(X)+' '+IntToStr(Y);
end;



procedure TBitmapDlg.FormCreate(Sender: TObject);
begin
Self.ScrollBox1.DoubleBuffered:=True;
end;

end.
