unit IDECompileMessages;

interface

uses Windows,Types,Graphics,Controls,StdCtrls,ComCtrls,SysUtils,ImgList,TextLibrary, XPLCompiler;

var MaxErrors: integer = 4;	{Maximum errors to display before aborting compile}

procedure HandleMessageSelect;
procedure IDEMessageHandler(Info: TMessageInfo; var Continue: boolean);
procedure MessagePostCompile;
procedure InitializeMessages;
procedure InsertMessage(Info: TMessageInfo; S: string);
procedure CustomDrawLBItem(Control: TWinControl; Index: Integer; Rect: TRect; State: TOwnerDrawState; ImageList: TImageList);
procedure ClearMessages(LB: TListBox; Search,Compiler: boolean);

implementation

uses I2LGlobal, EBXMain, IDEConfig, IDEGlobal;

var ErrorCount,WarningCount: integer;


function GetMessageTypeStr(MT: TXPLMessageType): string;
begin
case MT of
 xmError: Result:='Error';
 xmWarning: Result:='Warning';
 xmInfo: Result:='Info';
 xmRuntime: Result:='Runtime';
 xmSearch: Result:='Search';
 else Result:='Unknown';
 end;
end;


procedure ClearMessages(LB: TListBox; Search,Compiler: boolean);
var I: integer;
var MT: TXPLMessageType;
var MI: TMessageInfo;
begin
LB.Items.BeginUpdate;
try
for I:=LB.Items.Count-1 downto 0 do
	begin
	MI:=PMessageInfo(LB.Items.Objects[I])^;
	MT:=MI.MessageType;
	if Search and (MT=xmSearch) or
	   Compiler and (MT<>xmSearch) then
		begin
		{Data is disposed of in OnDeletion event}
		Dispose(PMessageInfo(LB.Items.Objects[I]));
		LB.Items.Delete(I);
		end;
	end;
finally LB.Items.EndUpdate; end;
end;




procedure HandleMessageSelect;
var Inx: integer;
var MI: TMessageInfo;
var Obj: TObject;
begin
Inx:=Form1.MessageView.ItemIndex;
if Inx<0 then exit;
MI:=PMessageInfo(Form1.MessageView.Items.Objects[Inx])^;
SetEditorPosition(MI.Info);
end;



procedure InsertMessage(Info: TMessageInfo; S: string);
var PM: PMessageInfo;
begin
New(PM);
PM^:=Info;
Form1.MessageView.Items.AddObject(S,TObject(PM));
end;



function LogError(MessType: TXPLMessageType; Num: integer): boolean;
begin
Result:=(MessType<>xmWarning) or IsWarningEnabled(Num);
end;




procedure IDEMessageHandler(Info: TMessageInfo; var Continue: boolean);
var S: string;
begin
Info.SearchCol1:=0;
Info.SearchCol2:=0;
S:=GetMessageTypeStr(Info.MessageType)+': ';
if Info.MessageType<>xmInfo then S:=S+ExtractFilename(Info.Info.Filename)+'(Ln:'+IntToStr(Info.Info.Line)+') ';
S:=S+Info.Description;
if LogError(Info.MessageType,Info.ErrorNumber) then InsertMessage(Info,S);
if Info.MessageType=xmError then Inc(ErrorCount);
if Info.MessageType=xmWarning then Inc(WarningCount);
Continue:=(ErrorCount<MaxErrors);
end;



procedure MessagePostCompile;
var I: integer;
var MI: TMessageInfo;
var Inx: integer;
begin
{Move errors to the top}
Inx:=0;
for I:=0 to Form1.MessageView.Items.Count-1 do
	begin
	MI:=PMessageInfo(Form1.MessageView.Items.Objects[I])^;
	if MI.MessageType<>xmSearch then
		begin
                Form1.MessageView.Items.Move(I,Inx);
		Inc(Inx);
		end;
	end;

{Move info to the top so compile summary items appear first}
Inx:=0;
for I:=1 to Form1.MessageView.Items.Count-1 do
	begin
	MI:=PMessageInfo(Form1.MessageView.Items.Objects[I])^;
	if MI.MessageType=xmInfo then
		begin
                Form1.MessageView.Items.Move(I,Inx);
		Inc(Inx);
		end;
	end;

{Find first actual error message and put highlight there}
for I:=0 to Form1.MessageView.Items.Count-1 do
	begin
	MI:=PMessageInfo(Form1.MessageView.Items.Objects[I])^;
	if (MI.MessageType=xmError) or
	   (MI.MessageType=xmRuntime) then
		begin
		Form1.MessageView.ItemIndex:=I;
		HandleMessageSelect;
		break;
		end;
	end;
end;


procedure InitializeMessages;
begin
ErrorCount:=0;
WarningCount:=0;
ClearMessages(Form1.MessageView,False,True);
end;


{----------------- Message Display Routines -----------------------------------}






procedure DrawImage(Canvas: TCanvas; Rect: TRect; ImageList: TCustomImageList; ImageIndex: Integer);
var CY: Integer;
begin
CY := Rect.Top + (Rect.Bottom - Rect.Top) div 2;
ImageList.Draw(Canvas, Rect.Left, CY - ImageList.Height div 2, ImageIndex, True);
end;



procedure DrawTreeNodeText(Text: string; Canvas: TCanvas; BackColor: TColor; Rect: TRect; X1,X2: integer);
var TW: TTextWriter;
var BM: TBitmap;
begin
BM:=TBitmap.Create;
try
BM.Canvas.Font:=Canvas.Font;
BM.Canvas.Brush.Color:=BackColor;

TW:=TTextWriter.Create;
try
BM.Width:=Rect.Right-Rect.Left;
BM.Height:=Rect.Bottom-Rect.Top;

TW.AddDefaultAttib(X1);
TW.AddAttribute(X2,BackColor,Canvas.Font.Color,[fsBold]);
TW.WriteText(BM.Canvas,Text,0,0);
Canvas.Draw(Rect.Left,Rect.Top,BM);
finally TW.Free; end;
finally BM.Free; end;
end;



procedure CustomDrawLBItem(Control: TWinControl; Index: Integer; Rect: TRect; State: TOwnerDrawState; ImageList: TImageList);
var TextRect,ImageRect: TRect;
var BackColor: TColor;
var Inx: integer;
var Info: TMessageInfo;
var LB: TListBox;
begin
LB:=TListBox(Control);
if Index>=LB.Items.Count then exit;
Info:=PMessageInfo(LB.Items.Objects[Index])^;
{we don't need to custom draw if it is not a search message}
if odSelected in State then BackColor:=clMenuHighlight else BackColor:=clWhite;

LB.Canvas.Pen.Color:=LB.Color;
LB.Canvas.Brush.Color:=LB.Color;
LB.Canvas.Rectangle(Rect);
TextRect:=Rect;
TextRect.Left:=ImageList.Width;

ImageRect:=Rect;
ImageRect.Right:=ImageList.Width;

DrawTreeNodeText(LB.Items[Index],LB.Canvas,BackColor,TextRect,Info.SearchCol1,Info.SearchCol2);

if odSelected in State then Inx:=1 else Inx:=0;
DrawImage(LB.Canvas,ImageRect, ImageList, Inx);

if odFocused in State then
	begin
	{The VCL will draw a focus rectangle, but it is done XOR}
	{So we draw it first so it is erased by the VCL}
	LB.Canvas.DrawFocusRect(Rect);
	{Then draw our own focus rectangle}
	LB.Canvas.DrawFocusRect(TextRect);
	end;
end;



end.
