unit IDEConfig;

interface

uses Windows, SysUtils, Classes, Graphics, Forms, Controls, StdCtrls, 
  Buttons, ExtCtrls, ComCtrls, Dialogs,IniFiles, Grids, CheckLst, DSupport,
  SpinButtons, intspin,BitmapLib,CodeEditor,ScrollEditor, CheckBoxExt,
  Autobttn,VCLLibrary, FancyGrids;

type
  TConfigDlg = class(TForm)
    Panel1: TPanel;
    OKBtn: TButton;
    CancelBtn: TButton;
    PageControl1: TPageControl;
    TabSheet1: TTabSheet;
    GroupBox4: TGroupBox;
    Label4: TLabel;
    EditorPathEdit: TEdit;
    BrowseEditBtn: TButton;
    OpenDialog1: TOpenDialog;
    TabSheet2: TTabSheet;
    MessageList: TCheckListBox;
    Label1: TLabel;
    TabSheet3: TTabSheet;
    GroupBox1: TGroupBox;
    CompileErrorSpin: TIntegerSpin;
    TabSheet4: TTabSheet;
    FontDialog1: TFontDialog;
    ColorDialog1: TColorDialog;
    GroupBox2: TGroupBox;
    ColorList: TListBox;
    ColorBtn: TBitBtn;
    FontBtn: TBitBtn;
    GroupBox3: TGroupBox;
    PreviewScroll: TScrollBox;
    PreviewBox: TPaintBox;
    BitBtn1: TBitBtn;
    GroupBox5: TGroupBox;
    AutoIntrinsicsChk: TCheckBox;
    PrototypeChk: TCheckBox;
    GroupBox6: TGroupBox;
    ScreenSizeBtn: TButton;
    TabSheet5: TTabSheet;
    Panel2: TPanel;
    Panel7: TPanel;
    DeleteBtn: TSpeedButton;
    FolderBtn: TSpeedButton;
    FileBtn: TSpeedButton;
    MoveUpBtn: TAutoButton;
    MoveDnBtn: TAutoButton;
    Panel3: TPanel;
    Label2: TLabel;
    OpenDialog2: TOpenDialog;
    ExamineBtn: TSpeedButton;
    EditBtn: TSpeedButton;
    GroupBox7: TGroupBox;
    Panel4: TPanel;
    HintSizeSpin: TIntegerSpin;
    LibList: TCheckGrid;
    procedure BrowseEditBtnClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure ColorListDrawItem(Control: TWinControl; Index: Integer;
      Rect: TRect; State: TOwnerDrawState);
    procedure FontBtnClick(Sender: TObject);
    procedure PreviewBoxPaint(Sender: TObject);
    procedure ColorBtnClick(Sender: TObject);
    procedure ScreenSizeBtnClick(Sender: TObject);
    procedure FileBtnClick(Sender: TObject);
    procedure MoveUpBtnClick(Sender: TObject);
    procedure MoveDnBtnClick(Sender: TObject);
    procedure DeleteBtnClick(Sender: TObject);
    procedure FolderBtnClick(Sender: TObject);
    procedure LibListDblClick(Sender: TObject);
    procedure EditBtnClick(Sender: TObject);
    procedure LibListDrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    function LibListGetCheck(ACol, ARow: Integer): Boolean;
    procedure LibListSetCheck(ACol, ARow: Integer; Checked,
      Update: Boolean);
    procedure LibListSelectCell(Sender: TObject; ACol, ARow: Integer;
      var CanSelect: Boolean);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  ConfigDlg: TConfigDlg;


procedure ShowConfigDlg;
procedure GetUserDefaults;
function IsWarningEnabled(Num: integer): boolean;
procedure ReadEditorDefaults(IniFile: TIniFile; Editor: TScrollEditor);

var EditorPath: string;

implementation

uses I2lIntrinsics, IDEGlobal, XPLCompiler, IDECompileMessages, EBXMain,
  IDEEditors, IDEDebugger, IDEVariableLib,OpenFolderDlg;

{$R *.dfm}


{------------------ Library Files ----------------------------------------------}



function GetLibTypeStr(LT: TLibType): string;
begin
case LT of
 ltFile: Result:='File';
 ltPath: Result:='Path';
 end;
end;



procedure SaveLibraryFiles(IniFile: TIniFile);
var I: integer;
begin
IniFile.EraseSection('Library');
IniFile.WriteInteger('Library','Count',LibFiles.Count);
for I:=0 to LibFiles.Count-1 do
	begin
	IniFile.WriteString('Library','File'+IntToStr(I),LibFiles[I]);
	IniFile.WriteBool('Library','Enabled'+IntToStr(I),TLibInfo(LibFiles.Objects[I]).Enabled);
	IniFile.WriteInteger('Library','LibType'+IntToStr(I),Integer(TLibInfo(LibFiles.Objects[I]).LibType));
	end;
end;


procedure SetRowCount;
var Count: integer;
begin
Count:=LibFiles.Count+1;
if Count<25 then Count:=25;
ConfigDlg.LibList.RowCount:=Count;
end;



procedure LoadLibraryFiles(IniFile: TIniFile);
var I,Count: integer;
begin
LibFiles.Clear;
Count:=IniFile.ReadInteger('Library','Count',LibFiles.Count);
for I:=0 to Count-1 do
	begin
	LibFiles.Add(IniFile.ReadString('Library','File'+IntToStr(I),''));
	TLibInfo(LibFiles.Objects[I]).Enabled:=IniFile.ReadBool('Library','Enabled'+IntToStr(I),True);
	TLibInfo(LibFiles.Objects[I]).LibType:=TLibType(IniFile.ReadInteger('Library','LibType'+IntToStr(I),Integer(ltFile)));
	end;
SetRowCount;
ConfigDlg.LibList.Invalidate;
end;



procedure TConfigDlg.FileBtnClick(Sender: TObject);
var I,Inx: integer;
begin
if OpenDialog2.Execute then
	begin
	for I:=0 to OpenDialog2.Files.Count-1 do
		begin
		Inx:=LibFiles.Add(OpenDialog2.Files[I]);
		TLibInfo(LibFiles.Objects[Inx]).Enabled:=True;
                TLibInfo(LibFiles.Objects[Inx]).LibType:=ltFile;
		end;
	SetRowCount;
	LibList.Invalidate;
	end;
end;



procedure TConfigDlg.FolderBtnClick(Sender: TObject);
var Inx: integer;
begin
FolderDlg.StartPath:=GetDefaultDir;
if FolderDlg.Execute then
	begin
	Inx:=LibFiles.Add(FolderDlg.NewPath);
	TLibInfo(LibFiles.Objects[Inx]).Enabled:=True;
	TLibInfo(LibFiles.Objects[Inx]).LibType:=ltPath;
	SetRowCount;
	LibList.Invalidate;
	end;
end;




procedure TConfigDlg.MoveUpBtnClick(Sender: TObject);
var Inx: integer;
begin
Inx:=LibList.Row-1;
if Inx<1 then exit;
LibFiles.Exchange(Inx,Inx-1);
LibList.Row:=Inx;
LibList.Invalidate;
end;


procedure TConfigDlg.MoveDnBtnClick(Sender: TObject);
var Inx: integer;
begin
Inx:=LibList.Row-1;
if (Inx<0) or (Inx>=LibFiles.Count-1) then exit;
LibFiles.Exchange(Inx,Inx+1);
LibList.Row:=Inx+2;
LibList.Invalidate;
end;




procedure TConfigDlg.DeleteBtnClick(Sender: TObject);
var Inx: integer;
var S: string;
begin
Inx:=LibList.Row-1;
if (Inx<0) or (Inx>=LibFiles.Count) then exit;
S:=ExtractFilename(LibFiles[Inx]);
if MessageDlg('Is it OK to delete: '+S, mtWarning, [mbYes, mbNo], 0)<>mrYes then exit;
LibFiles.Delete(Inx);
SetRowCount;
LibList.Invalidate;
end;


{==============================================================================}



procedure GetFactoryDefaults;
var I: integer;
begin
with ConfigDlg do
 begin
 for I:=0 to MessageList.Items.Count-1 do
	begin
	MessageList.Checked[I]:=I<>0;
	end;
 AutoIntrinsicsChk.Checked:=True;
 PrototypeChk.Checked:=True;
 end;
end;


const WarnSection = 'WarnMessage';


procedure GetWarningDefaults(IniFile: TIniFile);
var Cnt,I: integer;
begin
GetFactoryDefaults;
Cnt:=IniFile.ReadInteger(WarnSection,'Count',0);
with ConfigDlg do
 begin
 for I:=0 to Cnt-1 do
	begin
	MessageList.Checked[I]:=IniFile.ReadBool(WarnSection,'Message'+IntToStr(I),MessageList.Checked[I]);
	end;
 end;
end;



procedure SaveWarningDefaults(IniFile: TIniFile);
var I: integer;
begin
IniFile.EraseSection(WarnSection);
with ConfigDlg do
 begin
 IniFile.WriteInteger(WarnSection,'Count',MessageList.Items.Count);
 for I:=0 to MessageList.Items.Count-1 do
	begin
	IniFile.WriteBool(WarnSection,'Message'+IntToStr(I),MessageList.Checked[I]);
	end;
 end;
end;

{----------------- Initialization Routines ---------------------}


function EncodeFontStyle(S:TFontStyles): integer;
{Encode font style bit pattern }
begin
Result:=0;
if (S*[fsBold])=[fsBold] then Result:=1;
if (S*[fsItalic])=[fsItalic] then Result:=Result or 2;
if (S*[fsUnderline])=[fsUnderline] then Result:=Result or 4;
if (S*[fsStrikeout])=[fsStrikeout] then Result:=Result or 8;
end;


function DecodeFontStyle(S:integer): TFontStyles;
{Decode font style bit pattern }
begin
Result:=[];
if (S and 1) <> 0 then Result:=Result + [fsBold];
if (S and 2) <> 0 then Result:=Result + [fsItalic];
if (S and 4) <> 0 then Result:=Result + [fsUnderline];
if (S and 8) <> 0 then Result:=Result + [fsStrikeout];
end;


procedure ReadEditorDefaults(IniFile: TIniFile; Editor: TScrollEditor);
begin
Editor.Font.Color:=IniFile.ReadInteger('Screen','FontColor',clBlack);
Editor.Color:=IniFile.ReadInteger('Screen','ScreenColor',clWhite);
Editor.HighlightColor:=IniFile.ReadInteger('Screen','HiLiteColor',clNavy);
Editor.ControlColor:=IniFile.ReadInteger('Screen','ControlColor',clRed);
Editor.ShadowColor:=IniFile.ReadInteger('Screen','ShadowColor',RGB(230,240,250));
Editor.NonDeleteColor:=IniFile.ReadInteger('Screen','NonDeleteColor',clBlack);

Editor.Font.Size:=IniFile.ReadInteger('Screen','FontSize',9);
Editor.Font.Name:=IniFile.ReadString('Screen','FontName','FixedSys');
Editor.CursorWidth:=IniFile.ReadInteger('Screen','CursorWidth',3);
Editor.ShadowLine:=IniFile.ReadBool('Screen','ShadowLine',True);
Editor.SpacesPerTab:=IniFile.ReadInteger('Screen','SpacesPerTab',8);
Editor.Font.Style:=DecodeFontStyle(IniFile.ReadInteger('Screen','FontStyle',EncodeFontStyle([])));
GlobalHinter.FontSize:=IniFile.ReadInteger('Hinter','FontSize',GlobalHinter.FontSize);
Editor.Invalidate;
end;



procedure GetUserDefaults;
{Preset user configurable defaults  }
var XPLIDEIni: TIniFile;
begin
XPLIDEIni:=TIniFile.Create(IDEIniFilename);
with XPLIDEIni do
 begin
 EditorPath :=ReadString('Path','EditorPath', '');
 GetWarningDefaults(XPLIDEIni);
 MaxErrors:=ReadInteger('General','MaxErrors',4);
 ReadEditorDefaults(XPLIDEIni,Form1.SourceEdit.Editor);
 ConfigDlg.AutoIntrinsicsChk.Checked:=ReadBool('General','AutoIntrinsics',True);
 ConfigDlg.PrototypeChk.Checked:=ReadBool('General','PrototypeIntrinsics',True);
 Form1.Width:=ReadInteger('Form','FormWidth',636);
 Form1.Height:=ReadInteger('Form','FormHeight',526);
 LoadLibraryFiles(XPLIDEIni);
 end;
XPLIDEIni.Free;
end;





procedure SaveScreenSize;
{Save current screen size as default screen size }
var IniFile: TIniFile;
begin
IniFile:=TIniFile.Create(IDEIniFilename);
IniFile.WriteInteger('Form','FormWidth',Form1.Width);
IniFile.WriteInteger('Form','FormHeight',Form1.Height);
IniFile.Free;
end;



procedure SaveUserDefaults;
{Save user configurable defaults  }
var XPLIDEIni: TIniFile;
begin
XPLIDEIni:=TIniFile.Create(IDEIniFilename);
with XPLIDEIni do
 begin
 WriteString('Path','EditorPath',EditorPath);
 WriteInteger('General','MaxErrors',MaxErrors);
 WriteBool('General','AutoIntrinsics',ConfigDlg.AutoIntrinsicsChk.Checked);
 WriteBool('General','PrototypeIntrinsics',ConfigDlg.PrototypeChk.Checked);

 SaveWarningDefaults(XPLIDEIni);

 {We only need to save the one editors's settings }
 WriteInteger('Screen','FontColor',Form1.SourceEdit.Editor.Font.Color);
 WriteInteger('Screen','ScreenColor',Form1.SourceEdit.Editor.Color);
 WriteInteger('Screen','HiLiteColor',Form1.SourceEdit.Editor.HighlightColor);
 WriteInteger('Screen','ControlColor',Form1.SourceEdit.Editor.ControlColor);
 WriteInteger('Screen','ShadowColor',Form1.SourceEdit.Editor.ShadowColor);
 WriteInteger('Screen','NonDeleteColor',Form1.SourceEdit.Editor.NonDeleteColor);

 WriteInteger('Screen','FontSize',Form1.SourceEdit.Editor.Font.Size);
 WriteString('Screen','FontName',Form1.SourceEdit.Editor.Font.Name);
 WriteInteger('Screen','FontStyle',EncodeFontStyle(Form1.SourceEdit.Editor.Font.Style));
 WriteInteger('Hinter','FontSize',GlobalHinter.FontSize);

 SaveLibraryFiles(XPLIDEIni);
 end;
XPLIDEIni.Free;
end;






procedure BrowseExecutables(Caption: String;var FName: string);
begin
with ConfigDlg.OpenDialog1 do
 begin
 Filter:='Executables (*.EXE)|*.EXE|All Files(*.*)|*.*';
 Title:=Caption;
 InitialDir:=ExtractFilePath(FName);
 Filename:=ExtractFilename(FName);
 Options:=[ofHideReadOnly,ofPathMustExist,ofFileMustExist];
 if Execute then FName:=Filename;
 end;
end;




procedure TConfigDlg.BrowseEditBtnClick(Sender: TObject);
var T: String;
begin
T:=EditorPathEdit.Text;
BrowseExecutables('Locate Editor',T);
EditorPathEdit.Text:=T;
end;



function IsWarningEnabled(Num: integer): boolean;
begin
Result:=ConfigDlg.MessageList.Checked[Num-1];
end;

{==============================================================================}

type TColorInfo = record
 Color: TColor;
 Text: string;
 end;

type TPreviewInfo = record
 ForeColor,BackColor: TColor;
 Text: string;
 end;

var ColorArray: array [0..5] of TColorInfo = (
	(Color: clBlack; Text: 'Text'),
	(Color: clWhite; Text: 'Screen'),
	(Color: clNavy; Text: 'Highlight'),
	(Color: clRed; Text: 'Control Chars'),
	(Color: clSilver; Text: 'Shadow Line'),
	(Color: clSilver; Text: 'Search Highlight'));


var PreviewArray: array [0..5] of TPreviewInfo = (
	(ForeColor: clWhite; BackColor: clBlack; Text: 'Normal Text'),
	(ForeColor: clWhite; BackColor: clBlack; Text: 'Highlighted Text'),
	(ForeColor: clWhite; BackColor: clBlack; Text: 'Control Characters'),
	(ForeColor: clWhite; BackColor: clBlack; Text: 'Highlighted Control'),
	(ForeColor: clWhite; BackColor: clBlack; Text: 'Shadowed Line'),
	(ForeColor: clWhite; BackColor: clBlack; Text: 'Search Highlight'));


var FontTemp: TFont;


procedure UpdatePreview;
var H,W: integer;
var I: integer;
begin
with ConfigDlg do
 begin
 PreviewBox.Canvas.Font:=FontTemp;
 PreviewScroll.Color:=PreviewArray[0].BackColor;
 H:=PreviewBox.Canvas.TextHeight(PreviewArray[0].Text);
 W:=0;
 PreviewBox.Canvas.Font:=FontTemp;
 for I:=0 to High(PreviewArray) do
  if PreviewBox.Canvas.TextWidth(PreviewArray[I].Text)>W then W:=PreviewBox.Canvas.TextWidth(PreviewArray[I].Text);

 if (H*Length(PreviewArray))>PreviewScroll.Height then PreviewBox.Height:=H*Length(PreviewArray) else PreviewBox.Height:=PreviewScroll.Height;
 if W>PreviewScroll.Width then PreviewBox.Width:=W
 else PreviewBox.Width:=PreviewScroll.Width;

 for I:=0 to High(PreviewArray) do
	begin
	PreviewBox.Canvas.Brush.Color:=PreviewArray[I].BackColor;
	PreviewBox.Canvas.Font.Color:=PreviewArray[I].ForeColor;
	PreviewBox.Canvas.TextOut(3,I*H,PreviewArray[I].Text);
	end;
 end;
end;


procedure TConfigDlg.ColorListDrawItem(Control: TWinControl;
  Index: Integer; Rect: TRect; State: TOwnerDrawState);
var Offset: Integer;
const B=3;
begin
ColorList.Canvas.FillRect(Rect);
Offset := 2;
ColorList.Canvas.TextOut(Rect.Left + Offset, Rect.Top, ColorArray[Index].Text);
ColorList.Canvas.Brush.Color:=ColorArray[Index].Color;
ColorList.Canvas.Rectangle(Rect.Right-20,Rect.Top+B,Rect.Right-B,Rect.Bottom-B);
end;



procedure SetupPreview;
begin
PreviewArray[0].ForeColor:=ColorArray[0].Color;	{Normal Text}
PreviewArray[0].BackColor:=ColorArray[1].Color;

PreviewArray[1].ForeColor:=ContrastColor(ColorArray[2].Color);	{Highlighted Text}
PreviewArray[1].BackColor:=ColorArray[2].Color;

PreviewArray[2].ForeColor:=ContrastColor(ColorArray[3].Color);	{Control Characters}
PreviewArray[2].BackColor:=ColorArray[3].Color;

PreviewArray[3].ForeColor:=clWhite xor clWhite;	{Highlighted Control}
PreviewArray[3].BackColor:=ColorArray[3].Color xor clWhite;

PreviewArray[4].ForeColor:=ColorArray[0].Color;	{Shadowed Line}
PreviewArray[4].BackColor:=ColorArray[4].Color;

PreviewArray[5].ForeColor:=ContrastColor(ColorArray[5].Color);	{Search Highlight}
PreviewArray[5].BackColor:=ColorArray[5].Color;
end;




procedure GetColors;
begin
ColorArray[0].Color:=Form1.SourceEdit.Editor.Font.Color;	{Text}
ColorArray[1].Color:=Form1.SourceEdit.Editor.Color;		{Screen}
ColorArray[2].Color:=Form1.SourceEdit.Editor.HighlightColor;	{Highlight}
ColorArray[3].Color:=Form1.SourceEdit.Editor.ControlColor;	{Control Chars}
ColorArray[4].Color:=Form1.SourceEdit.Editor.ShadowColor;	{Shadow Line}
ColorArray[5].Color:=Form1.SourceEdit.Editor.NonDeleteColor;	{Shadow Line}

FontTemp:=Form1.SourceEdit.Editor.Font;
FontTemp.Color:=Form1.SourceEdit.Editor.Font.Color;
SetupPreview;
end;




procedure SetColors(var Editor: TScrollEditor);
begin
Editor.Font.Color:=ColorArray[0].Color;
Editor.Color:=ColorArray[1].Color;
Editor.HighlightColor:=ColorArray[2].Color;
Editor.ControlColor:=ColorArray[3].Color;
Editor.ShadowColor:=ColorArray[4].Color;
Editor.NonDeleteColor:=ColorArray[5].Color;
Editor.Font:=FontTemp;
Editor.Invalidate;
end;

procedure TConfigDlg.FontBtnClick(Sender: TObject);
begin
FontDialog1.Font:=FontTemp;
if FontDialog1.Execute then
	begin
	ColorArray[0].Color:=FontDialog1.Font.Color;
	FontTemp:=FontDialog1.Font;
	SetupPreview;
	ColorList.Invalidate;
	PreviewBox.Invalidate;
	end;
end;


procedure TConfigDlg.PreviewBoxPaint(Sender: TObject);
begin
UpdatePreview;
end;



procedure TConfigDlg.ColorBtnClick(Sender: TObject);
begin
ColorDialog1.Color:=ColorArray[ColorList.ItemIndex].Color;
if ColorDialog1.Execute then
	begin
	ColorArray[ColorList.ItemIndex].Color:=ColorDialog1.Color;
	SetupPreview;
	ColorList.Invalidate;
	PreviewBox.Invalidate;
	end;
end;



{==============================================================================}


procedure TConfigDlg.FormCreate(Sender: TObject);
var I: integer;
begin
PageControl1.ActivePageIndex:=0;
ColorList.ItemIndex:=0;
ColorList.Clear;
for I:=0 to High(ColorArray) do ColorList.Items.Add(ColorArray[I].Text);

MessageList.Clear;
for I:=1 to High(WarningStrings) do
	begin
	MessageList.Items.Add(WarningStrings[I]);
	end;
GetUserDefaults;
end;


procedure SetIncludeEditorColors;
var Iterator: TIncludeIterator;
var Editor: TCodeEditor;
begin
Iterator:=TIncludeIterator.Create(Form1.EditorPageControl);
try
while Iterator.GetIteration(Editor) do SetColors(Editor.Editor);
finally Iterator.Free; end;
end;





procedure ShowConfigDlg;
begin
GetColors;
ConfigDlg.EditorPathEdit.Text:=EditorPath;
ConfigDlg.CompileErrorSpin.Value:=MaxErrors;
ConfigDlg.HintSizeSpin.Value:=GlobalHinter.FontSize;
ConfigDlg.ShowModal;
if ConfigDlg.ModalResult=mrOK then
	begin
	EditorPath:=ConfigDlg.EditorPathEdit.Text;
	MaxErrors:=ConfigDlg.CompileErrorSpin.Value;
	GlobalHinter.FontSize:=ConfigDlg.HintSizeSpin.Value;
	SetColors(Form1.SourceEdit.Editor);
	SetIncludeEditorColors;
	SaveUserDefaults;
	end;
end;








procedure TConfigDlg.ScreenSizeBtnClick(Sender: TObject);
begin
SaveScreenSize;
end;



procedure TConfigDlg.LibListDblClick(Sender: TObject);
var Inx: integer;
var Info: TLibInfo;
begin
Inx:=LibList.Row-1;
if (Inx<0) or (Inx>=LibFiles.Count) then exit;
Info:=TLibInfo(LibFiles.Objects[Inx]);
case Info.LibType of
 ltFile:
	begin
	FindOrCreateEditor(LibFiles[Inx]);
	Form1.ReopenMenu2.AddMenuItem(LibFiles[Inx]);
	Form1.ReopenMenu2.SaveSubMenu;
	end;
 ltPath:
	begin
	OpenDialog2.InitialDir:=LibFiles[Inx];
	if OpenDialog2.Execute then
		begin
		FindOrCreateEditor(OpenDialog2.Filename);
		Form1.ReopenMenu2.AddMenuItem(LibFiles[Inx]);
		Form1.ReopenMenu2.SaveSubMenu;
		end;
	end;
 end;
end;



procedure TConfigDlg.EditBtnClick(Sender: TObject);
var Inx: integer;
var Info: TLibInfo;
begin
Inx:=LibList.Row-1;
if (Inx<0) or (Inx>=LibFiles.Count) then exit;
Info:=TLibInfo(LibFiles.Objects[Inx]);
case Info.LibType of
 ltFile:
	begin
	OpenDialog2.InitialDir:=LibFiles[Inx];
	if OpenDialog2.Execute then
		begin
		LibFiles[Inx]:=OpenDialog2.Filename;
		LibList.Invalidate;
		end;
	end;
 ltPath:
	begin
	FolderDlg.StartPath:=LibFiles[Inx];
	if FolderDlg.Execute then
		begin
		LibFiles[Inx]:=FolderDlg.NewPath;
		LibList.Invalidate;
		end;
	end;
 end;
end;


procedure ExcludeDrawCell(Grid: TDrawGrid; ACol, ARow: Integer; Rect: TRect; State: TGridDrawState);
var Inx: integer;
var Info: TLibInfo;
begin
Inx:=ARow-1;
if ARow = 0 then
	begin
	case ACol of
	 0: OwnerDrawItem(Grid.Canvas,'#',Rect);
	 1: OwnerDrawItem(Grid.Canvas,'Use',Rect);
	 2: OwnerDrawItem(Grid.Canvas,'Type',Rect);
	 3: OwnerDrawItem(Grid.Canvas,'File/Folder Name',Rect);
	 4: OwnerDrawItem(Grid.Canvas,'Path',Rect);
	 end;
	end
else
	begin
	if (Inx<0) or (Inx>=LibFiles.Count) then exit;
	Info:=TLibInfo(LibFiles.Objects[Inx]);

       	case ACol of
	 0: OwnerDrawItem(Grid.Canvas,IntToStr(ARow),Rect);
	 1: ;
	 2: OwnerDrawItem(Grid.Canvas,GetLibTypeStr(Info.LibType),Rect);
	 3: OwnerDrawItem(Grid.Canvas,ExtractFilename(LibFiles[Inx]),Rect);
	 4: OwnerDrawItem(Grid.Canvas,LibFiles[Inx],Rect);
       	 end;
	end;
end;


procedure TConfigDlg.LibListDrawCell(Sender: TObject; ACol,
  ARow: Integer; Rect: TRect; State: TGridDrawState);
begin
ExcludeDrawCell(LibList, ACol, ARow, Rect, State);
end;


function TConfigDlg.LibListGetCheck(ACol, ARow: Integer): Boolean;
var Inx: integer;
begin
Inx:=ARow-1;
Result:=False;
if Inx<0 then exit;
if Inx>=LibFiles.Count then exit;
Result:=TLibInfo(LibFiles.Objects[Inx]).Enabled;
end;



procedure TConfigDlg.LibListSetCheck(ACol, ARow: Integer; Checked,
  Update: Boolean);
var Inx: integer;
begin
Inx:=ARow-1;
TLibInfo(LibFiles.Objects[Inx]).Enabled:=Checked;
if Update then LibList.Invalidate; 
end;


procedure TConfigDlg.LibListSelectCell(Sender: TObject; ACol,
  ARow: Integer; var CanSelect: Boolean);
begin
CanSelect:=(ARow-1)<LibFiles.Count;
end;


end.
