unit IDEDebugFile;

interface

uses Forms,Classes,SysUtils,Graphics,ActiveX,XMLDoc,XMLIntf,XMLDocLibrary,IDEGlobal,
     IDEVariableLib,ScrollEditor;


function SaveXMLDebugInfo(Filename: string): boolean;
function LoadXMLDebugInfo(Filename: string): boolean;


implementation

uses IDEDebugger,IDEHeapAnalyzer,XPLDataTypes;


{------------------ Save Watches ----------------------------------------------}


procedure SaveUserFormat(Node: IXMLNode; Format: TVarFormat);
{Save specified format value}
var ChildXMLNode: IXMLNode;
var I: integer;
begin
ChildXMLNode:= Node.AddChild('UserFormat');
SaveTagXMLInteger(ChildXMLNode,'DisplayType',integer(Format.DisplayType));
SaveTagXMLInteger(ChildXMLNode,'Digits',Format.Digits);
SaveTagXMLInteger(ChildXMLNode,'Radix',integer(Format.Radix));
SaveTagXMLBoolean(ChildXMLNode,'ForcedArray',Format.ForcedArray);
ChildXMLNode:=ChildXMLNode.AddChild('Dimensions');
ChildXMLNode.Attributes['Count']:=Length(Format.DisplayDims);
for I:=0 to High(Format.DisplayDims) do
	begin
	SaveTagXMLInteger(ChildXMLNode,'Dim',Format.DisplayDims[I]);
	end;
end;



procedure SaveWatch(Node: IXMLNode; Watch: TWatchInfo);
{Save one watch item}
var ChildXMLNode: IXMLNode;
begin
ChildXMLNode:= Node.AddChild('Watch');
SaveTagXMLString(ChildXMLNode,'VarText',Watch.VarText);
SaveTagXMLString(ChildXMLNode,'Filename',Watch.Filename);
SaveTagXMLInteger(ChildXMLNode,'Line',Watch.Line);
SaveTagXMLBoolean(ChildXMLNode,'CustomFormat',Watch.CustomFormat);
SaveUserFormat(ChildXMLNode,Watch.UserFormat);
end;


procedure SaveWatches(Node: IXMLNode; Watches: TWatchArray);
{Save all watches in specified array}
var ChildXMLNode: IXMLNode;
var I: integer;
begin
ChildXMLNode:=Node.AddChild('Watches');
ChildXMLNode.Attributes['Count']:=Length(Watches);
for I:=0 to High(Watches) do
	begin
	SaveWatch(ChildXMLNode,Watches[I]);
	end;
end;


{----------------- Save Break points ------------------------------------------}


procedure SaveBreakPoint(Node: IXMLNode; BreakPoint: TBreakInfo);
{Save Single break point }
var ChildXMLNode: IXMLNode;
begin
ChildXMLNode:= Node.AddChild('BreakPoint');
SaveTagXMLString(ChildXMLNode,'Filename',BreakPoint.Filename);
SaveTagXMLInteger(ChildXMLNode,'Line',BreakPoint.Line);
SaveTagXMLInteger(ChildXMLNode,'PC',BreakPoint.PC);
end;




procedure SaveBreakPoints(Node: IXMLNode; BreakArray: TBreakArray);
{Save all break points}
var ChildXMLNode: IXMLNode;
var I: integer;
begin
ChildXMLNode:=Node.AddChild('BreakPoints');
ChildXMLNode.Attributes['Count']:=Length(BreakArray);
for I:=0 to High(BreakArray) do
	begin
	SaveBreakPoint(ChildXMLNode,BreakArray[I]);
	end;
end;

{------------------ Save Heap Info --------------------------------------------}

procedure SaveHeapInfo(Node: IXMLNode; HeapInfo: THeapInfo);
{Save Single Heap Info Item }
var ChildXMLNode: IXMLNode;
begin
ChildXMLNode:= Node.AddChild('HeapInfo');
SaveTagXMLString(ChildXMLNode,'Name',HeapInfo.Name);
SaveTagXMLInteger(ChildXMLNode,'Address',HeapInfo.Address);
SaveTagXMLInteger(ChildXMLNode,'DataType',Integer(HeapInfo.DataType));
SaveTagXMLBoolean(ChildXMLNode,'Indirect',HeapInfo.Indirect);
end;



procedure SaveHeapArray(Node: IXMLNode; HeapArray: THeapInfoArray);
{Save all break points}
var ChildXMLNode: IXMLNode;
var I: integer;
begin
ChildXMLNode:=Node.AddChild('HeapArray');
ChildXMLNode.Attributes['Count']:=Length(HeapArray);
for I:=0 to High(HeapArray) do
	begin
	SaveHeapInfo(ChildXMLNode,HeapArray[I]);
	end;
end;

{------------------------------------------------------------------------------}


function SaveXMLDebugInfo(Filename: string): boolean;
{Save Debug Info to specified XML file}
var XMLDoc: TXMLDocument;
var XMLNode: IXMLNode;
begin
Result:=False;
XMLDoc:= TXMLDocument.Create(nil);
XMLDoc.Options:=[doNodeAutoCreate,doNodeAutoIndent,doAttrNull,doAutoPrefix,doNamespaceDecl];
XMLDoc.Active:= True;
XMLNode:= XMLDoc.AddChild('XPLDebug');
XMLNode.Attributes['Version']:= '1.0';
SaveWatches(XMLNode,WatchArray);
SaveBreakPoints(XMLNode,BreakArray);
SaveHeapArray(XMLNode,HeapInfoArray);
XMLDoc.SaveToFile(Filename);
Result:=True;
end;

{=============== Read Routines ================================================}

{-------------- Read Watches --------------------------------------------------}

procedure LoadDimensions(XMLNode: IXMLNode; var Format: TVarFormat);
var ChildNode: IXMLNode;
begin
SetLength(Format.DisplayDims,0);
ChildNode:=XMLNode.ChildNodes.First;
while ChildNode <> nil do
	begin
	if ChildNode.LocalName='Dim' then
		begin
		SetLength(Format.DisplayDims,Length(Format.DisplayDims)+1);
		Format.DisplayDims[High(Format.DisplayDims)]:=ReadXMLInteger(ChildNode);
		ChildNode:=ChildNode.NextSibling;
		end;
	end;
end;


procedure LoadUserFormat(XMLNode: IXMLNode; var Format: TVarFormat);
{Save specified format value}
var ChildNode: IXMLNode;
begin
ChildNode:=XMLNode.ChildNodes.First;
while ChildNode <> nil do
	begin
	if ChildNode.LocalName='DisplayType' then Format.DisplayType:=TDisplayType(ReadXMLInteger(ChildNode))
	else if ChildNode.LocalName='Digits' then Format.Digits:=ReadXMLInteger(ChildNode)
	else if ChildNode.LocalName='Radix' then Format.Radix:=TIntegerFormat(ReadXMLInteger(ChildNode))
	else if ChildNode.LocalName='ForcedArray' then Format.ForcedArray:=ReadXMLBoolean(ChildNode)
	else if ChildNode.LocalName='Dimensions' then LoadDimensions(ChildNode,Format);
	ChildNode:=ChildNode.NextSibling;
	end;
end;


procedure LoadWatch(XMLNode: IXMLNode; var WatchArray: TWatchArray);
var ChildNode: IXMLNode;
var Inx: integer;
begin
SetLength(WatchArray,Length(WatchArray)+1);
Inx:=High(WatchArray);
ChildNode:=XMLNode.ChildNodes.First;
while ChildNode <> nil do
	begin
	if ChildNode.LocalName='VarText' then WatchArray[Inx].VarText:=ReadXMLString(ChildNode)
	else if ChildNode.LocalName='Filename' then WatchArray[Inx].Filename:=ReadXMLString(ChildNode)
	else if ChildNode.LocalName='Line' then WatchArray[Inx].Line:=ReadXMLInteger(ChildNode)
	else if ChildNode.LocalName='CustomFormat' then WatchArray[Inx].CustomFormat:=ReadXMLBoolean(ChildNode)
	else if ChildNode.LocalName='UserFormat' then LoadUserFormat(ChildNode,WatchArray[Inx].UserFormat);
	ChildNode:=ChildNode.NextSibling;
	end;
end;


procedure LoadWatches(XMLNode: IXMLNode);
var ChildNode: IXMLNode;
begin
SetLength(WatchArray,0);
ChildNode:=XMLNode.ChildNodes.First;
while (ChildNode <> nil) and (ChildNode.LocalName='Watch') do
	begin
	LoadWatch(ChildNode,WatchArray);
	ChildNode:= ChildNode.NextSibling;
	end;
end;



{----------------- Read break points ------------------------------------------}

procedure LoadBreakPoint(XMLNode: IXMLNode; var BreakArray: TBreakArray);
var ChildNode: IXMLNode;
var Inx: integer;
var ED: TScrollEditor;
begin
SetLength(BreakArray,Length(BreakArray)+1);
Inx:=High(BreakArray);
ChildNode:=XMLNode.ChildNodes.First;
while ChildNode <> nil do
	begin
	if ChildNode.LocalName='Filename' then BreakArray[Inx].Filename:=ReadXMLString(ChildNode)
	else if ChildNode.LocalName='Line' then BreakArray[Inx].Line:=ReadXMLInteger(ChildNode)
	else if ChildNode.LocalName='PC' then BreakArray[Inx].PC:=ReadXMLInteger(ChildNode);
	ChildNode:= ChildNode.NextSibling;
	end;
ED:=GetDebugEditor(BreakArray[Inx].Filename);
ED.RowMarks.AddLine('Breaks',BreakArray[Inx].Line,clRed);
end;


procedure LoadBreakPoints(XMLNode: IXMLNode);
var ChildNode: IXMLNode;
begin
SetLength(BreakArray,0);
ChildNode:=XMLNode.ChildNodes.First;
while (ChildNode <> nil) and (ChildNode.LocalName='BreakPoint') do
	begin
	LoadBreakPoint(ChildNode,BreakArray);
	ChildNode:= ChildNode.NextSibling;
	end;
end;

{------------------Read Heap Info ---------------------------------------------}

procedure LoadHeapInfo(XMLNode: IXMLNode; var HeapArray: THeapInfoArray);
var ChildNode: IXMLNode;
var Inx: integer;
begin
SetLength(HeapArray,Length(HeapArray)+1);
Inx:=High(HeapArray);
ChildNode:=XMLNode.ChildNodes.First;
while ChildNode <> nil do
	begin
	if ChildNode.LocalName='Name' then HeapArray[Inx].Name:=ReadXMLString(ChildNode)
	else if ChildNode.LocalName='Address' then HeapArray[Inx].Address:=ReadXMLInteger(ChildNode)
	else if ChildNode.LocalName='DataType' then HeapArray[Inx].DataType:=TDataType(ReadXMLInteger(ChildNode))
    	else if ChildNode.LocalName='Indirect' then HeapArray[Inx].Indirect:=ReadXMLBoolean(ChildNode);
	ChildNode:=ChildNode.NextSibling;
	end;
end;


procedure LoadHeapArray(XMLNode: IXMLNode);
var ChildNode: IXMLNode;
begin
SetLength(HeapInfoArray,0);
ChildNode:=XMLNode.ChildNodes.First;
while (ChildNode <> nil) and (ChildNode.LocalName='HeapInfo') do
	begin
	LoadHeapInfo(ChildNode,HeapInfoArray);
	ChildNode:= ChildNode.NextSibling;
	end;
end;


{------------------------------------------------------------------------------}


procedure LoadDebugInfo(ChildNode: IXMLNode);
var Rows: integer;
begin
while ChildNode <> nil do
	begin
	if ChildNode.LocalName='Watches' then LoadWatches(ChildNode)
	else if ChildNode.LocalName='BreakPoints' then LoadBreakPoints(ChildNode)
	else if ChildNode.LocalName='HeapArray' then LoadHeapArray(ChildNode);
	ChildNode:= ChildNode.NextSibling;
	end;
Rows:=Length(WatchArray)+1;
if Rows<30 then Rows:=30;
DebugDlg.WatchGrid.RowCount:=Rows+1;

Rows:=Length(BreakArray)+1;
if Rows<30 then Rows:=30;
DebugDlg.BreakGrid.RowCount:=Rows+1;

UpdateWatchGrid;
DebugDlg.BreakGrid.Invalidate;
end;



procedure LoadDebugInfoStream(Stream: TStream);
{Insert specified XML stream at specified Tree node }
var XMLNode: IXMLNode;
var XMLDoc: TXMLDocument;
begin
XMLDoc:=TXMLDocument.Create(Application);
try
XMLDoc.Options:=XMLDoc.Options + [doAttrNull];
XMLDoc.LoadFromStream(Stream);
XMLDoc.Active:= True;
if XMLDoc.ChildNodes.First = nil then raise Exception.Create('Empty File');

XMLNode:= XMLDoc.DocumentElement.ChildNodes.First;
LoadDebugInfo(XMLNode);
XMLDoc.Active:= False;
finally XMLDoc.Free; end;
end;



function LoadXMLDebugInfo(Filename: string): boolean;
{Load specified file into tree}
var Stream: TMemoryStream;
begin
Stream:=TMemoryStream.Create;
Stream.LoadFromFile(Filename);
try
LoadDebugInfoStream(Stream);
finally Stream.Free; end;
end;





initialization
CoInitialize(nil);
finalization
CoUnInitialize;
end.
