unit IDEExpression;

interface

uses SysUtils,DSupport;

function EvaluateExpression(ExpressStr,Filename: string; Line: integer): integer;


implementation

uses IDEVariableLib, I2LGlobal, I2LInterp, XPLDataTypes, XPLCompiler,
  IDEVariables;

var StatusStr: string;


{--------------- Get variable values, integer only ---------------------------}



function GetConstantValue(Info: TVariableInfo; var I: integer): boolean;
begin
Result:=True;
case Info.VarType of
 vtIntConst: I:=Info.Offset;
 vtRealConst: I:=Trunc(Info.RealConst);
 else Result:=False;
 end;
end;



function GetArrayValue(Inx: integer; Add: integer; Info: TVariableInfo): integer;
{Get single array element based on data type}
var Size: integer;
begin
case Info.DefFormat.DisplayType of
 dsCharString:
	begin
	if ProgramStore.RangeCheck(Add+Inx) then Result:=I2LProgram.DataByte[Add+Inx]
	else StatusStr:='Access Error';
	end;
 dsString:
	begin
	if Info.VarType=vtReal then Size:=RealSize else Size:=IntBytes;
	if ProgramStore.RangeCheck(Add+Inx*Size) then Result:=I2LProgram.GetDataWord(Add+Inx*Size)
	else StatusStr:='Access Error';
	end;
 dsInteger:
	begin
	if ProgramStore.RangeCheck(Add+Inx*IntBytes) then Result:=ProgramStore.GetDataWord(Add+Inx*IntBytes)
	else StatusStr:='Access Error';
	end;
 dsReal:
	begin
	if ProgramStore.RangeCheck(Add+Inx*RealSize) then Result:=ProgramStore.GetDataWord(Add+Inx*RealSize)
	else StatusStr:='Access Error';
	end;
 dsByte:
	begin
	if ProgramStore.RangeCheck(Add+Inx) then Result:=ProgramStore.DataByte[Add+Inx]
	else StatusStr:='Access Error';
	end;
 end;
end;



function GetVariableValue(Info: TVariableInfo): integer;
{Returns an integer representation of the specified watch item}
var Add: integer;
var Status: TVarStatus;
begin
if GetConstantValue(Info,Result) then exit;

Status:=GetVariableAddress(Info,Add);
case Status of
 vsScopeError: begin StatusStr:='Out of Scope'; exit; end;
 vsAccessError: begin StatusStr:='Access Error'; exit; end;
 end;
Result:=GetArrayValue(0,Add,Info);
end;



function GetVariableFactor(Ident, Filename: string; Line: integer): integer;
var VI: TVariableInfo;
begin
{Look up variable, recursing if there are subsripts}
if LookupVariable(Ident,Filename,Line,VI)=lrFound then Result:=GetVariableValue(VI);
end;


{==============================================================================}


function EvaluateExpression(ExpressStr,Filename: string; Line: integer): integer;
{ Evaluate Expression starting at EvalInx }
var Atom: char;
var EvalInx: integer;


procedure GetChar;
begin
if EvalInx > Length(ExpressStr) then
	begin
	StatusStr:='expr?';
	Atom:= ')';
	end
 else	begin
	Atom:= ExpressStr[EvalInx];
	Inc(EvalInx);
	end;
end;



procedure SkipWhiteSpace;
{ Skip Tabs And Spaces In Linebuf }
begin
while (Atom=TAB_CHAR) or (Atom=SP_char) do GetChar;
end;



procedure SkipSpaces;
{ Get Next Character, Ignoring Any Space Characters }
begin
repeat GetChar until Atom <> SP_CHAR;
end;




function GetASCII: integer;
begin
GetChar;
Result:= byte(Atom);
GetChar;
end;


function GetDecimal: integer;
{ Read In A Decimal String And Return Its Value }
var S: string;
begin
Result:=0;
S:='';
while True do
	begin
	if not IsDigit(Atom) then break;
	S:=S+Atom;
	GetChar;
	end;
if S='' then StatusStr:='number?'
else Result:=StrToInt(S);
end;

const IdentifierChars = ['A'..'Z','a'..'z','0'..'9','_'];



function GetIdentifier: integer;
{ Read Indentifier And Return Its Value }
var Ident: string;
var PC: integer;
begin
Result:=0;
Ident:=Atom;
while True do
	begin
	GetChar;
	if not (Atom in IdentifierChars) then break;
	Ident:=Ident+Atom;
	end;
if Atom='(' then
	begin
	Ident:=Ident+Atom;
	PC:=1;
	while true do
		begin
		GetChar;
		case Atom of
		 '(': Inc(PC);
		 ')': Dec(PC);
		 end;
		Ident:=Ident+Atom;
		if PC<1 then break;
		end;
	end;
Result:=GetVariableFactor(Ident, Filename,  Line);
end;



function GetHexadecimal: integer;
{ Return The Value Of A Hex String, E.g: $1ac4 }
var S: string;
begin
Result:=0;
S:='$';
GetChar;
while True do
	begin
	if not IsHexDigit(Atom) then break;
	S:=S+Atom;
	GetChar;
	end;
if S='$' then StatusStr:='hex?'
else Result:=StrToInt(S);
end;




function GetBinary: integer;
{ Return The Value Of A Hex String, E.g: $1ac4 }
var S: string;
begin
Result:=0;

GetChar;
while True do
	begin
	if not IsHexDigit(Atom) then break;
	S:=S+Atom;
	GetChar;
	end;
Result:=BinToInt(S);
end;


{ ---------------------------------------------------------------------- }

function Expression: integer;
{ Returns The Value Of An Expression }



function Factor: integer;
{ Returns The Value Of A Factor }
var NEG: boolean;
begin
Result:=0;
while Atom='+' do SkipSpaces;		{ Ignore Unary "+" }
NEG:= False;
while Atom ='-' do			{ Unary "-" }
	begin
	SkipSpaces;
	NEG:= not NEG;
	end;
if ((Atom>='A') and (Atom<='Z')) or (Atom='_') then Result:=GetIdentifier		{ Identifier }
else if (Atom>='0') and (Atom<='9') then Result:= GetDecimal	{ Unsigned Integer }
else case Atom of
  '(':	begin				{ Subexpression }
	SkipSpaces;
	Result:= Expression;
	if Atom<>')' then StatusStr:='Missing ")"';
	SkipSpaces;
	end;
  '$':	Result:= GetHexadecimal;
  '^':	Result:= GetASCII;
  '%':	Result:=GetBinary;
  else	StatusStr:='syntax';
  end;
{ Identifiers, Numbers, Etc. May Terminate With A Space Or Tab }
SkipWhiteSpace;
if NEG then Result:=-Result;
end;	{ Factor }



function Term: integer;
{ Returns Factor * Factor, Etc. }
begin
Result:= Factor;
while True do
	case Atom of
	  '*':	begin SkipSpaces; Result:= Result * Factor; end;
	  '/':	begin SkipSpaces; Result:= Result div Factor; end;
	  '\':	begin SkipSpaces; Result:= Result Mod Factor; end;
	  else	break;
	end;
end;
{ Term }



function AlgebraicExpression: integer;
{ Returns Term + Term, Etc. }
begin
Result:= Term;
while True do
	case Atom of
	  '+':	begin SkipSpaces; Result:= Result + Term; end;
	  '-':	begin SkipSpaces; Result:= Result - Term; end
	else	break;
	end;
end; { Algexp }



function LogicalExpression: integer;
{ Returns A Boolean, E.g: Algexp < Algexp }
begin
if Atom='~' then				{ Unary 'not' Operator }
	begin
	SkipSpaces;
	Result:= not LogicalExpression;
	end
 else	begin
	Result:= AlgebraicExpression;
	case Atom of
	 '=':	begin SkipSpaces; Result:= integer(Result = AlgebraicExpression); end;
	 '#':	begin SkipSpaces; Result:= integer(Result <> AlgebraicExpression); end;
	 '>':	begin
		SkipSpaces;
		if Atom<>'=' then Result:= integer(Result > AlgebraicExpression)
		else begin SkipSpaces; Result:= integer(Result >= AlgebraicExpression); end;
		end;
	 '<':	begin
		SkipSpaces;
		if Atom<>'=' then Result:= integer(Result < AlgebraicExpression)
		else begin SkipSpaces; Result:= integer(Result <= AlgebraicExpression); end;
		end;
	 end;
	end;
end; { Logexp }



function BooleanTerm: integer;
{ Returns Logexp & Logexp }
begin
Result:= LogicalExpression;
while True do	begin
	if Atom='&' then
		begin SkipSpaces; Result:= Result  and  LogicalExpression; end
	 else break;
	end;
end;	{ Boolterm }



begin	{ Expression }
SkipWhiteSpace;		{ Must Not Have A Tab Or Space }
Result:= BooleanTerm;
while True do	begin
	if Atom='!' then begin SkipSpaces; Result:= Result  or  BooleanTerm; end
	else break;
	end;
end;	{ Expression }



begin	{ EvaluateExpression }
try
EvalInx:=1;
GetChar;
Result:= Expression;
except StatusStr:='div by 0'; end;
if Atom <> ')' then StatusStr:='expr?';
end;	{ EvaluateExpression }




end.
