unit IDEFindAll;

interface

uses Windows, Messages, SysUtils, Classes, Graphics, Forms, Controls, StdCtrls,
  Buttons, ExtCtrls, Dialogs, FileCtrl, CheckedGroup,ScrollEditor, CodeEditor,
  Inifiles, VCLLibrary, DSupport,Masks;

type
  TFindAllDlg = class(TForm)
    Panel1: TPanel;
    OKBtn: TButton;
    CancelBtn: TButton;
    Panel2: TPanel;
    Label1: TLabel;
    TargetCombo: TComboBox;
    WhereRadio: TRadioGroup;
    GroupBox1: TGroupBox;
    CaseChk: TCheckBox;
    WholeWordChk: TCheckBox;
    DirOptionsGroup: TGroupBox;
    Label2: TLabel;
    MaskCombo: TComboBox;
    BrowseBtn: TButton;
    IncludeSubChk: TCheckBox;
    BitBtn1: TBitBtn;
    procedure WhereRadioClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure BrowseBtnClick(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure MaskComboDropDown(Sender: TObject);
    procedure MaskComboSelect(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  FindAllDlg: TFindAllDlg;

procedure DoFindAll;

implementation

uses IDEGlobal, IDEEditors, EBXMain, IDECompileMessages, XPLCompiler,
  IDESearchProgress;

{$R *.dfm}




procedure InsertSearchMsg(Info: TMessageInfo; Target: string);
var S: string;
var T: integer;
begin
S:='Search: ';
S:=S+ExtractFilename(Info.Info.Filename)+'('+IntToStr(Info.Info.Line)+') ';
T:=Length(S);
S:=S+Info.Description;
Info.SearchCol1:=Info.Info.Column+T+1;
Info.SearchCol2:=Info.SearchCol1+Length(Target);
InsertMessage(Info,S);
end;


{----------------- Editor Search ----------------------------------------------}


procedure SearchEditor(CodeEditor: TCodeEditor; Target: string; WholeWord, IgnoreCase: boolean);
var Loc,OldLoc: TPoint;
var Mark: TRect;
var Info: TMessageInfo;
var S: string;
begin
with CodeEditor do
 begin
 OldLoc:=Editor.WorkingCursor.Position;
 Loc:=Point(0,0);
 Editor.Searcher.IgnoreCase:=IgnoreCase;
 Editor.Searcher.WholeWordMode:=WholeWord;
 while Editor.DoForwardLoc(Loc) and Editor.DoForwardSearchMove(Target,Loc,Mark) do
	begin
	Info.MessageType:=xmSearch;
	S:=Editor.EditBuffer[Loc.Y];
	Info.Description:=S;
	Info.Info.Filename:=Editor.Filename;
	if CodeEditor = Form1.SourceEdit then Info.Info.SourceType:=stMain
	else Info.Info.SourceType:=stInclude;
	Info.Info.Column:=Mark.Left;
	Info.Info.Line:=Loc.Y;
	InsertSearchMsg(Info,Target);
	end;
 end;
end;



procedure DoEditorSearch;
var Iterator: TIncludeIterator;
var Editor: TCodeEditor;
var WholeWord,IgnoreCase: boolean;
var Target: string;
begin
WholeWord:=FindAllDlg.WholeWordChk.Checked;
IgnoreCase:=not FindAllDlg.CaseChk.Checked;
Target:=FindAllDlg.TargetCombo.Text;
SearchEditor(Form1.SourceEdit,Target,WholeWord,IgnoreCase);
Iterator:=TIncludeIterator.Create(Form1.EditorPageControl);
try
while Iterator.GetIteration(Editor) do
 SearchEditor(Editor,Target,WholeWord,IgnoreCase);
finally Iterator.Free; end;
end;


{----------------- Directory Search -------------------------------------------}

var FileCount: integer;
var FoundCount: integer;



function SearchOneFile(Path,Target: string; WholeWord,IgnoreCase: boolean): boolean;
var Info: TMessageInfo;
var I,Inx: integer;
var SL: TStringList;
var SourceType: TSourceType;

	function FindStr(SubStr, S: string; var Loc: integer): boolean;
	begin
	Result:=False;
	if IgnoreCase then Loc:=PosNoCase(SubStr,S)
	else Loc:=Pos(SubStr,S);
	if Loc<=0 then exit;
	if WholeWord and not IsWholeWord(S,Loc,Loc+Length(SubStr)-1) then exit;
	Result:=True;
	end;

begin
SL:=TStringList.Create;
try
if Form1.SourceEdit.Editor.Filename=Path then SourceType:=stMain
else SourceType:=stInclude;

SL.LoadFromFile(Path);
try
for I:=0 to SL.Count-1 do
	begin
	if FindStr(Target,SL[I],Inx) then
		begin
		Inc(FoundCount);
		Info.MessageType:=xmSearch;
		Info.Description:=SL[I];
		Info.Info.Filename:=Path;
		Info.Info.SourceType:=SourceType;
		Info.Info.Column:=Inx-1;
		Info.Info.Line:=I;
		InsertSearchMsg(Info,Target);
		end;
	end;
finally SL.Free; end;
except MessageDlg('Cannot open: '+ExtractFilename(Path),mtWarning,[mbOK],0); end;
end;




procedure DoDirectorySearch(FilePath,Target: string; WholeWord,IgnoreCase: boolean);
var SearchRecord: TSearchRec;
var FileAttrs: Integer;
var Filename,Path: string;
begin
Filename:=ExtractFilename(FilePath);
if Filename='' then Filename:='*.xpl';
if ExtractFileExt(Filename)='' then Filename:=Filename+'.xpl';
Path:=ExtractFilePath(FilePath);
Path:=IncludeTrailingBackslash(Path);
FileAttrs := faAnyFile;
if FindFirst(Path+'*.*', FileAttrs, SearchRecord) = 0 then
	begin
	repeat
		begin
		if AbortFlag then break;
		Inc(FileCount);
		if (FileCount mod 10)=0 then
			begin
			UpdateSearchProgress(Path,SearchRecord.Name,FileCount,FoundCount);
			Application.ProcessMessages;
			if Application.Terminated then break;
			end;
            	if (SearchRecord.Name='.') or (SearchRecord.Name='..') then Continue;
		if FindAllDlg.IncludeSubChk.Checked and ((SearchRecord.Attr and faDirectory)<>0) then
			begin
		 	DoDirectorySearch(Path+SearchRecord.Name+'\'+Filename,Target,WholeWord,IgnoreCase);
		  	end
		else if MatchesMask(SearchRecord.Name,Filename) then
			begin
			SearchOneFile(Path+SearchRecord.Name,Target,WholeWord,IgnoreCase);
			end;
		end
	until FindNext(SearchRecord) <> 0;
	FindClose(SearchRecord);
	end;
end;

procedure DirectorySearch;
var WholeWord,IgnoreCase: boolean;
begin
WholeWord:=FindAllDlg.WholeWordChk.Checked;
IgnoreCase:=not FindAllDlg.CaseChk.Checked;
ShowSearchProgressDlg;
try
FileCount:=0;
FoundCount:=0;
AbortFlag:=False;
DoDirectorySearch(FindAllDlg.MaskCombo.Text,FindAllDlg.TargetCombo.Text,WholeWord,IgnoreCase);
finally HideSearchProgressDlg; end;
end;

{------------------------------------------------------------------------------}


procedure DoFindAll;
var S: string;
var Editor: TCodeEditor;
var Inx: integer;
begin
Editor:=GetActiveEditor;
if Editor<>nil then
	begin
	S:=Editor.Editor.GetHighlightedText;
	if S='' then S:=Editor.Editor.GetNearestWord;
	if S<>'' then FindAllDlg.TargetCombo.Text:=S;
	end;

FindAllDlg.ShowModal;
if FindAllDlg.ModalResult=mrOK then
	begin
	ClearMessages(Form1.MessageView,True,False);
	Inx:=Form1.MessageView.Items.Count;
	UpdateAndSaveComboBox(IDEIniFilename,'FindTargets',FindAllDlg.TargetCombo);
	UpdateAndSaveComboBox(IDEIniFilename,'FindMasks',FindAllDlg.MaskCombo);
	case FindAllDlg.WhereRadio.ItemIndex of
	 0: DoEditorSearch;
	 1: DirectorySearch;
	 end;
	if Form1.MessageView.Items.Count>Inx then
		begin
		Form1.MessageView.ItemIndex:=Inx;
		Form1.MessageView.TopIndex:=Inx;
		Form1.MessageView.SetFocus;
		Form1.MessageView.Invalidate;
		end;
	end;
end;



procedure TFindAllDlg.FormCloseQuery(Sender: TObject;
  var CanClose: Boolean);
begin
if FindAllDlg.ModalResult<>mrOK then exit;
if FindAllDlg.TargetCombo.Text='' then
	begin
	MessageDlg('The search string cannot be blank.', mtError, [mbOK], 0);
	CanClose:=False;
	end;
end;


procedure TFindAllDlg.WhereRadioClick(Sender: TObject);
begin
EnableChildControls(DirOptionsGroup,WhereRadio.ItemIndex=1);
end;



procedure TFindAllDlg.FormCreate(Sender: TObject);
begin
EnableChildControls(DirOptionsGroup,False);
LoadStringsIni(IDEIniFilename,'FindTargets',FindAllDlg.TargetCombo.Items);
FindAllDlg.TargetCombo.ItemIndex:=0;
LoadStringsIni(IDEIniFilename,'FindMasks',FindAllDlg.MaskCombo.Items);
FindAllDlg.MaskCombo.ItemIndex:=0;
if FindAllDlg.MaskCombo.Items.Count=0 then
	begin
	FindAllDlg.MaskCombo.Text:=ExtractFilePath(Application.ExeName);
	UpdateAndSaveComboBox(IDEIniFilename,'FindMasks',FindAllDlg.MaskCombo);
	end;
MaskCombo.Hint:=MaskCombo.Items[MaskCombo.ItemIndex];
end;



procedure TFindAllDlg.BrowseBtnClick(Sender: TObject);
var Dir: string;
begin
Dir:=ExtractFilePath(MaskCombo.Text);
if SelectDirectory('Select A Directory','',Dir) then
	begin
	MaskCombo.Text:=Dir;
	UpdateAndSaveComboBox(IDEIniFilename,'FindMasks',FindAllDlg.MaskCombo);
	end;
end;


procedure TFindAllDlg.MaskComboDropDown(Sender: TObject);
var I: integer;
var Widest,W: integer;
var S: string;
begin
Widest:=0;
for I:=0 to MaskCombo.Items.Count-1 do
	begin
	S:=MaskCombo.Items[I];
	W:=MaskCombo.Canvas.TextWidth(S);
	if W>Widest then Widest:=W;
	end;
SendMessage(MaskCombo.Handle, CB_SETDROPPEDWIDTH, Widest+5, 0);
end;


procedure TFindAllDlg.MaskComboSelect(Sender: TObject);
begin
MaskCombo.Hint:=MaskCombo.Items[MaskCombo.ItemIndex];
end;


end.
