unit IDEGlobal;

interface

uses Types,SysUtils,Classes,Menus,Dialogs,DSupport,VCLLibrary,ScrollEditor,CodeEditor, XPLCompiler;

const BaseIniName = 'XPLIDE32.ini';
var ProgramTitle: string = 'E-XPL';

{Data structure to handle break point information}

type TBreakInfo = record
 Filename: string;
 Line,PC: integer;
 end;

{Define array of break points}

type TBreakArray = array of TBreakInfo;

var BreakArray: TBreakArray;


var IDEIniFilename: string;
var SourceIni,IncludeIni: string;

procedure ShowProgramTitle;
procedure ShowRunState(Running: boolean);
function GetFileAtCursor(Editor: TScrollEditor): string; overload;
function GetFileAtCursor(Sender: TObject): string; overload;

procedure SetEditorPosition(Info: TSourceInfo);
procedure SetupIniFilename;
function TypeToStr(VT: integer): string;
function DoCompileOnly: boolean;
function DoAutoCompile(Source: string; var I2LCode: string; MsgHan: TMessageHandler=nil; CompEvnt: TCompileEvent=nil): integer;


implementation

uses EBXMain, XPLDataTypes, IDECompileMessages, IDEEditors, I2lIntrinsics,
  IDEConfig,IDEStruncts;



procedure SetupIniFilename;
begin
IDEIniFilename:=GetIniPathName('EXPL32',True)+BaseIniName;
SourceIni:=GetIniPathName('EXPL32',True)+'XPLEdit1.ini';
IncludeIni:=GetIniPathName('EXPL32',True)+'XPLEdit2.ini';
end;



procedure ShowProgramTitle;
var Name,S: string;
begin
Name:=ExtractFilename(Form1.SourceEdit.CurrentFile);
S:=ProgramTitle+' - '+Name;
if Form1.RunOnOff.BoxOn then S:=S+' [Running]';
Form1.Caption:=S;
Form1.MainPage.Caption:='Main - '+Name;
end;


procedure ShowRunState(Running: boolean);
begin
Form1.RunOnOff.BoxOn:=Running;
Form1.RunOnOff.Update;
ShowProgramTitle;
end;


{------------------------------------------------------------------------------}


function NormalizeFilename(Filename: string): string;
{Make sure the filename and path are correct and fully expanded}
var Ext,Path: string;
begin
Result:=Filename;
Ext:=ExtractFileExt(Result);
if Ext='' then Result:=ChangeFileExt(Result,'.xpl');
Path:=ExtractFilePath(Result);
if Path='' then Result:=ExpandFileName(Result);
end;




function GetFileAtCursor(Editor: TScrollEditor): string;
const Delim = [#0..#$20,';'];
var Filename: string;
begin
Result:='';
Filename:=Editor.GetAndMarkBlock(Delim);
Editor.Invalidate;
Filename:=NormalizeFilename(Filename);
if FileExists(Filename) then Result:=Filename;
end;




function GetFileAtCursor(Sender: TObject): string;
var CE: TCodeEditor;
begin
Result:='';
if not (Sender is TMenuItem) then exit;
CE:=TCodeEditor(TMenuItem(Sender).Owner);
Result:=GetFileAtCursor(CE.Editor);
end;


{------------------------------------------------------------------------------}




procedure SetEditorPosition(Info: TSourceInfo);
var P: TPoint;
var Editor: TCodeEditor;
begin
Form1.IDEPageControl.ActivePageIndex:=0;
if Info.SourceType=stMain then
	begin
	Editor:=Form1.SourceEdit;
	Form1.EditorPageControl.ActivePageIndex:=0;
	end
else if Info.SourceType=stInclude then
	begin
	Editor:=FindOrCreateEditor(Info.Filename);
	ShowEditor(Editor);
	end;
P.X:=Info.Column;
P.Y:=Info.Line;
Editor.Editor.SetPosition(P);
Editor.Editor.HighlightRow:=P.Y;
Form1.IDEPageControl.ActivePageIndex:=0;
if Editor.Parent is TEditorTabsheet then
Form1.EditorPageControl.ActivePage:=TEditorTabsheet(Editor.Parent);

Editor.Editor.Invalidate;
if Editor.Editor.CanFocus then Editor.Editor.SetFocus;

end;


{------------------------------------------------------------------------------}





function TypeToStr(VT: integer): string;
begin
case VT of
 ADDRVAR: Result:='char';
 INVAR: Result:='int';
 RLVAR: Result:='real';
 INCON: Result:='int con';
 RLCON: Result:='real con';
 ADSEG: Result:='seg char';
 INSEG: Result:='seg int';
 RLSEG: Result:='seg real';
 SHSEG: Result:='seg short';
{$IFDEF STRUCTS}
 USRTYPE: Result:='user type';
{$ENDIF}
 end;
end;

procedure LoadInternalIncludes;
var I: integer;
var MS: TMemoryStream;
begin
MS:=TMemoryStream.Create;
try
SetLength(InternalIncludes,0);
if ConfigDlg.AutoIntrinsicsChk.Checked then
	begin
	SetLength(InternalIncludes,1);
	GetCodeDeclaration(InternalIncludes[0].FileStr,ConfigDlg.PrototypeChk.Checked);
	InternalIncludes[0].Title:='EXPLCodes.xpl';
	end;
for I:=0 to LibFiles.Count-1 do
 if TLibInfo(LibFiles.Objects[I]).Enabled and
    (TLibInfo(LibFiles.Objects[I]).LibType=ltFile) then
	begin
	SetLength(InternalIncludes,Length(InternalIncludes)+1);
	InternalIncludes[High(InternalIncludes)].Title:=LibFiles[I];
	MS.LoadFromFile(LibFiles[I]);
	SetString(InternalIncludes[High(InternalIncludes)].FileStr, PChar(MS.Memory), MS.Size div SizeOf(Char));
	end;
finally MS.Free; end;
end;


function DoAutoCompile(Source: string; var I2LCode: string; MsgHan: TMessageHandler=nil; CompEvnt: TCompileEvent=nil): integer;
{Do compile, automatically adding instrinsics}
begin
LoadInternalIncludes;
Result:=DoXPLCompile(Source,I2LCode,MsgHan,CompEvnt);
end;



function DoCompileOnly: boolean;
var ErrCnt: integer;
var I2LCode: string;
begin
Result:=False;
if Form1.SourceEdit.Editor.EditBuffer.Empty then
	begin
	MessageDlg('No Program', mtWarning, [mbOK], 0);
	exit;
	end;
InitializeMessages;
Form1.MessageView.Items.Clear;
try
ChangeDefaultDir(Form1.SourceEdit.CurrentFile);
ErrCnt:=DoAutoCompile(Form1.SourceEdit.Editor.Text,I2LCode,IDEMessageHandler,HandleCompilerEvents);
Result:=ErrCnt=0;
finally
 MessagePostCompile;
 if not Result then MessageDlg('Compile Errors, Cannot Debug.',mtWarning, [mbOK], 0);
 end;
end;



{------------------------------------------------------------------------------}



initialization
if IntBytes = 4 then ProgramTitle:='E-XPL32'
else ProgramTitle:='E-XPL16';
SetupIniFilename;
finalization
end.
