unit IDEHeapAnalyzer;

interface

uses Windows, SysUtils, Classes, Graphics, Forms, Controls, StdCtrls,
  Buttons, ExtCtrls, Grids, DSupport, SpinButtons, intspin, ScrollBarWindow,
  ScrollEditor, VCLLibrary,XPLDataTypes;


type THeapInfo = record
 Name: string;
 Address: integer;
 DataType: TDataType;
 Indirect: boolean;
 end;

type THeapInfoArray = array of THeapInfo;

var HeapInfoArray: THeapInfoArray;


procedure ShowHeapData;

procedure DrawHeapGrid(ACol, ARow: Integer;  Rect: TRect; State: TGridDrawState);
procedure ClearHeapInfo;
procedure AddHeapInfo(Add: integer; DataType: TDataType);
procedure SetHeapAddress(Inx,Value: integer);
procedure SetHeapDataType(Inx: integer; DT: TDataType);
procedure DeleteHeapItem(Inx: integer);
procedure MoveHeapItem(Inx: integer; Down: boolean);


implementation

uses I2LInterp,IDEDebugger;



procedure ShowHeapData;
begin
DebugDlg.HeapGrid.Invalidate;
end;

{====================================================================}

procedure ClearHeapInfo;
begin
SetLength(HeapInfoArray,0);
DebugDlg.HeapGrid.Invalidate;
end;


procedure AddHeapInfo(Add: integer; DataType: TDataType);
begin
SetLength(HeapInfoArray, Length(HeapInfoArray)+1);
HeapInfoArray[High(HeapInfoArray)].Address:=Add;
HeapInfoArray[High(HeapInfoArray)].DataType:=DataType;
DebugDlg.HeapGrid.Invalidate;
end;



procedure SetHeapAddress(Inx,Value: integer);
begin
if (Inx<0) or (Inx>High(HeapInfoArray)) then exit;
HeapInfoArray[Inx].Address:=Value;
DebugDlg.HeapGrid.Invalidate;
end;


procedure DeleteHeapItem(Inx: integer);
var I: integer;
begin
if Length(HeapInfoArray)<1 then exit;
if (Inx<0) or (Inx>High(HeapInfoArray)) then exit;
for I:=Inx to High(HeapInfoArray)-1 do
 HeapInfoArray[I]:=HeapInfoArray[I+1];
SetLength(HeapInfoArray,Length(HeapInfoArray)-1);
DebugDlg.HeapGrid.Invalidate;
end;

procedure MoveHeapItem(Inx: integer; Down: boolean);
var NewInx: integer;
var WI: THeapInfo;
begin
if Down then NewInx:=Inx+1 else NewInx:=Inx-1;
if (NewInx<0) or (NewInx>High(HeapInfoArray)) then exit;
WI:=HeapInfoArray[Inx];
HeapInfoArray[Inx]:=HeapInfoArray[NewInx];
HeapInfoArray[NewInx]:=WI;
DebugDlg.HeapGrid.Row:=NewInx+1;
DebugDlg.HeapGrid.Invalidate;
end;

procedure SetHeapDataType(Inx: integer; DT: TDataType);
begin
if (Inx<0) or (Inx>High(HeapInfoArray)) then exit;
HeapInfoArray[Inx].DataType:=DT;
end;



function GetHeapItemStr(Inx: integer; MSB, Indirect: boolean): string;
var Info: THeapInfo;
var Add: integer;
var P: Pointer;

	function GetString(P: Pointer; MSB: boolean): string;
	var I: integer;
	var S: shortstring;
	begin
	S:=PChar(P);
	if MSB then
		begin
		Result:='';
		for I:=1 to Length(S) do
			begin
			Result:=Result+Char(Byte(S[I]) and $7F);
			if (Byte(S[I]) and $80)<>0 then break;
			end;
		end
	else Result:=S;
	end;


begin
Result:='Inaccessible';
if Inx>High(HeapInfoArray) then exit;
Info:=HeapInfoArray[Inx];
if (Inx<0) or (Inx>High(I2LProgram.ProgData.Code)) then exit;
if Indirect then Add:=I2LProgram.ProgData.Code[Info.Address]
else Add:=Info.Address;
P:=@I2LProgram.ProgData.Code[Add];
case Info.DataType of
 dtChar: Result:=GetString(PChar(P),MSB);
 dtInteger: Result:=IntToStr(PInteger(P)^)+' $'+IntToHex(PInteger(P)^,8);
 dtReal: Result:=FloatToStrF(PDouble(P)^,ffFixed,18,1);
 end;
end;


procedure DrawHeapGrid(ACol, ARow: Integer;  Rect: TRect; State: TGridDrawState);
var Canvas: TCanvas;
var Inx: integer;
begin
Canvas:=DebugDlg.HeapGrid.Canvas;
if ARow=0 then
	begin
	case ACol of
	 0: OwnerDrawItem(Canvas, '#', Rect);
	 1: OwnerDrawItem(Canvas, 'Name', Rect);
	 2: OwnerDrawItem(Canvas, 'Address', Rect);
	 3: OwnerDrawItem(Canvas, 'Type', Rect);
	 4: OwnerDrawItem(Canvas, 'Indirect', Rect);
	 5: OwnerDrawItem(Canvas, 'Value', Rect);
	 end;
	end
else
	begin
	Inx:=ARow-1;
	if Inx>High(HeapInfoArray) then exit;
	case ACol of
	 0: OwnerDrawInteger(Canvas, Inx, Rect);
	 1: OwnerDrawItem(Canvas,HeapInfoArray[Inx].Name, Rect);
	 2: OwnerDrawItem(Canvas, '$'+IntToHex(HeapInfoArray[Inx].Address,IntNibbles), Rect);
	 3: OwnerDrawItem(Canvas, GetDataTypeStr(HeapInfoArray[Inx].DataType), Rect);
	 4: OwnerDrawItem(Canvas, PickString(HeapInfoArray[Inx].Indirect,'Yes','No'), Rect);
	 5: OwnerDrawItem(Canvas, GetHeapItemStr(Inx,UseMSBStrings,HeapInfoArray[Inx].Indirect), Rect);
	 end;
	end;
end;



end.
