unit IDEHints;

interface

uses Types, SysUtils, VCLLibrary, XPLCompiler, IDEGlobal, I2LGlobal, ScrollEditor;

procedure DoVariableHints(Editor: TScrollEditor;  X,Y: integer);
function DoIntrinsicHints(Editor: TScrollEditor; X,Y: integer): boolean;
procedure GotoDeclaration(CurEditor: TScrollEditor);
procedure HideHints;

implementation

uses EBXMain, I2lIntrinsics, IDEEditors;


var OldIdent: string;


function IsInComment(Editor: TScrollEditor;  P: TPoint): boolean;
{Check to see if }
var I,Cnt,Inx: integer;
var S: string;
begin
Result:=False;
{Get the current line}
S:=Editor.GetWholeLines(P.Y,P.Y,True);
if (P.X<0) or (P.X>Length(S)) then exit;
{Look for whole comment characters}
Result:=True;
Inx:=Pos('\\',S);
if (Inx>0) and (Inx<P.X) then exit;
{Count the number of back slashes}
Cnt:=0;
for I:=P.X downto 1 do
 if S[I]='\' then Inc(Cnt);
{Even number means in comment}
Result:=(Cnt and 1)<>0;
end;


function GetFileIdentifier(Editor: TScrollEditor): string;
{Get filename or file identifier from specified editor}
{The main file is not identified by the filename because}
{ it might have been streamed and streams don't have filenames}
begin
{The source editor is identified as Main}
if Editor = Form1.SourceEdit.Editor then Result:='Main'
else Result:=Editor.Filename;
end;



procedure DoVariableHints(Editor: TScrollEditor;  X,Y: integer);
{Display the value of variables if you mouse over them}
var P: TPoint;
var Ident: string;
var PC: integer;
var VS: string;
var PPI,VPI: PProcItem;
var SI: PSymbolItem;
var Filename: string;
begin
P:=Editor.ScreenToBufferPos(X,Y);
Ident:=Editor.GetMarkedText(Editor.MarkNearestWord(P));
if Ident='' then
	begin
	GlobalHinter.KillHint;
	OldIdent:='';
	exit;
	end;
if (Ident=OldIdent) then exit;
if IsInComment(Editor,P) then exit;

Filename:=GetFileIdentifier(Editor);
PC:=LineNumberToPC(Filename,P.Y+1,ProgramStore.ProgData.BaseAddress);
GetProcedureByName(PC,ProgramStore.ProgData.BaseAddress,Ident,PPI);
GetSymbolByName(PC,ProgramStore.ProgData.BaseAddress,Ident,VPI,SI);
GlobalHinter.KillHint;
if VPI<>nil then
	begin
	{Show variable info in the form: "var XYZ: int Test.XPL(232)"}
	VS:='var '+Ident+': '+TypeToStr(SI.SymType)+' '+ExtractFilename(SI.SourceInfo.Filename)+'('+IntToStr(SI.SourceInfo.Line)+')';
	GlobalHinter.ShowHint(Editor,X,Y+20,VS,5000);
	end
else if PPI<>nil then
	begin
	{Show procedure info in the form: "proc ABC: int Test.XPL(234)"}
	VS:='proc '+Ident+' '+ExtractFilename(PPI.SourceInfo.Filename)+'('+IntToStr(PPI.SourceInfo.Line)+')';
	GlobalHinter.ShowHint(Editor,X,Y+20,VS,5000);
	end;
OldIdent:=Ident;
end;


{----------------- Handle Instrinsic Hints ------------------------------------}

var OldHint: string;


function DoIntrinsicHints(Editor: TScrollEditor; X,Y: integer): boolean;
var P: TPoint;
var R: TRect;
var S,Proto: string;
var Num: integer;
begin
Result:=False;
with Form1 do
 begin
 P:=Editor.ScreenToBufferPos(X,Y);

 R:=Editor.MarkNearestWord(P);
 S:=Editor.GetMarkedText(R);
 if S='' then GlobalHinter.KillHint
 else if S<>OldHint then
	begin
	Proto:=NameToProtoType(S);
	if Proto='' then GlobalHinter.KillHint
	else
		begin
		Num:=NameToIntrinNumber(S);
		Proto:='Intrinsic: #'+IntToStr(Num)+' '+Proto;
		GlobalHinter.ShowHint(Editor,X,Y+20,Proto,5000);
		Result:=True;
		end;
	end
 else Result:=True;
 OldHint:=S;
 end;
end;



procedure GotoDeclaration(CurEditor: TScrollEditor);
var Ident: string;
var PC: integer;
var PPI,VPI: PProcItem;
var SI: PSymbolItem;
var Filename: string;
var Line,Col: integer;
var Editor: TScrollEditor;
var FileID: string;
begin
Ident:=CurEditor.GetMarkedText(CurEditor.FindNearestWord);
FileID:=GetFileIdentifier(CurEditor);
PC:=LineNumberToPC(FileID,CurEditor.WorkingCursor.Position.Y+1,ProgramStore.ProgData.BaseAddress);
GetProcedureByName(PC,ProgramStore.ProgData.BaseAddress,Ident,PPI);
GetSymbolByName(PC,ProgramStore.ProgData.BaseAddress,Ident,VPI,SI);
GlobalHinter.KillHint;
if VPI<>nil then
	begin
	Filename:=SI.SourceInfo.Filename;
	Line:=SI.SourceInfo.Line;
	Col:=SI.SourceInfo.Column;
	end
else if PPI<>nil then
	begin
	Filename:=PPI.SourceInfo.Filename;
	Line:=PPI.SourceInfo.Line;
	Col:=PPI.SourceInfo.Column;
	end
else exit;
if (Filename<>'Main') and (Form1.SourceEdit.Editor.Filename<>Filename) then
	begin
	Editor:=FindOrCreateEditor(Filename).Editor;
	end
else
	begin
	Editor:=Form1.SourceEdit.Editor;
	Form1.IDEPageControl.ActivePageIndex:=0;
	Form1.EditorPageControl.ActivePageIndex:=0;
	end;
Editor.SetPosition(Point(Col,Line));
Editor.HighlightLine;
Editor.Invalidate;
if Editor.CanFocus then Editor.SetFocus;
end;



procedure HideHints;
begin
GlobalHinter.KillHint;
end;


initialization
finalization
end.
