unit IDEProfiler;

interface

uses Windows, SysUtils, Classes, Graphics, Forms, Controls, StdCtrls, 
  Buttons, ExtCtrls,Dialogs,mmSystem, DSupport,Hash,IniFiles, ComCtrls,TreeSorter;

type
  TProfileDlg = class(TForm)
    Panel6: TPanel;
    RunBtn: TSpeedButton;
    StopBtn: TSpeedButton;
    Panel1: TPanel;
    CancelBtn: TButton;
    ProfileView: TListView;
    Help1Btn: TSpeedButton;
    procedure CancelBtnClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure RunBtnClick(Sender: TObject);
    procedure StopBtnClick(Sender: TObject);
    procedure ProfileViewDblClick(Sender: TObject);
    procedure ProfileViewColumnClick(Sender: TObject; Column: TListColumn);
    procedure Help1BtnClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  ProfileDlg: TProfileDlg;

procedure DoProfiler;

implementation

uses EBXMain, IDECompileMessages, XPLCompiler, I2LLoader, I2LGlobal,
  I2LInterp, IDEVariables, IDEGlobal, IDEStruncts;

{$R *.dfm}


var EndOfProgram: boolean;
var OldLine: integer;

var SourceInfo: TSourceInfo;
var I2LCode: string;

var ProfileAbort: boolean;

var ProfileList: TIntegerHashTable;



function CompileAndLoadProgram: boolean;
var ErrCnt: integer;
begin
Result:=False;
if Form1.RunOnOff.BoxOn then
	begin
	Form1.StopBtnClick(ProfileDlg);
	Application.ProcessMessages;
	end;
InitializeMessages;
Form1.MessageView.Items.Clear;
try
ChangeDefaultDir(Form1.SourceEdit.CurrentFile);
ErrCnt:=DoAutoCompile(Form1.SourceEdit.Editor.Text,I2LCode,IDEMessageHandler);
finally MessagePostCompile; end;
if ErrCnt>0 then
	begin
	MessageDlg('Compile Errors, Cannot Debug.',mtWarning, [mbOK], 0);
	exit;
	end;

Form1.ConsoleWnd.VGABox.Color:=clBlack;
Form1.ConsoleWnd.VGABox.Clear;

Form1.IDEPageControl.ActivePageIndex:=1;
if Form1.ConsoleWnd.CanFocus then Form1.ConsoleWnd.SetFocus;

if not LoadI2LCode(I2LCode,ProgramStore.ProgData) then
	begin
	MessageDlg('Loader Error, Cannot Debug.',mtWarning, [mbOK], 0);
	exit;
	end;
SetProgramCode(ProgramStore);
LoadVariables;
Result:=True;
end;





procedure InitializeProfiler;
begin
InitializeInterpreter(Form1.ConsoleWnd);
SourceInfo:=PCToSourceData(I2LProgram.ProgCntr,ProgramStore.ProgData.BaseAddress).Info;
EndOfProgram:=False;
OldLine:=-1;
end;

type TProfileItem = record
 Time: integer;
 Count: integer;
 end;
type PProfileItem = ^TProfileItem;

procedure AddItem(I: integer; Time: integer);
var Inx: integer;
var P: ^TProfileItem;
begin
Inx:=ProfileList.Search(I);
if Inx<0 then
	begin
        New(P);
	Inx:=ProfileList.InsertPointer(I,integer(P));
	P.Time:=0;
	P.Count:=0;
	end
else P:=ProfileList.Items.Pointers[Inx];
P.Time:=P.Time + Time;
Inc(P.Count);
end;


procedure ClearList;
var I: integer;
begin
for I:=0 to ProfileList.Count-1 do
 Dispose(ProfileList.Items.Pointers[I]);
ProfileList.Clear;
end;


procedure ShowList;
var I: integer;
var PI: TProfileItem;
var ListItem: TListItem;
begin
with ProfileDlg do
 begin
 ProfileList.Sort;
 ProfileView.Clear;
 for I:=0 to ProfileList.Count-1 do
	begin
	PI:=PProfileItem(ProfileList.Items.Pointers[I])^;
	ListItem:=ProfileView.Items.Add;
	ListItem.Caption:=IntToStr(I);
	ListItem.SubItems.Add(IntToStr(ProfileList[I]));
	ListItem.SubItems.Add(IntToStr(PI.Time));
	ListItem.SubItems.Add(FloatToStrF(PI.Time/PI.Count,ffFixed,18,3));
	ListItem.SubItems.Add(IntToStr(PI.Count));
	end;
 end;
end;


procedure DoProfile;
var OldTime,NewTime: integer;
begin
if EndOfProgram then exit;
ProfileAbort:=False;
ClearList;
try
SourceInfo:=PCToSourceData(I2LProgram.ProgCntr,ProgramStore.ProgData.BaseAddress).Info;
OldLine:=SourceInfo.Line;
AddItem(I2LProgram.ProgCntr,0);
Form1.ConsoleWnd.RefreshTimer.Enabled:=True;
repeat
	begin
	OldTime:=GetHiResTick;
	repeat
		begin
		InterpretLoop(True);
		SourceInfo:=PCToSourceData(I2LProgram.ProgCntr,ProgramStore.ProgData.BaseAddress).Info;
		end
	until ProfileAbort or (OldLine<>SourceInfo.Line) or (Opcode=0);
	NewTime:=GetHiResTick;
        AddItem(SourceInfo.Line,NewTime-OldTime);
	OldLine:=SourceInfo.Line;
	end
until ProfileAbort or (Opcode=0);
finally
 Form1.ConsoleWnd.RefreshTimer.Enabled:=False;
 end;
ShowList;
end;



procedure DoProfiler;
begin
if not CompileAndLoadProgram then exit;
ProfileDlg.Show;
InitializeProfiler;
end;



procedure TProfileDlg.CancelBtnClick(Sender: TObject);
begin
Close;
end;



procedure TProfileDlg.FormCreate(Sender: TObject);
begin
ProfileList:=TIntegerHashTable.Create(5);
end;



procedure TProfileDlg.FormDestroy(Sender: TObject);
begin
ProfileList.Free;
end;



procedure TProfileDlg.RunBtnClick(Sender: TObject);
begin
DoProfile;
end;



procedure TProfileDlg.StopBtnClick(Sender: TObject);
begin
ProfileAbort:=True;
end;



procedure TProfileDlg.ProfileViewDblClick(Sender: TObject);
var P: TPoint;
var Inx,Line: integer;
begin
Inx:=ProfileView.ItemIndex;
Line:=StrToIntDef(ProfileView.Items[Inx].SubItems[0],0);
P.X:=0;
P.Y:=Line;
Form1.IDEPageControl.ActivePageIndex:=0;
Form1.EditorPageControl.ActivePageIndex:=0;
Form1.SourceEdit.Editor.SetPosition(P);
Form1.SourceEdit.Editor.HighlightLine;
Form1.SourceEdit.Editor.Invalidate;
end;



procedure TProfileDlg.ProfileViewColumnClick(Sender: TObject;
  Column: TListColumn);
begin
{ determine the sort style }
case Column.Index of
 3: LvSortStyle:=cssFloat;
 else LvSortStyle := cssNumeric;
 end;

{ Call the CustomSort method }
ProfileView.CustomSort(@CustomSortProc, Column.Index -1);

{ Set the sort order for the column}
LvSortOrder[Column.Index] := not LvSortOrder[Column.Index];

end;

procedure TProfileDlg.Help1BtnClick(Sender: TObject);
begin
Application.HelpCommand(HELP_CONTEXT, 23);
end;

end.
