unit IDEStructDebug;

interface

uses SysUtils, Classes, DSUpport;

function GetStructValue(Symbol,Filename: string; Line: integer): string;
function GetStructStr(Symbol,Filename: string; Line: integer): string;


implementation

uses XPLCompiler, IDEVariableLib, IDEExpression, XPLDataTypes, XPLStructs,
     I2LGlobal, I2LInterp;

procedure DissectVariable(S: string; var Base,Indices,Comp: string);
{Extract all subsections of a variable }
var I1,I2: integer;
begin
Base:=''; Indices:=''; Comp:='';
I1:=Pos('(',S); I2:=Pos(')',S);
if (I1>0) and (I2>0) then
	begin
	Indices:=CopyRange(S, I1+1,I2-1);
	DeleteRange(S,I1,I2)
	end;
I1:=Pos('.',S);
if I1>0 then
	begin
	Base:=CopyRange(S,1,I1-1);
	Comp:=CopyRange(S,I1+1,Length(S));
	end
else Base:=S;
end;


procedure GetIndicesValues(S: string; Filename: string; Line: integer; var Indices: TDynDWordArray);
var SL: TStringList;
var I: integer;
begin
SL:=TStringList.Create;
try
SL.Delimiter:=',';
SL.DelimitedText:=S;
SetLength(Indices, SL.Count);
for I:=0 to SL.Count-1 do
	begin
	if IsIntegerStr(SL[I]) then Indices[I]:=StrToIntDef(SL[I],0)
	else Indices[I]:=EvaluateExpression(SL[I],Filename,Line);;
	end;
finally SL.Free; end;
end;


function AddIndiceOffsets(Base,Size,Offset: integer; Indices: TDynDWordArray): integer;
var I,Index,Inx: integer;
begin
Result:=Base;
if Length(Indices)<1 then exit;
{Handle all indices except last one}
for I:=0 to High(Indices)-1 do
	begin
	Index:=Indices[I];
	Inx:=Size * Index + Result;
	Result:=I2LProgram.GetDataWord(Inx);
	end;
Index:=Indices[High(Indices)];
Result:=Offset + (Size * Index) + Result;
end;



function GetDataItemStr(Add: integer; Format: TVarFormat): string;
var D: double;
const ErrorStr = 'Access Error';
begin
case Format.DisplayType of
 dsCharString:
	begin
	if ProgramStore.RangeCheck(Add) then Result:='"'+char(ProgramStore.DataByte[Add])+'"'
	else Result:=ErrorStr;
	end;
 dsString:
	begin
	if ProgramStore.RangeCheck(Add) and
	   ProgramStore.RangeCheck(I2LProgram.GetDataWord(Add)) then
	   	Result:='"'+ProgramStore.GetDataString(I2LProgram.GetDataWord(Add),UseMSBStrings)+'"'
	else Result:=ErrorStr;
	end;
 dsInteger:
	begin
	if ProgramStore.RangeCheck(Add) then Result:=GetIntegerRadixStr(ProgramStore.GetDataWord(Add),IntNibbles,Format.Radix)
	else Result:=ErrorStr;
	end;
 dsReal:
	begin
	if ProgramStore.RangeCheck(Add) then
		begin
		D:=ProgramStore.GetDataDouble(Add);
		Result:=GetRealStr(D,Format);
		end
	else Result:=ErrorStr;
	end;
 dsByte:
	begin
	if ProgramStore.RangeCheck(Add) then Result:=GetIntegerRadixStr(ProgramStore.DataByte[Add],1,Format.Radix)
	else Result:=ErrorStr;
	end;
 end;
end;


function GetEndTypeInfo(BaseIndex: integer; CompStr: string; var EndIndex: integer): TVariableType;
var EndType: integer;
begin
Result:=vtNone;
if GetVariableType(BaseIndex, CompStr, EndType, EndIndex) then
	begin
	case EndType of
	 ADDRVAR: Result:=vtChar;
	 INVAR: Result:=vtInteger;
	 RLVAR: Result:=vtReal;
	 USRTYPE: Result:=vtUserType;
	 end;
	end;
end;


function GetStructValue(Symbol,Filename: string; Line: integer): string;
var Inx,EndOff,EndIndex,InxOff,BaseSize,EndSize,BaseAdd: integer;
var SType,SIndex: integer;
var BaseStr,IndicesStr,CompStr: string;
var BaseType,EndType: TVariableType;
var S,TypeStr: string;
var Indices: TDynDWordArray;
var VI: TVariableInfo;
var R: TLookupResult;
var Add: integer;
var Format: TVarFormat;
begin
DissectVariable(Symbol,BaseStr,IndicesStr,CompStr);
GetIndicesValues(IndicesStr, Filename, Line, Indices);
R:=LookupVariable(BaseStr,Filename,Line,VI);
BaseType:=VI.VarType;
EndType:=GetEndTypeInfo(VI.StructIndex, CompStr, EndIndex);
EndOff:=GetVariableOffset(VI.StructIndex,CompStr);
BaseSize:=GetVariableSize(VI.StructIndex,'');
EndSize:=GetVariableSize(VI.StructIndex,CompStr);
Add:=DisplayVector[VI.Level]+VI.Offset;
if Length(Indices)>0 then Add:=I2LProgram.GetDataWord(Add)
else Add:=Add+EndOff;
Add:=AddIndiceOffsets(Add,BaseSize,EndOff,Indices);

Format.Digits:=5;
Format.Radix:=nrDecimal;
Format.DisplayType:=GetDisplayType(EndType,Length(Indices)>0);
Format.RealFormat:=rfGeneral;
Format.ForcedArray:=False;
Result:=GetDataItemStr(Add,Format);
end;

type TStringArray = array of string;

procedure GetIterateIndices(Dims: TDynWordArray; var SA: TStringArray);
{Create a list of strings containing all comibination of array indices}
var I: integer;
var Counters: array of integer;


	procedure SaveCount;
	{Save current iterated count in string array}
	var I: integer;
	var S: string;
	begin
	S:='(';
	for I:=0 to High(Counters) do
		begin
		S:=S+IntToStr(Counters[I]);
		if I<>High(Counters) then S:=S+',';
		end;
	S:=S+')';
	SetLength(SA,Length(SA)+1);
	SA[High(SA)]:=S;
	end;


	function Increment: boolean;
	{Increment each element of the count}
	var I: integer;
	begin
	{Increment each element, handling overflow and carry}
	for I:=High(Counters) downto 0 do
		begin
		Inc(Counters[I]);
		if Counters[I]<Dims[I] then break
		else Counters[I]:=0;
		end;
	Result:=True;
	{We are done when all elements roll back to zero}
	for I:=0 to High(Counters) do
	 if Counters[I]<>0 then exit;
	Result:=False;
	end;


begin
SetLength(SA,0);
{Zero the counter elements}
SetLength(Counters,Length(Dims));
for I:=0 to High(Counters) do Counters[I]:=0;
while true do
	begin
	SaveCount;
	if not Increment then break;
	end;
end;

function GetStructStr(Symbol,Filename: string; Line: integer): string;
var Inx,EndOff,EndIndex,InxOff,BaseSize,EndSize,BaseAdd: integer;
var SType,SIndex: integer;
var BaseStr,IndicesStr,CompStr: string;
var BaseType,EndType: TVariableType;
var S,TypeStr: string;
var Indices: TDynDWordArray;
var VI: TVariableInfo;
var R: TLookupResult;


	procedure ShowSubitems(Inx: integer; Name: string);
	var J, Start, Stop: integer;
	var S: string;
	begin
	Start:=SubItemIndices[Inx];
	Stop:=Start + SubItemCount[Inx]-1;
	for J:=Start to Stop do
		begin
		S:=Name+'.'+SubItemName[J];
		if SubItemType[J]=USRTYPE then ShowSubitems(SubUserIndex[J], S)
		else
			begin
			if J<>Start then Result:=Result+', ';
			Result:=Result+S+'='+GetStructValue(S,Filename, Line);
			end;
		end;
	end;


	procedure IterateDimensions;
	var SA: TStringArray;
	var I: integer;
	begin
	GetIterateIndices(VI.Dims^,SA);
	for I:=0 to High(SA) do
		begin
		if I>0 then Result:=Result+', ';
		ShowSubitems(EndIndex,Symbol+SA[I]);
		end;
	end;



begin
Result:='Invalid Variable';
if (Length(Symbol)>0) and
   (Symbol[Length(Symbol)]='.') then SetLength(Symbol,Length(Symbol)-1);
DissectVariable(Symbol,BaseStr,IndicesStr,CompStr);
if LookupVariable(BaseStr,Filename,Line,VI)<>lrFound then exit;
EndType:=GetEndTypeInfo(VI.StructIndex, CompStr,EndIndex);
if EndType=vtNone then exit;
if EndType=vtUserType then
	begin
	Result:='';
	if (IndicesStr='') and (Length(VI.Dims^)>0) then IterateDimensions
	else ShowSubitems(EndIndex,Symbol);
	end
else Result:=Symbol+': '+GetStructValue(Symbol,Filename, Line);
end;


end.
