unit IDESystemDlg;

interface

uses Windows, SysUtils, Classes, Graphics, Forms, Controls, StdCtrls, 
  Buttons, ExtCtrls, ComCtrls, ConsoleWindow, BitmapLib, IntVector,DSupport,
  SpinButtons, IDECanvasAttribs, Grids,TimeLibrary,FileLibrary;

type
  TSystemDlg = class(TForm)
    ConsoleSheet: TTabSheet;
    BitmapsSheet: TTabSheet;
    PageControl: TPageControl;
    GroupBox1: TGroupBox;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    AttribDis: TEditEx;
    Panel2: TPanel;
    CancelBtn: TButton;
    MemorySheet: TTabSheet;
    FilesSheet: TTabSheet;
    GroupBox2: TGroupBox;
    Label4: TLabel;
    VideoModeDis: TEditEx;
    Label5: TLabel;
    CardTypeDis: TEditEx;
    Label6: TLabel;
    ResolutionDis: TEditEx;
    Label9: TLabel;
    ModeTypeDis: TEditEx;
    BackColorDis: TEditEx;
    ForeColorDis: TEditEx;
    Label7: TLabel;
    TextResDis: TEditEx;
    Label10: TLabel;
    CellSizeDis: TEditEx;
    Label12: TLabel;
    FontTypeDis: TEditEx;
    Label11: TLabel;
    WinModeDis: TEditEx;
    Label13: TLabel;
    ClipWindDis: TEditEx;
    Label8: TLabel;
    BitsPerPixelDis: TEditEx;
    Label23: TLabel;
    TransColorDis: TEditEx;
    Label21: TLabel;
    CopyAlphaDis: TEditEx;
    Label22: TLabel;
    PerPixelDis: TEditEx;
    Label19: TLabel;
    CopyModeDis: TEditEx;
    Label14: TLabel;
    Panel4: TPanel;
    Panel3: TPanel;
    Label15: TLabel;
    BitmapGrid: TDrawGrid;
    Panel5: TPanel;
    Splitter1: TSplitter;
    Panel6: TPanel;
    Splitter2: TSplitter;
    ScrollBox1: TScrollBox;
    Image1: TImage;
    GroupBox3: TGroupBox;
    Panel7: TPanel;
    Label29: TLabel;
    MallocGrid: TDrawGrid;
    Panel8: TPanel;
    MallocSizeDis: TEditEx;
    Panel9: TPanel;
    Label16: TLabel;
    Label17: TLabel;
    Label18: TLabel;
    Label20: TLabel;
    Label24: TLabel;
    Label25: TLabel;
    Label26: TLabel;
    Label28: TLabel;
    ProgSizeDis: TEditEx;
    LocalStartDis: TEditEx;
    HeapStartDis: TEditEx;
    LocalSizeDis: TEditEx;
    HeapAllocDis: TEditEx;
    HeapPointerDis: TEditEx;
    HeapSizeDis: TEditEx;
    HeapMaxDis: TEditEx;
    Panel10: TPanel;
    Label27: TLabel;
    FileGrid: TDrawGrid;
    FullPathDis: TEditEx;
    Label30: TLabel;
    Label31: TLabel;
    CreateDis: TEditEx;
    AccessDis: TEditEx;
    Label32: TLabel;
    WriteDis: TEditEx;
    Label33: TLabel;
    Label34: TLabel;
    AttribsDis: TEditEx;
    PaletteBtn: TButton;
    // Warning: don't delete these two - they are components, not dialogs
    CanvasAttribFrm1: TCanvasAttribFrm;
    CanvasAttribFrm2: TCanvasAttribFrm;
    BitBtn1: TBitBtn;
    procedure BitmapGridDrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure BitmapGridSelectCell(Sender: TObject; ACol, ARow: Integer;
      var CanSelect: Boolean);
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure MallocGridDrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure MallocGridSelectCell(Sender: TObject; ACol, ARow: Integer;
      var CanSelect: Boolean);
    procedure FileGridDrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure FileGridSelectCell(Sender: TObject; ACol, ARow: Integer;
      var CanSelect: Boolean);
    procedure CancelBtnClick(Sender: TObject);
    procedure PaletteBtnClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  SystemDlg: TSystemDlg;

procedure ShowSystemDlg;
procedure UpdateSystemDisplay;


implementation

uses I2LGlobal, I2LConsoleIO, XPLDataTypes, I2LInterp, I2lIntrinsics,
  IDEVGAPalette;

{$R *.dfm}

function GetPenModeStr(PM: TPenMode): string;
begin
case PM of
 pmCopy: Result:='Normal';
 pmXOR: Result:='XOR';
 else Result:='Unknown';
 end;
end;



function GetFontStylesStr(FS: TFontStyles): string;
begin
Result:='';
if fsBold in FS then Result:=Result+'Bold,';
if fsItalic in FS then Result:=Result+'Italic,';
if fsUnderline in FS then Result:=Result+'Underline,';
if fsStrikeout in FS then Result:=Result+'StrikeOut';
end;



procedure DisplayColor(ED: TEditEx; Color: TColor);
var R,G,B: byte;
begin
ED.Color:=Color;
ED.Font.Color:=ContrastColor(Color);
ColorToRGBBytes(Color,R,G,B);
ED.Text:=IntToHex(R,2)+'-'+IntToHex(G,2)+'-'+IntToHex(B,2);
end;



procedure UpdateConsole;
var GM: TGraphicsMode;
begin
with SystemDlg do
 begin
 AttribDis.Text:='$'+IntToHex(GlobalConsole.VGABox.Attribute,4);
 DisplayColor(BackColorDis,GlobalConsole.VGABox.AttribInfo.BackColor);
 DisplayColor(ForeColorDis,GlobalConsole.VGABox.AttribInfo.FrontColor);
 GM:=GlobalConsole.VGABox.GraphicsMode;
 VideoModeDis.Text:='$'+IntToHex(GM.Mode,4);
 CardTypeDis.Text:=GetCardTypeStr(GM.Card);
 ResolutionDis.Text:=IntToStr(GM.Res.X)+' X '+IntToStr(GM.Res.Y);
 TextResDis.Text:=IntToStr(GM.TextRes.X)+' X '+IntToStr(GM.TextRes.Y);
 BitsPerPixelDis.Text:=IntToStr(GM.Bits);
 CellSizeDis.Text:=GetCharCellStr(GM.CharCell);
 ModeTypeDis.Text:=GetModeTypeStr(GM.ModeType);

 WinModeDis.Text:=GetWindowModeStr(GlobalConsole.VGABox.WinMode);
 FontTypeDis.Text:=GetFontTypeStr(GlobalConsole.VGABox.FontType);
 ClipWindDis.Text:=GetRectStr(WinCorners,False);

 CopyModeDis.Text:=IntToStr(GlobalConsole.VGABox.BitmapManager.CopyMode.Mode);
 CopyAlphaDis.Text:=IntToStr(GlobalConsole.VGABox.BitmapManager.CopyMode.Alpha);
 PerPixelDis.Text:=PickString(GlobalConsole.VGABox.BitmapManager.CopyMode.PerPixel,'Yes','No');
 DisplayColor(TransColorDis,GlobalConsole.VGABox.BitmapManager.CopyMode.TransColor);
 CanvasAttribFrm1.DisplayCanvasInfo(GlobalConsole.VGABox.Canvas);
 end;
end;

{------------------------------------------------------------------------------}


procedure DisplayBMGridItem(S: string; Rect: TRect);
{Display the string in the grid inside the rectangle}
begin
SystemDlg.BitmapGrid.Canvas.TextRect(Rect,Rect.Left+2,Rect.Top+2,S);
end;



procedure TSystemDlg.BitmapGridDrawCell(Sender: TObject; ACol,
  ARow: Integer; Rect: TRect; State: TGridDrawState);
var Inx: integer;
var Man: TBitmapManager;
var BM: TBitmap;
var Status: THandleStatus;
var S: string;
begin
Inx:=ARow - 2;
Man:=GlobalConsole.VGABox.BitmapManager;
if ARow = 0 then
	begin
	case ACol of
	 0: DisplayBMGridItem('Handle', Rect);
	 1: DisplayBMGridItem('Status', Rect);
	 2: DisplayBMGridItem('Width', Rect);
	 3: DisplayBMGridItem('Height', Rect);
	 end;
	end
else if Inx<Man.BitmapList.Count then
	begin
	Status:=Man.GetBitmapByHandle(Inx,BM);
	case ACol of
	 0: DisplayBMGridItem(IntToStr(Inx), Rect);
	 1: DisplayBMGridItem(PickString(Status=hsValid,'Valid','DeAlloc'),Rect);
	 2:
	 	begin
	 	if Status=hsValid then S:=IntToStr(BM.Width) else S:='-';
	 	DisplayBMGridItem(S,Rect);
	 	end;
	 3:
	 	begin
	 	if Status=hsValid then S:=IntToStr(BM.Height) else S:='-';
	 	DisplayBMGridItem(S,Rect);
		end;
	 end;
	end;
end;







procedure UpdateBitmapDisplay;
var Man: TBitmapManager;
var BM: TBitmap;
var Status: THandleStatus;
var Han: integer;
begin
Man:=GlobalConsole.VGABox.BitmapManager;
if Man.BitmapList.Count<10 then SystemDlg.BitmapGrid.RowCount:=10
else SystemDlg.BitmapGrid.RowCount:=Man.BitmapList.Count+3;
if (SystemDlg.BitmapGrid.Row-1)>Man.BitmapList.Count then SystemDlg.BitmapGrid.Row:=1;
SystemDlg.BitmapGrid.Invalidate;
Han:=SystemDlg.BitmapGrid.Row-2;
Status:=Man.GetBitmapByHandle(Han,BM);
if Status=hsValid then
	begin
	SystemDlg.Image1.Picture.Bitmap.Assign(BM);
	SystemDlg.CanvasAttribFrm2.DisplayCanvasInfo(BM.Canvas);
	end;
end;



procedure TSystemDlg.BitmapGridSelectCell(Sender: TObject; ACol,
  ARow: Integer; var CanSelect: Boolean);
var LastRow: integer;
begin
LastRow:=GlobalConsole.VGABox.BitmapManager.BitmapList.Count+1;
CanSelect:=ARow<=LastRow;
if CanSelect then UpdateBitmapDisplay;
end;


{------------------------------------------------------------------------------}



procedure DisplayMallocGridItem(S: string; Rect: TRect);
{Display the string in the grid inside the rectangle}
begin
SystemDlg.MallocGrid.Canvas.TextRect(Rect,Rect.Left+2,Rect.Top+2,S);
end;



procedure TSystemDlg.MallocGridDrawCell(Sender: TObject; ACol,
  ARow: Integer; Rect: TRect; State: TGridDrawState);
var Inx: integer;
begin
Inx:=ARow - 1;
if ARow = 0 then
	begin
	case ACol of
	 0: DisplayMallocGridItem('Seg', Rect);
	 1: DisplayMallocGridItem('Size', Rect);
	 end;
	end
else if Inx<Length(MemManager.SegArray) then
	begin
	case ACol of
	 0: DisplayMallocGridItem(IntToStr(Inx), Rect);
	 1: DisplayMallocGridItem(FloatToStrF(Length(MemManager.SegArray[Inx]),ffNumber,18,0),Rect);
	 end;
	end;
end;



procedure TSystemDlg.MallocGridSelectCell(Sender: TObject; ACol,
  ARow: Integer; var CanSelect: Boolean);
var LastRow: integer;
begin
LastRow:=GlobalConsole.VGABox.BitmapManager.BitmapList.Count;
CanSelect:=ARow<=LastRow;
end;




procedure UpdateMemoryDisplay;
var T: integer;
begin
with SystemDlg do
 begin
 ProgSizeDis.Text:=FloatToStrF(ProgramStore.LocalLow,ffNumber,18,0);
 LocalStartDis.Text:='$'+IntToHex(ProgramStore.LocalLow,IntNibbles);
 T:=ProgramStore.HeapLow-ProgramStore.LocalLow;
 LocalSizeDis.Text:=FloatToStrF(T,ffNumber,18,0);
 HeapStartDis.Text:='$'+IntToHex(ProgramStore.HeapLow,IntNibbles);
 HeapPointerDis.Text:='$'+IntToHex(ProgramStore.HeapPtr,IntNibbles);
 T:=ProgramStore.HeapPtr-ProgramStore.HeapLow;
 HeapSizeDis.Text:=FloatToStrF(T,ffNumber,18,0);
 T:= Length(ProgramStore.ProgData.Code)-ProgramStore.HeapLow;
 HeapAllocDis.Text:=FloatToStrF(T,ffNumber,18,0);
 HeapMaxDis.Text:=FloatToStrF(ProgramStore.HeapLimit,ffNumber,18,0);
 MallocSizeDis.Text:=FloatToStrF(MemManager.AllocSize,ffNumber,18,0);

 if Length(MemManager.SegArray)<20 then SystemDlg.MallocGrid.RowCount:=20
 else SystemDlg.MallocGrid.RowCount:=Length(MemManager.SegArray)+2;
 MallocGrid.Invalidate;
 end;
end;


{------------------------------------------------------------------------------}


procedure DisplayFileGridItem(S: string; Rect: TRect);
{Display the string in the grid inside the rectangle}
begin
SystemDlg.FileGrid.Canvas.TextRect(Rect,Rect.Left+2,Rect.Top+2,S);
end;



procedure TSystemDlg.FileGridDrawCell(Sender: TObject; ACol, ARow: Integer;
  Rect: TRect; State: TGridDrawState);
var Inx,H,Size: integer;
begin
Inx:=ARow - 1;
if ARow = 0 then
	begin
	case ACol of
	 0: DisplayFileGridItem('#', Rect);
	 1: DisplayFileGridItem('Name', Rect);
	 2: DisplayFileGridItem('Handle', Rect);
	 3: DisplayFileGridItem('Mode', Rect);
	 4: DisplayFileGridItem('Size', Rect);
	 end;
	end
else if Inx<Length(FileArray) then
	begin
	case ACol of
	 0: DisplayFileGridItem(IntToStr(Inx), Rect);
	 1: DisplayFileGridItem(ExtractFilename(FileArray[Inx].Name), Rect);
	 2: DisplayFileGridItem('$'+IntToHex(FileArray[Inx].Handle,IntNibbles), Rect);
	 3: DisplayFileGridItem(PickString(FileArray[Inx].Mode=0,'Read','Write'), Rect);
	 4:
		begin
		Size:=GetFileSize(FileArray[Inx].Handle,@H);
		DisplayFileGridItem(FloatToStrF(Size,ffNumber,18,0), Rect);
		end;
	 end;
	end;
end;




function DisplayFileInfo(Row: integer): boolean;
var Inx: integer;
var FileInfo: TByHandleFileInformation;
begin
Inx:=Row-1;
Result:=Inx<=High(FileArray);
if Result and (FileArray[Inx].Handle<>INVALID_HANDLE_VALUE) then
	begin
	GetFileInformationByHandle(FileArray[Inx].Handle,FileInfo);
	SystemDlg.FullPathDis.Text:=ExpandFileName(FileArray[Inx].Name);
	SystemDlg.CreateDis.Text:=GetFileDateStr(FileInfo.ftCreationTime,False,True,True);
	SystemDlg.AccessDis.Text:=GetFileDateStr(FileInfo.ftLastAccessTime,False,True,True);
	SystemDlg.WriteDis.Text:=GetFileDateStr(FileInfo.ftLastWriteTime,False,True,True);
	SystemDlg.AttribsDis.Text:=GetFileAttributeStr(FileInfo.dwFileAttributes);
	end
else
	begin
	SystemDlg.CreateDis.Text:='';
	SystemDlg.AccessDis.Text:='';
	SystemDlg.WriteDis.Text:='';
	end;
end;



procedure TSystemDlg.FileGridSelectCell(Sender: TObject; ACol,
  ARow: Integer; var CanSelect: Boolean);
begin
CanSelect:=DisplayFileInfo(ARow);
end;



procedure UpdateFileDisplay;
begin
if Length(FileArray)<10 then SystemDlg.FileGrid.RowCount:=10
else SystemDlg.FileGrid.RowCount:=Length(FileArray)+2;
SystemDlg.FileGrid.Invalidate;
end;


{------------------------------------------------------------------------------}



procedure UpdateSystemDisplay;
begin
if not SystemDlg.Visible then exit;
UpdateConsole;
UpdateBitmapDisplay;
UpdateMemoryDisplay;
UpdateFileDisplay;
DisplayFileInfo(SystemDlg.FileGrid.Row);
end;



procedure ShowSystemDlg;
begin
SystemDlg.Show;
end;



procedure TSystemDlg.FormCreate(Sender: TObject);
begin
ScrollBox1.DoubleBuffered:=True;
Self.PageControl.ActivePageIndex:=0;
end;

procedure TSystemDlg.FormShow(Sender: TObject);
begin
UpdateSystemDisplay;
end;


procedure TSystemDlg.CancelBtnClick(Sender: TObject);
begin
Close;
end;

procedure TSystemDlg.PaletteBtnClick(Sender: TObject);
begin
DoVGAPaletteDlg(True);
end;

end.
