unit IDEVGAPalette;

interface

uses Windows, SysUtils, Classes, Graphics, Forms, Controls, StdCtrls,
  Buttons, ExtCtrls, Grids, ConsoleWindow, ComCtrls,Math, DSupport,BitmapLib;

type
  TVGAPalDlg = class(TForm)
    CancelBtn: TButton;
    PageControl1: TPageControl;
    TabSheet1: TTabSheet;
    TabSheet2: TTabSheet;
    Label1: TLabel;
    VGAGrid: TStringGrid;
    Panel1: TPanel;
    PaletteGrid: TDrawGrid;
    CurrentModeBtn: TButton;
    Panel2: TPanel;
    ShowRGBChk: TCheckBox;
    BitBtn1: TBitBtn;
    procedure PaletteGridDrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure VGAGridDblClick(Sender: TObject);
    procedure CurrentModeBtnClick(Sender: TObject);
    procedure ShowRGBChkClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var VGAPalDlg: TVGAPalDlg;

procedure DoVGAPaletteDlg(Current: boolean);
procedure UpdatePaletteDisplay;


implementation

uses EBXMain;

{$R *.dfm}


var Palette: array of TColor;

var ColorBits: integer;


procedure ShowVGAModes;
var I,Inx,First,Last: integer;
begin
with VGAPalDlg do
 begin
 First:=0;
 Last:=High(GraphicsModes);
 VGAGrid.RowCount:=(Last-First)+1;
 VGAGrid.Cells[0,0]:='#';
 VGAGrid.Cells[1,0]:='Mode';
 VGAGrid.Cells[2,0]:='Card Type';
 VGAGrid.Cells[3,0]:='Base Res';
 VGAGrid.Cells[4,0]:='Text Res';
 VGAGrid.Cells[5,0]:='Char Cell';
 VGAGrid.Cells[6,0]:='Bits';
 VGAGrid.Cells[7,0]:='Type';

 for I:=First to Last do
	begin
	Inx:=I-First;
	VGAGrid.Cells[0,Inx+1]:=IntToStr(Inx);
	VGAGrid.Cells[1,Inx+1]:='$'+IntToHex(GraphicsModes[I].Mode,4);
	VGAGrid.Cells[2,Inx+1]:=GetCardTypeStr(GraphicsModes[I].Card);
	VGAGrid.Cells[3,Inx+1]:=IntToStr(GraphicsModes[I].Res.X)+' X '+IntToStr(GraphicsModes[I].Res.Y);
	VGAGrid.Cells[4,Inx+1]:=IntToStr(GraphicsModes[I].TextRes.X)+' X '+IntToStr(GraphicsModes[I].TextRes.Y);
	VGAGrid.Cells[5,Inx+1]:=GetCharCellStr(GraphicsModes[I].CharCell);
	VGAGrid.Cells[6,Inx+1]:=IntToStr(GraphicsModes[I].Bits);
	VGAGrid.Cells[7,Inx+1]:=PickString(GraphicsModes[I].ModeType=mtText,'Text','Graphics');
	end;
 end;
end;


procedure TVGAPalDlg.PaletteGridDrawCell(Sender: TObject; ACol,
  ARow: Integer; Rect: TRect; State: TGridDrawState);
var Inx,RowColors: integer;
var Color: TColor;
const Border=4;


	procedure DisplayColor(Rect: TRect; Color: TColor);
	var R,G,B: byte;
	var W: integer;
	var S: string;
	begin
	PaletteGrid.Canvas.Font.Color:=ContrastColor(Color);
	ColorToRGBBytes(Color,R,G,B);
	S:=IntToHex(R,2)+'-'+IntToHex(G,2)+'-'+IntToHex(B,2);
	W:=(Rect.Right-Rect.Left)-(Border*2);
	W:=W-PaletteGrid.Canvas.TextWidth(S);
	if W<0 then W:=0 else W:=W div 2;
	PaletteGrid.Canvas.TextOut(Rect.Left+Border+W,Rect.Top+Border+2,S);
	end;



	procedure ShowColor(Color: TColor; R: TRect);
	begin
	PaletteGrid.Canvas.Brush.Color:=Color;
	PaletteGrid.Canvas.Pen.Color:=clGray;
	PaletteGrid.Canvas.Rectangle(R.Left+Border, R.Top+Border, R.Right-Border, R.Bottom-Border);
	if Self.ShowRGBChk.Checked then DisplayColor(Rect,Color);
	end;

	procedure DrawText(Rect: TRect; S: String);
	const Border=2;
	begin
	PaletteGrid.Canvas.TextOut(Rect.Left+Border,Rect.Top+Border,S);
	end;

	procedure HandleHeaders;
	begin
	if (ACol mod 2)=0 then DrawText(Rect,'Inx')
	else DrawText(Rect,'Color');
	end;


begin
RowColors:=PaletteGrid.ColCount div 2;
Inx:=((ARow-1) * RowColors) + ACol div 2;
if ARow=0 then begin HandleHeaders; exit; end;
if ColorBits>8 then
	begin
	case ColorBits of
	 15: Color:=Color15ToColor(Inx);
	 16: Color:=Color16ToColor(Inx);
	 24: Color:=Color24ToColor(Inx);
	 end;
	end
else
	begin
	if Inx>High(Palette) then exit;
	Color:=Palette[Inx];
	end;
if (ACol mod 2)=0 then DrawText(Rect,IntToHex(Inx,4))
else ShowColor(Color,Rect);
end;



procedure SetColorPalette(P: array of TColor; Bits: integer);
var I,RowColors,RC: integer;
var ColorCount,Len: integer;
var CountChange: boolean;
begin
CountChange:=ColorBits<>Bits;
ColorBits:=Bits;
if ColorBits>24 then ColorBits:=24;
ColorCount:=trunc(IntPower(2,ColorBits));
if ColorCount>256 then Len:=Length(P)
else Len:=ColorCount;
SetLength(Palette,Len);
for I:=0 to Len-1 do Palette[I]:=P[I];
with VGAPalDlg do
 begin
 RowColors:=PaletteGrid.ColCount div 2;
 RC:=ColorCount div RowColors;
 if (ColorCount mod RowColors)<>0 then Inc(RC);
 if RC<10 then RC:=10;
 PaletteGrid.RowCount:=RC+1;
 if CountChange then PaletteGrid.TopRow:=1;
 PaletteGrid.Invalidate;
 end;
end;



procedure TVGAPalDlg.VGAGridDblClick(Sender: TObject);
var Inx: integer;
begin
Inx:=VGAGrid.Row-1;
if (Inx<0) or (Inx>High(GraphicsModes)) then exit;
case GraphicsModes[Inx].Bits of
 1: SetColorPalette(VGAPalette1,GraphicsModes[Inx].Bits);
 2: SetColorPalette(VGAPalette4[0],GraphicsModes[Inx].Bits);
 3: SetColorPalette(VGAPalette8,GraphicsModes[Inx].Bits);
 4: SetColorPalette(VGAPalette16,GraphicsModes[Inx].Bits);
 8: SetColorPalette(VGAPalette256,GraphicsModes[Inx].Bits);
 else SetColorPalette(VGAPalette256,GraphicsModes[Inx].Bits);
 end;
PageControl1.ActivePageIndex:=1;
end;



procedure UpdatePaletteDisplay;
var I: integer;
begin
if not VGAPalDlg.Visible then exit;
SetColorPalette(Form1.ConsoleWnd.VGABox.ColorPalette,Form1.ConsoleWnd.VGABox.GraphicsMode.Bits);
for I:=0 to High(GraphicsModes) do
 if GraphicsModes[I].Mode=Form1.ConsoleWnd.VGABox.VGAMode then
	begin
	VGAPalDlg.VGAGrid.Row:=I+1;
	end;
end;



procedure TVGAPalDlg.CurrentModeBtnClick(Sender: TObject);
begin
UpdatePaletteDisplay;
end;


procedure DoVGAPaletteDlg(Current: boolean);
begin
ShowVGAModes;
VGAPalDlg.Show;
if Current then
	begin
	VGAPalDlg.PageControl1.ActivePageIndex:=1;
	VGAPalDlg.CurrentModeBtnClick(VGAPalDlg)
	end
else SetColorPalette(VGAPalette256,8);
end;


procedure TVGAPalDlg.ShowRGBChkClick(Sender: TObject);
begin
PaletteGrid.Invalidate;
end;

end.
