unit IDEVariables;

interface

uses Windows, SysUtils, Classes, Graphics, Forms, Controls, StdCtrls,
  Buttons, ExtCtrls, ComCtrls, TreeViewEx, XPLCompiler, ImgList,ScrollEditor,
  Dialogs,Grids, VCLLibrary, IDEVariableLib,DSupport;

type
  TVariableDlg = class(TForm)
    Panel2: TPanel;
    VariableTree: TTreeViewEx;
    ImageList1: TImageList;
    Panel3: TPanel;
    CancelBtn: TButton;
    Panel1: TPanel;
    Panel6: TPanel;
    ProcDis: TLabel;
    Panel10: TPanel;
    OKBtn: TButton;
    Panel8: TPanel;
    CheckAllBtn: TSpeedButton;
    ClearAllBtn: TSpeedButton;
    ChkSelectedBtn: TSpeedButton;
    ClrSelectedBtn: TSpeedButton;
    Panel13: TPanel;
    ScopeDis: TLabel;
    Panel9: TPanel;
    Panel11: TPanel;
    TypeDis: TLabel;
    Panel5: TPanel;
    LevelDis: TLabel;
    Panel4: TPanel;
    OffDis: TLabel;
    Panel7: TPanel;
    NameDis: TLabel;
    Panel12: TPanel;
    FileDis: TLabel;
    SearchBtn: TSpeedButton;
    FindDialog1: TFindDialog;
    Panel25: TPanel;
    Panel14: TPanel;
    ProcCntDis: TLabel;
    Panel15: TPanel;
    SymbolCntDis: TLabel;
    CollapseBtn: TSpeedButton;
    ExpandBtn: TSpeedButton;
    Panel16: TPanel;
    VariablesChk: TCheckBox;
    BitBtn1: TBitBtn;
    procedure VariableTreeClick(Sender: TObject);
    procedure VariableTreeGetSelectedIndex(Sender: TObject;
      Node: TTreeNode);
    procedure CheckAllBtnClick(Sender: TObject);
    procedure ClearAllBtnClick(Sender: TObject);
    procedure ChkSelectedBtnClick(Sender: TObject);
    procedure ClrSelectedBtnClick(Sender: TObject);
    procedure VariableTreeDblClick(Sender: TObject);
    procedure VariableTreeCheckChange(Sender: TObject; Node: TTreeNode;
      Button: TMouseButton; Shift: TShiftState);
    procedure SearchBtnClick(Sender: TObject);
    procedure FindDialog1Find(Sender: TObject);
    procedure CollapseBtnClick(Sender: TObject);
    procedure ExpandBtnClick(Sender: TObject);
    procedure OKBtnClick(Sender: TObject);
    procedure CancelBtnClick(Sender: TObject);
    procedure VariablesChkClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  VariableDlg: TVariableDlg;




procedure DoVariableDlg(WatchMode: boolean);
procedure LoadVariables;
procedure AddWatchItem(Info: TWatchInfo);
procedure SetWatchItem(VarText,Filename: string; Line: integer);

procedure DeleteWatchItem(Inx: integer; ClearChecks: boolean); overload;
procedure DeleteWatchItem(Inx: integer); overload;
procedure MoveWatchItem(Inx: integer; Down: boolean);
procedure ClearWatches;
procedure RefreshVariables;
procedure ShowVarsInScope(SG: TStringGrid);

implementation

uses I2LDataTypes, I2LInterp, IDEDebugger, XPLDataTypes, I2LGlobal, IDEGlobal,
     EBXMain, XPLStructs;

{$R *.dfm}


var CurWatchMode: boolean;	{Flags if dialog is used to get watch info }


{------------------------------------------------------------------------------}



procedure GetNodeInfo(Node: TTreeNode; var PI: PProcItem; var SI: PSymbolItem);
begin
PI:=nil; SI:=nil;
if Node.ImageIndex=0 then PI:=PProcItem(Node.Data)
else
	begin
	PI:=PProcItem(Node.Parent.Data);
	SI:=PSymbolItem(Node.Data);
	end;
end;


function WatchToTreeItem(Info: TVariableInfo): integer;
var I: integer;
var PI: PProcItem;
var SI: PSymbolItem;
begin
Result:=-1;
for I:=0 to VariableDlg.VariableTree.Items.Count-1 do
	begin
	GetNodeInfo(VariableDlg.VariableTree.Items[I], PI, SI);
	if SI=nil then continue;
	if (Info.BaseName=SI.Name) and
	   (Info.Level=SI.Level) and
	   (Info.Offset=SI.Offset) then
		begin
		Result:=I;
		break;
		end;
	end;
end;

procedure SetTreeItemByInfo(Info: TVariableInfo; Check: boolean);
var Inx: integer;
begin
Inx:=WatchToTreeItem(Info);
if Inx>-1 then VariableDlg.VariableTree.Checked[Inx]:=Check;
end;


procedure AddWatchItem(Info: TWatchInfo);
var VI: TVariableInfo;
begin
SetLength(WatchArray,Length(WatchArray)+1);
WatchArray[High(WatchArray)]:=Info;
if LookupVariable(Info.VarText,Info.Filename,Info.Line,VI)=lrFound then
	begin
	SetTreeItemByInfo(VI,True);
	end;
end;


procedure SetWatchItem(VarText,Filename: string; Line: integer);
var Info: TWatchInfo;
var VI: TVariableInfo;
begin
Info.VarText:=Vartext;
Info.Filename:=Filename;
Info.Line:=Line;
Info.CustomFormat:=False;
Info.UserFormat.Digits:=5;
Info.UserFormat.Radix:=nrDecimal;
Info.UserFormat.RealFormat:=rfGeneral;
Info.UserFormat.ForcedArray:=False;
SetLength(Info.UserFormat.DisplayDims,0);
if LookupVariable(Info.VarText,Info.Filename,Info.Line,VI)=lrFound then
	Info.UserFormat.DisplayType:=GetDisplayType(VI)
else Info.UserFormat.DisplayType:=dsInteger;
AddWatchItem(Info);
end;




procedure RefreshVariables;
var I: integer;
var WA: TWatchArray;
var VI: TVariableInfo;
begin
WA:=WatchArray;
for I:=0 to High(WA) do
 if LookupVariable(WA[I].VarText,WA[I].Filename,WA[I].Line,VI)=lrFound then
	begin
	SetTreeItemByInfo(VI,True);
	end;
UpdateWatchGrid;
end;

{-----------------------------------------------------------------------------}


procedure SetAllChecks(Chk: boolean);
var I: integer;
begin
for I:=0 to VariableDlg.VariableTree.Items.Count-1 do
	begin
	if VariableDlg.VariableTree.Items[I].ImageIndex=0 then continue;
	VariableDlg.VariableTree.Checked[I]:=Chk;
	end;
end;



procedure SetAllSelected(Chk: boolean);
var I,I2: integer;
begin
for I:=0 to VariableDlg.VariableTree.SelectionCount-1 do
	begin
	I2:=VariableDlg.VariableTree.Selections[I].AbsoluteIndex;
	if VariableDlg.VariableTree.Items[I2].ImageIndex=0 then continue;
	VariableDlg.VariableTree.Checked[I2]:=Chk;
	end;
end;


procedure InvertAllSelected;
var I,I2: integer;
begin
for I:=0 to VariableDlg.VariableTree.SelectionCount-1 do
	begin
	I2:=VariableDlg.VariableTree.Selections[I].AbsoluteIndex;
	if VariableDlg.VariableTree.Items[I2].ImageIndex=0 then continue;
	VariableDlg.VariableTree.Checked[I2]:=not VariableDlg.VariableTree.Checked[I2];
	end;
end;


procedure TVariableDlg.CheckAllBtnClick(Sender: TObject);
begin
SetAllChecks(True);
end;

procedure TVariableDlg.ClearAllBtnClick(Sender: TObject);
begin
SetAllChecks(False);
end;

procedure TVariableDlg.ChkSelectedBtnClick(Sender: TObject);
begin
SetAllSelected(True);
end;

procedure TVariableDlg.ClrSelectedBtnClick(Sender: TObject);
begin
SetAllSelected(False);
end;


{-----------------------------------------------------------------------------}




function GetCurrentNodeInfo(var PI: PProcItem; var SI: PSymbolItem): boolean;
var Node: TTreeNode;
begin
Result:=False;
Node:=VariableDlg.VariableTree.Selected;
if Node.Data=nil then exit;
GetNodeInfo(Node,PI,SI);
Result:=True;
end;





procedure DeleteWatchItem(Inx: integer; ClearChecks: boolean);
var I: integer;
var WI: TWatchInfo;
var VI: TVariableInfo;
begin
if (Length(WatchArray)<1) or (Inx<0) or (Inx>=Length(WatchArray)) then exit;
{Because it is being deleted from watches, uncheck in the variable list}
if ClearChecks then
	begin
	WI:=WatchArray[Inx];
	if LookupVariable(WI.VarText,WI.Filename,WI.Line,VI)=lrFound then
		begin
		SetTreeItemByInfo(VI,False);
		end;
	end;
for I:=Inx to Length(WatchArray)-2 do WatchArray[I]:=WatchArray[I+1];
SetLength(WatchArray,Length(WatchArray)-1);
end;



procedure DeleteWatchItem(Inx: integer);
begin
DeleteWatchItem(Inx,True);
end;



procedure MoveWatchItem(Inx: integer; Down: boolean);
var NewInx: integer;
var WI: TWatchInfo;
begin
if Down then NewInx:=Inx+1 else NewInx:=Inx-1;
if (NewInx<0) or (NewInx>High(WatchArray)) then exit;
WI:=WatchArray[Inx];
WatchArray[Inx]:=WatchArray[NewInx];
WatchArray[NewInx]:=WI;
DebugDlg.WatchGrid.Row:=NewInx+1;
UpdateWatchGrid;
end;




procedure ClearWatches;
var I: integer;
begin;
SetLength(WatchArray,0);
UpdateWatchGrid;
for I:=0 to VariableDlg.VariableTree.Items.Count-1 do
	VariableDlg.VariableTree.Checked[I]:=False;
end;


procedure SetImageIndexes(Node: TTreeNode; Inx: integer);
begin
Node.ImageIndex:=Inx;
Node.StateIndex:=Inx;
end;



procedure RecurseVariables(ShowVariables: boolean; VariableTree: TTreeViewEx);
var TypeIndex,Inx: integer;
var Cnt: integer;


	procedure RecurseUserTypes(Index: integer; Node: TTreeNode);
	var J, Start, Stop: integer;
	var S: string;
	var SubNode: TTreeNode;
	begin
	Start:=SubItemIndices[Index];
	Stop:=Start + SubItemCount[Index]-1;
	for J:=Start to Stop do
		begin
		S:=SubItemName[J]+' = '+GetTypeString(SubItemType[J],SubUserIndex[J]);
		SubNode:=VariableTree.Items.AddChild(Node,S);
		SetImageIndexes(SubNode,4);
		if SubItemType[J]=USRTYPE then
			begin
                        //SubNode.Data:=@Proc.Symbols[J];
			RecurseUserTypes(SubUserIndex[J], SubNode);
			end;
		end;
	end;


	procedure AddSymbols(Proc: TProcItem; ProcNode: TTreeNode);
	var VarNode: TTreeNode;
	var Info: TSymbolItem;
	var J: integer;
	var S: string;
	var TypeDef: boolean;
	begin
	for J:=0 to High(Proc.Symbols) do
		begin
		Info:=Proc.Symbols[J];
		S:=GetTypeString(Info.SymType,Info.StructIndex);
		TypeDef:=Info.Name=S;
		if TypeDef then S:='Type Def';
		S:=Info.Name+'= '+S;
		VarNode:=VariableTree.Items.AddChild(ProcNode,S);
		VarNode.Data:=@Proc.Symbols[J];
                if Info.SymType=USRTYPE then
			begin
                        TypeIndex:=FindUserType(Info.Name);
                        if TypeIndex>=0 then
                        	begin
				RecurseUserTypes(TypeIndex, VarNode);
				end;
			end;
		if TypeDef then SetImageIndexes(VarNode,2)
		else SetImageIndexes(VarNode,1);
		Inc(Cnt);
		end;
	end;


	function AddProcedure(Proc: TProcItem; Parent: TTreeNode): TTreeNode;
	var S: string;
	begin
	S:=' ($'+IntToHex(Proc.BeginPC+1,4)+')';
	Result:=VariableTree.Items.AddChildFirst(Parent,Proc.ProcName+S);
	Result.Data:=@SymbolDump[Inx];
	SetImageIndexes(Result,0);
	if ShowVariables then AddSymbols(Proc,Result);
	end;


	procedure LoadNode(OldLevel: integer; Parent: TTreeNode);
	var ProcNode: TTreeNode;
	var Proc: TProcItem;
	begin
	while true do
		begin
		Proc:=SymbolDump[Inx];
		ProcNode:=AddProcedure(Proc,Parent);
		Dec(Inx);
		if Inx<0 then exit;
		if SymbolDump[Inx].Level<OldLevel then exit
		else if SymbolDump[Inx].Level>OldLevel then LoadNode(SymbolDump[Inx].Level,ProcNode);
		if Inx<0 then exit;
		end;
	end;


begin
Cnt:=0;
VariableTree.Items.BeginUpdate;
try
VariableTree.Items.Clear;
Inx:=High(SymbolDump);
LoadNode(0,nil);
with VariableDlg do
 begin
 ProcCntDis.Caption:=IntToStr(Length(SymbolDump));
 SymbolCntDis.Caption:=IntToStr(Cnt);
 end;
finally VariableTree.Items.EndUpdate; end;
end;




procedure LoadVariables;
begin
RecurseVariables(VariableDlg.VariablesChk.Checked, VariableDlg.VariableTree);
end;



procedure ClearVariableDisplay;
begin
VariableDlg.NameDis.Caption:='';
VariableDlg.TypeDis.Caption:='';
VariableDlg.LevelDis.Caption:='';
VariableDlg.OffDis.Caption:='';
end;



procedure TVariableDlg.VariableTreeClick(Sender: TObject);
var PI: PProcItem;
var SI: PSymbolItem;
begin
ClearVariableDisplay;
if not GetCurrentNodeInfo(PI,SI) then exit;
ScopeDis.Caption:='$'+IntToHex(PI.BeginPC,SizeOf(TI2LWord)*2)+' $'+IntToHex(PI.EndPC,SizeOf(TI2LWord)*2);
ProcDis.Caption:=PI.ProcName;
FileDis.Caption:=ExtractFilename(PI.SourceInfo.Filename);
if SI<>nil then
	begin
	NameDis.Caption:=SI.Name;
	if SI.SymType=USRTYPE then
		begin
		TypeDis.Caption:=UserTypeNames[SI.StructIndex];
		Caption:=IntToStr(GetVariableSize(SI.StructIndex, SI.Name));
		end
	else TypeDis.Caption:=TypeToStr(SI.SymType);
	LevelDis.Caption:=IntToStr(SI.Level);
	OffDis.Caption:=IntToStr(SI.Offset);
	FileDis.Caption:=ExtractFilename(SI.SourceInfo.Filename);
	end
else
	begin
	NameDis.Caption:=PI.ProcName;
	TypeDis.Caption:='Proc';
	LevelDis.Caption:=IntToStr(PI.Level);
	OffDis.Caption:='0';
	FileDis.Caption:=ExtractFilename(PI.SourceInfo.Filename);
	end;
end;

procedure TVariableDlg.VariableTreeGetSelectedIndex(Sender: TObject;
  Node: TTreeNode);
begin
Node.SelectedIndex:=Node.ImageIndex;
end;


procedure DeleteNonArrays;
{Delete all variables that aren't arrays variables}
var I: integer;
begin
for I:=High(WatchArray) downto 0 do
 if Pos('(',WatchArray[I].VarText)<1 then
	begin
	DeleteWatchItem(I, False);
	end;
end;


procedure CopySelectedVariables;
var I: integer;
var PI: PProcItem;
var SI: PSymbolItem;
begin
DeleteNonArrays;
for I:=0 to VariableDlg.VariableTree.Items.Count-1 do
	begin
	if not VariableDlg.VariableTree.Checked[I] then continue;
	if VariableDlg.VariableTree.Items[I].ImageIndex=0 then continue;
	GetNodeInfo(VariableDlg.VariableTree.Items[I],PI,SI);
	SetWatchItem(SI.Name,PI.SourceInfo.Filename,SI.SourceInfo.Line);
	end;
UpdateWatchGrid;
end;



procedure TVariableDlg.VariableTreeDblClick(Sender: TObject);
var PI: PProcItem;
var SI: PSymbolItem;
var Info: TSourceInfo;
var ED: TScrollEditor;
begin
if not GetCurrentNodeInfo(PI,SI) then exit;
if SI<>nil then Info:=SI.SourceInfo
else Info:=PI.SourceInfo;
if CurWatchMode then ED:=SetDebugEditor(Info.Filename)
else ED:=Form1.SourceEdit.Editor;
ED.SetPosition(Point(0,Info.Line));
ED.HighlightRow:=Info.Line;
ED.Invalidate;
end;



procedure TVariableDlg.VariableTreeCheckChange(Sender: TObject;
  Node: TTreeNode; Button: TMouseButton; Shift: TShiftState);
begin
if Node.ImageIndex<>0 then exit;
VariableTree.SetAllChildren(Node,TTreeNodeEx(Node).Checked);
end;



procedure TVariableDlg.SearchBtnClick(Sender: TObject);
begin
FindDialog1.Execute;
end;


procedure TVariableDlg.FindDialog1Find(Sender: TObject);
begin
SearchTree(VariableTree,FindDialog1);
end;



procedure TVariableDlg.CollapseBtnClick(Sender: TObject);
begin
VariableTree.FullCollapse;
end;




procedure TVariableDlg.ExpandBtnClick(Sender: TObject);
begin
VariableTree.FullExpand;
end;

{-----------------------------------------------------------------------------}



procedure TVariableDlg.OKBtnClick(Sender: TObject);
begin
if CurWatchMode then CopySelectedVariables;
VariableDlg.Close;
end;




procedure TVariableDlg.CancelBtnClick(Sender: TObject);
begin
VariableDlg.Close;
end;


procedure SetupDialogMode;
begin
with VariableDlg do
 begin
 Caption:=PickString(CurWatchMode,'Watch Variables','Program Structure');
 CancelBtn.Caption:=PickString(CurWatchMode,'Cancel','Close');
 OkBtn.Visible:=CurWatchMode;
 CheckAllBtn.Visible:=CurWatchMode;
 ClearAllBtn.Visible:=CurWatchMode;
 ChkSelectedBtn.Visible:=CurWatchMode;
 ClrSelectedBtn.Visible:=CurWatchMode;
 VariableTree.CheckBoxes:=CurWatchMode;
 Panel16.Visible:=not CurWatchMode;
 if CurWatchMode then VariablesChk.Checked:=True;
 end;
end;



procedure DoVariableDlg(WatchMode: boolean);
begin
CurWatchMode:=WatchMode;
SetupDialogMode;
ClearVariableDisplay;
VariableDlg.Show;
end;


procedure TVariableDlg.VariablesChkClick(Sender: TObject);
begin
LoadVariables;
end;


function VarsInScope(PC: integer): integer;
var I,J,K: integer;
var Proc: TProcItem;
var Symbol: TSymbolItem;
begin
Result:=0;
for I:=0 to High(SymbolDump) do
	begin
	Proc:=SymbolDump[I];
	if (PC<Proc.BeginPC) or (PC>Proc.EndPC) then continue;
	for J:=0 to High(Proc.Symbols) do
		begin
		Symbol:=Proc.Symbols[J];
                if Symbol.IsTypeDef then continue;
		Inc(Result);
                if Length(Symbol.Dims^)>0 then Inc(Result,Symbol.Dims^[0]);
		end;
	end;
end;


procedure ShowVarsInScope(SG: TStringGrid);
var I,J,K,L,PC,Row,Count,Size: integer;
var LastAdd: integer;
var Proc: TProcItem;
var Symbol: TSymbolItem;
var S: string;
const ExtraMem = 1024;

	procedure GetDataDisplay(Add: integer; var IntStr,HexStr: string);
	var Data: integer;
	begin
	if (Add<DisplayVector[0]) or not I2LProgram.RangeCheck(Add) then
		begin
		IntStr:='-';
		HexStr:='-';
		end
	else
		begin
		Data:=I2LProgram.GetDataWord(Add);
		IntStr:=IntToStr(Data);
		HexStr:=IntToHexBytes(Data,2);
		end;
	end;

	procedure DisplayItem(Name: string; Add,Level,Offset: integer);
	var IntStr,HexStr: string;
	begin
	SG.Cells[0,Row]:=IntToStr(Row);
	SG.Cells[1,Row]:=IntToHexBytes(Add,2);
	SG.Cells[2,Row]:=Name;
	GetDataDisplay(Add,IntStr,HexStr);
	SG.Cells[3,Row]:=HexStr;
	SG.Cells[4,Row]:=IntStr;
	SG.Cells[5,Row]:=IntToStr(Level);
	SG.Cells[6,Row]:=IntToHexBytes(Offset,2);
	Inc(Row);
	end;


	procedure ShowItem(Name: string; Level,Offset,Index: integer);
	var Add, Data: integer;
	var IntStr,HexStr: string;
	begin
	SG.Cells[0,Row]:=IntToStr(Row);
	Add:=DisplayVector[Level]+Offset;
	if Index>=0 then Add:=I2LProgram.GetDataWord(Add) + Size * Index;
	SG.Cells[1,Row]:=IntToHexBytes(Add,2);
	if Index>=0 then Name:=Name+'('+IntToStr(Index)+')';
	DisplayItem(Name,Add,Level,Offset);
	if Add>LastAdd then LastAdd:=Add;
	end;

	function GetSymbolSize: integer;
	begin
	case Symbol.SymType of
	 RLVAR: Result:=RealSize;
	 USRTYPE: Result:=GetVariableSize(Symbol.StructIndex, '');
	 else Result:=IntBytes;
	 end;
	end;


begin
LastAdd:=0;
ClearStringGrid(SG,False);
Row:=1;
Size:=IntBytes;
ShowItem('Global(0)',0,0,-1);
ShowItem('Global(1)',0,0,-1);
Count:=2;
PC:=ProgramStore.ProgCntr;
Count:=Count+VarsInScope(PC)+ExtraMem;
SG.RowCount:=Count+1;
for I:=0 to High(SymbolDump) do
	begin
	Proc:=SymbolDump[I];
	if (PC<Proc.BeginPC) or (PC>Proc.EndPC) then continue;
	for J:=0 to High(Proc.Symbols) do
		begin
		Symbol:=Proc.Symbols[J];
                if Symbol.IsTypeDef then continue;
		Size:=GetSymbolSize;
		ShowItem(Symbol.Name,Symbol.Level,Symbol.Offset,-1);
                if Length(Symbol.Dims^)>0 then
		 for K:=0 to Symbol.Dims^[0]-1 do
			begin
			ShowItem(Symbol.Name,Symbol.Level,Symbol.Offset,K);
			for L:=1 to (Size div IntBytes)-1 do
				DisplayItem('Mem',LastAdd+(IntBytes*L),Symbol.Level,Symbol.Offset);
			end;
		end;
	end;
for K:=0 to ExtraMem-1 do
	begin
	Inc(LastAdd,IntBytes);
	DisplayItem('Mem',LastAdd,Symbol.Level,Symbol.Offset);
	end;

end;

end.
