unit XPLCompiler;
{ XPL0 Compiler for embedding in other programs }
{ Original Compiler P.J.R. Boyle }
{ Improvements By Loren Blaney and Larry Fish }
{  Embedded version by Larry Fish - 2004 }

interface

uses Forms,SysUtils,Controls,Contnrs,Classes,Types,Dialogs,
     DSupport,Hash,VCLLibrary,StreamLib,CircularBuffer, XPLStructs;

{Note: STRUCTS are turned on or off under:
Project->Options->Directories/Conditions->Conditionals
This is necessary to make it global to all modules.
Rebuild is necessary for changes to take effect.
}

{Information to hold search path and file path information}

type TLibType = (ltFile,ltPath);

type TLibInfo = class(TObject)
 Enabled: boolean;
 LibType: TLibType;
 end;

{Array to hold search path and file information}
{Compiler will scan this list to find include files}

var LibFiles: TAutoString;

{Type defines type of Source file were dealing with}

type TSourceType = (stMain,stInclude,stInternalInclude);

{Internal automatic includes}

type TInternalInclude = record
 FileStr: string;
 Title: string;
 end;

var I2LText,ListText: string;

var InternalIncludes: array of TInternalInclude;

{Record to define a location in the source code}

type TSourceInfo = record
 Filename: string;
 SourceType: TSourceType;
 Line,Column: integer;
 end;

{Compiler event used by external programs to monitor compiler state}

type TCompEventType = (etCompileStart,etCompileEnd,etAfterVarDec,etProcedureEnd);

type TCompileEvent=procedure(Info: TCompEventType; ProcName: string);

var CompileEvent: TCompileEvent;
var ProcedureName: string;

{Record containing XPL Error/Warning data}

type TXPLMessageType = (xmError,xmWarning,xmInfo,xmRuntime,xmSearch);

type TMessageInfo = record
 MessageType: TXPLMessageType;
 ErrorNumber: integer;
 Description: string;
 SearchCol1,SearchCol2: integer;
 Info: TSourceInfo;
 end;
type PMessageInfo = ^TMessageInfo;


{Symbol Dump Arrays}

type TSymbolItem = record
 Name: string;
 SymType: integer;
 Offset: integer;
 Level: integer;
 Dims: PDynWordArray;
 {$IFDEF STRUCTS}
 FullName: string;
 StructIndex: integer;
 IsTypeDef: boolean;
 {$ENDIF}
 SourceInfo: TSourceInfo;
 RealConst: double;
 Used: boolean;
 end;

type PSymbolItem = ^TSymbolItem;

type TSymbolList = array of TSymbolItem;

type TProcItem = record
 ProcName: string;
 SourceInfo: TSourceInfo;
 BeginPC,EndPC,Level: integer;
 Symbols: TSymbolList;
 end;

type PProcItem = ^TProcItem;

type TSymbolDump = array of TProcItem;

var SymbolDump: TSymbolDump;


{Array to correlate the PC with source line number for debugging purposes}

type TPCInfo = record
 Info: TSourceInfo;
 PC: integer;
 end;

type TPCArray = array of TPCInfo;



{Souce code array - holds main file and all includes}

type TSourceItem = record
 Text: string;
 TextInx: integer;
 ArrayInx: integer;
 Info: TSourceInfo;
 PCArray: TPCArray;
 end;

type PSourceItem = ^TSourceItem;

var SourceArray: array of TSourceItem;	{Array of files that will be compiled}
var FileStack: TIntStack;		{Handles include file nesting }

var CurrentSource: PSourceItem;		{Current file being compiled}
var OldInfo: TSourceInfo;

{These are User Supplied subroutines to handle messages for errors and warnings}
{To handle errors/warnings pass a custom subroutine to the compiler}
{if the user doesn't pass a subroutine to the compiler they are ignored}

{Procedure Definition for error handler}

type TMessageHandler=procedure(MessageInfo: TMessageInfo; var Continue: boolean);

var MessageHandler: TMessageHandler;	{User-supplied Message handler - if nil it is ignored}

const ErrorStrings: array [0..80] of string =
 ('',						{0}
 'Too Many Variables ',				{1}
 'Too Many Real Constant Names ',		{2}
 'Too Many Names ',				{3}
 'Too Many "quits" ',				{4}
 'Too Many Static Levels ',			{5}
 'Number Out Of Range ',			{6}
 'Number Out Of Range ',			{7 - For Intrinsic Declarations }
 '? ', '? ',					{8 - 9}
 'Undeclared Name ',				{10}
 'Name Already Declared ',			{11}
 '? ','? ','? ','? ','? ','? ','? ','? ',	{12 - 19}
 'Illegal Start Of A Statement ',		{20 - In "assign" }
 '":=" Expected But Not Found ',		{21}
 '"Then" Expected But Not Found ',		{22}
 '"Do" Expected But Not Found ',		{23}
 '",", "to" or "downto" Expected But Not Found ',	{24}
 '? ',						{25}
 'Illegal Factor ',				{26 - Unrecognizable Special Factor }
 'Statement Starting With A Constant ', 	{27 - In "assign" }
 '"Until'''' Expected But Not Found ',		{28}
 '"Other'''' Expected But Not Found ',		{29}
 '"Else'''' Expected But Not Found ',		{30}
 'Digit Expected But Not Found ',		{31}
 'Integer Variable Expected But Not Found ',	{32 - In A 'for' Statement }
 '? ','? ','? ','? ','? ',			{33 - 37}
 '">" Expected But Not Found ',			{38 - ASR ->>}
 '"(" Expected But Not Found ',			{39}
 '"=" Expected But Not Found ',			{40}
 '";" Expected But Not Found ',			{41}
 'Constant Expected But Not Found ',		{42 -  In "getcon" }
 'Variable Expected But Not Found ',		{43 -  For An 'addr' Operator }
 '")" Expected But Not Found ',			{44}
 'Name Expected But Not Found ',		{45}
 'Mixed Mode ',					{46}
 'Integer Expected But Not Found ',		{47}
 '"Of'''' Expected But Not Found ',		{48}
 '":" Expected But Not Found ',			{49}
 '"]" Expected But Not Found ',			{50}
 'No Arguments Declared ',			{51}
 'Statement Starting With "ELSE" ',		{52}
 'Statement Starting With "OTHER" ',		{53}
 '? ','? ','? ','? ','? ','? ',			{54 - 59}
 '"QUIT" Not In A "loop" ',			{60}
 'EOF Expected But Not Found ',			{61}
 'EOF Inside A Block ',				{62}
 'EOF Inside A String ',			{63}
 '? ',						{64}
 '"FPROC" & Its "PROC" Not At Same Level ',	{65}
 '"FPROC" Reference Not Found ',		{66}
 '"PROC" Or "FUNC" Expected But Not Found ',	{67}
 '"EPROC`S And "PUBLIC`S Must Be Global ',	{68}
 '"Include`s Nested Too Deep ',			{69}
 'Bad File Spec Syntax ',			{70}
 'File Not Found ',				{71}
 {$IFDEF STRUCTS}
 '"int", "real", "char," "addr" "User Type" Expected But Not Found ',	{72}
 {$ELSE}
 '"int", "real", "char" or "addr" Expected But Not Found ',	{72}
 {$ENDIF}
 'Divide By Zero In A Constant Expression ',	{73}
 'Math Error In A Constant Expression ',	{74}
 'Too Many Intrinsic Arguments',		{75}	//New
 'Too Few Intrinsic Arguments',			{76} 	//New
 'Argument Type Mismatch',			{77} 	//New
 'Intrinic Procedure used as a function',	{78} 	//New
 'Intrinsic Function used as a Procedure',	{79} 	//New
 'Squareroot of Negative Number'		{80}	//New
 );


const WarningStrings: array [0..16-1] of string =
 ('',							{0}
 'Local Name Same As More-Global Name ',		{1}
 'Rounding Errors Make Equality Unlikely ',		{2}
 'Control Character ',					{3}
 'Use "other" Instead Of "else" ',			{4}
 'Commas Missing In "define" ',				{5}
 'Unknown Warning',					{6}
 'Doubled Unary Operator: ++, --, Or ~~ ',		{7}
 'Declared But Not Used ',				{8}
 '"addr" Of "for" Loop Control Variable ',		{9}
 'Value Assigned But Never Used ',			{10}
 '"eproc" & "ext" Not Supported In This Version.',	{11}
 'Possible Type Mismatch',				{12} 	//New
 'Unused Symbol in include File/library',		{13}	//New
 'Constant Defined But Never Used ',			{14}	//New
 'Unused Public Procedure/Function'			{15}	//New
 );



{Variable Type Definitions}

const 	UNDEF=0;	{ Undefined Id (no. Order Is Critical) }
	ADDRVAR=1;	{ Address Variable Id (type = Integer) }
	INVAR=3;	{ Integer Variable Id (odd Nos.=integer) }
	RLVAR=4;	{ Real Variable Id }
	INCON=5;	{ Integer Constant Id }
	RLCON=6;	{ Real Constant Id }
	INPROC=7;	{ Integer Procedure Id }
	RLPROC=8;	{ Real Procedure Id }
	INFPROC=9;	{ Integer Forward Procedure Id }
	RLFPROC=10;	{ Real Forward Procedure Id }
	INOPT=11;	{ Integer Optimized Procedure Id }
	RLOPT=12;	{ Real Optimized Procedure Id }
	INEPRO=13;	{ Integer External I2l Procedure Id }
	RLEPRO=14;	{ Real External I2l Procedure Id }
	ININT=15;	{ Integer Intrinsic Id }
	RLINT=16;	{ Real Intrinsic Id }
	INEXT=17;	{ Integer External Assembly Procedure Id }
	RLEXT=18;	{ Real External Assembly Procedure Id }
	INSEG=19;	{ Integer Segment Variable Id }
	RLSEG=20;	{ Real Segment Variable Id }
	ADSEG=21;	{ Address Variable Id }
	SHSEG=22;	{ Short Segment Variable Id }
{$IFDEF STRUCTS}
	USRTYPE=23;	{User defined variable type (struct)}
	USRPROC=24;	{function that returns a struct}
	USROPROC=25;	{Optimised function that returns a struct}
{$ENDIF}


function DoXPLCompile(Source: string; var I2LCode: string; MsgHan: TMessageHandler=nil; CompEvnt: TCompileEvent=nil): integer;
function DoXPLCompileStream(Source, I2LCode: TStream; MsgHan: TMessageHandler=nil; CompEvnt: TCompileEvent=nil): integer;
function DoXPLCompileFile(SourceFile,I2LFile: string; MsgHan: TMessageHandler=nil; CompEvnt: TCompileEvent=nil): integer;

procedure GetSymbolTableDump(var ST: TSymbolList);
procedure ClearSymbolDump;
function PCToSourceData(TargetPC, Offset: integer): TPCInfo;
function LineNumberToPC(Filename: string; LineNumber, Offset: integer): integer;
procedure GetSymbolByName(PC,Offset: integer; Name: string; var PI: PProcItem; var SI: PSymbolItem);
procedure GetProcedureByName(PC,Offset: integer; Name: string; var PI: PProcItem);

procedure AddressToProcName(Add: integer; var Filename,ProcName: string); overload;
function AddressToProcName(Add: integer): string; overload;

implementation

uses XPLDataTypes, XPLIntrinValidate;
                                   
const 	SYMAX=1600;	{ Size Of The Symbol Table }
	EMTPTR=-1;	{ Empty Symbol Pointer }

	RLMAX=1600;	{ Size Of Real-constant Symbol Table }

	RLSIZE=8;	{ No. Of Bytes In A Real Number }

{$IFDEF STRUCTS}
	SIGCHAR=64;	{ No. Of Significant Chars In An Ident/Struct }
{$ELSE}
	SIGCHAR=16;	{ No. Of Significant Chars In An Ident }
{$ENDIF}
	QUITMAX=160;	{ Maximum No. Of 'quit's In A 'loop' }

var	ERRCNT,		{ Error Counter }
	ATYPE: Integer;	{ Present Atom Type Descriptor }
var	ATOM: word;	{ Present Atom Descriptor }
			{ Contains Reserved Word Hash Or The Ascii For A Special Character;}
			{ 0 If The Atom Is A Constant Or An Identifier }

var	InChar: char;	{ Current Character.  Most Of The Time It }
			{  Contains The Terminator Of The Current Atom }

var	CONDITIONAL,		{ Flag: Conditional Compile }
	STRTERM,		{ FLAG: String Termination (False=NULL, True=MSB)}
	DEBFLG: boolean;	{ Flags Debug Mode Is On }


const 	SPECIAL=0;	{ Atom type constants}
	IDENTIFIER=1;
	INTCON=2;
	REALCON=3;

var	IDENT: string[SIGCHAR];	{ Array--current Identifier Name }
var	IATOM: Integer;		{ Value Of Current Integer Constant }
var	CRemainder: integer;	{ Remainder from last Constant Expression divide}
var	RLATOM: double;		{ Real Constant From Proc "ratom" }

var	IDTYPE: Integer;	{ Present Identifier Type Descriptor }

{$IFDEF STRUCTS}
var 	StructIdent: string;		{ Struct identifier with spaces trimmed away}
var	StructBase: string;		{ Root part of struct variable}
var	StructComp: string;		{ Struct components }
var	StructIndex: integer;		{ Contains current struct index}
var	StructSize: integer;		{ Size of current struct}
var	StructOffset: integer;		{ Offset from beginning of struct to component}
var	StructType: integer;		{ Type of current struct component - can ordinary int or real type, or another struct}
var	StructCompIndex: integer;       { If current component is itself a struct, holds index to struct}
var	FuncStructInx: integer;		{ Return type of struct function}
{$ENDIF}

var	LEV,			{ Static Level Of Current Identifier }
	VAL,			{ Value Or Address Of Current Identifier }
	SYMNUM,			{ Position In "symtbl" Of Current Identifer }
	FACTYP: Integer;	{ Factor (or Operand) Type (real Or Integer) }

const 	RealType = 0;		{ Factor types }
	IntegerType = 1;

var	PC,			{ The I2l Program Counter }
	OLDPC,			{ Pc Variable Used By Procedure "GEN" }
	LEVEL,			{ Static Level Of Current Procedure }
	FIXCNT,			{ Count Of The Number Of Outstanding 'quit's }
	STKLOD,			{ No. Of Integers Left On Stack By 'for' ! 'case' }
	NORLSY: integer;	{ Current Number Of Real Constants In Table }

var	OPTPROC: boolean;	{ Boolean--generate An Optimized Procedure Call }

var	FIXES: array [0..QUITMAX-1] of integer;	{ Array--'quit' Fixes Still Outstanding }

const HEXDIGIT: pchar ='0123456789ABCDEF'; 	{ Array Of Hex Digits (0 - F) }


{  -- Symbol Table Arrays -- }


type TSymbolInfo = record
 DeclarePos: TSourceInfo;
 LastUsePos: TSourceInfo;
 end;

var SymbolTable: THashTable;			{Holds symbol names and hashes them}

var SymbolInfo: array [0..SYMAX-1] of TSymbolInfo;	{ Additional information about the Symbol}

var	SYMTYP,					{ Type Descriptors (idtype) }
	SYMLEV,					{ Level (lev) }
	SYMTAG: array [0..SYMAX-1] of byte;	{ Tag For Warnings }

{ Tag values:   0 = unused, 1 = Used 2 = Address, 4 = For Symbol  }
{		4..7 = number of declared array dimensions
{		8 = Looked Up }

var	SYMPUB: array [0..SYMAX-1] of boolean;	{ Flags public procedures}

var	ArrayDims: array [0..SYMAX] of TDynWordArray;

var	SYMVAL: array [0..SYMAX] of Integer;	{ Value or Address (VAL) }

var	RLTBL: array [0..RLMAX-1] of double;	{ Real Constant Table }

{$IFDEF STRUCTS}
var	StructIndices: array [0..SYMAX] of Integer;	{ If SymTyp = USRTYPE, this the index into the struct table}
var	IsTypeDef: array [0..SYMAX] of boolean;		{ Flags if the item is a Type Defintion or a Struct Variable}
{$ENDIF}

{ Reserved Word Hashes }

const
	ADRSYM=$88E4;	BEGSYM=$84C7;	CASEYM=$8053;	CODSYM=$8184;
	DEFSYM=$9CC6;	DOSYM=$0CEF;	ELSEYM=$99F3;	ENDSYM=$99A4;
	EXITYM=$9B69;	EXTNYM=$9B74;	FALSYM=$944C;	FFUNYM=$94B5;
	FORSYM=$9592;	FPRSYM=$9672;	FUNSYM=$96CE;	GESYM=$0C85;
	GETSYM=$90D4;	IFSYM=$0D46;	INTSYM=$A9B4;	LESYM=$0DE5;
	LOOPYM=$BD8F;	NOTSYM=$B594;	OFSYM=$0D86;	PROCYM=$CE2F;
	QUITYM=$CAC9;	REALYM=$C4C1;	REPSYM=$C4D0;	RETSYM=$C4D4;
	THENYM=$DD65;	TRUSYM=$DE35;	UNTSYM=$D9B4;	VOIDSYM=$D589;
	WHILYM=$D169;

	CHARYM=$8161;	EPRSYM=$9A72;	EFUNYM=$98B5;	PUBSYM=$CEC2;
	OTHSYM=$B2E8;	INCSYM=$A9A3;	LSLSYM=$BE0C;	LSRSYM=$BE92;
	CONSYM=$818E;	ABSSYM=$8833;	REMSYM=$C4CD;	SWAPYM=$C281;
	EXTSYM=$9B74;	PORTYM=$CD92;	FIXSYM=$9558;	FLOSYM=$95EF;
	SQRTYM=$C252;	STRSYM=$C2F2;	ASMSYM=$8A0D;	SQSYM=$0E11;
	TOSYM=$0EEF;	DOWNYM=$9D97;	ASRSYM=$8A12;	ANDSYM=$89A4;
	ORSYM=$0D92;	XORSYM=$ED92;	SIZSYM=$C15A;

       	SEGSYM=$C0C7;	SHTSYM=$C16F;

{$IFDEF STRUCTS}
const
	TYPESYM = 57168;
	RECORDSYM = 50371;
{$ENDIF}




{----------------- Utility Routines -------------------------------------------}


procedure FilenameToSource(Filename: string; var Source: TSourceItem);
var I: integer;
begin
for I:=0 to High(SourceArray) do
 if Filename = SourceArray[I].Info.Filename then
	begin
	Source:=SourceArray[I];
	break;
	end;
end;



procedure FilenameToPCArray(Filename: string; var PCArray: TPCArray);
var Source: TSourceItem;
begin
FilenameToSource(Filename,Source);
PCArray:=Source.PCArray;
end;


procedure FindNearestPC(PC, Offset: integer; var PCArray: TPCArray);
var I,L,H,Len: integer;
begin
for I:=0 to High(SourceArray) do
	begin
	Len:=Length(SourceArray[I].PCArray);
	if Len<1 then continue;
	L:=SourceArray[I].PCArray[0].PC+Offset;
	H:=SourceArray[I].PCArray[Len-1].PC+Offset;
	if (PC>=L) and (PC<=H) then begin PCArray:=SourceArray[I].PCArray; break; end;
	end;
end;


function GetMaxPC: integer;
var I,H,Len: integer;
begin
Result:=0;
for I:=0 to High(SourceArray) do
	begin
	Len:=Length(SourceArray[I].PCArray);
	if Len<1 then continue;
	H:=SourceArray[I].PCArray[Len-1].PC;
	if H>Result then Result:=H;
	end;
end;

function PCToSourceData(TargetPC, Offset: integer): TPCInfo;
{Find the nearest source line to specified PC and return associated data}
var PC: integer;
var L,H,I,Max: integer;
var PCArray: TPCArray;
begin
Max:=GetMaxPC;
if TargetPC<Offset then TargetPC:=Offset;
if TargetPC>Max then TargetPC:=Max;
FindNearestPC(TargetPC,Offset,PCArray);
L:=0; H:=High(PCArray);
while L <= H do
	begin
	I:=(L + H) shr 1;
	PC:=PCArray[I].PC+Offset;
	if TargetPC<PC then H:=I-1
	else if TargetPC>PC then L:=I+1
	else break;
	end;
while (I<High(PCArray)) and ((PCArray[I+1].PC+Offset)=TargetPC) do Inc(I);
Result:=PCArray[I];
end;



function LineNumberToPC(Filename: string; LineNumber, Offset: integer): integer;
{Find the PC that corresponds to a particular line of code}
var Line: integer;
var L,H,I: integer;
var PCArray: TPCArray;
begin
Result:=-1;
FileNameToPCArray(Filename,PCArray);
L:=0; H:=High(PCArray);
if H<0 then exit;
while L <= H do
	begin
	I:=(L + H) shr 1;
	Line:=PCArray[I].Info.Line;
	if LineNumber<Line then H:=I-1
	else if LineNumber>Line then L:=I+1
	else break;
	end;
{Move to lowest matching line number}
while (I>0) and (PCArray[I-1].Info.Line=LineNumber) do Dec(I);
{If we are beyond the end of one procedures line numbers move to the next}
if (I<High(PCArray)) and (LineNumber>PCArray[I].Info.Line) then Inc(I);
Result:=PCArray[I].PC+Offset;
end;



procedure GetSymbolByName(PC,Offset: integer; Name: string; var PI: PProcItem; var SI: PSymbolItem);
{Given the specified PC, find the most "Local" matching variable}
{Only returns variable symbols, not procedure sysmbols}
{Returns nil pointers if no matching variable is in scope.}
var I,J: integer;
var InScope: array of integer;
var BaseName: string;
begin
PI:=nil; SI:=nil;
if Name='' then exit;
PC:=PC - Offset;
SetLength(InScope,0);
{Find all procedures in scope}
for I:=0 to High(SymbolDump) do
 if (PC>=SymbolDump[I].BeginPC) and (PC<=SymbolDump[I].EndPC) then
	begin
	SetLength(InScope,Length(InScope)+1);
	InScope[High(Inscope)]:=I;
	end;
{Since XPL ignores chars past the SIGCHAR size }
{ truncate the Name to maximum indentifier length }
if Length(Name)>SIGCHAR then SetLength(Name,SIGCHAR);
{Trim off component part of struct}
BaseName:=Trim(GetRootName(Name));
{Find the most local matching symbol}
{The list is order from most local to most global}
{So the first matching symbol will be most local }
for I:=0 to High(InScope) do
 for J:=0 to High(SymbolDump[InScope[I]].Symbols) do
	begin
	if not SameText(SymbolDump[InScope[I]].Symbols[J].Name,BaseName) then continue;
	PI:=@SymbolDump[InScope[I]];
	SI:=@SymbolDump[InScope[I]].Symbols[J];
	{$IFDEF STRUCTS}
	SI.FullName:=Name;
	{$ENDIF}
	exit;
	end;
end;



function FirstProcInScope(PC,Offset: integer): integer;
var I: integer;
begin
PC:=PC - Offset;
Result:=-1;
for I:=High(SymbolDump) downto 0 do
  if PC<=SymbolDump[I].EndPC then Result:=I
  else exit;
end;



procedure GetProcedureByName(PC,Offset: integer; Name: string; var PI: PProcItem);
{Given the specified PC, find the most "Local" matching procedure}
{Returns nil pointer if no matching procedure is in scope.}
var I: integer;
var Inx: integer;
var InScope: array of integer;
begin
PI:=nil;
if (PC<0) or (Name='') then exit;
PC:=PC - Offset;
SetLength(InScope,0);

Inx:= FirstProcInScope(PC,Offset);
if Inx<0 then exit;
Level:=SymbolDump[Inx].Level;

{Create a list of procedures in scope}
if SymbolDump[Inx].ProcName='Main' then
	begin
	{Special case the "main" procedure because its scoping rules are}
	{ different. Everything at level zero is in scope and nothing else}
	for I:=Inx downto 0 do
	 if SymbolDump[I].Level=0 then
	 	begin
		{Use every proc at level zero}
	 	Level:=SymbolDump[I].Level;
	 	SetLength(InScope,Length(InScope)+1);
	 	InScope[High(Inscope)]:=I;
	 	end;
	 end
else
	begin
	{Move to procedure with the most global scope i.e. at Level 0}
	while (Inx<High(SymbolDump)) and (SymbolDump[Inx].Level>0) do Inc(Inx);
	{Look for nearest procedure in scope}
	for I:=Inx downto 0 do
		begin
		{look symbols 2 levels lower and above}
		if (SymbolDump[I].Level=0) or (SymbolDump[I].Level<=(Level+2)) then
			begin
			Level:=SymbolDump[I].Level;
			SetLength(InScope,Length(InScope)+1);
			InScope[High(Inscope)]:=I;
			end;
		{Only move the level up in scope}
		if SymbolDump[I].Level<Level then Level:=SymbolDump[I].Level;
		end;
	end;


{Since XPL ignores chars past the SIGCHAR size }
{ truncate the Name to maximum indentifier length }
if Length(Name)>SIGCHAR then SetLength(Name,SIGCHAR);
{Find the most local matching symbol}
{The list is order from most local to most global}
{So the first matching symbol will be most local }
for I:=0 to High(InScope) do
	begin
	if not SameText(SymbolDump[InScope[I]].ProcName,Name) then continue;
	PI:=@SymbolDump[InScope[I]];
	exit;
	end;
end;


procedure AddressToProcName(Add: integer; var Filename,ProcName: string);
{Get Procedure name from Program Counter address}
var I,W,Width,H: integer;
begin
H:=High(SymbolDump);
if H>High(SymbolDump) then exit;
Width:=SymbolDump[H].EndPC-SymbolDump[H].BeginPC;
Filename:=SymbolDump[H].SourceInfo.Filename;
ProcName:=SymbolDump[H].ProcName;
for I:=H-1 downto 0 do
	begin
	if Add<=SymbolDump[I].BeginPC then continue;
	if Add>SymbolDump[I].EndPC then continue;
	{The procedure with narrowest width will be the most deeply nested}
	W:=SymbolDump[I].EndPC-SymbolDump[I].BeginPC;
	if W>=Width then continue;
	Width:=W;
	Filename:=SymbolDump[I].SourceInfo.Filename;
	ProcName:=SymbolDump[I].ProcName;
	end;
end;


function AddressToProcName(Add: integer): string;
{Get Procedure name from Program Counter address}
var I,W,Width,H: integer;
begin
Result:='';
H:=High(SymbolDump);
Width:=SymbolDump[H].EndPC-SymbolDump[H].BeginPC;
Result:=SymbolDump[H].ProcName;
for I:=H-1 downto 0 do
	begin
	if Add<=SymbolDump[I].BeginPC then continue;
	if Add>SymbolDump[I].EndPC then continue;
	{The procedure with narrowest width will be the most deeply nested}
	W:=SymbolDump[I].EndPC-SymbolDump[I].BeginPC;
	if W>=Width then continue;
	Width:=W;
	Result:=SymbolDump[I].ProcName;
	end;
end;



{------------------------------------------------------------------------------}



procedure ClearSymbolDump;
{Clear the symbol dump array}
var I: integer;
begin
for I:=0 to High(SymbolDump) do
	begin
	SetLength(SymbolDump[I].Symbols,0);
	end;
SetLength(SymbolDump,0);
end;




procedure AddSymbolSpace(ProcName: string; Level,BeginPC,EndPC, Inx1,Inx2: integer);
{Add a symbol to the Symbol Dump}
{Because of recursion symbols are added most local first and most global last}
var I,SI,Inx,PI: integer;
{$IFDEF STRUCTS}
const VariableIDs = [ADDRVAR,INVAR,RLVAR,INCON,RLCON,ADSEG,INSEG,RLSEG,SHSEG,USRTYPE];
{$ELSE}
const VariableIDs = [ADDRVAR,INVAR,RLVAR,INCON,RLCON,ADSEG,INSEG,RLSEG,SHSEG];
{$ENDIF}
begin
SetLength(SymbolDump,Length(SymbolDump)+1);
SI:=High(SymbolDump);
SymbolDump[SI].ProcName:=Trim(ProcName);
SymbolDump[SI].BeginPC:=BeginPC;
SymbolDump[SI].EndPC:=EndPC;
SymbolDump[SI].Level:=Level;

PI:=SymbolTable.Search(ProcName);
if PI>=0 then SymbolDump[SI].SourceInfo:=SymbolInfo[PI].DeclarePos
else SymbolDump[SI].SourceInfo:=CurrentSource.Info;
Inx:=0;
for I:=Inx1 to Inx2-1 do
 with SymbolDump[SI] do
	begin
	if not (SYMTYP[I] in VariableIDs) then continue;
	SetLength(Symbols,Length(Symbols)+1);
	Symbols[Inx].Name:=Trim(SymbolTable[I]);
	Symbols[Inx].SymType:=SYMTYP[I];
	Symbols[Inx].Offset:=SYMVAL[I];
	Symbols[Inx].Level:=SYMLEV[I];
	Symbols[Inx].Dims:=@ArrayDims[I];
	{$IFDEF STRUCTS}
	Symbols[Inx].StructIndex:=StructIndices[I];
	Symbols[Inx].IsTypeDef:=IsTypeDef[I];
	{$ENDIF}
	Symbols[Inx].Used:=SYMTAG[I]<>0;
	Symbols[Inx].SourceInfo:=SymbolInfo[I].DeclarePos;
	if Symbols[Inx].SymType=RLCON then Symbols[Inx].RealConst:=RLTBL[SYMVAL[I]]
        else Symbols[Inx].RealConst:=0;
	Inc(Inx);
	end;
end;




procedure GetSymbolTableDump(var ST: TSymbolList);
{Dump symbol table for debugging purposes}
var I: integer;
begin
SetLength(ST,SymbolTable.Count);
for I:=0 to SymbolTable.Count-1 do
	begin
	ST[I].Name:=SymbolTable[I];
	ST[I].SymType:=SYMTYP[I];
	ST[I].Offset:=SYMVAL[I];
	ST[I].Level:=SYMLEV[I];
	{$IFDEF STRUCTS}
	ST[I].StructIndex:=StructIndices[I];
	ST[I].IsTypeDef:=IsTypeDef[I];
	{$ENDIF}
	ST[I].Used:=SYMTAG[I]<>0;
	end;
end;

{----------------- Input Routines -------------------------------------------}

var Depth:integer;	{Debug}


procedure ClearSource;
{Clear the array file to be compiled}
var I: integer;
begin
FileStack.Clear;
for I:=0 to High(SourceArray) do SetLength(SourceArray[I].PCArray,0);
SetLength(SourceArray,0);
end;



procedure AddSource(Text,Filename: string; SourceType: TSourceType);
{Add a source file to the list files that will be compiled}
var Inx: integer;
begin
SetLength(SourceArray,Length(SourceArray)+1);
Inx:=High(SourceArray);
SourceArray[Inx].Text:=Text;
SourceArray[Inx].TextInx:=1;
SourceArray[Inx].Info.Filename:=Filename;
SourceArray[Inx].Info.Line:=0;
SourceArray[Inx].Info.Column:=0;
SourceArray[Inx].Info.SourceType:=SourceType;
SourceArray[Inx].ArrayInx:=Inx;
CurrentSource:=@SourceArray[Inx];
end;



function Fallback: boolean;
{Fall back to the previous file}
{After an include file has been processed}
{Returns false if there on more files to process}
var Inx: integer;
begin
Result:=FileStack.Pop(Inx);
if Result then CurrentSource:=@SourceArray[Inx];
{for explicit 'code' declarations}
if LEVEL<0 then LEVEL:= 0;
end;


var LastLineLength: integer;



function GetSourceChar: char;
{Get one character from the current source file}
begin
{Check if we have an EOF or are past end of file}
if (CurrentSource.TextInx>Length(CurrentSource.Text)) or
   (CurrentSource.Text[CurrentSource.TextInx]=EOF_Char) then
	begin
	{Don't allow EOF through unless we are at the end of all Includes}
	if not FallBack then Result:=EOF_Char
	else Result:=SP_Char;
	exit;
	end;
Result:=CurrentSource.Text[CurrentSource.TextInx];
Inc(CurrentSource.TextInx);
if (Result>=SP_Char) or (Result=Tab_Char) then Inc(CurrentSource.Info.Column);
if Result=CR_Char then
	begin
	LastLineLength:=CurrentSource.Info.Column;
	Inc(CurrentSource.Info.Line);
	CurrentSource.Info.Column:=0;
	end;
end;


{----------------- Output Routines -------------------------------------------}


var IgnoreI2L: boolean;



type PInteger = ^integer;

procedure PutI2LChar(C: char);
begin
if IgnoreI2L then exit;
I2LText:=I2LText + C;
end;


procedure I2LOut(S: string);
begin
if IgnoreI2L then exit;
I2LText:=I2LText + S;
end;


{--------------------------------------------------------------}


procedure PutListChar(C: string);
begin
ListText:=ListText+C;
end;


procedure ListDeviceOut(S: string);
var I: integer;
begin
for I:=1 to Length(S) do PutListChar(S[I]);
end;


{----------------------------------------------------------------}



procedure AdjustLineNumber(var MsgInfo: TMessageInfo);
{Adjust line numbers for messages, because the compiler is
{ often pointing at the next line when it discovers the error }
begin
{if we are on the first char of a line, the error occurred in the previous line}
if (MsgInfo.Info.Column=0) and (MsgInfo.Info.Line>0) then
	begin
	Dec(MsgInfo.Info.Line);
	MsgInfo.Info.Column:=LastLineLength;
	end;
end;


procedure HandleError(N : Integer; More: string);
{ Send Error Message to User}
var Info: TMessageInfo;
var S: string;
var Continue: boolean;
begin
IgnoreI2L:=True;
ERRCNT:=ERRCNT+1;
if Assigned(MessageHandler) then
	begin
	Info.MessageType:=xmError;
	Info.Info:=CurrentSource.Info;
	Info.ErrorNumber:=N;
	Info.Description:=ErrorStrings[N];
	AdjustLineNumber(Info);
	if More<>'' then Info.Description:=Info.Description+' - '+More;
	MessageHandler(Info,Continue);
	if not Continue then abort;
	end
else
	begin
	Beep;
	S:='Error: '+IntToStr(N)+': '+ErrorStrings[N]+CRLF_Char;
	S:=S+'Line: '+IntToStr(CurrentSource.Info.Line)+' Col: '+IntToStr(CurrentSource.Info.Column)+CRLF_Char;
	S:=S+'Attempt to Continue?';
	if not MessageDlg(S,mtConfirmation, [mbYes, mbNo], 0) = mrYes then abort;
	end;
end;	{ Error }



procedure DisplayWarning(N : Integer; More: string);
{ Send Warning Message To User}
var Info: TMessageInfo;
var Continue: boolean;
begin
if Assigned(MessageHandler) then
	begin
	Info.MessageType:=xmWarning;
	Info.Info:=CurrentSource.Info;
	Info.ErrorNumber:=N;
	Info.Description:=WarningStrings[N];
	if More<>'' then Info.Description:=Info.Description+' - '+More;
	MessageHandler(Info,Continue);
	if not Continue then abort;
	end
end;	{ Warning }




procedure DisplayMessage(S: string);
{ Send Infomrational Message To User }
var Info: TMessageInfo;
var Continue: boolean;
begin
if Assigned(MessageHandler) then
	begin
	Info.MessageType:=xmInfo;
	Info.Info:=CurrentSource.Info;
	Info.ErrorNumber:=0;
	Info.Description:=S;
	MessageHandler(Info,Continue);
	end
end;


procedure CHKUSE(LIM: Integer);
{ Display any symbols that have }
{ been defined but are unused. }
var I: Integer;
var S: string;
var Info: TMessageInfo;
var Con: boolean;
begin
if not Assigned(MessageHandler) then exit;
{Ignore unused variable when we get an error}
{The error may cause us to miss where it is used}
if ERRCNT>0 then exit;
for I:=SymbolTable.Strings.ItemIndex-1 downto LIM do
	begin
	{Ignore intrinsics - they are often unused}
	if (SYMTYP[I]=ININT) or (SYMTYP[I]=RLINT) then continue;
	S:=Trim(SymbolTable.Strings.Value[I]);
	if S='' then continue;
	{$IFDEF STRUCTS}
	{Struct symbols aren't assignable}
	if (SYMTAG[I]=8) and (SYMTYP[I]<>USRTYPE) then
	{$ELSE}
	if SYMTAG[I]=8 then
	{$ENDIF}
		begin
		Info.Info:=SymbolInfo[I].LastUsePos;
		if (SYMTYP[I]=INCON) or (SYMTYP[I]=RLCON) then Info.ErrorNumber:=14
		else Info.ErrorNumber:=10;
		Info.Description:='Value assigned to "'+S+'" never used';
		end
	else if SYMTAG[I]=0 then
		begin
		Info.Info:=SymbolInfo[I].DeclarePos;
		if Info.Info.SourceType=stInclude then Info.ErrorNumber:=13	{Unused in library}
		else if SYMTYP[I] in [INCON,RLCON] then Info.ErrorNumber:=14	{Unused define}
		else if SYMPUB[I] then Info.ErrorNumber:=15			{Unused public}
		else Info.ErrorNumber:=8;					{Unused}
		Info.Description:='Unused Symbol: '+S;
		end
	else Continue;
	Info.MessageType:=xmWarning;
	MessageHandler(Info,Con);
	end;
end;	{ Chkuse }

type TCharSet = set of #0..#255;

procedure GETCH;
{ Get A Character From The Source Device }
{  Filters Out Comments And Does Case Shift. }
{ (this Procedure Is Optimized For Speed.) }
var StopChar: TCharSet;
begin
InChar:=GetSourceChar;
if InChar<>EOF_CHAR then PutListChar(InChar);
{ Filter Out Comments }
while InChar='\' do
	begin
	InChar:=GetSourceChar;
	PutListChar(InChar);
	{if double backslash, eat rest of line}
	if InChar='\' then StopChar:=[LF_Char,EOF_Char]
	else  StopChar:=[LF_Char,EOF_Char,'\'];
	while True do
		begin
		InChar:=GetSourceChar;
		PutListChar(InChar);
		if InChar in StopChar then break;
		end;
	InChar:= ' ';
	end;
end;	{ Getch }



procedure GETCH_;
{Get a non-underline character from the source}
begin
repeat GETCH until InChar<>'_';
end;


procedure StoreIdentChar(C: char; Reset: boolean);
begin
if Reset then IDENT:=C
else IDENT:=IDENT + C;
end;


{-------------- Routines to handle Includes ---------------}


procedure INCLUDE;	{SET UP AN INCLUDE FILE}
var Filename: string;
var IncludeStr: string;
var ErrorStr: string;


	procedure GETC;
	{RETURN A CHARACTER WITH NO FILTERING}
	begin
	InChar:=GetSourceChar;
	PutListChar(InChar)
	end;


	function NormalizeFilename(Filename: string): string;
	{Make sure the filename and path are correct and fully expanded}
	var Ext,Path: string;
	begin
	Result:=Trim(Filename);
	Ext:=ExtractFileExt(Result);
	if Ext='' then Result:=ChangeFileExt(Result,'.xpl');
	Path:=ExtractFilePath(Result);
	if Path='' then Result:=ExpandFileName(Result);
	end;


	function GetFilename: string;
	{READ A FILENAME}
	begin
	Result:='';
	while true do
		begin
		if InChar=';' then break;
		Result:=Result+InChar;
		GETC;
		end;
	Result:=Trim(Result);
	GetC;
	end;


	function FindFile(var Filename: string): boolean;
	{Find include file, searching through all search pathes.}
	var TS: string;
	var I: integer;
	begin
	Result:=True;
	{Does the unmodified file exist?}
	if FileExists(Filename) then Exit;
	{Does the normalized file exits?}
	TS:=NormalizeFilename(Filename);
	if FileExists(TS) then begin Filename:=TS; exit; end;
	{If no path specified, try library search pathes.}
	if ExtractFilePath(Filename)='' then
		begin
		for I:=0 to LibFiles.Count-1 do
		 if TLibInfo(LibFiles.Objects[I]).Enabled and
		    (TLibInfo(LibFiles.Objects[I]).LibType=ltPath) then
			begin
			TS:=LibFiles[I]+'\'+Filename;
			if FileExists(TS) then
				begin
				Filename:=TS;
				exit;
				end;
			end;
		end;
	Result:=False;
	end;


begin
Filename:=GetFilename;
if (Filename='') then ErrorStr:='No file specified'
else if not FindFile(Filename) then ErrorStr:='File: '+ExtractFilename(Filename)+' does not exist'
else if not FileToString(Filename,IncludeStr) then ErrorStr:='Cannot Read File: '+ExtractFilename(Filename)
else ErrorStr:='';
if ErrorStr<>'' then
	begin
	HandleError(70,ErrorStr);
	while (InChar<>';') and (InChar<>EOF_Char) do GETC;
	exit;
	end;
{Trim Ctrl-Z otherwise it will abort compile before original file is complete}
IncludeStr:=TrimRight(IncludeStr);
FileStack.Push(CurrentSource.ArrayInx);
Depth:=FileStack.Depth;
AddSource(IncludeStr,Filename,stInclude);
end;	{INCLUDE}


{----------------------------------------------------------------------}


procedure CONEXPRESS; forward;
procedure GEN(OP,LEV,VAL,CL: Integer); forward;

procedure RATOM;
{ Read An Atom }
{ Outputs:  Atom, Atype, Ident, Hash, Iatom, , RLAtom. }
{  (this Procedure Is Optimized For Speed.) }
var NEG,INTOVF: boolean;
var EXP,I: Integer;
var FRACT,DENOM: double;

	procedure RFRACT;
	{ Read The Fractional Part Of A Real Number }
	begin
	ATYPE:=REALCON;
	ATOM:=0;
	GETCH_;
	FRACT:=0; DENOM:=10;
	while (InChar>='0')  and (InChar<='9') do
		begin
		FRACT:=FRACT + (byte(InChar)-byte('0')) / DENOM;
		DENOM:=DENOM*10;
		GETCH_;
		end;
	RLATOM:=RLATOM +FRACT;
	end;	{ Rfract }


	procedure REXP;
	{ Read An Exponent If Any }
	begin
	if (InChar='E') or (InChar='e') then
		begin
		ATYPE:=REALCON;
		GETCH_;
		if InChar='-' then
			begin
			NEG:=True;
			GETCH_;
			end
		else NEG:=False;
		if InChar='+' then GETCH_;
		EXP:=0;
		if (InChar<'0') or (InChar>'9') then HandleError(31,'');
		while (InChar>='0') and (InChar<='9') do
			begin
			EXP:=EXP *10 + byte(InChar)- byte('0');
			GETCH_;
			end;
		if NEG then EXP:= -EXP;
		while EXP>0 do
			begin
			RLATOM:=RLATOM * 10;
			EXP:=EXP-1;
			end;
		while EXP<0 do
			begin
			RLATOM:=RLATOM / 10;
			EXP:=EXP+1;
			end;
		end;
	end;	{ Rexp }

begin	{ Ratom }
OldInfo:=CurrentSource.Info;
{ Skip Spaces, Tabs, Returns, LF's, & FF's, Etc. Don't go past EOF }
while InChar<=SP_Char do
	begin
	case InChar of
	 EOF_Char:
		begin
		ATYPE:=SPECIAL;
		ATOM:=integer(EOF_Char);
		exit;
		end;
	 LF_Char: if DEBFLG then
			begin
			ListDeviceOut('<$');
			ListDeviceOut(IntToHex(PC,IntNibbles));
			ListDeviceOut('> - ');
			end;
	 end;
	GETCH;
	end;
if (InChar>='a') then
 if InChar<='z' then			{ Reserved Word }
	begin
	ATYPE:=SPECIAL;
	ATOM:=byte(InChar); GETCH;
	if (InChar<'a') or (InChar>'z') then HandleError(26,'Invalid Reserved Word');
	ATOM:= ATOM shl 5 xor byte(InChar); GETCH;
	if (InChar>='a') and (InChar<='z') then
		begin
		ATOM:=ATOM shl 5 xor byte(InChar);
		GETCH;
		end;

	ATOM:= ATOM and $FFFF;
	while (InChar>='a') and (InChar<='z') do GETCH;
	case ATOM of
	  TRUSYM:
		begin
		ATYPE:=INTCON;
		ATOM:=0;
		IATOM:=XPLTrue;
		end;
	  FALSYM:
		begin
		ATYPE:=INTCON;
		ATOM:=0;
		IATOM:=XPLFalse;
		end;
	  CONSYM:
		begin
		RATOM;
		CONEXPRESS;
		if FACTYP=IntegerType then CONDITIONAL:=boolean(IATOM) else HandleError(47,'');
		while ATOM=integer(';') do RATOM;		{ Eat Semi, If Any }
		while True do
			begin				{ Eat Atoms Until Cond=true }
			if CONDITIONAL { #0 } then break;
			if ATYPE=SPECIAL then
			if ATOM=byte(EOF_Char) then break
			   else if ATOM=byte('"') then	{ Ignore 'con' In Strings }
				begin
				while True do
					begin
					InChar:=GetSourceChar; PutListChar(InChar);
					if InChar=EOF_Char then begin HandleError(63,''); abort; end;
					if InChar='^' then
						begin
						InChar:=GetSourceChar;
						PutListChar(InChar);
						end
					else if InChar='"' then break;
					end;
				GETCH;
				end;
			RATOM;
			end;
		end;
	  STRSYM:
	  	begin
	  	RATOM;
	  	CONEXPRESS;
	  	if FACTYP=IntegerType then STRTERM:=IATOM<>0 else HandleError(47,'');
	  	while ATOM=byte(';') do RATOM;		{Eat Semicolon, if any}
	  	end;
	 INCSYM: begin INCLUDE; GETCH; RATOM; end;
	 end;
	exit;
	end;
{ Identifier }
if InChar>='A' then if (InChar<='Z') or (InChar='_') then
	begin
	ATYPE:=IDENTIFIER; ATOM:=0;
	StoreIdentChar(InChar,True);
	GETCH;
	while True do
		begin
		if (InChar>='A') and (InChar<='Z') or
                   (InChar>='a') and (InChar<='z') or
		   (InChar>='0') and (InChar<='9') or
		   (InChar='_')
		   {$IFDEF STRUCTS}
		   or (InChar='.')
		   {$ENDIF}
		   then
			begin
			StoreIdentChar(InChar,false);
			GETCH;
			end
		else break;
		end;
	while Length(IDENT)<SIGCHAR do StoreIdentChar(' ',false);
	exit;
	end;
if InChar>='0' then if InChar<='9' then			{ Unsigned Integer }
	begin
	ATYPE:=INTCON;		{ Assume Integer Until Shown Otherwise }
	ATOM:=0;
	INTOVF:=False;
	IATOM:=byte(InChar)-byte('0');
	GETCH_;
	while True do	begin
		I:=IATOM;
		if (InChar<'0') or (InChar>'9') then break;
		I:=IATOM*10+byte(InChar)-byte('0');
		if (IATOM>MaxInt10) or (IATOM=MaxInt10) and (InChar>'7') then	{ "i" Overflowed }
			begin INTOVF:=True; break end;			{ Ok If It's Real }
		IATOM:=I;
		GETCH_;
		end;
	RLATOM:=IATOM;
	IATOM:=I;	{ Get Possible Legal Value Of 32768 (as -32768) }
	while (InChar>='0') and (InChar<='9') do	{ More Digits Must Be Real }
		begin
		RLATOM:=RLATOM*10 + byte(InChar)-byte('0');
		GETCH_;
		end;
	if InChar='.' then RFRACT;				{ Unsigned Real }
	REXP;
	if (ATYPE=INTCON) and INTOVF and (IATOM<>MinInteger) then
		if CONDITIONAL then HandleError(6,'');
	exit;
	end;
case InChar of
 '.':	begin RLATOM:=0;				{ Unsigned Real }
	RFRACT;
	REXP;
	exit end;
 '$':	begin						{ Unsigned Hex Integer }
	ATYPE:=INTCON; ATOM:=0;
	GETCH_;
	if (InChar>='0') and (InChar<='9') then IATOM:=byte(InChar)-byte('0')
	else if (InChar>='A') and (InChar<='F') then IATOM:=byte(InChar)-$37
	else if (InChar>='a') and (InChar<='f') then IATOM:=byte(InChar)-$57
	else begin { Digit Expected } HandleError(31,''); exit end;
	while True do
		begin
		GETCH_;
		if (InChar>='0') and (InChar<='9') then I:=byte(InChar)-byte('0')
		else if (InChar>='A') and (InChar<='F') then I:=byte(InChar)-$37
		else if (InChar>='a') and (InChar<='f') then I:=byte(InChar)-$57
		else exit;
		if (IATOM and MaxInt16)<>0 then
			if CONDITIONAL then HandleError(6,'');
		IATOM:=IATOM*16+I end;
	end;
 '%':	begin						{ Unsigned binary integer }
	ATYPE:=INTCON; ATOM:=0;
	GETCH_;
	if (InChar>='0') and (InChar<='1') then IATOM:=byte(InChar)-byte('0')
	else
		begin
		{DIGIT EXPECTED}
		HandleError(31,'');
		exit;
		end;
	while true do
		begin GETCH_;
		if (InChar>='0') and (InChar<='1') then I:=byte(InChar)-byte('0') else exit;
                {(if IATOM > $7FFF ... unsigned)}
		if IATOM<0 then
		 if CONDITIONAL then HandleError(6,'');
		IATOM:=IATOM*2+I;
		end ;
	end;

 '^':	begin
	ATYPE:=INTCON;	{ Meta Character = Integer Constant }
	ATOM:=0;
	InChar:=GetSourceChar; PutListChar(InChar);
	if (InChar<SP_Char) or (InChar=#$7F) then DisplayWarning(3,'');
	IATOM:=byte(InChar);
	GETCH;
	exit;
	end;
 '"':	begin
	ATYPE:=SPECIAL;				{ Special Character }
	ATOM:=byte(InChar);			{ (' And Backslash Have No Effect In Strings) }
	InChar:=GetSourceChar; PutListChar(InChar);
	exit;
	end
 end;
ATYPE:=SPECIAL;						{ Special Character }
ATOM:=byte(InChar);
GETCH;
case InChar of
 '=': case char(ATOM) of
	 ':':	begin GETCH; ATOM:=GETSYM end;
	 '>':	begin GETCH; ATOM:=GESYM end;
	 '<':	begin GETCH; ATOM:=LESYM end;
	 end;
 '<': if char(ATOM) = '<' then begin GETCH; ATOM:= LSLSYM; end;
 '>':	begin
	if char(ATOM) = '>' then begin GETCH; ATOM:= LSRSYM end
	else if char(ATOM) = '-' then
		begin
		GETCH;
		if InChar<>'>' then HandleError(38,'');
		GETCH;
		ATOM:= ASRSYM;
		end;
	end
 else;
 end;
end;	{ Ratom }




procedure 	SKIPIT;
{ Skip The Rest Of A Statement For Error Recovery }
begin
while (char(ATOM)<>EOF_Char) and (char(ATOM)<>';') and
      (ATOM<>ENDSYM) and (char(ATOM)<>']') and
      (ATOM<>BEGSYM) and (char(ATOM)<>'[') do RATOM;
end;


procedure HEXB(B: byte);
{ Output Hex Byte }
begin
I2LOut(IntToHex(B,2));
end;	{ Hexb }



procedure OutputHex(Bytes: array of byte; Num: integer);
{Output a hex array of bytes}
{Optimised for speed because we are using a local string}
var I: integer;
var S: string;
begin
S:='';
for I:=0 to Num-1 do
    S:=S+IntToHex(Bytes[I],2);
I2LOut(S);
end;




procedure SYMOUT(SYM: Integer);
{ Output Symbol Name At 'sym' }
begin
I2LOut(SymbolTable.Strings.Value[Sym]);
I2LOut(CRLF_Char);
end;	{ Symout }



procedure SavePCInfo;
begin
SetLength(CurrentSource.PCArray,Length(CurrentSource.PCArray)+1);
CurrentSource.PCArray[High(CurrentSource.PCArray)].Info:=OldInfo;
CurrentSource.PCArray[High(CurrentSource.PCArray)].PC:=PC;
end;


procedure OutputBytes(Value: TI2LWord; Count: integer);
{Output integer as bytes in low byte, high byte order}
var I: integer;
begin
for I:=0 to Count-1 do HexB(PByteArray(@Value)[I]);
Inc(PC,Count);
end;


{ The Encoding Of The Class, CL, Is An 8-bit Binary Code. }

{ 0 -  0000 0000 = Opcode Only, 1 Byte }
{ 2 -  0000 0010 = Opcode With One Byte Value }
{ 3 -  0000 0011 = Opcode With Value, 1 + IntBytes output}
{ 7 -  0000 0111 = Opcode With Address (to Be Relocated By Loader), 1 + IntBytes output}
{ 10 - 0000 1010 = Opcode With Level And One Byte Offset (in Val) }
{ 15 - 0000 1111 = Opcode With Level And Address, 2 + IntBytes output }
{ 16 - 0001 0000 = IMMF Opcode With "RLSIZE" Bytes Of Value }
{ 32 - 0010 0000 = Constant Real Array Pointer }
{ 64 - 0100 0000 = Opcode With Base Relative Address, 1 + IntBytes output}

procedure GEN(OP,LEV,VAL,CL: Integer);
{ Generate Op Code }
var	I : Integer;
var A: TByteArray;
begin
SavePCInfo;
if PC<>OLDPC then
	begin
	I2LOut(CRLF_Char);
	PutI2LChar(';');
	I2LOut(IntToHex(PC,IntNibbles));
	end;

if (OP=1) and (LEV=0) and (VAL>=0) and (VAL<$80) then  { Short Global Load (magic Opcode) }
	begin
	OutputBytes(VAL or $80,1);
	end
 else	begin
	OutputBytes(OP,1);
	if (CL and 8)<>0 then OutputBytes(LEV,1);
	if (CL and 4)<>0 then
		begin
		PutI2LChar('*');
		I2LOut(IntToHex(VAL,IntNibbles));
		PC:=PC+IntBytes;
		end
	else
		begin
		if CL=10 then OutputBytes(VAL,IntBytes)
		else if (CL and 3)=3 then OutputBytes(VAL,IntBytes)
		else if (CL and 2)=2 then OutputBytes(VAL,1);
		end;
	end;
if CL=16 then		{ IMMF--load Immediate Real No. (rlatom) }
	begin
	A:=PByteArray(@RLATOM)^;  { To Access Individual Bytes In Rlatom }
	for I:=0 to RLSIZE-1 do HEXB(A[I]);
	PC:=PC+RLSIZE;
	end;
if (CL and 32)<>0 then	{ Output Dummy Bytes For Constant Real Array Pointers }
 for I:=1 to RLSIZE-IntBytes do
	begin
	HEXB(0);
	PC:=PC+1;
	end;
if (CL and 64)<>0 then
	begin
	PutI2LChar('#');
	SYMOUT(VAL);
	PC:=PC+IntBytes;
	end;
OLDPC:=PC;
end;	{ Gen }



procedure GENIMM(VAL: integer);
{ Generate Imm Op Code }
begin
if (ABS(VAL)<128) and (cardinal(VAL)<>$80000000) then GEN({ IMS } 36,0,VAL,2)
else GEN({ IMM } 11,0,VAL,3);
end;


procedure FIX(I : Integer);
{ Fix Forward References }
begin
I2LOut(CRLF_Char);
if PC<>OLDPC then
	begin PutI2LChar(';'); I2LOut(IntToHex(PC,IntNibbles));
	OLDPC:=PC;
	I2LOut(CRLF_Char) end;
PutI2LChar('^'); I2LOut(IntToHex(I+1,IntNibbles));
end;	{ Fix }


{----------------- Symbol Table Routines --------------------------------------}


procedure LOOKUP;
{ Lookup Identifier In Symbol Table }
{ Inputs: Ident }
{ Outputs: Idtype, Val, Lev, Symnum. }
{ If Two Identifiers Of The Same Name Are In The Symbol Table }
{  Then The Most Recent Entry Is Used. }
var PNTR : Integer;
begin
{$IFDEF STRUCTS}
PNTR:=SymbolTable.Search(GetRootName(IDENT));
{$ELSE}
PNTR:=SymbolTable.Search(IDENT);
{$ENDIF}

if PNTR>=0 then
	begin
	IDTYPE:=SYMTYP[PNTR];
	VAL:=SYMVAL[PNTR];
	LEV:=SYMLEV[PNTR];

	{$IFDEF STRUCTS}
	StructType:=0;
	if (IDTYPE=USRTYPE) or (IDTYPE=USRPROC) then
		begin
		StructIndex:=StructIndices[PNTR];
		StructIdent:=Trim(IDENT);
		StructBase:=Trim(GetRootName(IDENT));
		StructComp:=GetComponentString(StructBase,StructIdent);
		StructOffset:=GetVariableOffset(StructIndex,StructComp);
		StructSize:=GetVariableSize(StructIndex,StructComp);
		GetVariableType(StructIndex,StructComp,StructType,StructCompIndex);
		if StructOffset<0 then IDType:=UNDEF;
		VAL:=VAL + StructOffset;
		end;
	{$ENDIF}

	SymbolInfo[PNTR].LastUsePos:=CurrentSource.Info;
	SYMTAG[PNTR]:=SYMTAG[PNTR] or 8;
	SYMNUM:=PNTR;		{ (for Forward Proc) }
	end
else IDTYPE:=UNDEF;
end;	{ Lookup }



{$IFDEF STRUCTS}
procedure INSERT(STYP,SLEV,SVAL: Integer; StructIndex: integer = 0; TypeDef: boolean = False);
{$ELSE}
procedure INSERT(STYP,SLEV,SVAL: Integer);
{$ENDIF}
{ Insert The Current Identifier Into The Symbol Table }
{ Inputs:  Styp, Slev, Sval, Ident }
var Inx : Integer;
begin
LOOKUP;
if IDTYPE<>UNDEF then
 if LEV=LEVEL then HandleError(11,IDENT)	{ Collision }
 else DisplayWarning(1,IDENT);
Inx:=SymbolTable.Strings.ItemIndex;
SYMTYP[Inx]:=STYP;
SYMLEV[Inx]:=SLEV;
SYMVAL[Inx]:=SVAL;
SYMTAG[Inx]:=0;
SYMPUB[Inx]:=False;
SymbolInfo[Inx].DeclarePos:=CurrentSource.Info;
{$IFDEF STRUCTS}
StructIndices[Inx]:=StructIndex;
IsTypeDef[Inx]:=TypeDef;
StructOffset:=GetVariableOffset(StructIndex,'');
StructSize:=GetVariableSize(StructIndex,'');
{$ENDIF}
SymbolTable.Insert(IDENT);
end;	{ Insert }



//====== Global Struct Routines ==========
{$IFDEF STRUCTS}

procedure BOOLEXP; forward;



procedure GetStructIndices;
{Leaves Base, Size, Index on stack}
var IDT: string;
var Size: integer;
begin
IDT:=IDENT;
Size:=StructSize;
{Addr/Char accessed as an array has a size of one byte}
if StructType=ADDRVAR then Size:=1;
GENIMM(Size);
while True do
	begin
	RATOM;
	BOOLEXP;	{Puts index is on the stack}
	if FACTYP<>IntegerType then HandleError(47,'');
	if char(ATOM)<>',' then break;
	GEN({PINS} $6A,0,0,0);
	GENIMM(Size);
	end;
if char(ATOM)<>')' then HandleError(44,'');
IDENT:=IDT;
end;


function GetStructComponents: boolean;
{Get components of struct i.e. ".L2.P1.X" }
{Leaves Base, Size, Index on stack}
{return true if last component is indexed}
var S: string;
var SType: integer;
begin
Result:=False;
S:=Trim(IDENT)+'.';
GetCH;
RATOM;
{Combine base name and components}
IDENT:=S+Trim(IDENT);
while Length(IDENT)<SIGCHAR do StoreIdentChar(' ',false);
LOOKUP;
if IDTYPE=UNDEF then HandleError(10,IDENT);
if InChar='(' then //***
	begin
	Result:=True;
	SType:=StructType;
	RATOM;
	GENIMM(0);
	GEN({PISA} $69,0,0,0);
	GetStructIndices;	{Leaves Base, Size, Index on stack}
	StructType:=SType;
	end;
end;

procedure PushStructAddress(var BaseType,Endtype: integer; var Indexed: boolean);
{Pushes struct address on the stack}
{ Assumes LEV, VAL, and StructOffset has been set by LOOKUP}
{Returns the type of Base struct and the End component}
{Indexed is true if end component is indexed}
begin
Indexed:=False;
BaseType:=StructType;
EndType:=StructType;
{Put Base address on stack = DiplayVector[Lev] + Value}
GEN({ LOD } 1,LEV,VAL,10);
{Leaves Base, Size, Index on stack}
GetStructIndices;
if InChar='.' then
	begin
	{Leaves Base, Size, Index on stack}
	Indexed:=GetStructComponents;
	EndType:=StructType;
	end;
{Put offset on stack}
GENIMM(StructOffset);
{Returns with Base, Size, Index and Offset on stack}
end;

{$ENDIF}

{============= Routines To Handle Recursive Expression Evaluation =============}


procedure GETCON;
{ Get A Constant--either By Value Or By Name }
var NEG: boolean;
begin
if char(ATOM)='+' then RATOM;
if char(ATOM)='-' then begin NEG:=True; RATOM end  else NEG:=False;
case ATYPE of
 INTCON: begin if NEG then IATOM:=-IATOM; FACTYP:=IntegerType end;
 REALCON: begin if NEG then RLATOM:=-RLATOM; FACTYP:=RealType end;
 IDENTIFIER:
	begin
	LOOKUP;
	if IDTYPE<>UNDEF then SYMTAG[SYMNUM]:=SYMTAG[SYMNUM] or $01;
	case IDTYPE of
	 INCON:	begin
		if NEG then IATOM:=-VAL else IATOM:=VAL;
		FACTYP:=IntegerType;
		end;
	 RLCON:	begin
		if NEG then RLATOM:=-RLTBL[VAL] else RLATOM:=RLTBL[VAL];
		FACTYP:=RealType;
		end
	 else HandleError(42,'');
	 end;
	end
 else HandleError(42,'');
 end;
end;	{ Getcon }



function GetSizeOfValue: integer;
{Return sizeof a variable type, a constant, a procedure or a variable}
begin
RATOM;
if char(ATOM)<>'(' then HandleError(39,'') else RATOM;
case ATOM of
 ADRSYM,CHARYM: Result:= 1;
 INTSYM: Result:= IntBytes;
 REALYM: Result:= RealSize;
 SHTSYM: Result:= SingleSize
 else case AType of
  INTCON: Result:=IntBytes;
  REALCON: Result:=RealSize;
  IDENTIFIER:
  	begin
	LookUp;
	case IDType of
	 UNDEF: HandleError(10,'');
	 ADDRVAR,ADSEG: Result:=1;
	 INVAR,INCON,INPROC,INOPT,INEPRO,ININT,INEXT,INSEG: Result:=IntBytes;
	 RLVAR,RLCON,RLPROC,RLOPT,RLEPRO,RLINT,RLEXT,RLSEG: Result:=RealSize;
	 SHSEG: Result:= SingleSize;
	 {$IFDEF STRUCTS}
	 USRTYPE,USRPROC,USROPROC: Result:=StructSize;
	 {$ENDIF}
	 else HandleError(26,'Has no size');
	 end;
	end
  else HandleError(26,'Has no size');
  end;
 end;
end;




{================ Constant Expression ========================================}



procedure CONEXPRESS;
{ Evaluate Constant Expressions }
{ Outputs Factyp, Iatom, Rlatom }
var	SFACTYP,ITEMP,IFVAR : Integer;
var	RTEMP : double;


	procedure  INTTEST;
	{ Test For Integer Errors }
	begin
	if FACTYP <> IntegerType then HandleError(47,'');
	end;

	procedure  MIXTEST(FTYPE: Integer);
	{ Test For Mixed Mode Errors }
	begin
	if FTYPE<>FACTYP then HandleError(46,'');
	end;

	procedure 	CFACTOR;
	begin
	if char(ATOM)='(' then
		begin
		RATOM;
		CONEXPRESS;
		if char(ATOM)<>')' then HandleError(44,'');
		end
	else if ATOM=ABSSYM then
		begin					{absolute value function}
		RATOM;
		if char(ATOM)<>'(' then HandleError(39,'') else RATOM;
		CONEXPRESS;
		if FACTYP=IntegerType then IATOM:= abs(IATOM) else RLATOM:= abs(RLATOM);
		if char(ATOM)<>')' then HandleError(44,'');
		end
	else if ATOM=REMSYM then
		begin					{Remainder function}
		RATOM;
		if char(ATOM)<>'(' then HandleError(39,'') else RATOM;
		CONEXPRESS;
		if FACTYP<>IntegerType then HandleError(47,'');
		IATOM:= CRemainder;
		if char(ATOM)<>')' then HandleError(44,'');
		end
	else if ATOM=EXTSYM then
		begin					{Remainder function}
		RATOM;
		if char(ATOM)<>'(' then HandleError(39,'') else RATOM;
		CONEXPRESS;
		if FACTYP<>IntegerType then HandleError(47,'');
		IATOM:= ShortInt(Lo(IATOM));
		if char(ATOM)<>')' then HandleError(44,'');
		end
	else if ATOM=SQRTYM then
		begin					{square-root function}
		RATOM;
		if char(ATOM)<>'(' then HandleError(39,'') else RATOM;
		CONEXPRESS;
		if FACTYP=IntegerType then
			begin
			if IATOM<0 then HandleError(80,'')
			else IATOM:= round(sqrt(IATOM));
			end
		else
			begin
			if RLATOM<0 then HandleError(80,'')
			else RLATOM:= sqrt(RLATOM);
			end;
		if char(ATOM)<>')' then HandleError(44,'');
		end
	else if ATOM=SQSYM then
		begin					{square function}
		RATOM;
		if char(ATOM)<>'(' then HandleError(39,'') else RATOM;
		CONEXPRESS;
		if FACTYP=IntegerType then IATOM:= sqr(IATOM) else RLATOM:= sqr(RLATOM);
		if char(ATOM)<>')' then HandleError(44,'');
		end
	else if ATOM=SWAPYM then
		begin					{swap bytes function}
		RATOM;
		if char(ATOM)<>'(' then HandleError(39,'') else RATOM;
		CONEXPRESS;
		if FACTYP<>IntegerType then HandleError(47,'');
		IATOM:= SwapByteArgs(IATOM);
		if char(ATOM)<>')' then HandleError(44,'');
		end
	else if ATOM=FIXSYM then
		begin					{fix function}
		RATOM;
		if char(ATOM)<>'(' then HandleError(39,'') else RATOM;
		CONEXPRESS;
		if FACTYP<>RealType then HandleError(37,'');
		IATOM:= round(RLATOM);
		FACTYP:= IntegerType;
		if char(ATOM)<>')' then HandleError(44,'');
		end
	else if ATOM=FLOSYM then
		begin					{float function}
		RATOM;
		if char(ATOM)<>'(' then HandleError(39,'') else RATOM;
		CONEXPRESS;
		if FACTYP<>IntegerType then HandleError(47,'');
		RLATOM:= IATOM;
		FACTYP:= RealType;
		if char(ATOM)<>')' then HandleError(44,'');
		end
	else if ATOM=SIZSYM then
		begin					{sizeof function}
		IATOM:=GetSizeOfValue;
		FACTYP:= IntegerType;
		RATOM;
		if char(ATOM)<>')' then HandleError(44,'');
		end
	else GETCON;
	RATOM;
	end;	{ Cfactor }


	procedure 	CSHIFTEXP;
	var	ITEMP : Integer;
	begin
	CFACTOR;
	ITEMP:=IATOM;
	case ATOM of
	 LSLSYM:
	 	begin
		INTTEST;
		RATOM;
		CFACTOR;
		INTTEST;
		IATOM:=ITEMP shl IATOM;
		end;
	 LSRSYM:
	 	begin
		INTTEST;
		RATOM;
		CFACTOR;
		INTTEST;
		IATOM:=ITEMP shr IATOM;
		end;
	ASRSYM:	begin
		INTTEST;
		RATOM;
		CFACTOR;
		INTTEST;
		IATOM:=DoASR(ITEMP,IATOM);
		end;
	 end;
	end;	{ CSHIFTEXP }


	procedure 	CTERM;
	var	SFACTYP : Integer;
	var	ITEMP : Integer;
	var	RTEMP : double;
	begin
	CSHIFTEXP;
	SFACTYP:=FACTYP;
	while True do
		begin
		ITEMP:=IATOM;RTEMP:=RLATOM;
		case char(ATOM) of
		  '*':	begin
			RATOM;
			CSHIFTEXP;
			MIXTEST(SFACTYP);
			if FACTYP=IntegerType then IATOM:=ITEMP * IATOM
				 else RLATOM:=RTEMP * RLATOM;
			end;
		  '/':	begin
			RATOM;
			CSHIFTEXP;
			MIXTEST(SFACTYP);
			if (IATOM=0) or (RLATOM=0.0) then HandleError(73,'') { Divide By Zero }
			else if FACTYP=IntegerType then
				begin
				CRemainder:=ITEMP mod IATOM;
				IATOM:=ITEMP div IATOM;
				end
			 else RLATOM:=RTEMP / RLATOM;
			end
		 else break;
		 end;
		end;
	end;	{ Cterm }


	procedure 	CALGEXP;	{ Algebriac Expression }
	var	SFACTYP : Integer;
	var	ITEMP : Integer;
	var	RTEMP : double;
	begin
	CTERM;
	SFACTYP:=FACTYP;
	while True do
		begin
		ITEMP:=IATOM; RTEMP:=RLATOM;
		case char(ATOM) of
		  '+':	begin
			RATOM;
			CTERM;
			MIXTEST(SFACTYP);
			if FACTYP=IntegerType then IATOM:=ITEMP + IATOM
			 else RLATOM:=RTEMP + RLATOM;
			end;
		  '-':	begin
			RATOM;
			CTERM;
			MIXTEST(SFACTYP);
			if FACTYP=IntegerType then IATOM:=ITEMP - IATOM
				 else RLATOM:=RTEMP - RLATOM;
			end
		 else break;
		 end;
		end;
	end;	{ Calgexp }


	procedure 	CLOGEXP;	{ 'not' And Comparisons }
	var	SFACTYP : Integer;
	var	ITEMP : Integer;
	var	RTEMP : double;
	begin
	if (ATOM=NOTSYM) or (char(ATOM)='~') then
		begin
		RATOM;
		CLOGEXP;
		INTTEST;
		IATOM:=  not IATOM;
		end
	 else	begin
		CALGEXP;
		SFACTYP:=FACTYP;
		ITEMP:=IATOM; RTEMP:=RLATOM;
		case char(ATOM) of
		 '=':	begin
			RATOM;
			CALGEXP;
			MIXTEST(SFACTYP);
			if FACTYP=IntegerType then IATOM:= integer(ITEMP = IATOM)
			 else begin IATOM:= integer(RTEMP = RLATOM); DisplayWarning(2,'') end;
			FACTYP:=IntegerType;
			end;
		 '#':	begin
			RATOM;
			CALGEXP;
			MIXTEST(SFACTYP);
			if FACTYP=IntegerType then IATOM:= integer(ITEMP <> IATOM)
			 else begin IATOM:= integer(RTEMP <> RLATOM); DisplayWarning(2,'') end;
			FACTYP:=IntegerType;
			end;
		 '>':	begin
			RATOM;
			CALGEXP;
			MIXTEST(SFACTYP);
			if FACTYP=IntegerType then IATOM:= integer(ITEMP > IATOM)
			 else IATOM:= integer(RTEMP > RLATOM);
			FACTYP:=IntegerType;
			end;
		 '<':	begin
			RATOM;
			CALGEXP;
			MIXTEST(SFACTYP);
			if FACTYP=IntegerType then IATOM:= integer(ITEMP < IATOM)
			 else IATOM:= integer(RTEMP < RLATOM);
			FACTYP:=IntegerType;
			end;
		 char(GESYM):
			begin
			RATOM;
			CALGEXP;
			MIXTEST(SFACTYP);
			if FACTYP=IntegerType then IATOM:= integer(ITEMP >= IATOM)
			 else IATOM:= integer(RTEMP >= RLATOM);
			FACTYP:=IntegerType;
			end;
		 char(LESYM):
			begin
			RATOM;
			CALGEXP;
			MIXTEST(SFACTYP);
			if FACTYP=IntegerType then IATOM:= integer(ITEMP <= IATOM)
			 else IATOM:= integer(RTEMP <= RLATOM);
			FACTYP:=IntegerType
			end;
		 end;
		end;
	end;	{ Clogexp }


	procedure 	CBOOLTERM;	{ Boolean "&" Expressions }
	var	ITEMP : Integer;
	begin
	CLOGEXP;
	while True do	begin
		ITEMP:=IATOM;
		if (char(ATOM)='&') or (ATOM=ANDSYM) then
			begin
			INTTEST;
			RATOM;
			CLOGEXP;
			INTTEST;
			IATOM:=ITEMP  and  IATOM;
			end
		 else break;
		INTTEST;
		end;
	end;	{ Cboolterm }


begin	{ Conexpress }
try
if ATOM=IFSYM then		{ 'if' Expression }
	begin
	RATOM;
	CONEXPRESS;
	INTTEST;
	IFVAR:=IATOM;
	if ATOM<>THENYM then HandleError(22,'');
	RATOM;
	CONEXPRESS;
	SFACTYP:=FACTYP;
	ITEMP:=IATOM; RTEMP:=RLATOM;
	if ATOM<>ELSEYM then HandleError(30,'');
	RATOM;
	CONEXPRESS;
	MIXTEST(SFACTYP);
	if boolean(IFVAR) then
	 if FACTYP=IntegerType then IATOM:=ITEMP else RLATOM:=RTEMP;
	end
 else	begin
	CBOOLTERM;
	while True do	begin
		ITEMP:=IATOM;
		case ATOM of
		  byte('!'),ORSYM:
		  	begin
			INTTEST;
			RATOM;
			CBOOLTERM;
			INTTEST;
			IATOM:=ITEMP  or  IATOM;
			end;
		  byte('|'),XORSYM:
			begin
			INTTEST;
			RATOM;
			CBOOLTERM;
			INTTEST;
			IATOM:=ITEMP  xor  IATOM;
			end
		 else break;
		 end;
		end;
	end;

except on EMathError do HandleError(74,''); end;
end;	{ Conexpress }


{================= Boolean Expression ========================================}

{$IFNDEF STRUCTS}
procedure BOOLEXP; forward;
{$ENDIF}


procedure TestIntrinsic(IDType,IntrinNum: integer; Args: string);	//New
{Test for intrinsic argument errors }
var ErrorMsg: string;
var ArgError: TIntrinsicErrors;
begin
if IDType in [ININT,RLINT] then
	begin
	ArgError:=TestIntrinsicArgs(IntrinNum,Args,ErrorMsg);
	case ArgError of
	 ieNone: ;
	 ieTooMany: HandleError(75,'');
	 ieTooFew: HandleError(76,'');
	 ieMismatch: HandleError(77,ErrorMsg);
	 ieWarning: DisplayWarning(12,ErrorMsg);
	 end;
	end;
end;


type TPointerType = (ptNone,ptAtSymbol,ptAddrSymbol);


function GetPointerType(Atom: integer): TPointerType;
{Get pointer type for current Atom}
begin
case Atom of
 ADRSYM: Result:=ptAddrSymbol;
 byte('@'): Result:=ptAtSymbol;
 else Result:=ptNone;
 end;
end;


procedure PROCAL(IsFunction: boolean);
var	SVAL,SLEV,ARGCNT,SID,CURSYM : Integer;
{New variables for intrinsic type checking}
var Arg: string;
var Args: string;
var ErrorMsg: string;
var PointerType: TPointerType;
const IntegerIDs=[ADDRVAR, INVAR, INCON, INPROC, INFPROC, INOPT, INEPRO, ININT, INEXT, INSEG, RLSEG, ADSEG];
begin
Args:='';
SVAL:=VAL; SLEV:=LEV; SID:=IDTYPE; CURSYM:=SYMNUM;
RATOM;
ARGCNT:=0;
if char(ATOM)='(' then
	begin
	repeat
		begin
		RATOM;
		{Is pointer argument?} 					//New
		PointerType:=GetPointerType(Atom);
		BOOLEXP;
		{Get argument type based pointer type and the underlying variable}
		{Strucks and @-structs can't be sent to intrinsics}
		case PointerType of
		 ptNone: Arg:=PickString(FACTYP=IntegerType,'I',PickString(FACTYP=RealType,'R','S'));
		 ptAtSymbol: Arg:=PickString(FACTYP in IntegerIDs,'i',PickString(FACTYP=USRTYPE,'S','r'));
		 ptAddrSymbol: Arg:=PickString(FACTYP in IntegerIDs,'i','d')
		 end;
		Args:=Args+Arg;
		{$IFDEF STRUCTS}
		if FACTYP=USRTYPE then ARGCNT:=ARGCNT+StructSize
		else
		{$ENDIF}
		if FACTYP=IntegerType then ARGCNT:=ARGCNT + IntBytes else ARGCNT:=ARGCNT + RLSIZE;
		end
	until char(ATOM)<>',';
	{Test here so errors fall on the right line/column}
	TestIntrinsic(SID,SVAL,Args);
	if char(ATOM)<>')' then HandleError(44,'')  else RATOM;
	end
else TestIntrinsic(SID,SVAL,Args);	{Test here for zero argument calls}

if (SID>=INPROC) and (SID<=RLFPROC)
{$IFDEF STRUCTS} or (SID=USRPROC) {$ENDIF} then	{ Normal Procedure Call }
	begin
	if ARGCNT>0 then GEN({ ARG } 10,0,ARGCNT-1,2);
	GEN({ CAL } 5,SLEV+1,SVAL,15);
	end
else case SID of
 ININT,RLINT:
	begin
	case TestReturnErrors(SVAL,IsFunction,ErrorMsg) of
	 reNone: ;
	 reNeedsFunc: HandleError(78,'');
	 reNeedsProc: HandleError(79,'');
	 end;
	GEN($60,0,Integer(STRTERM),2);			{ STRMOD }
	GEN({ CML } 12,0,SVAL,2);			{ Intrinsic Procedure Call }
	end;
 {$IFDEF STRUCTS} USROPROC, {$ENDIF}
 INOPT,RLOPT:				{ Optimized Procedure Call }
	begin
	if ARGCNT>0 then HandleError(51,'');	{ No Args Declared }
	GEN({ JSR } 38,0,SVAL,7);
	end;
 INEXT,RLEXT:
	begin
	GEN({ ECL } 41,0,CURSYM,64);	{ External Assembly Procedure Call }
	end;
 INEPRO,RLEPRO:				{ External I2l Procedure Call }
	begin
	if ARGCNT>0 then GEN({ ARG } 10,0,ARGCNT-1,2);
	GEN({ CAL } 5,SLEV+1,CURSYM,72);
	end;
 end;
end;	{ Procal }




{================== Boolean Expression ======================================}



procedure BOOLEXP;
{ Boolean Expression }
{ Outputs Factor Type (factyp) }
var P1,P2,SFACTYP : Integer;



procedure FACTOR;


function STRCON: integer;
{ String Constant Function }
var SPC,SCHAR : Integer;
var HCS: string;
begin
HCS:='';
SPC:=PC;
I2LOut(CRLF_Char); PutI2LChar(';'); I2LOut(IntToHex(PC,IntNibbles));
while InChar<>'"' do
	begin			{ (getch--optimized For Speed) }
	case InChar of
	  '^':	begin
	  	InChar:=GetSourceChar; PutListChar(InChar);
		if (InChar <SP_Char) or (InChar=#$7F) then
			if (InChar<>CR_Char) and (InChar<>LF_Char) and (InChar<>Tab_Char) then DisplayWarning(3,'');
		if (InChar>='@') and (InChar<='_') and (InChar<>'^') then InChar:=char(byte(InChar)-byte('@'));
		if (InChar>='`') and (InChar<='~') then InChar:=char(byte(InChar)-byte('`'));
		end;
	  EOF_Char: begin HandleError(63,''); Abort; end
	  end;
	{Save constant  hex I2L string to dump later}
	HCS:=HCS+IntToHex(byte(InChar),2);
	PC:=PC+1;
	SCHAR:=byte(InChar);
	InChar:=GetSourceChar;
	PutListChar(InChar);
	if (InChar <SP_Char) or (InChar=#$7F) then
		if (InChar<>CR_Char) and (InChar<>LF_Char) and (InChar<>Tab_Char) then DisplayWarning(3,'');
	end;
I2LOut(HCS);
if STRTERM then
	begin
	I2LOut(CRLF_Char); PutI2LChar(';'); I2LOut(IntToHex(PC-1,IntNibbles));
	HEXB(SCHAR or $80);
	end
else begin HEXB(0); PC:=PC+1; end;
GETCH;
FACTYP:=IntegerType;
Result:= SPC;	{ Return Starting Address Of String }
end;		{ Strcon }


type TArrayValue = record
 case integer of
  0: (Real: double);
  1: (Int: TI2LWord);
  2: (Bytes: array [0..SizeOf(double)-1] of byte);
 end;

type PArrayLink = ^TArrayLink;
 TArrayLink = record
 NextLink: PArrayLink;	{0}
 Entry: TArrayValue;	{1}
 Indirect: boolean;	{2}
 end;

type TArrayConInfo = record
 Entry: TArrayValue;
 Indirect: boolean;
 end;

type TConArray = array of TArrayConInfo;


function ARRAYCON: integer;	{ Constant Arrays }
var AC: TConArray;
var SPC,I,J,SFACTYP : Integer;
var INDIRECT: boolean;
var ENTRY: TArrayValue;


	procedure  ARRAYX;	{ (this Must Be An Optimized Procedure }
	begin		{  For The Reserve To Work Properly.) }
	RATOM;
	INDIRECT:=True;
	case char(ATOM) of
	  '[':	begin ENTRY.Int:=ARRAYCON; RATOM end;
	  '"':	begin ENTRY.Int:=STRCON; RATOM end
	  else	begin
		INDIRECT:=False;
		CONEXPRESS;
		if FACTYP=IntegerType then ENTRY.Int:=IATOM
		else ENTRY.Real:=RLATOM;  { Factyp=real }
		end;
	  end;
	SetLength(AC,Length(AC)+1);
	AC[High(AC)].Entry:=Entry;
	AC[High(AC)].Indirect:=Indirect;
	end;



begin	{ Arraycon }
SetLength(AC,0);
ARRAYX;
while char(ATOM)=',' do
	begin
	SFACTYP:=FACTYP;
	ARRAYX;
	if FACTYP<>SFACTYP then HandleError(46,'');
	end;   { Mixed Mode }
if char(ATOM)<>']' then HandleError(50,'');
SPC:=PC;
I2LOut(CRLF_Char); PutI2LChar(';'); I2LOut(IntToHex(PC,IntNibbles));
for J:=0 to High(AC) do		{ Dump List }
	begin
	ENTRY:=AC[J].Entry;
	if FACTYP=IntegerType then
		begin
		if AC[J].Indirect then { Indirect }
			begin
			PutI2LChar('*');
			I2LOut(IntToHex(ENTRY.Int,IntNibbles));
			end
		else OutputHex(ENTRY.Bytes, IntBytes);
		PC:=PC+IntBytes;
		end
	 else	begin	{ (factyp=real) }
		if AC[J].Indirect then  { Indirect }
			begin
			PutI2LChar('*');
			I2LOut(IntToHex(ENTRY.Int,IntNibbles));
			for I:=1 to RLSIZE-IntBytes do HEXB(0);
			end
		 else OutputHex(ENTRY.Bytes, RLSIZE);
		PC:=PC+RLSIZE;
		end;
	end;
Result:=SPC;	{ Return Starting Address Of Array }
end;	{ Arraycon }






procedure SPECFAC;
{ Special Character Factor }
var SVAL, SPC, SID, SATOM: Integer;
var I: integer;

	{$IFDEF STRUCTS}
	procedure DoAtOperator(Indexed: boolean);
	{@ Operator version - Stores address of variable on the stack}
	{If the final component is a real/struct, the integer address}
	{is packed into the first word of the full real/struct}
	var I: integer;
	begin
	FACTYP:=IntegerType;
	case SID of
	  ADDRVAR,INVAR:
		begin
	 	if Indexed then GEN({ PIIA } $6C,0,0,0)
		else GEN({ PISA } $69,0,0,0);
		end;
	  RLVAR: begin
	  	 FACTYP:=RealType;
	 	 GEN({PISA} $69,0,0,0);
	  	 for I:=1 to IntsPerReal-1 do GEN({IMS} 36,0,0,2);
		 end;
	  USRTYPE:
	  	begin
	 	FACTYP:=USRTYPE;
	 	GEN({PISA} $69,0,0,0);
	 	for I:=1 to (StructSize div IntBytes)-1 do GEN({IMS} 36,0,0,2);
	 	end;
	  end;
	 end;


	procedure DoAddrOperator;
	{"addr" Operator version - Stores address of variable on the stack}
	{Always stores the address as a single integer on the stack}
	begin
	FACTYP:=IntegerType;
	case SID of
	  INVAR: GEN({PISA} $69,0,0,0);
	  RLVAR: begin
	 	 GEN({PISA} $69,0,0,0);
		 GENIMM(StructOffset);
		 GEN({ ADD } $0D,0,0,0);
		 end;
	  ADDRVAR: GEN({PISA} $69,0,0,0);
	  USRTYPE:
	  	begin
	 	GEN({PISA} $69,0,0,0);
	 	end;
	  end;
	end;


	procedure StructsArray;
	var BaseType,EndType: integer;
	var Indexed: boolean;
	begin
	PushStructAddress(BaseType,Endtype,Indexed);
	SID:=EndType;
	RATOM;
	if SATOM=byte('@') then DoAtOperator(Indexed)
	else DoAddrOperator;
	end;


	procedure NonArrayStruct;
	var I: integer;
	begin
	{Non-Indexed pointer factor}
	GEN({ADR} $21,LEV,VAL,10);
	{"addr" with struct is always an integer pointer}
	FACTYP:=IntegerType;
	if (SID=RLVAR) and (SATOM=byte('@')) then
		begin
		FACTYP:=RealType;
		for I:=1 to IntsPerReal-1 do GEN({IMS} 36,0,0,2);
		end;
	if (SID=USRTYPE) and (SATOM=byte('@')) then
		begin
		FACTYP:=USRTYPE;
		for I:=1 to (StructSize div IntBytes)-1 do GEN({IMS} 36,0,0,2);
		end;
	end;



	procedure HandleStructs;
	begin
	SID:= IDTYPE;
	RATOM;
	if char(ATOM)='(' then StructsArray			{ Indexed }
	else NonArrayStruct;
	end;
        {$ENDIF}

begin
case ATOM of
 integer('('):
	begin 				{ Parenthesized Expression }
	RATOM;				{ Parenthesized Expression }
	BOOLEXP;			{ (factor Type Is Unchanged) }
	if char(ATOM)<>')' then HandleError(44,'');
	RATOM
	end;
 integer('"'):
	begin SPC:=PC;		{ String Constant }
	GEN({ JMP } 7,0,0,7);
	SVAL:=STRCON;
	FIX(SPC);
	GEN({ IMM } 11,0,SVAL,7);
	RATOM;
	end;
 integer('['):
	begin
	SPC:=PC;		{ Constant Array }
	GEN({ JMP } 7,0,0,7);
	SVAL:=ARRAYCON;
	FIX(SPC);
	if FACTYP=IntegerType then GEN({ IMM } 11,0,SVAL,7)
	else	{ Factyp=real }
		begin
		PI2LWordArray(@RLATOM)^[0]:=SVAL;	{Pack integer into real}
		GEN({ IMMF } $2C,0,SVAL,36);
		end;{  Pointer In RLATOM }
	RATOM;
	end;
  ADRSYM,byte('@'):
	begin				{ Get Absolute Heap Address }
	SATOM:=ATOM;
	RATOM;
	if ATYPE<>IDENTIFIER then HandleError(45,'');
	LOOKUP;
	if IDTYPE<>UNDEF then
		begin
		SYMTAG[SYMNUM]:=SYMTAG[SYMNUM] or $02;
		if (SYMTAG[SYMNUM] and $06) = $06 then DisplayWarning(9,'');
		end;
	case IDTYPE of
	  {$IFDEF STRUCTS}
	  USRTYPE: HandleStructs;
	  {$ENDIF}
	  INVAR, RLVAR, ADDRVAR:
		begin
		SID:= IDTYPE;
		RATOM;
		if char(ATOM)='(' then			{ Indexed }
			begin
			GEN({ LOD } 1,LEV,VAL,10);	{ (even For Reals) }
			RATOM;
			BOOLEXP;		{ 1st Index }
			if FACTYP<>IntegerType then HandleError(47,'');
			while char(ATOM)=',' do	{ Multiple Indexing }
				begin
				case SID of
				  INVAR, ADDRVAR: GEN({ DBX } $20,0,0,0);
				  RLVAR: GEN({ TRX } $39,0,0,0)
				  end;
				RATOM;
				BOOLEXP;
				if FACTYP<>IntegerType then HandleError(47,'');
				end;
			case SID of
			  INVAR: GEN({ DBA } $1E,0,0,0);
			  RLVAR: begin
				 GEN({TRA} $38,0,0,0);
				 if SATOM=byte('@') then
					begin
					FACTYP:=RealType;
					for I:=1 to IntsPerReal-1 do GEN({IMS} 36,0,0,2);
					end;
				 end;
			  ADDRVAR: GEN({ ADD } $0D,0,0,0);
			  end;
			if char(ATOM)<>')' then HandleError(44,'')  else RATOM;
			end
		 else
		 	begin
			{Non-Indexed pointer factor}
			GEN({ADR} $21,LEV,VAL,10);
			FACTYP:=IntegerType;
			if (SID=RLVAR) and (SATOM=byte('@')) then
				begin
				FACTYP:=RealType;
				for I:=1 to IntsPerReal-1 do GEN({IMS} 36,0,0,2);
				end;
			end;
		end;
	  UNDEF: HandleError(10,IDENT)	{ (undeclared Name) }
	  else HandleError(43,'');	{ (variable Expected) }
	  end;
	end;
  ABSSYM:
	begin					{absolute value function}
	RATOM;
	if char(ATOM)<>'(' then HandleError(39,'') else RATOM;
	BOOLEXP;
	if FACTYP=IntegerType then GEN({ABS} $50,0,0,0) else GEN({FABS} $58,0,0,0);
	if char(ATOM)<>')' then HandleError(44,'') else RATOM;
	end;
  REMSYM:
	begin					{remainder of last divide}
	RATOM;
	if char(ATOM)<>'(' then HandleError(39,'') else RATOM;
	BOOLEXP;
	if FACTYP<>IntegerType then HandleError(47,'');
	GEN({REM)} $51,0,0,0);
	if char(ATOM)<>')' then HandleError(44,'') else RATOM;
	end;
  SWAPYM:
	begin					{swap bytes function}
	RATOM;
	if char(ATOM)<>'(' then HandleError(39,'') else RATOM;
	BOOLEXP;
	if FACTYP<>IntegerType then HandleError(47,'');
	GEN({SWAP}  $52,0,0,0);
	if char(ATOM)<>')' then HandleError(44,'') else RATOM;
	end;
  EXTSYM:
	begin					{sign extend function}
	RATOM;
	if char(ATOM)<>'(' then HandleError(39,'') else RATOM;
	BOOLEXP;
	if FACTYP<>IntegerType then HandleError(47,'');
	GEN({EXT}  $53,0,0,0);
	if char(ATOM)<>')' then HandleError(44,'') else RATOM;
	end;
  FIXSYM:
	begin					{fix function}
	RATOM;
	if char(ATOM)<>'(' then HandleError(39,'') else RATOM;
	BOOLEXP;
	if FACTYP<>RealType then HandleError(37,'');
	GEN({FIX}  $56,0,0,0);
	FACTYP:= IntegerType;
	if char(ATOM)<>')' then HandleError(44,'') else RATOM;
	end;
  FLOSYM:
	begin					{float function}
	RATOM;
	if char(ATOM)<>'(' then HandleError(39,'') else RATOM;
	BOOLEXP;
	if FACTYP<>IntegerType then HandleError(47,'');
	GEN({FLOAT}  $57,0,0,0);
	FACTYP:= RealType;
	if char(ATOM)<>')' then HandleError(44,'') else RATOM;
	end;
  SQRTYM:
	begin					{sqrt function}
	RATOM;
	if char(ATOM)<>'(' then HandleError(39,'') else RATOM;
	BOOLEXP;
	if FACTYP=IntegerType then GEN({SQRT}  $54,0,0,0) else GEN({FSQRT} $59,0,0,0);
	if char(ATOM)<>')' then HandleError(44,'') else RATOM;
	end;
  SQSYM:
	begin					{square function}
	RATOM;
	if char(ATOM)<>'(' then HandleError(39,'') else RATOM;
	BOOLEXP;
	if FACTYP=IntegerType then GEN({SQ} $55,0,0,0) else GEN({FSQ} $5A,0,0,0);
	if char(ATOM)<>')' then HandleError(44,'') else RATOM;
	end;
  SIZSYM:
	begin					{sizeof function}
	IATOM:=GetSizeOfValue;
        GENIMM(IATOM);
	RATOM;
	FACTYP:= IntegerType;
	if char(ATOM)<>')' then HandleError(44,'');
	RATOM;
	end
 else HandleError(26,'');	{ Illegal Factor }
 end;
end;	{ Specfac }



procedure IDFAC;
{ Identifier Factor }
var SID,SINDX: Integer;

	{$IFDEF STRUCTS}
	procedure StructsArray;
	var BaseType,EndType: integer;
	var Indexed: boolean;
	begin
	PushStructAddress(BaseType,Endtype,Indexed);

	case EndType of
	 ADDRVAR: if Indexed then GEN({ LIBS } $64,0,0,0)
	 	  else GEN({ LIIS } $63,0,0,2);
	 INVAR: GEN({ LIIS } $63,0,byte(Indexed),2);
	 RLVAR: GEN({ LIRS } $62,0,byte(Indexed),2);
	 USRTYPE:
		begin
		GENIMM(StructSize);
		GEN({ LSX } $61,0,0,0);
		end;
	 end;
	RATOM;
	SID:=EndType;
	end;

	procedure NonArrayStruct;
	var I: Integer;
	begin
	case StructType of
	 USRTYPE:
		begin
		for I:=0 to (StructSize div IntBytes)-1  do
		 GEN($01,LEV,VAL+(I * IntBytes),10);	{ LOD }
		end;
	 RLVAR:
		begin
		SID:=RLVAR;
		GEN($2A,LEV,VAL,10);	{ FLOD }
		end;
	 else
		begin
		SID:=INVAR;
		GEN($01,LEV,VAL,10);	{ LOD }
		end;
	 end;
	end;

	procedure HandleStructs;
	begin
	RATOM;
	{ Is it Indexed? }
	if char(ATOM)='(' then StructsArray
	else NonArrayStruct;
	end;

	procedure HandleStructFunction;
	var FuncSize: integer;
	begin
	{Arguments to PROCAL may change struct info so save values}
	FuncStructInx:=StructCompIndex;
	FuncSize:=StructSize;
	PROCAL(True);
	{Load address of struct on heap from Global Zero}
	GEN({ LOD } $01,0,0,10);
	GENIMM(FuncSize);
	GENIMM(0);
	GENIMM(0);
	GENIMM(FuncSize);
	GEN({ LSX } $61,0,0,0);
	{So restore Struct proc/func info}
	StructCompIndex:=FuncStructInx;
	StructSize:=FuncSize;
	end;
	{$ENDIF}


begin	{ IDFAC }
LOOKUP;
if IDTYPE<>UNDEF then SYMTAG[SYMNUM]:=SYMTAG[SYMNUM] or $01;
SID:=IDTYPE;

case IDTYPE of
 UNDEF:	HandleError(10,IDENT);
 {$IFDEF STRUCTS}
 USRTYPE: HandleStructs;
 USRPROC,USROPROC: HandleStructFunction;
 {$ENDIF}
 INVAR, RLVAR, ADDRVAR:		{ Variable }
	begin
	if SID=RLVAR then GEN($2A,LEV,VAL,10)	{ LODF }
	else GEN($01,LEV,VAL,10);		{ LOD }
	RATOM;
	if char(ATOM)='(' then				{ It Is Indexed }
		begin
		while True do
			begin
			RATOM;
			BOOLEXP;
			if FACTYP<>IntegerType then HandleError(47,'');
			if char(ATOM)<>',' then break;
			if SID=RLVAR then GEN($3A,0,0,0)	{ Tri }
			else GEN($20,0,0,0);			{ Dbx }
			end;
		if char(ATOM)<>')' then HandleError(44,'')  else RATOM;
		case SID of
		 INVAR:   GEN({ DBX } $20,0,0,0);
		 RLVAR:   GEN({ TRI } $3A,0,0,0);
		 ADDRVAR: GEN({ LDX } $02,0,0,0)
		 end;
		end;
	end;

 INSEG, RLSEG, ADSEG, SHSEG:	{ Segment Variable }
	begin
	SINDX:=0;			{ Watch Index Level }
	GEN({ LOD } 1,LEV,VAL,10);
	RATOM;
	if char(ATOM)='(' then		{ Handle First Index }
		begin
		RATOM;
		BOOLEXP;
		if FACTYP<>IntegerType then HandleError(47,'');
		GEN({ DBX } 32,0,0,0);
		if char(ATOM)=',' then	{ Handle Second Index }
			begin
			SINDX:=2;	{ Flag Two Indexes }
			RATOM;
			BOOLEXP;
			if FACTYP<>IntegerType then HandleError(47,'');
			case SID of
			  INSEG: GEN({ LDSI } $40,0,0,0);
			  ADSEG: GEN({ LDSB } $41,0,0,0);
			  RLSEG: GEN({ LDSR } $42,0,0,0);
			  SHSEG: GEN({ LSHORT } $46,0,0,0)
			  end;
			end;
		if char(ATOM)<>')' then HandleError(44,'')  else RATOM;
		end;
	{ Force Real To Be Integer When There Are Less Than 2 Indexes }
	case SID of
	 RLSEG,SHSEG: begin if SINDX<2 then SID:=INSEG end;
	 end;
	end;

 INCON:	begin	{ Integer Constant Identifier }
	GENIMM(VAL);
	RATOM;
	end;

 RLCON:	begin	{ Real Constant Identifier }
	RLATOM:=RLTBL[VAL]; GEN({ IMMF } $2C,0,0,16);
	RATOM;
	end;

 else	begin	{ Procedures Used As Functions (by Default) }
	PROCAL(True);
	if SID<ININT then
		begin
		if (SID and 1)<>0 then GEN(1,0,0,10)	{ LOD }
		else GEN($2A,0,0,10);		{ LODF }
		end;
	end;
 end;
{$IFDEF STRUCTS}
if (SID=USRTYPE) or (SID=USRPROC) or (SID=USROPROC) then FACTYP:=USRTYPE else
{$ENDIF}
if (SID and 1)<>0 then FACTYP:=IntegerType else FACTYP:=RealType; { Odd Ids Are Integer }
end;	{ Idfac }


begin	{ Factor }
while char(ATOM)='+' do
	begin
	RATOM;
	if char(ATOM)='+' then DisplayWarning(7,'');
	end;	{ Ignore Unary "+" }
if char(ATOM)='-' then				{ Unary "-" }
	begin
	RATOM;
	if char(ATOM)='-' then DisplayWarning(7,'');
	FACTOR;
	if FACTYP=IntegerType then GEN(17,0,0,0)	{ Neg }
	else GEN($31,0,0,0);				{ Negf }
	end
else case ATYPE of
	 SPECIAL: SPECFAC;
	 INTCON:
		begin
		FACTYP:=IntegerType;	{ Integer Constant }
		GENIMM(IATOM);
		RATOM;
		end;
	 REALCON:
		begin
		FACTYP:=RealType;		{ Real Constant }
		GEN({ IMMF } $2C,0,0,16);
		RATOM;
		end;
	  else IDFAC;	{ Atype = Identifier (by Default) }
	 end;
end;	{ Factor }


procedure SHIFTEXP;

	procedure  SHIFTX(INOP: Integer);
	begin
	if FACTYP <> IntegerType then { Integer Expected } HandleError(47,'');
	RATOM; FACTOR;
	if FACTYP <> IntegerType then HandleError(47,'');
	GEN(INOP, 0, 0, 0);
	end;

begin	{ Shiftexp }
FACTOR;
case ATOM of
  LSLSYM: SHIFTX({LSL} $3E);	{ << }
  LSRSYM: SHIFTX({LSR} $3F);	{ >> }
  ASRSYM: SHIFTX({ASR} $3D);	{->> }
  end;
end;	{ Shiftexp }



procedure TERM;
var SFACTYP: Integer;

	procedure TERMX(INOP,RLOP: Integer);
	begin RATOM; SHIFTEXP;
	if SFACTYP<>FACTYP then { Mixed Mode } HandleError(46,'');
	if FACTYP=IntegerType then GEN(INOP,0,0,0)
	else GEN(RLOP,0,0,0);
	end;

begin	{ Term }
SHIFTEXP;
SFACTYP:=FACTYP;
while True do
	case char(ATOM) of
	  '*':	TERMX({ Mul }15,{ Mulf }$2F);
	  '/':	TERMX({ Div }16,{ Divf }$30)
	 else	break;
	 end;
end;	{ Term }


procedure 	ALGEXP;		{ Algebriac Expression }
var	SFACTYP : Integer;

	procedure  ALGX(INOP,RLOP: Integer);
	begin
	RATOM;
	TERM;
	if SFACTYP<>FACTYP then { Mixed Mode } HandleError(46,'');
	if FACTYP=IntegerType then GEN(INOP,0,0,0)
	else GEN(RLOP,0,0,0);
	end;

begin	{ Algexp }
TERM;
SFACTYP:=FACTYP;
while True do
 	case char(ATOM) of
	  '+':	ALGX({ Add }13,{ Addf }$2D);
	  '-':	ALGX({ Sub }14,{ Subf }$2E)
	  else	break;
	 end;
end;	{ Algexp }


procedure 	LOGEXP;		{ Logical Expression }
var	SFACTYP : Integer;

	procedure LOGX(INOP,RLOP: Integer);
	begin
	RATOM;
	ALGEXP;
	if SFACTYP<>FACTYP then { Mixed Mode } HandleError(46,'');
	if FACTYP=IntegerType then GEN(INOP,0,0,0)
	else GEN(RLOP,0,0,0);
	FACTYP:=IntegerType;
	end;

begin	{ Logexp }
if (ATOM=NOTSYM) or (char(ATOM)='~') then		{ Unary 'not' Operator }
	begin
	RATOM;
	if (ATOM=NOTSYM) or (char(ATOM)='~') then DisplayWarning(7,'');
	LOGEXP;
	if FACTYP<>IntegerType then HandleError(47,'');
	GEN({ NOT } 28,0,0,0);
	end
else
	begin
	ALGEXP;
	SFACTYP:=FACTYP;
	case ATOM of
	 integer('='):	begin LOGX({ Eq }18,{ Eqf }$32); if FACTYP=RealType then DisplayWarning(2,'') end;
	 integer('#'):	begin LOGX({ Ne }19,{ Nef }$33); if FACTYP=RealType then DisplayWarning(2,'') end;
	 integer('>'):	LOGX({ Gt }21,{ Gtf }$35);
	 integer('<'):	LOGX({ Lt }23,{ Ltf }$37);
	 GESYM:	LOGX({ Ge }20,{ Gef }$34);
	 LESYM:	LOGX({ Le }22,{ Lef }$36)
	 end;
	end;
end;	{ Logexp }


procedure BOOLTERM;
{ Boolean "&" Expressions }
begin
LOGEXP;
while True do
	begin
	if (char(ATOM)='&') or (ATOM=ANDSYM) then
		begin RATOM;
		LOGEXP; GEN({ AND } 27,0,0,0);
		end
	else break;
	if FACTYP<>IntegerType then HandleError(47,'');
	end;
end;


procedure  BEXPX(INOP: Integer);
begin
if FACTYP <> IntegerType then { Integer Expected } HandleError(47,'');
RATOM; BOOLTERM;
if FACTYP <> IntegerType then HandleError(47,'');
GEN(INOP, 0, 0, 0);
end;	{ Bexpx }



begin	{ Boolexp }
if ATOM=IFSYM then			{ 'if' Expression }
	begin
	RATOM; BOOLEXP;
	P1:=PC; GEN({ JPC } 8,0,0,7);
	if ATOM<>THENYM then HandleError(22,'');
	RATOM; BOOLEXP; SFACTYP:=FACTYP;
	if ATOM<>ELSEYM then HandleError(30,'');
	P2:=PC; GEN({ JMP } 7,0,0,7); FIX(P1);
	RATOM; BOOLEXP;
	if SFACTYP<>FACTYP then { Mixed Mode } HandleError(46,'');
	FIX(P2);
	end
 else	begin			{ Boolean "!" (or) And "|" (xor) Expressions }
	BOOLTERM;
	while True do
	 case ATOM of
	  byte('!'),ORSYM: BEXPX({ Or } $1A);
	  byte('|'),XORSYM: BEXPX({ Eor } $1D)
	  else break;
	 end;
	end;
end;	{ Boolexp }

{==================== Handle Statements =======================================}

procedure SSTATEMENT(SSTK: Integer);	{ (for 'quit's In 'case' Stmnts) }


procedure 	STATEMENT;
var P2,P3,SFIXS,SLEV,SVAL,SFACTYP,I : Integer;
var DownToFlag: boolean;

procedure 	ASSIGN;		{ Assignment Statement }
var	SID : Integer;
{  (also Includes Procedure Calls) }

	procedure 	ASSX;
	begin if ATOM<>GETSYM then HandleError(21,'');
	RATOM;
	BOOLEXP;
	end;	{ Right-hand Side Of Assignment }



	{$IFDEF STRUCTS}
	procedure StructsArray;
	var BaseType,EndType: integer;
	var Indexed: boolean;
	begin
	PushStructAddress(BaseType,Endtype,Indexed);

	{Put the address of the struck on the stack}
	if Indexed then GEN({ PIIA } $6C,0,0,0)
	else GEN({ PISA } $69,0,0,0);
	{Put the right side of the assignment on the stack}
	RATOM;
	ASSX;
	{Now store it in the address}
	case EndType of
 	 RLVAR:
		begin
		if FACTYP<>RealType then HandleError(46,'');
		GEN({ SIRS } $65,0,0,0);
		end;
	 INVAR:
		begin
		if FACTYP<>IntegerType then HandleError(46,'');
		GEN({ SIIS } $66,0,0,0);
		end;
	 ADDRVAR:
		begin
		if FACTYP<>IntegerType then HandleError(46,'');
		{If the last struct item is indexed - use bytes }
		{ Otherwise use integer}
		if Indexed then GEN({ SIBS } $67,0,0,0)
		else GEN({ SIIS } $66,0,0,0);
		end;
	 USRTYPE:
		begin
		if FACTYP<>USRTYPE then  HandleError(77,'');	{Argument type mismatch}
		GENIMM(StructSize);
		{Store data in variable}
		GEN({ SSX } $68,0,0,0);
		end;
	 end;
	end;


	procedure NonArrayStruct;
	var I: integer;
	var LeftIndex,LeftType: integer;
	begin
	LeftIndex:=StructCompIndex;
	LeftType:=StructType;
	ASSX;
	case LeftType of
	 USRTYPE:
		begin
		if FACTYP<>USRTYPE then  HandleError(77,'');	{Argument type mismatch}
		if LeftIndex<>StructCompIndex then HandleError(77,'');	{Argument type mismatch}
		{Store Struct by moving from stack to variable (heap) }
		for I:=(StructSize div IntBytes)-1 downto 0 do
		 GEN({ STO } $03,SLEV,SVAL+(I * IntBytes),10);
		end;
	 RLVAR:
		begin
		if FACTYP<>RealType then HandleError(46,'');
		GEN({ STOF } $2B,SLEV,SVAL,10);
		end;
	 else
		begin
		if FACTYP<>IntegerType then HandleError(46,'');
		GEN({ STO } $03,SLEV,SVAL,10)
		end;
	 end;
	end;


	procedure HandleStructs;
	begin
	SYMTAG[SYMNUM]:=SYMTAG[SYMNUM] or $01;
	if StructType=RLVAR then SFACTYP:=RealType else SFACTYP:=IntegerType;
	RATOM;
	{Is it indexed - i.e. an array entry}
	if char(ATOM)='(' then StructsArray
	else NonArrayStruct;
	end;
	{$ENDIF}


begin	{ Assign }
if ATYPE<>IDENTIFIER then begin HandleError(20,''); SKIPIT; exit end;	{ Bad Start Of A Statement }
LOOKUP;
if IDTYPE=UNDEF then begin HandleError(10,IDENT); SKIPIT; exit end;
SLEV:=LEV; SVAL:=VAL;	{ Save These For "gen." An Interim }
SID:=IDTYPE;		{  Boolexp May Change Lev & Val }

{$IFDEF STRUCTS}
if IDTYPE>=USRTYPE then HandleStructs else
{$ENDIF}
if (IDTYPE>=INPROC) and (IDTYPE<=RLEXT) then
	begin
	SYMTAG[SYMNUM]:=SYMTAG[SYMNUM] or $01;
	PROCAL(False);
	end
else if (IDTYPE=INVAR) or (IDTYPE=RLVAR) or (IDTYPE=ADDRVAR) then
	begin
	if IDTYPE=RLVAR then SFACTYP:=RealType else SFACTYP:=IntegerType;
	RATOM;
	if char(ATOM)='(' then			{ Indexed }
		begin
		SYMTAG[SYMNUM]:=SYMTAG[SYMNUM] or $01;
		GEN({ LOD } 1,SLEV,SVAL,10);
		RATOM;
		BOOLEXP;		{ 1st Index }
		if FACTYP<>IntegerType then HandleError(47,'');

		while char(ATOM)=',' do	{ Multiple Indexing }
			begin
			if SFACTYP=IntegerType then GEN({ DBX } $20,0,0,0)
			else GEN({ TRX } $39,0,0,0);
			RATOM;
			BOOLEXP;
			if FACTYP<>IntegerType then HandleError(47,'');
			end;
		if char(ATOM)<>')' then HandleError(44,'')  else RATOM;
		case SID of
		 INVAR:   GEN({ DBA } $1E,0,0,0);
		 RLVAR:   GEN({ TRA } $38,0,0,0);
		 ADDRVAR: GEN({ ADD } $0D,0,0,0)
		 end;

		ASSX;			{ Tos Now Points To Array Element }
		case SID of
		 INVAR:   GEN({ STD } $1F,0,0,0);
		 RLVAR:   GEN({ STT } $3B,0,0,0);
		 ADDRVAR: GEN({ STX } $04,0,0,0)
		 end;
		end

	 else
		begin
		{Non-indexed}
		ASSX;
		if SFACTYP=IntegerType then GEN({ STO } $03,SLEV,SVAL,10)
		else GEN({ STOF } $2B,SLEV,SVAL,10);
		end;
	if FACTYP<>SFACTYP then { Mixed Mode } HandleError(46,'');
	end
else if (IDTYPE=INSEG) or (IDTYPE=RLSEG) or
	(IDTYPE=ADSEG) or (IDTYPE=SHSEG) then	{ Segment Variables }
	begin
	SFACTYP:=IntegerType;
	RATOM;
	if char(ATOM)<>'(' then	begin ASSX;GEN({ STO } 3,SLEV,SVAL,10) end
	 else	begin			{ 1st Index }
		SYMTAG[SYMNUM]:=SYMTAG[SYMNUM] or $01;
		GEN({ LOD } 1,SLEV,SVAL,10);
		RATOM;
		BOOLEXP;
		if FACTYP<>IntegerType then HandleError(47,'');
		if char(ATOM)=',' then		{ 2nd Index }
			begin
			GEN({ DBX } 32,0,0,0);
			RATOM;BOOLEXP;
			if FACTYP<>IntegerType then HandleError(47,'');
			if char(ATOM)<>')' then HandleError(44,'')  else RATOM;
			ASSX;	{ Get Right Side Of Assignment }
			case SID of
			  INSEG: GEN({ STSI } $43,0,0,0);
			  RLSEG: begin GEN({ STSR } $45,0,0,0); SFACTYP:=RealType; end;
			  ADSEG: GEN({ STSA } $44,0,0,0);
			  SHSEG: begin GEN({ SSHORT } $47,0,0,0); SFACTYP:=RealType; end;
			  end;
			end
		 else	begin
			GEN({ DBA } 30,0,0,0);
			if char(ATOM)<>')' then HandleError(44,'')  else RATOM;
			ASSX;	{ Get Right Side Of Assignment }
			GEN({ STD } 31,0,0,0);
			end;
		end;
	if FACTYP<>SFACTYP then { Mixed Mode } HandleError(46,'');
	end
else { Statement Starting With A Constant }
	begin
	HandleError(27,'');
	SKIPIT;
	end;
end;	{ Assign }


procedure CASER(OpType: Integer);
var SPC1,SPC2,SPC3 : Integer;


procedure CASER2;
begin
RATOM;
BOOLEXP;
if FACTYP<>IntegerType then HandleError(47,'');
SPC1:=PC;
GEN(OpType,0,0,7);
if char(ATOM)=',' then				{ Multiple Labels }
	begin SPC3:=PC;
	repeat	RATOM;	{ For More Than 2 Labels }
		GEN({ JMP } 7,0,SPC3,7);	{ 2 Jumps To Statement }
		FIX(SPC1);
		BOOLEXP;
		if FACTYP<>IntegerType then HandleError(47,'');
		SPC1:=PC;
		GEN(OpType,0,0,7);
	until char(ATOM)<>',';
	FIX(SPC3) end;
if char(ATOM)<>':' then begin HandleError(49,''); SKIPIT; exit end;
RATOM;
STATEMENT;
end;	{ Caser2 }


begin	{ Caser }
CASER2;
SPC2:=PC;
GEN({ JMP } 7,0,0,7);		{ Jump Out Of Case Statement }
FIX(SPC1);
while char(ATOM)=';' do
	begin
	CASER2;
	GEN({ JMP } 7,0,SPC2,7);	{ 2 Jump Exit (back Then Out) }
	FIX(SPC1);
	end;
if (ATOM<>ELSEYM) and (ATOM<>OTHSYM) then HandleError(29,'');
if ATOM=ELSEYM then DisplayWarning(4,'');
RATOM;
STATEMENT;
FIX(SPC2);
end;	{ Caser }


begin	{ Statement }
case ATOM of
 BEGSYM, integer('['):
	begin
	RATOM;
	while True do
		begin
		if ATOM=ELSEYM then begin HandleError(52,''); RATOM end;
		if ATOM=OTHSYM then begin HandleError(53,''); RATOM end;
		STATEMENT;
		case ATOM of
		 integer(';'):	RATOM;
		 ENDSYM: break;
		 integer(']'):	break;
		 integer(EOF_Char): begin HandleError(62,''); abort; end;
		 else HandleError(41,''); 	{ Semi Expected }
		 end;
		end;
	RATOM;		{ Read Past The 'end' }
	end;
 CASEYM:
	begin				{ Case Statement }
	RATOM;
	if ATOM=OFSYM then CASER({ Jpc }8)
	 else	begin
		BOOLEXP;
		if FACTYP<>IntegerType then HandleError(47,'');
		if ATOM<>OFSYM then begin HandleError(48,''); SKIPIT; exit end;
		STKLOD:=STKLOD+1;
		CASER({ Cjp }37);
		GEN({ DRP } 40,0,0,0);
		STKLOD:=STKLOD-1;
		end;
	end;
 QUITYM:
	begin
	for I:=SSTK to STKLOD-1 do GEN({ DRP } 40,0,0,0);
	if FIXCNT>=QUITMAX then begin HandleError(4,''); FIXCNT:=QUITMAX-1 end;
	FIXES[FIXCNT]:=PC;		{ Quit Statement }
	FIXCNT:=FIXCNT+1;
	GEN({ JMP } 7,0,0,7);   { (will Be "fixed" At End Of 'loop') }
	RATOM;
	end;
 IFSYM:
	begin				{ If Statement }
	RATOM;
	BOOLEXP;
	P3:=PC;
	GEN({ JPC } 8,0,0,7);
	if ATOM<>THENYM then begin HandleError(22,''); SKIPIT; exit end;
	RATOM;
	STATEMENT;
	if ATOM=ELSEYM then
		begin P2:=PC;
		GEN({ JMP } 7,0,0,7);
		FIX(P3);
		P3:=P2;
		RATOM;
		STATEMENT end;
	FIX(P3);
	end;
 REPSYM:
	begin
	P2:=PC;			{ Repeat Statement }
	repeat RATOM; STATEMENT until char(ATOM)<>';';
	if ATOM<>UNTSYM then begin HandleError(28,''); SKIPIT; exit end;
	RATOM;
	BOOLEXP;
	GEN({ JPC } 8,0,P2,7);
	end;
 WHILYM:
	begin RATOM;				{ While Statement }
	P2:=PC;
	BOOLEXP;
	P3:=PC;
	GEN({ JPC } 8,0,0,7);
	if ATOM<>DOSYM then begin HandleError(23,''); SKIPIT; exit end;
	RATOM;
	STATEMENT;
	GEN({ JMP } 7,0,P2,7);
	FIX(P3);
	end;
 RETSYM:
	begin				{ Return Statement }
	RATOM;
	for I:=0 to STKLOD-1 do GEN({ DRP } 40,0,0,0);
	if (ATOM<>ELSEYM) and (ATOM<>OTHSYM) and (char(ATOM)<>';') and (char(ATOM)<>']') and
	   (ATOM<>ENDSYM) and (ATOM<>UNTSYM) then	{ Store The Returned }
		begin
		BOOLEXP;		{  Value In Global #0 }
		{$IFDEF STRUCTS}
		if FACTYP=USRTYPE then
			begin
			{Handle returning a struct from a function}
			if FuncStructInx<>StructIndex then HandleError(77,'');	{Argument type mismatch}
			{Put struct-size on stack}
			GENIMM(StructSize);
			{Store TOS @ top of heap, and put address in global-zero}
			GEN({ SSHP } $6B,0,0,0);
			end
		else
		{$ENDIF}
		if FACTYP=IntegerType then GEN({ STO } 3,0,0,10)
		else GEN({ STOF } $2B,0,0,10);
		end;
	if OPTPROC then GEN({ RTS } 39,0,0,0)
	else GEN({ RET } 6,0,0,0);
	end;
 LOOPYM:
	begin				{ Loop Statement }
	SFIXS:=FIXCNT;
	RATOM;
	P2:=PC;
	SSTATEMENT(STKLOD);
	GEN({ JMP } 7,0,P2,7);
	while FIXCNT>SFIXS do	{ Fix The Jumps For The 'quit's }
		begin FIXCNT:=FIXCNT-1; FIX(FIXES[FIXCNT]) end;
	end;
 FORSYM:
	begin				{ For Statement }
	RATOM;
	if ATYPE<>IDENTIFIER then begin HandleError(33,''); SKIPIT; exit end;
	LOOKUP;
	if IDTYPE=UNDEF then HandleError(10,IDENT)
	 else	begin
		if (IDTYPE<>INVAR) and (IDTYPE<>ADDRVAR) then HandleError(33,'');
		SYMTAG[SYMNUM]:=SYMTAG[SYMNUM] or $04;
		if (SYMTAG[SYMNUM] and $06) = $06 then DisplayWarning(9,'');
		end;
	SLEV:=LEV; SVAL:=VAL;
	RATOM;
	if ATOM<>GETSYM then begin HandleError(21,''); SKIPIT; exit end;
	RATOM;
	BOOLEXP;
	if FACTYP<>IntegerType then HandleError(47,'');
	GEN({ STO } 3,SLEV,SVAL,10);

	DownToFlag:= false;
	if (ATOM=TOSYM) or (ATOM=byte(',')) then
	else if ATOM=DOWNYM then DownToFlag:= true
	else begin HandleError(24,''); SKIPIT; exit end;

	RATOM;
	BOOLEXP;
	if FACTYP<>IntegerType then HandleError(47,'');
	if ATOM<>DOSYM then begin HandleError(23,''); SKIPIT; exit end;
	GEN({ LOD } 1,SLEV,SVAL,10);
	P2:=PC;

	if DownToFlag then  GEN({ FORD } $49,0,0,7)
	else GEN({ FOR } 24,0,0,7);

	RATOM;
	STKLOD:=STKLOD+1;
	STATEMENT;
	STKLOD:=STKLOD-1;
	if DownToFlag then GEN({DEC} $4A, SLEV, SVAL, 10)
	else GEN({INC} 25, SLEV, SVAL, 10);
	GEN({JMP} 7,0,P2,7);
	FIX(P2);
	{ Note: The Interpreter Speeds Up Things By Assuming That "inc"  or "dec"}
	{  Will Always Be Followed By A "jmp" Going Back To The "for" }
	end;
 EXITYM:
	begin
	RATOM;				{ Exit Statement }
	if (ATOM<>ELSEYM) and (ATOM<>OTHSYM) and (char(ATOM)<>';') and (char(ATOM)<>']') and
	   (ATOM<>ENDSYM) and (ATOM<>UNTSYM) then	{ Store The Returned }
		begin
		BOOLEXP;			{  Value In Global #0 }
		if FACTYP=IntegerType then GEN({ STO } 3,0,0,10)
		else GEN({ STOF } $2B,0,0,10);
		end;
	GEN({ EXT } 0,0,0,0);
	end;
 ELSEYM: ;				{ Null Statement }
 OTHSYM: ;
 integer(';'): ;
 integer(']'): ;
 ENDSYM: ;
 UNTSYM: ;
 integer(EOF_Char): ;		{ (this Is Mostly An Academic Point) }
 else ASSIGN;
 end;
end;	{ Statement }


begin	{ SStatement }
{ Trick To Adjust Stack (with Drp's) When A 'quit' Is In  A 'case' Statement. }
STATEMENT;
end;


{================= Handle Procedure Declarations =============================}


procedure DoProcedure(SSNOX: Integer);
{ "ssno" From "procdec" For Optimized Procs }
var	SLEVEL,		{ Save Level (complicated By Optimized Procedures) }
	P1,DX,		{ Heap Space Requirement Counter, Unsigned }
	DXOFF,		{ Variable's Offset From Base, Unsigned }
	FPBASE,		{ Pc At End Of Declarations }
	FPROCNT : Integer;	{ Count Of Pending Forward Procedures }
var BaseProcName: string;	{ Local procedure name for each level recursed}
{$IFDEF STRUCTS}
var StartUserType: integer;	{ Struct index at start of procedure}
{$ENDIF}


procedure 	EATARGS;	{ Skip Arguments In Parenthesis }
begin
if char(ATOM)='(' then
	begin
	while True do
		begin if InChar=')' then break;
		if InChar=CR_Char then
			begin
			HandleError(44,'');
			break;
			 end;
		GETCH;
		end;
	GETCH;
	RATOM;
	end;
end;	{ Eatargs }





function GetIntrinsicArgs(var ReturnType: char; var ArgTypes: string): boolean;
//New
{ Get and arguments that might be in Parenthesis }
{ Returns true if there were arguments }
var ReturnFlag: boolean;


	procedure ParseArgument;
	{Parse one argument}
	var Arg: char;
	begin
	RATOM;
	case ATOM of
	 ADRSYM:
		begin
		{addr always indicates integer pointer type}
		RATOM;
		Arg:='i';
	 	end;
	 byte('@'):
		begin
		{@ indicates real or integer pointer according to the underlying type}
		RATOM;
		case ATOM of
		 CHARYM,INTSYM: Arg:= 'i';
		 REALYM,SHTSYM: Arg:= 'r';
	 	 end;
		end;
	 CHARYM,INTSYM: Arg:= 'I';
	 REALYM,SHTSYM: Arg:= 'R';
	 VOIDSYM: Arg:= 'V';
	 else Arg:=#0;
	 end;
	if ReturnFlag then ReturnType:=Arg
	else ArgTypes:=ArgTypes + Arg;
	end;



begin
Result:=False;
ReturnType:='V';
ArgTypes:='';
ReturnFlag:=true;

if char(ATOM)<>'(' then exit;

while True do
	begin
	case InChar of
	 ')': break;
	 ':':
		begin
		ReturnFlag:=False;
		GETCH;
		end;
	 ',': GETCH;
	 CR_Char:
		begin
		HandleError(44,'');
		break;
		end
	 else ParseArgument;
	 end;
	end;
GETCH;
RATOM;

Result:=True;
end;	{ GetIntrinsicArgs }




procedure 	CODDEC;		{ Declare Intrinsic Names }
var	SID : Integer;
var ReturnType: char;
var ArgTypes: string;	 //New
var HasArgs: boolean;   //New
begin
SID:=ININT;		{ Default Is Integer Intrinsic }
RATOM;
if ATOM=REALYM then begin SID:=RLINT; RATOM end
	 else if ATOM=INTSYM then RATOM;
if ATYPE<>IDENTIFIER then HandleError(45,'');
while ATYPE=IDENTIFIER do
	begin
	RATOM;
	HasArgs:=GetIntrinsicArgs(ReturnType,ArgTypes);	//New
	if char(ATOM)<>'=' then HandleError(40,'');
	RATOM;
	GETCON; if FACTYP<>IntegerType then HandleError(47,'');
	if (IATOM<0) or (IATOM>255) then HandleError(7,'');
	INSERT(SID,LEVEL,IATOM);
	SetInstrinsicInfo(IATOM,IDENT,HasArgs,ReturnType,ArgTypes);	//New
	RATOM;
	if char(ATOM)=',' then RATOM;
	end;
if char(ATOM)<>';' then HandleError(41,'')  else RATOM;
end;	{ Coddec }


procedure 	CONDEC;		{ Declare Constant Names }
var CNTR,SSNO: integer;
begin
RATOM;
CNTR:=0;
if ATYPE<>IDENTIFIER then HandleError(45,'');
while ATYPE=IDENTIFIER do
	begin
	RATOM;
	if char(ATOM)<>'=' then begin INSERT(INCON,LEVEL,CNTR); CNTR:=CNTR+1; end
	else	begin
		SSNO:=SymbolTable.Strings.ItemIndex;
		INSERT(INCON,LEVEL,NORLSY);	{ Insert Id Now }
		RATOM;				{  Fix Up Parms Later }
		CONEXPRESS;
		if FACTYP=IntegerType then begin SYMVAL[SSNO]:=IATOM; CNTR:=IATOM+1; end
		else	{ Factyp=real }
			begin SYMTYP[SSNO]:=RLCON;
			if NORLSY>=RLMAX then
				begin HandleError(2,''); NORLSY:=RLMAX-1 end;
			RLTBL[NORLSY]:=RLATOM;
			NORLSY:=NORLSY+1 end;
		end;
	if char(ATOM)=',' then RATOM  else if char(ATOM)<>';' then DisplayWarning(5,'');
	end;
if char(ATOM)<>';' then HandleError(41,'')  else RATOM;
end;	{ Condec }


{$IFDEF STRUCTS}
procedure HandleTypeDef;
{Handle user defined types i.e. structs/records}
var TypeIndex: integer;
var TypeName: string;


	procedure GetVariableList(DataType: integer);
	{Get list of variables for specified type}
	begin
	RATOM;
	if ATYPE<>IDENTIFIER then HandleError(45,'');
	while ATYPE=IDENTIFIER do
		begin
		case AddSubitem(TypeIndex, IDENT, DataType, StructIndex) of
		 ueArrayFull: HandleError(3,'Struct Subitem Table Full');
		 end;
		RATOM;
		if char(ATOM)=',' then RATOM;
		end;
	{if char(ATOM)<>';' then HandleError(41,'')  else} RATOM;
	end;

	function HandleUserType: boolean;
	{Handle user-defined type as the subtype}
	begin
	Lookup;
	Result:=Idtype=USRTYPE;
	if Result then GetVariableList(IdType)
	else HandleError(72,'');
	end;

	procedure HandleComponents(VarType: integer);
	{Retrieve record components declarations}
	begin
	GetVariableList(VarType);
	end;

	procedure HandleRecord;
	{Handle "record" declaration}
	begin
	case AddUserType(IDENT,TypeIndex) of
	 ueArrayFull: HandleError(3,'Struct Table Full');
	 ueReDeclaration: HandleError(11,'Struct redeclared');
	 end;
	RATOM;
	if (ATOM=BEGSYM) or (char(ATOM)='[') then RATOM;
	while true do
		begin
		case ATOM of
		 ADRSYM, CHARYM: HandleComponents(ADDRVAR);
		 REALYM: HandleComponents(RLVAR);
		 INTSYM: HandleComponents(INVAR);
		 ENDSYM, integer(']'):
			begin
			RATOM;
			break;
			end;
		 else if (ATYPE<>IDENTIFIER) or not HandleUserType then break;
		 end;
		end;
	if char(ATOM)<>';' then HandleError(41,'');
	end;

	procedure HandleAlias;
	{Handle Aliases - currently unfinished}
	var R: TUserTypeErrors;
	var VarType: integer;
	begin
	case ATOM of
	 ADRSYM, CHARYM: VarType:=ADDRVAR;
	 REALYM: VarType:=RLVAR;
	 INTSYM: VarType:=INVAR;
	 else
		begin
		Lookup;
		VarType:=Idtype;
		end;
	 end;
	case AddAlias(TypeName,TypeIndex,VarType,StructIndex) of
	 ueArrayFull: HandleError(3,'Struct Table Full');
	 ueReDeclaration: HandleError(11,'Struct redeclared');
	 end;
	RATOM;
	end;


begin
RATOM;
if ATYPE<>IDENTIFIER then HandleError(45,'');
INSERT(USRTYPE,LEVEL,DXOFF, UserTypeIndex, True);
TypeName:=IDENT;

RATOM;
if char(ATOM)<>'=' then HandleError(45,'');
RATOM;
case ATOM of
 RECORDSYM: HandleRecord;
 ADRSYM, REALYM,INTSYM,USRTYPE: HandleAlias;
 else if ATYPE=IDENTIFIER then HandleAlias;
 end;

RATOM;
end;

{$ENDIF}

procedure VARDEC(VarType: Integer; StructInx: integer = -1);
{ Declare Variables--int, Real & Addr }
var	SDX,DIMS,T,ST, SymInx, CurSize: Integer;
var DimArray: PDynWordArray;
begin
RATOM;
if ATYPE<>IDENTIFIER then HandleError(45,'');
OPTPROC:=False;
while ATYPE=IDENTIFIER do
	begin
	SymInx:=SymbolTable.Strings.ItemIndex;
	{$IFDEF Structs}
	INSERT(VarType,LEVEL,DXOFF,StructInx);
	{$ELSE}
	INSERT(VarType,LEVEL,DXOFF);
	{$ENDIF}
	DimArray:=@ArrayDims[SymInx];
	SetLength(DimArray^,0);
	SDX:=DXOFF;
	case VarType of
	 RLVAR: CurSize:=RLSIZE;
         INVAR: CurSize:=IntBytes;
	{$IFDEF Structs}
         USRTYPE:
		begin
		CurSize:=StructSize;
		end;
	{$ENDIF}
         else CurSize:=IntBytes;
         end;
        DXOFF:=DXOFF + CurSize;
        DX:=DX + CurSize;
        T:=CurSize;
	ST:= T;
	if VarType=ADDRVAR then ST:=1;
	RATOM;
	if char(ATOM)='(' then		{ Get Dimensions Of An Array }
		begin
		if DX<>0 then begin GEN({ HPI } 9,0,DX,3); DX:=0 end;
		DIMS:= 0;
		while True do
			begin
			RATOM;
			CONEXPRESS;
			if FACTYP<>IntegerType then HandleError(47,'');
			SetLength(DimArray^,Length(DimArray^)+1);
			DimArray^[High(DimArray^)]:=IAtom;
			{ Push Sizes Of Each Dimension On Stack }
			GENIMM(IATOM);
			DIMS:= DIMS +1;
			if char(ATOM) <> ',' then	{ Last Dimension Of Char Array Is }
				begin			{  Only A Single Byte Per Entry }
				if ST=1 then T:=T div IntBytes;
				T:=T*IATOM;
				DXOFF:=DXOFF+T;
				break;
				end;
			T:=T*IATOM;	{ Accumulate Byte Count For Each Dimension }
			DXOFF:=DXOFF+T;	{ Offset From Base }
			end;
		if char(ATOM)<>')' then HandleError(44,'')  else RATOM;
		{ Gen Code To Set Up Array At Run Time }
		{ Gen Call Makearray([3, 5, 7, 10], 4, 2, Addr Arrayname); }
		GENIMM(DIMS);		{ Number Of Dimensions }
		GENIMM(ST);		{ Number Of Bytes In Each Element }
		GEN({ ADR } 33,LEVEL,SDX,10); { Address Of Pointer To Array ($21) }
		GEN({ ARY } $48,0,0,0);
		end;
	if char(ATOM)=',' then RATOM;
	end;
if char(ATOM)<>';' then HandleError(41,'')  else RATOM;
end;	{ Vardec }



procedure  SEGDEC;	{ Declare Segment Variables }
begin
RATOM;
case ATOM of
  SHTSYM: VARDEC(SHSEG);
  INTSYM: VARDEC(INSEG);
  REALYM: VARDEC(RLSEG);
  ADRSYM,CHARYM: VARDEC(ADSEG)
  else HandleError(72,'');
  end;
end;


procedure EXTDEC(I2L: boolean);
{ Declare External Procedures }
var SID : Integer;
begin
{ Handle I2l Vs. Assembly Type External }
{  The Default Is Integer External }
if I2L then SID:=INEPRO else SID:=INEXT;

{ Handle Real Vs Integer Procedure }
RATOM;
if ATOM=REALYM then
	begin
	if I2L then SID:=RLEPRO else SID:=RLEXT;
	RATOM;
	end
else if ATOM=INTSYM then RATOM;
if ATYPE<>IDENTIFIER then HandleError(45,'');
while ATYPE=IDENTIFIER do
	begin
	DisplayWarning(11,': '+IDENT);
	INSERT(SID,LEVEL,IATOM);
	RATOM;
	EATARGS;
	if char(ATOM)=',' then RATOM;
	end;
if char(ATOM)<>';' then HandleError(41,'')  else RATOM;
end;	{ Extdec }


procedure FPRDEC(OPTYPE: Integer);
{ Declare Forward Referenced Procedures }
var SID : Integer;
begin
SID:=INFPROC;		{ Default Is Integer Forward Procedure }
RATOM;
if ATOM=REALYM then begin SID:=RLFPROC; RATOM end
	 else if ATOM=INTSYM then RATOM;
if ATYPE<>IDENTIFIER then HandleError(45,'');
while ATYPE=IDENTIFIER do
	begin RATOM;
	INSERT(SID,LEVEL,PC);
	GEN({ JMP } 7,0,PC,OPTYPE);
	FPROCNT:=FPROCNT+1;
	EATARGS;
	if char(ATOM)=',' then RATOM end;
if char(ATOM)<>';' then HandleError(41,'')  else RATOM;
end;	{ Fprdec }


procedure PROCDEC(CANOPT,PublicFlag: boolean);
{ Declare Procedure Names }
var SNOSYM,SID,SSNO,SNORL : Integer;
var StartPC: integer;
var ProcName: string;
begin
SID:=INPROC;			{ Typed Procedure (for Functions) }
RATOM;
if ATOM=REALYM then begin SID:=RLPROC; RATOM end
else if ATOM=INTSYM then RATOM;
if ATYPE<>IDENTIFIER then HandleError(45,'') else LOOKUP;
{$IFDEF Structs}
if IDTYPE=USRTYPE then
	begin
	RATOM;
	if ATYPE<>IDENTIFIER then HandleError(45,'') else LOOKUP;
	//*** SID:=USRPROC;
	{Since this is a function, change type a function type}	
	case StructType of
	 ADDRVAR,INVAR: SID:=INPROC;
	 RLVAR: SID:=RLPROC;
	 else SID:=USRPROC;
	 end;
	SSNO:=SymbolTable.Strings.ItemIndex;
	INSERT(SID,LEVEL,PC,StructIndex);
	FuncStructInx:=StructIndex;
	OPTPROC:=CANOPT;
	end
else
{$ENDIF }
if (IDTYPE=INFPROC) or (IDTYPE=RLFPROC) then
	{ Procedure Has Been Previously Declared By A 'fproc' Or 'ffunct' }
	begin
	if LEVEL<>LEV then HandleError(65,'');
	FIX(VAL);	{ ('fproc' & 'proc' Must Be Same Scope) }
	SYMVAL[SYMNUM]:=PC;
	SYMPUB[SYMNUM]:=PublicFlag;
	if IDTYPE=INFPROC then SYMTYP[SYMNUM]:=INPROC else SYMTYP[SYMNUM]:=RLPROC;
	if SID<>SYMTYP[SYMNUM] then { Mixed Mode } HandleError(46,'');
	if VAL>=FPBASE then FPROCNT:=FPROCNT-1;
	OPTPROC:=False;
	end
else
	begin
	SSNO:=SymbolTable.Strings.ItemIndex;
	INSERT(SID,LEVEL,PC);
	SYMPUB[SSNO]:=PublicFlag;
	OPTPROC:=CANOPT;
	end;

{ Handle Public Procedures }
if PublicFlag then
	begin
	PutI2LChar('%');
	SYMOUT(SymbolTable.Strings.ItemIndex-1);
	end;

{ Assume for now that Procedure is not optimized and move down a Level }
SLEVEL:=LEVEL;
LEVEL:=LEVEL+1; if LEVEL>7 then HandleError(5,'');
{ Eat The Argument List As A Comment }
while (InChar<>';') and (InChar<>CR_Char) do GETCH;	{ Special Comment Stops On Cr }
if InChar<>';' then HandleError(41,'');
ProcName:=IDENT;
ProcedureName:=IDENT;
GETCH; RATOM;

{Save symbol table index so we can remove all local entries when procedure is done}
SNOSYM:=SymbolTable.Strings.ItemIndex;
SNORL:=NORLSY;
StartPC:=PC;
DoProcedure(SSNO);	{ Pass Location Of Proc Name In Case It's Optimized }
CHKUSE(SNOSYM);
if char(ATOM)<>';' then HandleError(41,'') else RATOM;
AddSymbolSpace(ProcName,SLevel,StartPC,PC,SNOSYM,SymbolTable.Count);

{Restore symbol table to state before procedure call}
{in other words, delete all local symbols}
SymbolTable.Strings.ItemIndex:=SNOSYM;
SymbolTable.Rehash;
NORLSY:=SNORL;
LEVEL:=SLEVEL;
end;	{ Procdec }


{$IFDEF STRUCTS}
function IsTypeIdentifier: boolean;
{Test if current Atom is a user defined type}
begin
Result:=False;
if ATYPE<>IDENTIFIER then exit;
Lookup;
Result:=IDTYPE=USRTYPE;
end;


function UserVarDec: boolean;
{Handle the declaration of a user type variable (struct)}
begin
Result:=IsTypeIdentifier;
if Result then
	begin
	VARDEC(USRTYPE, StructIndex);
	end;
end;
{$ENDIF}



begin	{ DoProcedure }
{$IFDEF STRUCTS}
StartUserType:=UserTypeIndex;
{$ENDIF}
BaseProcName:=ProcedureName;
if LEVEL<=0 then DXOFF:= RLSIZE else DXOFF:= 0; { Save Heap Space For Return }
DX:= DXOFF;
repeat
	begin
	while True do
		begin
		case ATOM of
		 INTSYM:	VARDEC(INVAR);
		 ADRSYM:	VARDEC(ADDRVAR);
		 CHARYM:	VARDEC(ADDRVAR);
		 REALYM:	VARDEC(RLVAR);
		 CODSYM:	CODDEC;
		 EXTNYM:	EXTDEC(false);
		 DEFSYM:	CONDEC;
		 {$IFDEF STRUCTS}
		 TYPESYM: 	HandleTypeDef;
		 {$ENDIF}
		 SEGSYM:	SEGDEC;
		 EPRSYM, EFUNYM:
			begin
			if LEVEL=0 then EXTDEC(true)
			else
				begin
				HandleError(68,'');
				SKIPIT;
				end;
			end;
		 else
		 {$IFDEF STRUCTS}
		  if not UserVarDec then
		 {$ENDIF}
		     break;
		 end;
		end;
	P1:=PC;
	GEN({ JMP } 7,0,0,7);

	FPROCNT:=0;
	FPBASE:=PC;
	while True do
		case ATOM of
		 PUBSYM:
		 	begin
		 	if LEVEL<>0 then HandleError(68,'');
			RATOM;
			case ATOM of
			  PROCYM,FUNSYM:
				begin
				PROCDEC(false,true);
				end;
			  else HandleError(67,'');
			 end;
			end;
		 PROCYM,FUNSYM:	PROCDEC(true,false);
		 FPRSYM,FFUNYM:	FPRDEC(7);
		 else break;
		end;
	{ If There Is Nothing To Jump Over Then Eliminate The Jump }
	if PC=P1+3 then PC:=P1  else FIX(P1);
	end
until
	(ATOM<>INTSYM) and (ATOM<>ADRSYM) and (ATOM<>CHARYM) and (ATOM<>REALYM) and (ATOM<>CODSYM) and
	(ATOM<>EXTNYM) and (ATOM<>DEFSYM) and (ATOM<>SEGSYM) and (ATOM<>EPRSYM) and (ATOM<>EFUNYM)
	{$IFDEF STRUCTS}
	and (ATOM<>TYPESYM)
	and not IsTypeIdentifier
	{$ENDIF}
	;

{ Reserve Space For Local Variables }
if DXOFF<>0 then		{note that a dimensioned array can make DX # DXOFF}
	begin
	OPTPROC:=False;
	if DX<>0 then GEN({ HPI } 9,0,DX,3);
	end;

if OPTPROC then
	begin
	{$IFDEF STRUCTS}
	if SYMTYP[SSNOX]=USRPROC then SYMTYP[SSNOX]:=USROPROC else
	{$ENDIF}
	if SYMTYP[SSNOX]=INPROC then SYMTYP[SSNOX]:=INOPT
	else SYMTYP[SSNOX]:=RLOPT;
	LEVEL:=LEVEL-1;
	end;	{ Same Level As Nesting Procedure (tricky!) }

if Assigned(CompileEvent) then CompileEvent(etAfterVarDec,BaseProcName);

SSTATEMENT(STKLOD);	{ (STKLOD will always be zero here) }

if OPTPROC then GEN({ RTS } 39,0,0,0)
else GEN({ RET }  6,0,0,0);

if FIXCNT<>0 then { Some 'quit's Not In A 'loop' } HandleError(60,'');
if FPROCNT<>0 then { Unresolved Fwd References } HandleError(66,'');
if Assigned(CompileEvent) then CompileEvent(etProcedureEnd,ProcedureName);
{IFDEF STRUCTS}
{Remove structs that have gone out of scope}
TrimUserTypes(StartUserType);
{ENDIF}
end;	{ Procedure }


procedure DoInternalIncludes;
var I: integer;
begin
if Length(InternalIncludes)<1 then exit;
LEVEL:= -1;
FileStack.Push(CurrentSource.ArrayInx);
Depth:=FileStack.Depth;
for I:=0 to High(InternalIncludes) do
	begin
	AddSource(InternalIncludes[I].FileStr,InternalIncludes[I].Title,stInclude);
	end;
end;


procedure InitializeArrays;
var I: integer;
begin
for I:=0 to High(ArrayDims) do
 SetLength(ArrayDims[I],0);
SymbolTable.Strings.ItemIndex:=0;
SymbolTable.Rehash;
ClearSymbolDump;
end;


function DoXPLCompile(Source: string; var I2LCode: string; MsgHan: TMessageHandler=nil; CompEvnt: TCompileEvent=nil): integer;
{Compile the specified XPL code specified in "Source" string}
{Return I2L code in "I2LCode" string}
{MsgHandler should contain the address of subroutine to handle message/errors}
{MsgHandler can be "nil" if no routines are available to handle messages}
begin
try
{$IFDEF STRUCTS}
ClearUserTypes;
{$ENDIF}

ClearIntrinsicTable;
I2LText:='';
ListText:='';
IgnoreI2L:=False;
MessageHandler:=MsgHan;
CompileEvent:=CompEvnt;
ClearSource;
LEVEL:=0;
AddSource(Source,'Main',stMain);
ProcedureName:='Main';
if Assigned(CompileEvent) then CompileEvent(etCompileStart,ProcedureName);

{ Initialize Some Stuff }
InitializeArrays;

DEBFLG:=False;
PC:=0;
OLDPC:=$FFFFFFFF;
STKLOD:=0;
NORLSY:=0; FIXCNT:=0;
ERRCNT:=0;
CONDITIONAL:= True;
STRTERM:=True;		{ Default to MSB String Termination (True) }

DoInternalIncludes;
GETCH; RATOM;
OPTPROC:=False;		{ (for 2 Reasons) }
DoProcedure(0);		{ Compile Main Procedure, I.e. The Program }
CHKUSE(0);
while char(ATOM)=';' do RATOM;
if char(ATOM)<>EOF_Char then { More Code After End }
	begin
	HandleError(61,'');
	DoProcedure(0);
	end;
PutI2LChar('$');
PutI2LChar(EOF_Char);
PutListChar(EOF_Char);
Result:=ERRCNT;
AddSymbolSpace('Main',0,0,PC,0,SymbolTable.Count);
if Assigned(CompileEvent) then CompileEvent(etCompileEnd,ProcedureName);
finally
 if IgnoreI2L then
	begin
	DisplayMessage('Fatal Error(s) - No Code Generated');
	DisplayMessage('Errors Detected: '+IntToStr(ERRCNT));
	I2LCode:='';
	end
 else
	begin
	DisplayMessage('Sucessful Compile - Program Length:  '+IntToStr(PC+1));
	I2LCode:=I2LText;
	end;
 end;
end;




function DoXPLCompileStream(Source, I2LCode: TStream; MsgHan: TMessageHandler=nil; CompEvnt: TCompileEvent=nil): integer;
var SS,CS: string;
begin
SetLength(SS, Source.Size);
Source.ReadBuffer(Pointer(Source)^, Source.Size);
Result:=DoXPLCompile(SS,CS,MsgHan,CompEvnt);
I2LCode.WriteBuffer(Pointer(CS)^, Length(CS));
end;



function DoXPLCompileFile(SourceFile,I2LFile: string; MsgHan: TMessageHandler=nil; CompEvnt: TCompileEvent=nil): integer;
var SF,CF: TMemoryStream;
begin
SF:=TMemoryStream.Create;
CF:=TMemoryStream.Create;
SF.LoadFromFile(SourceFile);
Result:=DoXPLCompileStream(SF,CF,MsgHan,CompEvnt);
CF.SaveToFile(I2LFile);
end;



initialization
SymbolTable:=THashTable.Create(5,SIGCHAR);
SymbolTable.CaseSensitive:=False;
FileStack:=TIntStack.Create;
LibFiles:=TAutoString.Create(TLibInfo);
LibFiles.Clear;
finalization
LibFiles.Free;
ClearSource;
FileStack.Free;
SymbolTable.Free;
end.
