unit XPLDataTypes;

interface

uses SysUtils;

{These declaration define basic data type used by XPL}

type TI2LWord = integer;	{Main Defintion -This defines the integer type Used by XPL}

{Derrivatives from main integer type }

type PI2LWord = ^TI2LWord;				{Pointer to integer type}
type TI2LWordArray = array [0..8192] of TI2LWord;	{Array of integers}
type PI2LWordArray = ^TI2LWordArray;

const WordMask =  $FFFFFFFF shr (32 - (SizeOf(TI2LWord)*8));	{Mask that covers integer type}

const MaxInteger = High(TI2LWord);	{Maximum integer size (Signed)}
const MinInteger = Low(TI2LWord);	{Minimum integer size (Signed)}
const MaxInt10 = MaxInteger div 10;	{Maximum integer div 10 - used in overflow testing}
const MaxInt16 = not (WordMask div 16);	{ Unsigned Maximum integer div 16 - used in overflow testing}

const XPLTrue = -1;			{True/False value - different from Pascal}
const XPLFalse = 0;

{Defines the size of integers and reals in bytes and nibbles}

const RealSize = SizeOf(Double);		{Defines Real Size in bytes}
const SingleSize = SizeOf(Single);		{Defines Single Size in bytes}
const IntBytes = SizeOf(TI2LWord);		{Defines integer Size in bytes}
const IntNibbles = IntBytes * 2;		{Defines integer Size in nibbles}
const IntsPerReal = RealSize div IntBytes;	{The number integers that can be packed in a real}

type TDataType = (dtChar,dtInteger,dtReal,dtStruct);	{Used to specify the type of a data item}

function DoASR(V,N: integer): integer;
function SwapByteArgs(V: TI2LWord): TI2LWord;

function GetDataTypeStr(DT: TDataType): string;

implementation



{----------- Math Handling Routines -----------------------------}



function DoASR(V,N: integer): integer;
{Preform Arithmetic Shift Right operation}
{Shift V, N-number of times right}
asm
	mov	eax,V
	mov	ecx,N
	sar	eax,cl
end;


function SwapByteArgs(V: TI2LWord): TI2LWord;
{Swap each pair of byte in an integer argument of any size}
var PW: PWordArray;
var W: word;
var I: integer;
begin
Result:=V;
PW:=PWordArray(@Result);
for I:=0 to (SizeOf(TI2LWord) div 2)-1 do
	begin
	PW[I]:=Swap(PW[I]);
	end;
end;

function GetDataTypeStr(DT: TDataType): string;
begin
case DT of
 dtChar: Result:='Char';
 dtInteger: Result:='Integer';
 dtReal: Result:='Real';
 dtStruct: Result:='Struct';
 else Result:='Unknown';
 end;
end;

end.
