unit XPLStructs;

interface

uses Classes,SysUtils, DSupport;

{

Arrays to hold structs data

   UserType      Subitem  Subitem          Subitem  Subitem     SubUser
   Names         Count    Indices          Name     Type        Index

0 "MyType1" ---> 3 -----> 0 -----------> 0 "X" ---> Real ------> NA
                                    |--> 1 "Y" ---> Real ------> NA
                                    |--> 2 "Z" ---> Real ------> NA
1 "MyType2" ---> 4 -----> 3 -----|-----> 3 "A" ---> Real ------> NA
                                 |-----> 4 "B" ---> Real ------> NA
                                 |-----> 5 "C" ---> Real ------> NA
                                 |-----> 6 "D" ---> UserType --> 0 (MyType1)
2 "MyType3" ---> 2 -----> 7 --|--------> 7 "I" ---> Integer ---> NA
                              |--------> 8 "J" ---> Integer ---> NA
3 "MyTypeN" ---> ...


Alias type signified by:

1. A subitem count of zero
}


{Debug}

const
	ADDRVAR=1;	{ Address Variable Id (type = Integer) }
	INVAR=3;	{ Integer Variable Id (odd Nos.=integer) }
	RLVAR=4;	{ Real Variable Id }
	USRTYPE=23;	{User defined variable type}

{=====}

const UserMax = 500;		{Number of user defined variable types}
const SubitemMax = UserMax * 3;	{Number of total subitems available for user types}

{All base data types are negative. A postive data type
 indicates a user-defined data type and is an index to
 a previously defined user data type}

var UserTypeNames: array [0..UserMax-1] of string;	{Name of each user defined type}
var SubitemCount: array [0..UserMax-1] of integer;	{Number of subitems in each user type}
var SubItemIndices:  array [0..UserMax-1] of integer;	{Index into subitem arrays for each user type}

var SubItemName: array [0..SubitemMax-1] of string;	{Name of each subItem}
var SubItemType: array [0..SubitemMax-1] of integer;	{variable type of each subItem}
var SubUserIndex: array [0..SubitemMax-1] of integer;	{Index into user-defined type list for subitems that are user defined}


var UserTypeIndex: integer;	{Index into type array, points to next available space}
var SubItemIndex: integer;	{Index into subitem array,  points to next available space}

{Error types for user-defined data types }

type TUserTypeErrors = (ueNoError,ueArrayFull, ueRecursiveDef, ueReDeclaration);

{Used by compiler}

procedure ClearUserTypes;
procedure TrimUserTypes(Index: integer);
function AddUserType(Name: string; var Index: integer): TUserTypeErrors;
function AddSubitem(Index: integer; Name: string; DataType,UserType: integer): TUserTypeErrors;
function AddAlias(Name: string; var Index: integer; DataType,UserType: integer): TUserTypeErrors;
function GetVariableSize(TypeIndex: integer; Target: string): integer;
function GetVariableOffset(TypeIndex: integer; Target: string): integer;
function GetVariableType(TypeIndex: integer; Target: string; var SType, SIndex: integer): boolean;
function GetRootName(S: string): string;
function GetComponentString(Base,S: string): string;

{Used by debugger}

function GetTypeString(DataType, UserIndex: Integer): string;
function FindUserType(Name: string): integer;
function GetDataTypeSize(DataType: integer): integer;


implementation

uses XPLDataTypes;


{Currently Unused: }

function GetLastComponent(S: string): string;
{Get last component of struct, not including dot}
{i.e. "Frog.Cat.X" returns "X" }
var Inx: integer;
begin
Inx:=FindLastOccurrence(S,'.');
if Inx<1 then Result:=S
else Result:=Copy(S,Inx+1,Length(S)-Inx);
end;



function GetRootName(S: string): string;
{Remove the dot and everything past it.}
{i.e. "Frog.Cat.X" returns "Frog" }
var I: integer;
var Dot: boolean;
begin
Dot:=False;
SetLength(Result,Length(S));
For I:=1 to Length(S) do
	begin
	if S[I]='.' then Dot:=True;
	if Dot then Result[I]:=' '
	else Result[I]:=S[I];
	end;
end;


function GetComponentString(Base,S: string): string;
{Remove the base string and the trailing dot if present}
{Which leaves all the components of the struct}
{i.e. "Frog.Cat.X" returns "Cat.X" }
var Inx: integer;
begin
Result:=S;
DeleteSubstring(Base,Result,False);
Inx:=Pos(')',Result);
if Inx>0 then Delete(Result,1,Inx);
if (Length(Result)>0) and (Result[1]='.') then Delete(Result,1,1);
end;



function GetTypeString(DataType, UserIndex: Integer): string;
{Get user type names - used for debugging tools}
begin
case DataType of
 INVAR: Result:='int';
 RLVAR: Result:='real';
 ADDRVAR: Result:='char';
 USRTYPE: Result:=UserTypeNames[UserIndex];
 end;
end;



function GetDataTypeSize(DataType: integer): integer;
{Get the size in bytes of each data type}
begin
case DataType of
 INVAR: Result:=IntBytes;
 RLVAR: Result:=RealSize;
 ADDRVAR: Result:=IntBytes;
 else Result:=0;
 end
end;



procedure ClearUserTypes;
{Clear user-type array}
var I: integer;
begin
for I:=0 to High(UserTypeNames) do UserTypeNames[I]:='';
for I:=0 to High(SubItemName) do SubItemName[I]:='';
UserTypeIndex:=0;
SubItemIndex:=0;
end;


procedure TrimUserTypes(Index: integer);
{Trim user types back to a previous location specified by Index}
{Used when local variables and declarations go out of scope}
begin
{Don't trim if array is shorter than index}
if Index>=UserTypeIndex then exit;
UserTypeIndex:=Index;
SubItemIndex:=SubItemIndices[Index]+SubitemCount[Index];
end;


function FindUserType(Name: string): integer;
{Find user type by Type name - non-case sensitve)
{Return index of matching location}
{Returns -1 if name is not found in list}
var I: integer;
begin
for Result:=0 to UserTypeIndex-1 do
 if SameText(UserTypeNames[Result],Name) then exit;
Result:=-1;
end;



function AddUserType(Name: string; var Index: integer): TUserTypeErrors;
{Insert new user type sympbol. Returns array-full error if array is full}
{Returns Re-declaration error if the symbol already exists}
begin
Result:=ueArrayFull;
if UserTypeIndex>=UserMax then exit;
Result:=ueReDeclaration;
if FindUserType(Name)>=0 then exit;
Result:=ueNoError;
Index:=UserTypeIndex;
{XPL identifiers are padded with spaces, so trim them off}
UserTypeNames[UserTypeIndex]:=Trim(Name);
SubItemIndices[Index]:=SubItemIndex;
SubItemCount[Index]:=0;
UserTypeIndex:=UserTypeIndex+1;
end;




function AddSubitem(Index: integer; Name: string; DataType,UserType: integer): TUserTypeErrors;
{Add subitem to user type specified by Index}
{Returns false is array is full}
begin
Result:=ueArrayFull;
if SubItemIndex>=SubItemMax then exit;
SubItemCount[Index]:=SubItemCount[Index]+1;
{XPL identifiers are padded with spaces, so trim them off}
SubItemName[SubItemIndex]:=Trim(Name);
SubItemType[SubItemIndex]:=DataType;
SubUserIndex[SubItemIndex]:=UserType;
SubItemIndex:=SubItemIndex+1;
Result:=ueNoError;
end;


function AddAlias(Name: string; var Index: integer; DataType,UserType: integer): TUserTypeErrors;
{Add alias type to list. Aliases have one subitem, with no name }
begin
Result:=AddUserType(Name, Index);
if Result<>ueNoError then exit;
Result:=AddSubitem(Index, '', DataType, UserType);
end;

{---------------------------------------------------------------------}

function GetVariableSize(TypeIndex: integer; Target: string): integer;
{Find the size of a specified variable, including all subitems}
{Data Type specified by "TypeIndex," "Target" specifies the variable i.e. "P.A.X"}
{Returns -1 if the specified target is not found}

	function FindItemSize(Inx: integer): integer;
	{Find total item size by recursing all subitems }
	{Top-level type is specified by Inx}
	var J, Start, Stop: integer;
	begin
	Result:=0;
	{Get the range of indices for this type}
	Start:=SubItemIndices[Inx];
	Stop:=Start + SubItemCount[Inx]-1;
	for J:=Start to Stop do
		begin
		if SubItemType[J]=USRTYPE then Result:=Result+FindItemSize(SubUserIndex[J])
		else Result:=Result+GetDataTypeSize(SubItemType[J]);
		end;
	end;

	procedure FindSubitem(Inx: integer; Name: string);
	{Scan all subitems of data type specified by Inx, recursing user-types}
	{"Name" is the accumulated variable name as the recursion progresses }
	var J, Start, Stop: integer;
	var S: string;
	begin
	{Get the range of indices for this type}
	Start:=SubItemIndices[Inx];
	Stop:=Start + SubItemCount[Inx]-1;
	for J:=Start to Stop do
		begin
		{Accumulate variable name items}
		if Name='' then S:=SubItemName[J]
		else S:=Name+'.'+SubItemName[J];
		{Get item size when target found}
		if Target=S then
			begin
			if SubItemType[J]=USRTYPE then Result:=FindItemSize(SubUserIndex[J])
			else Result:=GetDataTypeSize(SubItemType[J]);
			end;
		{Recurse if the subitem is user defined}
		if SubItemType[J]=USRTYPE then FindSubitem(SubUserIndex[J],S);
		end;
	end;

begin
Result:=-1;
if Target='' then Result:=FindItemSize(TypeIndex)
else FindSubitem(TypeIndex,'');
end;



function GetVariableOffset(TypeIndex: integer; Target: string): integer;
{Find offset to a specified variable by recursing through all subitems of a data type}
{User data type specified by "TypeIndex," "Target" specifies the variable i.e. "P.A.X"}
{Returns -1 if the specified target is not found}
var Offset: integer;

	procedure TraverseSubitems(Inx: integer; Name: string);
	{Scan all subitems of data type specified by Inx, recursing on user-types}
	{"Name" is the accumulated variable name as the recursion progresses }
	var J, Start, Stop: integer;
	var S: string;
	begin
	{Get the range of indices for this type}
	Start:=SubItemIndices[Inx];
	Stop:=Start + SubItemCount[Inx]-1;
	for J:=Start to Stop do
		begin
		{Accumulate variable name items}
		if Name='' then S:=SubItemName[J]
		else S:=Name+'.'+SubItemName[J];
		{Exit when target found}
		if Target=S then
			begin
			Result:=Offset;
			Exit;
			end;
		{Recurse if the subitem is user defined, otherwise increment offset}
		if SubItemType[J]=USRTYPE then TraverseSubitems(SubUserIndex[J],S)
		else Offset:=Offset + GetDataTypeSize(SubItemType[J]);
		end;
	end;

begin
Offset:=0;
Result:=-1;
if Target='' then Result:=0
else TraverseSubitems(TypeIndex,'');
end;



function GetVariableType(TypeIndex: integer; Target: string; var SType, SIndex: integer): boolean;
{Find the Type of a specified variable, Data Type specified by "TypeIndex,"}
{"Target" specifies the variable i.e. "P.A.X". SType Returns the data type.\
{If SType = USRTYPE then SIndex will hold an index to the struct}
{Returns False if the specified target is not found}


	procedure FindSubitem(Inx: integer; Name: string);
	{Scan all subitems of data type specified by Inx, recursing user-types}
	{"Name" is the accumulated variable name as the recursion progresses }
	var J, Start, Stop: integer;
	var S: string;
	begin
	{Get the range of indices for this type}
	Start:=SubItemIndices[Inx];
	Stop:=Start + SubItemCount[Inx]-1;
	for J:=Start to Stop do
		begin
		{Accumulate variable name items}
		if Name='' then S:=SubItemName[J]
		else S:=Name+'.'+SubItemName[J];
		{Get item size when target found}
		if Target=S then
			begin
			Result:=True;
			if SubItemType[J]=USRTYPE then
				begin
				SType:=USRTYPE;
				SIndex:=SubUserIndex[J];
				end
			else SType:=SubItemType[J];
			Exit;
			end;
		{Recurse if the subitem is user defined}
		if SubItemType[J]=USRTYPE then FindSubitem(SubUserIndex[J],S);
		end;
	end;

begin
Result:=False;
if Target='' then
	begin
	Result:=True;
	{If the component name is empty, then it is an Alias}
	if SubItemName[SubItemIndices[TypeIndex]] = '' then
		SType:=SubItemType[SubItemIndices[TypeIndex]]
	else SType:=USRTYPE;
	SIndex:=TypeIndex
	end
else FindSubitem(TypeIndex,'');
end;



end.
